/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LAPACK
#define COMMON_LAPACK

#ifndef ASSEMBLER

/* Lapack Library */

blasint sgetf2_k(long, long, float  *a, long, blasint *, long, float *);
blasint dgetf2_k(long, long, double *a, long, blasint *, long, double *);
blasint qgetf2_k(long, long, xdouble *a, long, blasint *, long, xdouble *);
blasint cgetf2_k(long, long, float  *a, long, blasint *, long, float *);
blasint zgetf2_k(long, long, double *a, long, blasint *, long, double *);
blasint xgetf2_k(long, long, xdouble *a, long, blasint *, long, xdouble *);

blasint sgetrf_single(long, long, float  *a, long, blasint *, long, float *);
blasint dgetrf_single(long, long, double *a, long, blasint *, long, double *);
blasint qgetrf_single(long, long, xdouble *a, long, blasint *, long, xdouble *);
blasint cgetrf_single(long, long, float  *a, long, blasint *, long, float *);
blasint zgetrf_single(long, long, double *a, long, blasint *, long, double *);
blasint xgetrf_single(long, long, xdouble *a, long, blasint *, long, xdouble *);

blasint sgetrf_parallel(long, long, float  *a, long, blasint *, long, float *);
blasint dgetrf_parallel(long, long, double *a, long, blasint *, long, double *);
blasint qgetrf_parallel(long, long, xdouble *a, long, blasint *, long, xdouble *);
blasint cgetrf_parallel(long, long, float  *a, long, blasint *, long, float *);
blasint zgetrf_parallel(long, long, double *a, long, blasint *, long, double *);
blasint xgetrf_parallel(long, long, xdouble *a, long, blasint *, long, xdouble *);

int slaswp_plus(long, long, long, float,  float  *, long, float  *, long, blasint *, long);
int slaswp_minus(long, long, long, float,  float  *, long, float  *, long, blasint *, long);
int dlaswp_plus(long, long, long, double, double *, long, double *, long, blasint *, long);
int dlaswp_minus(long, long, long, double, double *, long, double *, long, blasint *, long);
int qlaswp_plus(long, long, long, xdouble, xdouble *, long, xdouble *, long, blasint *, long);
int qlaswp_minus(long, long, long, xdouble, xdouble *, long, xdouble *, long, blasint *, long);
int claswp_plus(long, long, long, float, float, float *, long, float *, long, blasint *, long);
int claswp_minus(long, long, long, float, float, float *, long, float *, long, blasint *, long);
int zlaswp_plus(long, long, long, double, double, double *, long, double *, long, blasint *, long);
int zlaswp_minus(long, long, long, double, double, double *, long, double *, long, blasint *, long);
int xlaswp_plus(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, blasint *, long);
int xlaswp_minus(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, blasint *, long);

blasint sgetrs_N(long, long, float  *, long, float  *, long, blasint *, float  *);
blasint sgetrs_T(long, long, float  *, long, float  *, long, blasint *, float  *);

blasint dgetrs_N(long, long, double *, long, double *, long, blasint *, double *);
blasint dgetrs_T(long, long, double *, long, double *, long, blasint *, double *);

blasint qgetrs_N(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint qgetrs_T(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);

blasint cgetrs_N(long, long, float  *, long, float  *, long, blasint *, float  *);
blasint cgetrs_T(long, long, float  *, long, float  *, long, blasint *, float  *);
blasint cgetrs_R(long, long, float  *, long, float  *, long, blasint *, float  *);
blasint cgetrs_C(long, long, float  *, long, float  *, long, blasint *, float  *);

blasint zgetrs_N(long, long, double *, long, double *, long, blasint *, double *);
blasint zgetrs_T(long, long, double *, long, double *, long, blasint *, double *);
blasint zgetrs_R(long, long, double *, long, double *, long, blasint *, double *);
blasint zgetrs_C(long, long, double *, long, double *, long, blasint *, double *);

blasint xgetrs_N(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_T(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_R(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_C(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);

blasint spotf2_U(long, float  *, long, float *);
blasint spotf2_L(long, float  *, long, float *);
blasint dpotf2_U(long, double *, long, double *);
blasint dpotf2_L(long, double *, long, double *);
blasint qpotf2_U(long, xdouble *, long, xdouble *);
blasint qpotf2_L(long, xdouble *, long, xdouble *);
blasint cpotf2_U(long, float  *, long, float *);
blasint cpotf2_L(long, float  *, long, float *);
blasint zpotf2_U(long, double *, long, double *);
blasint zpotf2_L(long, double *, long, double *);
blasint xpotf2_U(long, xdouble *, long, xdouble *);
blasint xpotf2_L(long, xdouble *, long, xdouble *);

blasint spotrf_U(long, float  *, long, float *);
blasint spotrf_L(long, float  *, long, float *);
blasint dpotrf_U(long, double *, long, double *);
blasint dpotrf_L(long, double *, long, double *);
blasint qpotrf_U(long, xdouble *, long, xdouble *);
blasint qpotrf_L(long, xdouble *, long, xdouble *);
blasint cpotrf_U(long, float  *, long, float *);
blasint cpotrf_L(long, float  *, long, float *);
blasint zpotrf_U(long, double *, long, double *);
blasint zpotrf_L(long, double *, long, double *);
blasint xpotrf_U(long, xdouble *, long, xdouble *);
blasint xpotrf_L(long, xdouble *, long, xdouble *);

blasint slauu2_U(long, float  *, long, float *);
blasint slauu2_L(long, float  *, long, float *);
blasint dlauu2_U(long, double *, long, double *);
blasint dlauu2_L(long, double *, long, double *);
blasint qlauu2_U(long, xdouble *, long, xdouble *);
blasint qlauu2_L(long, xdouble *, long, xdouble *);
blasint clauu2_U(long, float  *, long, float *);
blasint clauu2_L(long, float  *, long, float *);
blasint zlauu2_U(long, double *, long, double *);
blasint zlauu2_L(long, double *, long, double *);
blasint xlauu2_U(long, xdouble *, long, xdouble *);
blasint xlauu2_L(long, xdouble *, long, xdouble *);

blasint slauum_U(long, float  *, long, float *);
blasint slauum_L(long, float  *, long, float *);
blasint dlauum_U(long, double *, long, double *);
blasint qlauum_L(long, xdouble *, long, xdouble *);
blasint qlauum_U(long, xdouble *, long, xdouble *);
blasint dlauum_L(long, double *, long, double *);
blasint clauum_U(long, float  *, long, float *);
blasint clauum_L(long, float  *, long, float *);
blasint zlauum_U(long, double *, long, double *);
blasint zlauum_L(long, double *, long, double *);
blasint xlauum_U(long, xdouble *, long, xdouble *);
blasint xlauum_L(long, xdouble *, long, xdouble *);

blasint strti2_UU(long, float  *, long, float  *);
blasint strti2_UN(long, float  *, long, float  *);
blasint strti2_LU(long, float  *, long, float  *);
blasint strti2_LN(long, float  *, long, float  *);
blasint dtrti2_UU(long, double *, long, double *);
blasint dtrti2_UN(long, double *, long, double *);
blasint dtrti2_LU(long, double *, long, double *);
blasint dtrti2_LN(long, double *, long, double *);
blasint qtrti2_UU(long, xdouble *, long, xdouble *);
blasint qtrti2_UN(long, xdouble *, long, xdouble *);
blasint qtrti2_LU(long, xdouble *, long, xdouble *);
blasint qtrti2_LN(long, xdouble *, long, xdouble *);
blasint ctrti2_UU(long, float  *, long, float  *);
blasint ctrti2_UN(long, float  *, long, float  *);
blasint ctrti2_LU(long, float  *, long, float  *);
blasint ctrti2_LN(long, float  *, long, float  *);
blasint ztrti2_UU(long, double *, long, double *);
blasint ztrti2_UN(long, double *, long, double *);
blasint ztrti2_LU(long, double *, long, double *);
blasint ztrti2_LN(long, double *, long, double *);
blasint xtrti2_UU(long, xdouble *, long, xdouble *);
blasint xtrti2_UN(long, xdouble *, long, xdouble *);
blasint xtrti2_LU(long, xdouble *, long, xdouble *);
blasint xtrti2_LN(long, xdouble *, long, xdouble *);

blasint strtri_UU(long, float  *, long, float  *);
blasint strtri_UN(long, float  *, long, float  *);
blasint strtri_LU(long, float  *, long, float  *);
blasint strtri_LN(long, float  *, long, float  *);
blasint dtrtri_UU(long, double *, long, double *);
blasint dtrtri_UN(long, double *, long, double *);
blasint dtrtri_LU(long, double *, long, double *);
blasint dtrtri_LN(long, double *, long, double *);
blasint qtrtri_UU(long, xdouble *, long, xdouble *);
blasint qtrtri_UN(long, xdouble *, long, xdouble *);
blasint qtrtri_LU(long, xdouble *, long, xdouble *);
blasint qtrtri_LN(long, xdouble *, long, xdouble *);
blasint ctrtri_UU(long, float  *, long, float  *);
blasint ctrtri_UN(long, float  *, long, float  *);
blasint ctrtri_LU(long, float  *, long, float  *);
blasint ctrtri_LN(long, float  *, long, float  *);
blasint ztrtri_UU(long, double *, long, double *);
blasint ztrtri_UN(long, double *, long, double *);
blasint ztrtri_LU(long, double *, long, double *);
blasint ztrtri_LN(long, double *, long, double *);
blasint xtrtri_UU(long, xdouble *, long, xdouble *);
blasint xtrtri_UN(long, xdouble *, long, xdouble *);
blasint xtrtri_LU(long, xdouble *, long, xdouble *);
blasint xtrtri_LN(long, xdouble *, long, xdouble *);

#define SGETF2	sgetf2_k
#define DGETF2	dgetf2_k
#define QGETF2	qgetf2_k
#define CGETF2	cgetf2_k
#define ZGETF2	zgetf2_k
#define XGETF2	xgetf2_k

#ifndef SMP
#define SGETRF	sgetrf_single
#define DGETRF	dgetrf_single
#define QGETRF	qgetrf_single
#define CGETRF	cgetrf_single
#define ZGETRF	zgetrf_single
#define XGETRF	xgetrf_single
#else
#define SGETRF	sgetrf_parallel
#define DGETRF	dgetrf_parallel
#define QGETRF	qgetrf_parallel
#define CGETRF	cgetrf_parallel
#define ZGETRF	zgetrf_parallel
#define XGETRF	xgetrf_parallel
#endif

#define SGETRS_N	sgetrs_N
#define SGETRS_T	sgetrs_T
#define DGETRS_N	dgetrs_N
#define DGETRS_T	dgetrs_T
#define QGETRS_N	qgetrs_N
#define QGETRS_T	qgetrs_T
#define CGETRS_N	cgetrs_N
#define CGETRS_T	cgetrs_T
#define ZGETRS_N	zgetrs_N
#define ZGETRS_T	zgetrs_T
#define XGETRS_N	xgetrs_N
#define XGETRS_T	xgetrs_T

#define SLASWP_PLUS	slaswp_plus
#define SLASWP_MINUS	slaswp_minus
#define DLASWP_PLUS	dlaswp_plus
#define DLASWP_MINUS	dlaswp_minus
#define QLASWP_PLUS	qlaswp_plus
#define QLASWP_MINUS	qlaswp_minus
#define CLASWP_PLUS	claswp_plus
#define CLASWP_MINUS	claswp_minus
#define ZLASWP_PLUS	zlaswp_plus
#define ZLASWP_MINUS	zlaswp_minus
#define XLASWP_PLUS	xlaswp_plus
#define XLASWP_MINUS	xlaswp_minus

#ifndef COMPLEX
#ifdef XDOUBLE
#define GETF2	QGETF2
#define GETRF	QGETRF
#define LASWP_PLUS	QLASWP_PLUS
#define LASWP_MINUS	QLASWP_MINUS
#define GETRS_N		QGETRS_N
#define GETRS_T		QGETRS_T
#define GETRF_SINGLE	qgetrf_single
#elif defined(DOUBLE)
#define GETF2	DGETF2
#define GETRF	DGETRF
#define LASWP_PLUS	DLASWP_PLUS
#define LASWP_MINUS	DLASWP_MINUS
#define GETRS_N		DGETRS_N
#define GETRS_T		DGETRS_T
#define GETRF_SINGLE	dgetrf_single
#else
#define GETF2	SGETF2
#define GETRF	SGETRF
#define LASWP_PLUS	SLASWP_PLUS
#define LASWP_MINUS	SLASWP_MINUS
#define GETRS_N		SGETRS_N
#define GETRS_T		SGETRS_T
#define GETRF_SINGLE	sgetrf_single
#endif
#else
#ifdef XDOUBLE
#define GETF2	XGETF2
#define GETRF	XGETRF
#define LASWP_PLUS	XLASWP_PLUS
#define LASWP_MINUS	XLASWP_MINUS
#define GETRS_N		XGETRS_N
#define GETRS_T		XGETRS_T
#define GETRF_SINGLE	xgetrf_single
#elif defined(DOUBLE)
#define GETF2	ZGETF2
#define GETRF	ZGETRF
#define LASWP_PLUS	ZLASWP_PLUS
#define LASWP_MINUS	ZLASWP_MINUS
#define GETRS_N		ZGETRS_N
#define GETRS_T		ZGETRS_T
#define GETRF_SINGLE	zgetrf_single
#else
#define GETF2	CGETF2
#define GETRF	CGETRF
#define LASWP_PLUS	CLASWP_PLUS
#define LASWP_MINUS	CLASWP_MINUS
#define GETRS_N		CGETRS_N
#define GETRS_T		CGETRS_T
#define GETRF_SINGLE	cgetrf_single
#endif
#endif

#ifndef COMPLEX
#ifdef XDOUBLE
#define  POTF2_U  qpotf2_U
#define  POTF2_L  qpotf2_L
#define  LAUU2_U  qlauu2_U
#define  LAUU2_L  qlauu2_L
#define  TRTI2_UU qtrti2_UU
#define  TRTI2_UN qtrti2_UN
#define  TRTI2_LU qtrti2_LU
#define  TRTI2_LN qtrti2_LN
#define  TRTRI_UU qtrtri_UU
#define  TRTRI_UN qtrtri_UN
#define  TRTRI_LU qtrtri_LU
#define  TRTRI_LN qtrtri_LN
#elif defined(DOUBLE)
#define  POTF2_U  dpotf2_U
#define  POTF2_L  dpotf2_L
#define  LAUU2_U  dlauu2_U
#define  LAUU2_L  dlauu2_L
#define  TRTI2_UU dtrti2_UU
#define  TRTI2_UN dtrti2_UN
#define  TRTI2_LU dtrti2_LU
#define  TRTI2_LN dtrti2_LN
#define  TRTRI_UU dtrtri_UU
#define  TRTRI_UN dtrtri_UN
#define  TRTRI_LU dtrtri_LU
#define  TRTRI_LN dtrtri_LN
#else
#define  POTF2_U  spotf2_U
#define  POTF2_L  spotf2_L
#define  LAUU2_U  slauu2_U
#define  LAUU2_L  slauu2_L
#define  TRTI2_UU strti2_UU
#define  TRTI2_UN strti2_UN
#define  TRTI2_LU strti2_LU
#define  TRTI2_LN strti2_LN
#define  TRTRI_UU strtri_UU
#define  TRTRI_UN strtri_UN
#define  TRTRI_LU strtri_LU
#define  TRTRI_LN strtri_LN
#endif
#else
#ifdef XDOUBLE
#define  POTF2_U  xpotf2_U
#define  POTF2_L  xpotf2_L
#define  LAUU2_U  xlauu2_U
#define  LAUU2_L  xlauu2_L
#define  TRTI2_UU xtrti2_UU
#define  TRTI2_UN xtrti2_UN
#define  TRTI2_LU xtrti2_LU
#define  TRTI2_LN xtrti2_LN
#define  TRTRI_UU xtrtri_UU
#define  TRTRI_UN xtrtri_UN
#define  TRTRI_LU xtrtri_LU
#define  TRTRI_LN xtrtri_LN
#elif defined(DOUBLE)
#define  POTF2_U  zpotf2_U
#define  POTF2_L  zpotf2_L
#define  LAUU2_U  zlauu2_U
#define  LAUU2_L  zlauu2_L
#define  TRTI2_UU ztrti2_UU
#define  TRTI2_UN ztrti2_UN
#define  TRTI2_LU ztrti2_LU
#define  TRTI2_LN ztrti2_LN
#define  TRTRI_UU ztrtri_UU
#define  TRTRI_UN ztrtri_UN
#define  TRTRI_LU ztrtri_LU
#define  TRTRI_LN ztrtri_LN
#else
#define  POTF2_U  cpotf2_U
#define  POTF2_L  cpotf2_L
#define  LAUU2_U  clauu2_U
#define  LAUU2_L  clauu2_L
#define  TRTI2_UU ctrti2_UU
#define  TRTI2_UN ctrti2_UN
#define  TRTI2_LU ctrti2_LU
#define  TRTI2_LN ctrti2_LN
#define  TRTRI_UU ctrtri_UU
#define  TRTRI_UN ctrtri_UN
#define  TRTRI_LU ctrtri_LU
#define  TRTRI_LN ctrtri_LN
#endif
#endif

#endif
#endif
