/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifndef COMPLEX
#ifdef XDOUBLE
#define ERROR_NAME "QGESV  "
#elif defined(DOUBLE)
#define ERROR_NAME "DGESV  "
#else
#define ERROR_NAME "SGESV  "
#endif
#else
#ifdef XDOUBLE
#define ERROR_NAME "XGESV  "
#elif defined(DOUBLE)
#define ERROR_NAME "ZGESV  "
#else
#define ERROR_NAME "CGESV  "
#endif
#endif

int NAME(blasint *N, blasint *NRHS, FLOAT *a, blasint *LDA, blasint *ipiv,
         FLOAT *b, blasint *LDB, blasint *Info){

  blasint n    = *N;
  blasint nrhs = *NRHS;
  blasint lda = *LDA;
  blasint ldb = *LDB;
  blasint info;
  FLOAT *buffer;
  
  info  = 0;
  if (ldb < MAX(1,n)) info = 7;
  if (lda < MAX(1,n)) info = 4;
  if (nrhs < 0)       info = 2;
  if (n < 0)          info = 1;

  if (info) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    *Info = - info;
    return 0;
  }

  *Info = 0;

  if (n == 0 || nrhs == 0) return 0;

#ifdef SMP
  if (blas_cpu_number == 0) blas_get_cpu_number();
#ifdef SMP_SERVER
  if (blas_server_avail == 0) blas_thread_init();
#endif
#endif

  buffer = (FLOAT *)blas_memory_alloc(0);

#ifdef USE_BUSYWAIT
  blas_busywait(1);
#endif

  info = GETRF(n, n, a, lda, ipiv, 0, buffer);
  
  if (info == 0){
    GETRS_N(n, nrhs, a, lda, b, ldb, ipiv, buffer);
  }

#ifdef USE_BUSYWAIT
  blas_busywait(0);
#endif

  blas_memory_free(buffer);

  *Info = info;

  return 0;
}
