/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "QGETF2"
#elif defined(DOUBLE)
#define ERROR_NAME "DGETF2"
#else
#define ERROR_NAME "SGETF2"
#endif

int NAME(blasint *M, blasint *N, FLOAT *a, blasint *LDA, blasint *ipiv, blasint *Info){

  blasint m = *M;
  blasint n = *N;
  blasint lda = *LDA;
  blasint info;
  FLOAT *buffer;
  
  info  = 0;
  if (lda < MAX(1,m)) info = 4;
  if (n < 0)          info = 2;
  if (m < 0)          info = 1;
  if (info) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    *Info = - info;
    return 0;
  }

  *Info = 0;

  if (m == 0 || n == 0) return 0;
  
  buffer = (FLOAT *)blas_memory_alloc(0);

  info = GETF2(m, n, a, lda, ipiv, 0, buffer);

  *Info = info;

  blas_memory_free(buffer);

  return 0;
}
