/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "QGETRS"
#elif defined(DOUBLE)
#define ERROR_NAME "DGETRS"
#else
#define ERROR_NAME "SGETRS"
#endif

static blasint (*getrs[])(long, long, FLOAT *, long, FLOAT *, long, blasint *, FLOAT *) ={

#ifdef XDOUBLE
  qgetrs_N, qgetrs_T,
#elif defined(DOUBLE)
  dgetrs_N, dgetrs_T,
#else
  sgetrs_N, sgetrs_T,
#endif
  };

int NAME(char *TRANS, blasint *N, blasint *NRHS, FLOAT *a, blasint *LDA,
	    blasint *ipiv, FLOAT *b, blasint *LDB, blasint *Info){

  char trans_arg = *TRANS;

  blasint n = *N;
  blasint nrhs = *NRHS;
  blasint lda = *LDA;
  blasint ldb = *LDB;
  blasint info;
  int trans;
  FLOAT *buffer;

  info = 0;

  TOUPPER(trans_arg);
  trans = -1;

  if (trans_arg == 'N') trans = 0;
  if (trans_arg == 'T') trans = 1;
  if (trans_arg == 'R') trans = 0;
  if (trans_arg == 'C') trans = 1;

  if (ldb  < MAX(1, n)) info = 8;
  if (lda  < MAX(1, n)) info = 5;
  if (nrhs < 0) info = 3;
  if (n    < 0) info = 2;
  if (trans< 0) info = 1;

  if (info != 0) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    return 0;
  }
  
  *Info = info;

  if (n == 0 || nrhs == 0) return 0;

  buffer = (FLOAT *)blas_memory_alloc(0);

  (getrs[trans])(n, nrhs, a, lda, b, ldb, ipiv, buffer);
  
  blas_memory_free(buffer);

  return 0;
  
}
