/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 = 1.;

blasint CNAME(long n, long nrhs, FLOAT *a, long lda, FLOAT *b, long ldb, blasint *ipiv, FLOAT *buffer){

#ifndef TRANS
    LASWP_PLUS(nrhs, 1,    n, ZERO, b, ldb, NULL, 0, ipiv, 1);
    if (nrhs == 1){
      TRSV_NLU (n, a, lda, b, 1, buffer);
      TRSV_NUN (n, a, lda, b, 1, buffer);
    } else {
      TRSM_LNLU (n,    nrhs, 0, dp1,  a, lda, NULL, 0, b,    ldb, buffer);
      TRSM_LNUN (n,    nrhs, 0, dp1,  a, lda, NULL, 0, b,    ldb, buffer);
    }
#else
    if (nrhs == 1){
      TRSV_TUN (n, a, lda, b, 1, buffer);
      TRSV_TLU (n, a, lda, b, 1, buffer);
    } else {
      TRSM_LTUN  (n,    nrhs, 0, dp1,  a, lda, NULL, 0, b,    ldb, buffer);
      TRSM_LTLU  (n,    nrhs, 0, dp1,  a, lda, NULL, 0, b,    ldb, buffer);
    }
    LASWP_MINUS(nrhs, 1,    n, ZERO, b, ldb, NULL, 0, ipiv, -1);
#endif

    return 0;
  }
