/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#ifdef EV6
#define  UNROLL4
#endif

	.set noat
	.set noreorder
.text
	.align 5
	.globl NAME
	.ent NAME
NAME:
	.frame $sp,0,$26,0
#ifdef PROFILE
	ldgp	$gp, 0($27)
	lda	$28, _mcount
	jsr	$28, ($28), _mcount
#endif

	ldq	$7,   16($sp)		# $7 = ipiv
	unop
	ldl	$19,  24($sp)		# incx
	unop
#ifndef PROFILE
	.prologue 0
#else
	.prologue 1
#endif

#ifndef MINUS
	lda	$1,   -1($17)		# $1 = k1 - 1
	s4addq	$1,  $7, $7		# ipiv += k1 - 1
#else
	s4subq	$18, 4,  $1		# $1 = k2 - 1
	mull	$1, $19, $1
	subq	$7, $1,  $7
#endif

	subl	$18, $17, $5		# count  = k2 - k1
	addl	$5,   1, $5		# count += 1
#ifdef UNROLL8
	sra	$16,   3, $0		# k = (n >> 2)
#elif defined(UNROLL4)
	sra	$16,   2, $0		# k = (n >> 2)
#else
	sra	$16,   1, $0		# k = (n >> 2)
#endif
	ble	$5, $End
	.align 4

	lda	$20, -SIZE($20)
	SXADDQ	$21, 0, $21
	ble	$0,  $Sub		# if (k =< 0) goto $L3
	.align 4

$MainPart:
	ldl	$28,  0($7)		# ip = *p
	mov	$5, $23		# i = count
	s4addq	$19, $7, $24		# p += incx

#ifndef MINUS
	mov	$17, $22		# ii = k1
	SXADDQ	$17, $20, $6		# fromA = a + k1
#else
	mov	$18, $22		# ii = k2
	SXADDQ	$18, $20, $6		# fromA = a + k2
#endif
	.align 4

$MainLoop:
	cmpeq	$28, $22, $1		# ip == ii ?
	SXADDQ	$28, $20, $3		# toA = a + ip

	ldl	$28,  0($24)		# ip = *p
	s4addq	$19, $24, $24		# p += incx

#ifdef EV6
	ldl	$31,  8*4($24)
#else
	ldl	$31,  4*4($24)
#endif
	lda	$23,  -1($23)		# i++
#ifndef MINUS
	lda	$22,  1($22)		# ii++
#else
	lda	$22,  -1($22)		# ii--
#endif
	bne	$1,  $MainSkip		# if (ip == ii)
	.align 4

#ifdef UNROLL8
	LD	$f10, 0($3)
	addq	$3,  $21, $4		# toB = toA + lda
	LD	$f11, 0($4)
	addq	$4,  $21, $4		# toC = toB + lda

	LD	$f12, 0($4)
	addq	$4,  $21, $4		# toD = toC + lda
	LD	$f13, 0($4)
	addq	$4,  $21, $4		# toD = toC + lda

	LD	$f14, 0($4)
	addq	$4,  $21, $4		# toB = toA + lda
	LD	$f15, 0($4)
	addq	$4,  $21, $4		# toC = toB + lda

	LD	$f16, 0($4)
	addq	$4,  $21, $4		# toD = toC + lda
	LD	$f17, 0($4)
	addq	$4,  $21, $4		# toD = toC + lda
	nop

	LD	$f18, 0($6)
	addq	$6,  $21, $8
	LD	$f19, 0($8)
	addq	$8,  $21, $8

	LD	$f20, 0($8)
	addq	$8,  $21, $8
	LD	$f21, 0($8)
	addq	$8,  $21, $8

	LD	$f22, 0($8)
	addq	$8,  $21, $8
	LD	$f23, 0($8)
	addq	$8,  $21, $8

	LD	$f24, 0($8)
	addq	$8,  $21, $8
	LD	$f25, 0($8)
	unop

	ST	$f10,  0($6)
	addq	$6,  $21, $8
	ST	$f11,  0($8)
	addq	$8,  $21, $8

	ST	$f12,  0($8)
	addq	$8,  $21, $8
	ST	$f13,  0($8)
	addq	$8,  $21, $8

	ST	$f14,  0($8)
	addq	$8,  $21, $8
	ST	$f15,  0($8)
	addq	$8,  $21, $8

	ST	$f16,  0($8)
	addq	$8,  $21, $8
	ST	$f17,  0($8)
	unop

	ST	$f18,  0($3)
	addq	$3,  $21, $4
	ST	$f19,  0($4)
	addq	$4,  $21, $4

	ST	$f20,  0($4)
	addq	$4,  $21, $4
	ST	$f21,  0($4)
	addq	$4,  $21, $4

	ST	$f22,  0($4)
	addq	$4,  $21, $4
	ST	$f23,  0($4)
	addq	$4,  $21, $4

	ST	$f24,  0($4)
	addq	$4,  $21, $4
	ST	$f25,  0($4)
	unop
#elif defined(UNROLL4)
	LD	$f10, 0($3)
	addq	$3,  $21, $4		# toB = toA + lda
	LD	$f11, 0($4)
	addq	$4,  $21, $4		# toC = toB + lda

	LD	$f12, 0($4)
	addq	$4,  $21, $4		# toD = toC + lda
	LD	$f13, 0($4)
	nop

	LD	$f14, 0($6)
	addq	$6,  $21, $8
	LD	$f15,0($8)
	addq	$8,  $21, $8

	LD	$f16,0($8)
	addq	$8,  $21, $8
	LD	$f17,0($8)
	unop
	
	ST	$f10, 0($6)
	addq	$6,  $21, $8
	ST	$f11,0($8)
	addq	$8,  $21, $8

	ST	$f12,0($8)
	addq	$8,  $21, $8
	ST	$f13,0($8)
	unop

	ST	$f14,0($3)
	addq	$3,  $21, $4		# toD = toC + lda
	ST	$f15,0($4)
	addq	$4,  $21, $4		# toD = toC + lda

	ST	$f16,0($4)
	addq	$4,  $21, $4		# toD = toC + lda
	ST	$f17,0($4)
	unop
#else
	LD	$f10, 0($3)
	addq	$3,  $21, $4		# toB = toA + lda
	LD	$f11, 0($4)
	unop

	LD	$f14, 0($6)
	addq	$6,  $21, $8
	LD	$f15, 0($8)
	unop

	ST	$f10, 0($6)
	addq	$6,  $21, $8
	ST	$f11,0($8)
	unop

	ST	$f14,0($3)
	addq	$3,  $21, $4		# toD = toC + lda
	ST	$f15,0($4)
	unop
#endif
	.align 4

$MainSkip:
#ifndef MINUS
	lda	$6,  SIZE($6)
#else
	lda	$6,  -SIZE($6)
#endif
	bgt	$23, $MainLoop
	.align 4

#ifdef UNROLL8
	s8addq	$21, $20, $20		# a+= 4 * lda
#elif defined(UNROLL4)
	s4addq	$21, $20, $20		# a+= 8 * lda
#else
	addq	$21, $20, $20
	addq	$21, $20, $20
#endif
	lda	$0,  -1($0)		# k --
	bgt	$0,$MainPart
	.align 4

$Sub:
#ifdef UNROLL8
	and	$16,  7, $0
#elif defined(UNROLL4)
	and	$16,  3, $0
#else
	and	$16,  1, $0
#endif
	ble	$0,  $End
	.align 4

$SubPart:
	mov	$5, $23		#i = 0
	mov	$7,  $24		# p  = ipiv
	ldl	$2,   0($24)		# ip = *p
#ifndef MINUS
	mov	$17, $22		# ii = k1
	SXADDQ	$17, $20, $6		# fromA = a + k1
#else
	mov	$18, $22		# ii = k1
	SXADDQ	$18, $20, $6		# fromA = a + k1
#endif
	.align 4

$SubLoop:
	s4addq	$19, $24, $24		# p += incx
	cmpeq	$2,  $22, $1		# ip == ii ?
	SXADDQ	$2,  $20, $3		# toA = a + ip
	bne	$1,  $SubSkip		# if (ip == ii)
	.align 4

	LD	$f14,0($6)
	LD	$f13,0($3)
	ST	$f13,0($6)
	ST	$f14,0($3)
	.align 4

$SubSkip:
	ldl	$2,   0($24)		# ip = *p
#ifndef MINUS
	lda	$6,  SIZE($6)
	lda	$22,   1($22)		# ii++
#else
	lda	$6,  -SIZE($6)
	lda	$22,  -1($22)		# ii--
#endif
	lda	$23,  -1($23)		# i++
	bgt	$23, $SubLoop

	addq	$21, $20, $20		# a+= lda
	lda	$0,  -1($0)		# k --
	bgt	$0,$SubPart
	.align 4

$End:
	clr	$0
	ret
	.end NAME
