/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#ifdef EV6
#define UNROLL4
#endif

	.set noat
	.set noreorder
.text
	.align 5
	.globl NAME
	.ent NAME
NAME:
	.frame $sp,16,$26,0
#ifdef PROFILE
	ldgp	$gp, 0($27)
	lda	$28, _mcount
	jsr	$28, ($28), _mcount
#endif
	ldq	$5,   24($sp)		# $5 = ipiv
	ldl	$19,  32($sp)		# incx
#ifndef PROFILE
	.prologue 0
#else
	.prologue 1
#endif
	mov	$21, $20
	ldl	$21,   0($sp)

#ifndef MINUS
	lda	$1,   -1($17)		# $1 = k1 - 1
	s4addq	$1,  $5, $5		# ipiv += k1 - 1
#else
	s4subq	$18, 4,  $1		# $1 = k2 - 1
	mull	$1, $19, $1
	subq	$5, $1,  $5
#endif

	subl	$18, $17, $25		# count  = k2 - k1
	addl	$25,   1, $25		# count += 1
#ifdef UNROLL4
	sra	$16,   2, $0		# k = (n >> 2)
#else
	sra	$16,   1, $0		# k = (n >> 1)
#endif
	ble	$25, $End

	addl	$21, $21, $21
	lda	$20, -2 * SIZE($20)
	SXADDQ	$21, 0, $21
	ble	$0,  $Sub		# if (k =< 0) goto $L3
	.align 4

$MainPart:
	ldl	$28,  0($5)		# ip = *p
	mov	$25, $23		# i = count
	s4addq	$19, $5, $24		# p += incx

#ifndef MINUS
	mov	$17, $22		# ii = k1
	addl	$17, $17, $6
	SXADDQ	$6, $20, $6		# fromA = a + k1
#else
	mov	$18, $22		# ii = k2
	addl	$18, $18, $6
	SXADDQ	$6, $20, $6		# fromA = a + k2
#endif
	.align 4

$MainLoop:
	cmpeq	$28, $22, $1		# ip == ii ?
	addl	$28, $28, $28
	SXADDQ	$28, $20, $3		# toA = a + ip

	ldl	$28,  0($24)		# ip = *p
	s4addq	$19, $24, $24		# p += incx

	ldl	$31,  4*4($24)
	lda	$23,  -1($23)		# i++
	bne	$1,  $MainSkip		# if (ip == ii)

	LD	$f10, 0*SIZE($3)
	unop
	LD	$f11, 1*SIZE($3)
	addq	$3,  $21, $4		# toB = toA + lda

#ifdef UNROLL4
	LD	$f12, 0*SIZE($4)
	unop
	LD	$f13, 1*SIZE($4)
	addq	$4,  $21, $4		# toB = toA + lda

	LD	$f14, 0*SIZE($4)
	unop
	LD	$f15, 1*SIZE($4)
	addq	$4,  $21, $4		# toB = toA + lda

#endif
	LD	$f16, 0*SIZE($4)
	unop
	LD	$f17, 1*SIZE($4)
	unop

	LD	$f18,  0*SIZE($6)
	nop
	LD	$f19,  1*SIZE($6)
	addq	$6,  $21, $8

#ifdef UNROLL4

	LD	$f20,  0*SIZE($8)
	nop
	LD	$f21,  1*SIZE($8)
	addq	$8,  $21, $8

	LD	$f22,  0*SIZE($8)
	nop
	LD	$f23,  1*SIZE($8)
	addq	$8,  $21, $8

#endif

	LD	$f24,  0*SIZE($8)
	nop
	LD	$f25,  1*SIZE($8)
	nop

	ST	$f10,  0*SIZE($6)
	nop
	ST	$f11,  1*SIZE($6)
	addq	$6,  $21, $8

#ifdef UNROLL4

	ST	$f12, 0*SIZE($8)
	nop
	ST	$f13, 1*SIZE($8)
	addq	$8,  $21, $8

	ST	$f14, 0*SIZE($8)
	nop
	ST	$f15, 1*SIZE($8)
	addq	$8,  $21, $8

#endif

	ST	$f16, 0*SIZE($8)
	nop
	ST	$f17, 1*SIZE($8)
	nop

	ST	$f18, 0*SIZE($3)
	nop
	ST	$f19, 1*SIZE($3)
	addq	$3,  $21, $4		# toB = toA + lda

#ifdef UNROLL4

	ST	$f20, 0*SIZE($4)
	nop
	ST	$f21, 1*SIZE($4)
	addq	$4,  $21, $4		# toB = toA + lda

	ST	$f22, 0*SIZE($4)
	nop
	ST	$f23, 1*SIZE($4)
	addq	$4,  $21, $4		# toB = toA + lda

#endif

	ST	$f24, 0*SIZE($4)
	nop
	ST	$f25, 1*SIZE($4)
	nop
	.align 4

$MainSkip:
#ifndef MINUS
	lda	$6,   2*SIZE($6)
#else
	lda	$6,  -2*SIZE($6)
#endif

#ifndef MINUS
	lda	$8,  2*SIZE($8)
	lda	$22,  1($22)		# ii++
#else
	lda	$8,  -2*SIZE($8)
	lda	$22,  -1($22)		# ii--
#endif
	bgt	$23, $MainLoop
	.align 4

#ifdef UNROLL4
	s4addq	$21, $20, $20		# a+= 4 * lda
#else
	addq	$21, $20, $20		# a+= lda
	addq	$21, $20, $20		# a+= lda

#endif
	lda	$0,  -1($0)		# k --
	unop
	bgt	$0,$MainPart
	.align 4

$Sub:
#ifdef UNROLL4
	and	$16,  3, $0
#else
	and	$16,  1, $0
#endif
	ble	$0,  $End
	.align 4

$SubPart:
	mov	$25, $23		#i = 0
	mov	$5,  $24		# p  = ipiv
#ifndef MINUS
	mov	$17, $22		# ii = k1
	addl	$17, $17, $6
	SXADDQ	$6, $20, $6		# fromA = a + k1
#else
	mov	$18, $22		# ii = k1
	addl	$18, $18, $6
	SXADDQ	$6, $20, $6		# fromA = a + k1
#endif
	.align 4

$SubLoop:
	ldl	$2,   0($24)		# ip = *p
	s4addq	$19, $24, $24		# p += incx
	cmpeq	$2,  $22, $1		# ip == ii ?
	addl	$2,  $2,  $2
	SXADDQ	$2,  $20, $3		# toA = a + ip
	bne	$1,  $SubSkip		# if (ip == ii)

	LD	$f10, 0*SIZE($6)
	LD	$f11, 1*SIZE($6)

	LD	$f12, 0*SIZE($3)
	LD	$f13, 1*SIZE($3)

	ST	$f12, 0*SIZE($6)
	ST	$f13, 1*SIZE($6)

	ST	$f10, 0*SIZE($3)
	ST	$f11, 1*SIZE($3)
	.align 4

$SubSkip:
#ifndef MINUS
	lda	$6,   2*SIZE($6)
	lda	$22,   1($22)		# ii++
#else
	lda	$6,  -2*SIZE($6)
	lda	$22,  -1($22)		# ii--
#endif
	lda	$23,  -1($23)		# i++
	bgt	$23, $SubLoop

	addq	$21, $20, $20		# a+= lda
	lda	$0,  -1($0)		# k --
	bgt	$0,$SubPart
	.align 4

$End:
	clr	$0
	ret	$31,($26),1
	.end NAME
