/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

int CNAME(long n, long k1, long k2, FLOAT dummy1, FLOAT dummy4, 
	  FLOAT *a, long lda, 
	  FLOAT *dummy2, long dumy3, blasint *ipiv, long incx){

  long i, j, ip;
  blasint *piv;
  FLOAT *dx, *dy;
  FLOAT atemp1, atemp2, btemp1, btemp2;

  a -= 2;
  lda *= 2;

#ifndef MINUS
  ipiv += k1 - 1;
#else
  ipiv -= (k2 - 1) * incx;
#endif

  if (n  <= 0) return 0;
  if (k1 > k2) return 0;

  j = (n >> 1);
  if (j > 0) {
    do {

      piv = ipiv;

#ifndef MINUS
      i = k1;
#else
      i = k2;
#endif
      do {
	ip = *piv;
	piv += incx;
	if (ip != i) {
	  dx = a + i * 2;
	  dy = a + ip * 2;
	  atemp1 = *(dx + 0);
	  atemp2 = *(dx + 1);
	  btemp1 = *(dy + 0);
	  btemp2 = *(dy + 1);
	  *(dx + 0) = btemp1;
	  *(dx + 1) = btemp2;
	  *(dy + 0) = atemp1;
	  *(dy + 1) = atemp2;

	  dx = a + i  * 2 + lda;
	  dy = a + ip * 2 + lda;
	  atemp1 = *(dx + 0);
	  atemp2 = *(dx + 1);
	  btemp1 = *(dy + 0);
	  btemp2 = *(dy + 1);
	  *(dx + 0) = btemp1;
	  *(dx + 1) = btemp2;
	  *(dy + 0) = atemp1;
	  *(dy + 1) = atemp2;

	}
#ifndef MINUS
	i++;
#else
	i--;
#endif
      } while
#ifndef MINUS
	(i <= k2);
#else
        (i >= k1);
#endif
      
      a += 2 * lda;
      j --;
    } while (j > 0);
  }

  if (n & 1) {
    do {

      piv = ipiv;

#ifndef MINUS
      i = k1;
#else
      i = k2;
#endif
      do {
	ip = *piv;
	piv += incx;
	if (ip != i) {
	  dx = a + i * 2;
	  dy = a + ip * 2;
	  atemp1 = *(dx + 0);
	  atemp2 = *(dx + 1);
	  btemp1 = *(dy + 0);
	  btemp2 = *(dy + 1);
	  *(dx + 0) = btemp1;
	  *(dx + 1) = btemp2;
	  *(dy + 0) = atemp1;
	  *(dy + 1) = atemp2;
	}
#ifndef MINUS
	i++;
#else
	i--;
#endif
      } while
#ifndef MINUS
	(i <= k2);
#else
        (i >= k1);
#endif
      
      a += lda;
      j --;
    } while (j > 0);
  }

  return 0;
} 

