/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define PREFETCH_SIZE 128

#define SP	r12

#define N	r32
#define K1	r33
#define K2	r34
#define A	r37
#define LDA	r38
#define IPIV	r39
#define INCX	r36

#define I	r14
#define II	r15
#define J	r16
#define PIV	r17
#define IP	r18
#define A1	r19
#define DX1	r20
#define DY1	r21
#define A2	r22
#define DX2	r23
#define DY2	r24
#define A3	r25
#define DX3	r26
#define DY3	r27
#define DX4	r29
#define DY4	r30

#define ARLC	r31

#define XDATA1	loc0
#define XDATA2	loc1
#define XDATA3	loc2
#define XDATA4	loc3
#define XDATA5	loc4
#define XDATA6	loc5
#define XDATA7	loc6
#define XDATA8	loc7

#define YDATA1	loc8
#define YDATA2	loc9
#define YDATA3	loc10
#define YDATA4	loc11
#define YDATA5	loc12
#define YDATA6	loc13
#define YDATA7	loc14
#define YDATA8	loc15

	PROLOGUE
	.prologue
	PROFCODE

	alloc	r28 = ar.pfs, 8, 16, 0, 0
	adds	r15 = 24, SP
	adds	r16 = 32, SP
	cmp.lt	p6, p7 = 0, N

	adds	A = -2 * SIZE, A		// A--
	.save	ar.lc, ARLC
	mov	ARLC = ar.lc
	;;	
	.body
	ld8	IPIV = [r15]
	ld8	INCX = [r16]
	cmp.ge	p0, p8 = K2, K1
	(p7) br.cond.dpnt .L999
	(p8) br.cond.dpnt .L999
	;;

	shladd	LDA = LDA, ZBASE_SHIFT, r0
	shr	J = N, 2
	;;
#ifndef MINUS
	shladd	IPIV = K1, INTSHIFT, IPIV
 	;;
	adds	IPIV = -INTSIZE, IPIV		// ipiv += k1 - 1
	shladd	INCX = INCX, INTSHIFT, r0	// scaling for INCX
#else
	mov	r2 = 1
	sxt4	INCX = INCX
	;;
	shladd	INCX = INCX, INTSHIFT, r0	// scaling for INCX
	sub	r2 = r2, K2	// 1 - K2
	;;
	setf.sig f6 = r2
	setf.sig f7 = INCX
	;;
	xmpy.l	f6 = f6, f7
	;;
	getf.sig r2 = f6
	;;	
	add	IPIV = IPIV, r2
#endif
	cmp.ge	p6, p7 = 0, J
	(p6)	br.cond.dpnt .L20
	;;
	.align 32

.L12:
	mov	PIV = IPIV
	sub	II = K2, K1		// II = K1 - K2
#ifndef MINUS
	mov	I = K1
#else
	mov	I = K2
#endif
	add	A1 = A, LDA
	shladd	A2 = LDA, 1, A
	;;
	add	A3 = A2, LDA
	mov	ar.lc = II
	LDINT	IP = [PIV], INCX		// ip = *ipiv
	;;
	shladd	DX1 = I,  ZBASE_SHIFT, A		// dx = a + i
	shladd	DY1 = IP, ZBASE_SHIFT, A
	shladd	DX2 = I,  ZBASE_SHIFT, A1	// dx = a + i
	shladd	DY2 = IP, ZBASE_SHIFT, A1
	shladd	DX3 = I,  ZBASE_SHIFT, A2	// dx = a + i
	shladd	DY3 = IP, ZBASE_SHIFT, A2
	shladd	DX4 = I,  ZBASE_SHIFT, A3	// dx = a + i
	shladd	DY4 = IP, ZBASE_SHIFT, A3
	CMP4NE	p8, p0 = I, IP
	;;
	.align 32

.L15:
	(p8) LD8   XDATA1 = [DX1], SIZE
	(p8) LD8   YDATA1 = [DY1], SIZE
	(p8) LD8   XDATA2 = [DX2], SIZE
	(p8) LD8   YDATA2 = [DY2], SIZE
	(p8) LD8   XDATA3 = [DX3], SIZE
	(p8) LD8   YDATA3 = [DY3], SIZE
	(p8) LD8   XDATA4 = [DX4], SIZE
	(p8) LD8   YDATA4 = [DY4], SIZE
	;;
	(p8) LD8   XDATA5 = [DX1], -SIZE
	(p8) LD8   YDATA5 = [DY1], -SIZE
	(p8) LD8   XDATA6 = [DX2], -SIZE
	(p8) LD8   YDATA6 = [DY2], -SIZE
	(p8) LD8   XDATA7 = [DX3], -SIZE
	(p8) LD8   YDATA7 = [DY3], -SIZE
	(p8) LD8   XDATA8 = [DX4], -SIZE
	(p8) LD8   YDATA8 = [DY4], -SIZE

	LDINT	IP = [PIV], INCX		// ip = *ipiv
#ifndef MINUS
	adds	I = 1, I
#else
	adds	I =-1, I
#endif
	;;
	(p8) ST8 [DX1] = YDATA1, SIZE
	(p8) ST8 [DY1] = XDATA1, SIZE
	;;
	(p8) ST8 [DX2] = YDATA2, SIZE
	(p8) ST8 [DY2] = XDATA2, SIZE
	;;
	(p8) ST8 [DX3] = YDATA3, SIZE
	(p8) ST8 [DY3] = XDATA3, SIZE
	;;
	(p8) ST8 [DX4] = YDATA4, SIZE
	(p8) ST8 [DY4] = XDATA4, SIZE
	;;
	(p8) ST8 [DX1] = YDATA5
	shladd	DX1 = I,  ZBASE_SHIFT, A		// dx = a + i
	(p8) ST8 [DY1] = XDATA5
	shladd	DY1 = IP, ZBASE_SHIFT, A
	;;
	(p8) ST8 [DX2] = YDATA6
	shladd	DX2 = I,  ZBASE_SHIFT, A1	// dx = a + i
	(p8) ST8 [DY2] = XDATA6
	shladd	DY2 = IP, ZBASE_SHIFT, A1
	;;
	(p8) ST8 [DX3] = YDATA7
	shladd	DX3 = I,  ZBASE_SHIFT, A2	// dx = a + i
	(p8) ST8 [DY3] = XDATA7
	shladd	DY3 = IP, ZBASE_SHIFT, A2
	;;
	(p8) ST8 [DX4] = YDATA8
	shladd	DX4 = I,  ZBASE_SHIFT, A3	// dx = a + i
	(p8) ST8 [DY4] = XDATA8
	shladd	DY4 = IP, ZBASE_SHIFT, A3
	CMP4NE	p8, p0 = I, IP
	br.cloop.sptk.few .L15
	;;
	adds	J = -1, J
	shladd	A  = LDA, 2, A
	shladd	A1 = LDA, 2, A1
	shladd	A2 = LDA, 2, A2
	shladd	A3 = LDA, 2, A3
	;;
	cmp.lt	p6, p0 = 0, J
	(p6) br.cond.dptk .L12
	.align 32
	
.L20:
	and	J = 3, N
	;;
	cmp.eq	p6, p0 = 0, J
	(p6) br.cond.dpnt .L999
	;;
.L22:
	sub	II = K2, K1
#ifndef MINUS
	mov	I = K1
#else
	mov	I = K2
#endif
	mov	PIV = IPIV
	;;
	mov	ar.lc = II
	.align 32

.L25:
	LDINT IP = [PIV], INCX
	;;
	shladd DX1 = I,  ZBASE_SHIFT, A
	shladd DY1 = IP, ZBASE_SHIFT, A
	CMP4NE	p7, p0 = I, IP
#ifndef MINUS
	adds I = 1, I
#else
	adds I =-1, I
#endif
	;;
	(p7) LD8 XDATA1 = [DX1], SIZE
	(p7) LD8 YDATA1 = [DY1], SIZE
	;;
	(p7) LD8 XDATA2 = [DX1], -SIZE
	(p7) LD8 YDATA2 = [DY1], -SIZE
	;;
	(p7) ST8 [DX1] = YDATA1, SIZE
	(p7) ST8 [DY1] = XDATA1, SIZE
	;;
	(p7) ST8 [DX1] = YDATA2
	(p7) ST8 [DY1] = XDATA2
	br.cloop.sptk.few .L25

	adds J = -1, J
	add A = A, LDA
	;;
	cmp4.lt p6, p7 = 0, J
	(p6) br.cond.dptk .L22
	.align 32

.L999:
	mov	r8 = r0
	mov	ar.lc = ARLC
	br.ret.sptk.many b0
	EPILOGUE
