/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	16
#define ARGS	20
	
#define LDAS	  0 + STACK(%esp)
#define INCX4	  4 + STACK(%esp)
#define LDAS2	  8 + STACK(%esp)
#define PIV	 16 + STACK(%esp)

#define N	 4 + STACK + ARGS(%esp)
#define K1	 8 + STACK + ARGS(%esp)
#define K2	12 + STACK + ARGS(%esp)
#define ALPHA	16 + STACK + ARGS(%esp)
#ifdef DOUBLE
#define A	24 + STACK + ARGS(%esp)
#define LDA	28 + STACK + ARGS(%esp)
#define DM1	32 + STACK + ARGS(%esp)
#define DM2	36 + STACK + ARGS(%esp)
#define IPIV	40 + STACK + ARGS(%esp)
#define INCX	44 + STACK + ARGS(%esp)
#else
#define A	20 + STACK + ARGS(%esp)
#define LDA	24 + STACK + ARGS(%esp)
#define DM1	28 + STACK + ARGS(%esp)
#define DM2	32 + STACK + ARGS(%esp)
#define IPIV	36 + STACK + ARGS(%esp)
#define INCX	40 + STACK + ARGS(%esp)
#endif

	PROLOGUE

	subl	$ARGS, %esp
	pushl	%ebp
	pushl	%edi
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	A,    %edi
	subl	$SIZE,%edi		# a--
#ifndef MINUS
	movl	K1, %esi
#else
	movl	K2, %esi
#endif
	leal	-4(,%esi,4),%esi	# k1 - 1(scaled)

#ifndef MINUS
	addl	IPIV, %esi
	movl	%esi, PIV
#else
	imull	INCX, %esi

	movl	IPIV, %eax
	subl	%esi, %eax
	movl	%eax, PIV
#endif


	cmpl	$0, N
	jle	.L43

	movl	K2, %esi
	cmpl	%esi, K1
	jg	.L43

	movl	INCX, %esi
	sall	$2,   %esi
	movl	%esi, INCX4

	movl	LDA, %esi
	sall	$BASE_SHIFT, %esi
	movl	%esi, LDAS
	addl	%esi, %esi
	movl	%esi, LDAS2

	movl	N, %ebp
	sarl	$1,%ebp
	jle	.L31
	ALIGN_3

.L32:
	movl	PIV, %ebx
	movl	LDA, %esi
	sall	$BASE_SHIFT, %esi
	
#ifndef MINUS
	movl	K1,  %ecx
#else
	movl	K2,  %ecx
#endif
	ALIGN_3

.L35:
	movl	(%ebx), %eax
	leal	(%edi, %ecx, SIZE), %edx
#ifdef PENTIUM4
       prefetchnta	24 * SIZE(%edx, %esi, 1)
       prefetchnta	24 * SIZE(%eax, %esi, 1)
#endif
	addl	INCX4,  %ebx
	cmpl	%ecx,   %eax
	je	.L38

	sall	$BASE_SHIFT, %eax
	leal	(%edi, %eax, 1), %eax
	FLD	(%edx, %esi, 1)
	FLD	(%eax, %esi, 1)
	FLD	(%edx)
	FLD	(%eax)
	FST	(%edx)
	FST	(%eax)
	FST	(%edx, %esi, 1)
	FST	(%eax, %esi, 1)
	ALIGN_3

.L38:
#ifndef MINUS
	incl	%ecx
	cmpl	K2, %ecx
	jle	.L35
#else
	decl	%ecx
	cmpl	K1, %ecx
	jge	.L35
#endif

	addl	LDAS2, %edi
	decl	%ebp
	jg	.L32
	ALIGN_3

.L31:
	movl	N, %esi
	testl	$1,%esi
	je	.L43
	ALIGN_3

.L44:
	movl	PIV, %ebx
#ifndef MINUS
	movl	K1,  %ecx
#else
	movl	K2,  %ecx
#endif
	ALIGN_3

.L47:
	movl	(%ebx),%eax
	leal	(%edi,%ecx,SIZE),%edx
	addl	INCX4, %ebx
	cmpl	%ecx,%eax
	je	.L50

	leal	(%edi,%eax,SIZE),%eax
	FLD	(%edx)
	FLD	(%eax)
	FST	(%edx)
	FST	(%eax)
	ALIGN_3

.L50:
#ifndef MINUS
	incl	%ecx
	cmpl	K2, %ecx
	jle	.L47
#else
	decl	%ecx
	cmpl	K1, %ecx
	jge	.L47
#endif
	ALIGN_3

.L43:
	xorl	%eax,%eax
	popl	%ebx
	popl	%esi
	popl	%edi
	popl	%ebp
	addl	$ARGS, %esp
	ret

	EPILOGUE
