/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define ARGS	128
	
#define N	 0(%rsp)
#define K1	 8(%rsp)
#define K2	16(%rsp)
#define A	24(%rsp)
#define LDA	32(%rsp)
#define IPIV	40(%rsp)
#define INCX	48(%rsp)
#define LDAS	56(%rsp)
#define INCX4	64(%rsp)
#define LDAS2	72(%rsp)
#define PIV	80(%rsp)

	PROLOGUE
	PROFCODE
	
	pushq	%r12
	pushq	%r13
	pushq	%r14
	pushq	%r15
	pushq	%rbx
	pushq	%rbp

	subq	$ARGS, %rsp

	movq	ARG1, N
	movq	ARG2, K1
	movq	ARG3, K2
	movq	ARG4, A
	movq	ARG5, LDA

	movq	16 + 48 + ARGS(%rsp), %rax
	movq	%rax, IPIV
	movslq	24 + 48 + ARGS(%rsp), %rax
	movq	%rax, INCX

	movq	A,    %rdi
	subq	$2 * SIZE,%rdi		# a--
#ifndef MINUS
	movq	K1, %rsi
#else
	movq	K2, %rsi
#endif
	leaq	-4(,%rsi,4),%rsi	# k1 - 1(scaled)

#ifndef MINUS
	addq	IPIV, %rsi
	movq	%rsi, PIV
#else
	imulq	INCX, %rsi

	movq	IPIV, %rax
	subq	%rsi, %rax
	movq	%rax, PIV
#endif


	cmpq	$0, N
	jle	.L43

	movq	K2, %rsi
	cmpq	%rsi, K1
	jg	.L43

	movq	INCX, %rsi
	salq	$2,   %rsi
	movq	%rsi, INCX4

	movq	LDA, %rsi
	addq	%rsi, %rsi
	leaq	(, %rsi, SIZE), %rsi
	movq	%rsi, LDAS
	addq	%rsi, %rsi
	movq	%rsi, LDAS2

	movq	N, %rbp
	sarq	$1,%rbp
	jle	.L31

	.align	8

.L32:
	movq	PIV, %rbx
	movq	LDA, %rsi
	addq	%rsi, %rsi
#ifndef MINUS
	movq	K1,  %rcx
#else
	movq	K2,  %rcx
#endif
	.align	8

.L35:
	movslq	(%rbx), %rax
	leaq	(%rdi, %rcx, SIZE), %rdx
	addq	INCX4,  %rbx
	leaq	(%rdx, %rcx, SIZE), %rdx
	cmpq	%rcx,   %rax
	je	.L38

	addq	%rax, %rax
	leaq	(%rdi, %rax, SIZE), %rax
	FLD	1 * SIZE(%rdx, %rsi, SIZE)
	FLD	0 * SIZE(%rdx, %rsi, SIZE)
	FLD	1 * SIZE(%rax, %rsi, SIZE)
	FLD	0 * SIZE(%rax, %rsi, SIZE)
	FLD	1 * SIZE(%rdx)
	FLD	0 * SIZE(%rdx)
	FLD	1 * SIZE(%rax)
	FLD	0 * SIZE(%rax)
	FST	0 * SIZE(%rdx)
	FST	1 * SIZE(%rdx)
	FST	0 * SIZE(%rax)
	FST	1 * SIZE(%rax)
	FST	0 * SIZE(%rdx, %rsi, SIZE)
	FST	1 * SIZE(%rdx, %rsi, SIZE)
	FST	0 * SIZE(%rax, %rsi, SIZE)
	FST	1 * SIZE(%rax, %rsi, SIZE)
	.align 8

.L38:
#ifndef MINUS
	incq	%rcx
	cmpq	K2, %rcx
	jle	.L35
#else
	decq	%rcx
	cmpq	K1, %rcx
	jge	.L35
#endif

	addq	LDAS2, %rdi
	decq	%rbp
	jg	.L32
	.align 8

.L31:
	movq	N, %rsi
	testq	$1,%rsi
	je	.L43
	.align	8

.L44:
	movq	PIV, %rbx
#ifndef MINUS
	movq	K1,  %rcx
#else
	movq	K2,  %rcx
#endif
	.align	8

.L47:
	movslq	(%rbx),%rax
	leaq	(%rdi, %rcx, SIZE),%rdx
	addq	INCX4, %rbx
	leaq	(%rdx, %rcx, SIZE),%rdx
	cmpq	%rcx,  %rax
	je	.L50

	addq	%rax, %rax
	leaq	(%rdi,%rax, SIZE),%rax
	FLD	1 * SIZE(%rdx)
	FLD	0 * SIZE(%rdx)
	FLD	1 * SIZE(%rax)
	FLD	0 * SIZE(%rax)
	FST	0 * SIZE(%rdx)
	FST	1 * SIZE(%rdx)
	FST	0 * SIZE(%rax)
	FST	1 * SIZE(%rax)
	.align 8

.L50:
#ifndef MINUS
	incq	%rcx
	cmpq	K2, %rcx
	jle	.L47
#else
	decq	%rcx
	cmpq	K1, %rcx
	jge	.L47
#endif

	addq	LDAS, %rdi
	decq	%rbp
	jg	.L44
	.align 8

.L43:
	xorq	%rax,%rax
	addq	$ARGS, %rsp

	popq	%rbp
	popq	%rbx
	popq	%r15
	popq	%r14
	popq	%r13
	popq	%r12
	ret

	EPILOGUE

