/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 = 1.;

blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *buffer){

  long bk, i, blocking;
  int mode;

#ifndef COMPLEX
#ifdef DOUBLE
  mode  =  BLAS_DOUBLE | BLAS_REAL;
#else
  mode  =  BLAS_SINGLE | BLAS_REAL;
#endif  
#else
#ifdef DOUBLE
  mode  =  BLAS_DOUBLE | BLAS_COMPLEX;
#else
  mode  =  BLAS_SINGLE | BLAS_COMPLEX; 
#endif  
#endif

  if (n <= DTB_ENTRIES) {
    LAUU2_U(n, a, lda, buffer);
    return 0;
  }

  blocking = GEMM_Q;
  if (n <= 4 * GEMM_Q) blocking = n / 4;

  for (i = 0; i < n; i += blocking) {
    
    bk = MIN(blocking, n - i);
    
    syrk_thread(mode | BLAS_TRANSA_N | BLAS_TRANSB_T,
		i, bk, dp1, ZERO,
		a + (    i * lda) * COMPSIZE, lda, 
		NULL, 0,
		a                           , lda,
		(void *)HERK_UN,
		buffer);
    
    trsm_thread(mode | BLAS_TRANSA_T | BLAS_RSIDE, 
		i, bk, dp1, ZERO,
		a + (i + i * lda) * COMPSIZE, lda,
		a + (    i * lda) * COMPSIZE, lda, 
		(void *)TRMM_RCUN,
		buffer);
    
    CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, buffer);
  }
  
  return 0;
}
