/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dm1 = -1.;

#ifndef COMPLEX
#define TRSM_KERNEL   TRSM_KERNEL_RN
#else
#define TRSM_KERNEL   TRSM_KERNEL_RR
#undef SYRK_KERNEL_L
#ifdef XDOUBLE
#define SYRK_KERNEL_L xherk_kernel_LN
#elif defined(DOUBLE)
#define SYRK_KERNEL_L zherk_kernel_LN
#else
#define SYRK_KERNEL_L cherk_kernel_LN
#endif
#endif

#define GEMM_PQ  MAX(GEMM_P, GEMM_Q)

blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *buffer){

  long info;
  long bm, bk, i, blocking;
  long is, min_i;
  long ls, min_l;

  FLOAT *sa1, *sa2, *sb;
  sa1 = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sa2 = (FLOAT *)((((long)sa1
		    + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + GEMM_ALIGN) & ~GEMM_ALIGN)
		  + GEMM_BUFFER_A_OFFSET);
  sb  = (FLOAT *)((((long)sa2
		    + GEMM_P  * GEMM_Q * COMPSIZE * SIZE + GEMM_ALIGN) & ~GEMM_ALIGN)
		  + GEMM_BUFFER_B_OFFSET);

  FLOAT *subbuffer = (FLOAT *)((long)buffer + BUFFER_LIMIT);
#if (GEMM_UNROLL_M == GEMM_UNROLL_N)
  FLOAT *aa;
#endif

  if (n <= DTB_ENTRIES) {
    info = POTF2_L(n, a, lda, buffer);
    return info;
  }

  blocking = GEMM_Q;
  if (n <= 4 * GEMM_Q) blocking = n / 4;

  for (i = 0; i < n; i += blocking) {

    bk = n - i;
    if (bk > blocking) bk = blocking;

    info = CNAME(bk, a, lda, buffer);
    if (info) return info + i;

    bm = n - i - bk;

    if (bm > 0) {
     
      TRSM_OLTCOPY(bk, bk, a, lda, 0, sa1);

      for(ls = 0; ls < bm; ls += GEMM_R){
	  min_l = bm - ls;
	  if (min_l > GEMM_R) min_l = GEMM_R;

	  for (is = ls; is < bm; is += GEMM_P) {
	    min_i = bm - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
#if (GEMM_UNROLL_M == GEMM_UNROLL_N)
	    if (is < ls + min_l) {
	      aa = sb + bk * (is - ls) * COMPSIZE;
	    } else {
	      aa = sa2;
	    }
	    GEMM_ITCOPY(bk, min_i, a + (is + bk) * COMPSIZE, lda, aa);
#else
	    GEMM_ITCOPY(bk, min_i, a + (is + bk) * COMPSIZE, lda, sa2);
#endif

	    if (ls == 0) {
	      TRSM_KERNEL(min_i, bk, bk, dm1,
#ifdef COMPLEX
			  ZERO,
#endif

#if (GEMM_UNROLL_M == GEMM_UNROLL_N)
			  aa,
#else
			  sa2,
#endif
			  sa1,
			  a + (is + bk) * COMPSIZE, lda, 0);
	    }

#if (GEMM_UNROLL_M != GEMM_UNROLL_N)
	    if (is < ls + min_l) {
	      GEMM_OTCOPY(bk, min_i, a + (is + bk) * COMPSIZE, lda,
			  sb + bk * (is - ls) * COMPSIZE);
	    }
#endif
	    
	    SYRK_KERNEL_L(min_i, min_l, bk, dm1,
#if (GEMM_UNROLL_M == GEMM_UNROLL_N)
			  aa,
#else
			  sa2,
#endif
			  sb,
			  a + (is + bk + (ls + bk) * lda) * COMPSIZE, lda,
			  is - ls, subbuffer);
	  }
      }
    }

    a += (1 + lda) * blocking * COMPSIZE;
  }

  return 0;
}
