/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "XTRTI2"
#elif defined(DOUBLE)
#define ERROR_NAME "ZTRTI2"
#else
#define ERROR_NAME "CTRTI2"
#endif

static blasint (*trti2[])(long, FLOAT *, long, FLOAT *) ={
#ifdef XDOUBLE
  xtrti2_UU, xtrti2_UN, xtrti2_LU, xtrti2_LN,
#elif defined(DOUBLE)
  ztrti2_UU, ztrti2_UN, ztrti2_LU, ztrti2_LN,
#else
  ctrti2_UU, ctrti2_UN, ctrti2_LU, ctrti2_LN,
#endif
  };

int NAME(char *UPLO, char *DIAG, blasint *N, FLOAT *a, blasint *LDA, blasint *Info){

  blasint uplo_arg = *UPLO;
  blasint diag_arg = *DIAG;
  blasint n    = *N;
  blasint lda  = *LDA;
  blasint uplo, diag;
  blasint info;
  FLOAT *buffer;
  
  TOUPPER(uplo_arg);
  TOUPPER(diag_arg);

  uplo = -1;
  if (uplo_arg == 'U') uplo = 0;
  if (uplo_arg == 'L') uplo = 1;
  diag = -1;
  if (diag_arg == 'U') diag = 0;
  if (diag_arg == 'N') diag = 1;

  info  = 0;
  if (lda  < MAX(1,n)) info = 5;
  if (n    < 0)        info = 3; 
  if (diag < 0)        info = 2;
  if (uplo < 0)        info = 1;
  if (info) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    *Info = - info;
    return 0;
  }

  *Info = 0;

  if (n == 0) return 0;
  
#ifdef SMP
  if (blas_cpu_number == 0) blas_get_cpu_number();
#ifdef SMP_SERVER
  if (blas_server_avail == 0) blas_thread_init();

#endif
#endif

  buffer = (FLOAT *)blas_memory_alloc(0);

#ifdef USE_BUSYWAIT
  blas_busywait(1);
#endif

  info = (trti2[(uplo << 1) | diag])(n, a, lda, buffer);

  *Info = info;

#ifdef USE_BUSYWAIT
  blas_busywait(0);
#endif

  blas_memory_free(buffer);

  return 0;
}
