/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 =  1.;
static FLOAT dm1 = -1.;

#ifdef UNIT
#define TRTI2	TRTI2_LU
#else
#define TRTI2	TRTI2_LN
#endif

#define GEMM_PQ  MAX(GEMM_P, GEMM_Q)

blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *buffer){

  long i, is, min_i, start_i;
  long ls, min_l;
  long bn, bk;
  long blocking;
  FLOAT *sa_trsm1, *sa_trsm2, *sa_gemm, *sb;
  sa_trsm1 = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sa_trsm2 = (FLOAT *)((((long)sa_trsm1
		    + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
		  + GEMM_BUFFER_A_OFFSET);
  sa_gemm  = (FLOAT *)((((long)sa_trsm2
		    + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
		  + GEMM_BUFFER_A_OFFSET);
  sb  = (FLOAT *)((((long)sa_gemm
		    + GEMM_P  * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
		  + GEMM_BUFFER_B_OFFSET);
  
  if (n <= DTB_ENTRIES) {
    TRTI2(n, a, lda, buffer);
    return 0;
  }

  blocking = n / 4;
  if (blocking > GEMM_Q) blocking = GEMM_Q;

  start_i = 0;
  while (start_i < n) start_i += blocking;
  start_i -= blocking;

  for (i = start_i; i >= 0; i -= blocking) {

    bk = MIN(blocking, n - i);
    bn = n - bk - i;
    
    if (bn > 0) TRSM_OLNCOPY(bk, bk, a + (i + i * lda) * COMPSIZE, lda, 0, sa_trsm1);
	
    if (i > 0) {

      TRSM_ILTCOPY(bk, bk, a + (i + i * lda) * COMPSIZE, lda, 0, sa_trsm2);

      for (ls = 0; ls < i; ls += GEMM_R) {
	min_l = i - ls;
	if (min_l > GEMM_R) min_l = GEMM_R;
	
	GEMM_ONCOPY (bk, min_l, a + (i + ls * lda) * COMPSIZE, lda, sb);
	
	if (bn > 0) {
	  if (ls == 0) {
	    GEMM_BETA(bn, bk, 0,  dm1, 
#ifdef COMPLEX
			     ZERO,
#endif
		      NULL, 0, NULL, 0,
		      a + (i + bk + i * lda) * COMPSIZE, lda);
	  }

	  for (is = 0; is < bn; is += GEMM_P) {
	    min_i = bn - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	    GEMM_ITCOPY (bk, min_i, a + (i + bk + is + i * lda) * COMPSIZE, lda, sa_gemm);
	    
	    if (ls == 0) {
	      TRSM_KERNEL_RT(min_i, bk, bk, dm1, 
#ifdef COMPLEX
			     ZERO,
#endif
			     sa_gemm, sa_trsm1,
			     a + (i + bk + is + i * lda) * COMPSIZE, lda, 0);
	    }

	    GEMM_KERNEL_N(min_i, min_l, bk, dp1, 
#ifdef COMPLEX
			  ZERO,
#endif
			  sa_gemm, sb,
			  a + (i + bk + is + ls * lda) * COMPSIZE, lda);
	  }
	}
	
	for (is = 0; is < bk; is += GEMM_P) {
	  min_i = bk - is;
	  if (min_i > GEMM_P) min_i = GEMM_P;
	  
	  TRSM_KERNEL_LT(min_i, min_l, bk, dp1, 
#ifdef COMPLEX
			 ZERO,
#endif
			 sa_trsm2 + is * bk * COMPSIZE, sb,
			 a + (i + is + ls * lda) * COMPSIZE, lda, is);
	}
      }

      CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, buffer);

    } else {
      if (bn > 0) {
	GEMM_BETA(bn, bk, 0,  dm1, 
#ifdef COMPLEX
		  ZERO,
#endif
		  NULL, 0, NULL, 0,
		  a + (i + bk + i * lda) * COMPSIZE, lda);
	
	for (is = 0; is < bn; is += GEMM_P) {
	  min_i = bn - is;
	  if (min_i > GEMM_P) min_i = GEMM_P;
	  
	  GEMM_ITCOPY (bk, min_i, a + (i + bk + is + i * lda) * COMPSIZE, lda, sa_gemm);
	  
	  TRSM_KERNEL_RT(min_i, bk, bk, dm1, 
#ifdef COMPLEX
			 ZERO,
#endif
			 sa_gemm, sa_trsm1,
			 a + (i + bk + is + i * lda) * COMPSIZE, lda, 0);
	}
      }
      
      CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, buffer);
    }
  }

  return 0;
}
