/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 =  1.;
static FLOAT dm1 = -1.;

#ifdef ITANIUM2
#undef GEMM_Q
#define GEMM_Q GEMM_P
#endif

#ifdef UNIT
#define TRTI2	TRTI2_UU
#define TRMM	TRMM_LNUU
#define TRMM2	TRMM_RNUU
#define TRSM	TRSM_RNUU
#define TRSM2	TRSM_LNUU
#else
#define TRTI2	TRTI2_UN
#define TRMM	TRMM_LNUN
#define TRMM2	TRMM_RNUN
#define TRSM	TRSM_RNUN
#define TRSM2	TRSM_LNUN
#endif

static void inner_thread1(long m, long n, long k, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *a, long lda, FLOAT *b, long ldb,
			 FLOAT *c, long ldc, void *buffer){

  GEMM_BETA(m, n, 0, dm1,
#ifdef COMPLEX
	    ZERO,
#endif
	    NULL, 0, NULL, 0, c, ldc);

  TRSM(m, n, 0, ZERO,
#ifdef COMPLEX
	    ZERO,
#endif
       a, lda, NULL, 0, c, ldc, buffer);

}

static void inner_thread2(long m, long n, long k, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *a, long lda, FLOAT *b, long ldb,
			 FLOAT *c, long ldc, void *buffer){


  GEMM_NN(m, n, k, dp1,
#ifdef COMPLEX
	    ZERO,
#endif
	  a, lda, b,    ldb, c, ldc, buffer);

  TRSM2  (k, n, 0,
#ifdef COMPLEX
	    ZERO,
#endif
	  ZERO, a + m * COMPSIZE, lda, NULL, 0,   b, ldb, buffer);
}

blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *buffer){

  long info;
  long bk, i, blocking;
  int mode;

#ifndef COMPLEX
#ifdef DOUBLE
  mode  =  BLAS_DOUBLE | BLAS_REAL;
#else
  mode  =  BLAS_SINGLE | BLAS_REAL;
#endif  
#else
#ifdef DOUBLE
  mode  =  BLAS_DOUBLE | BLAS_COMPLEX;
#else
  mode  =  BLAS_SINGLE | BLAS_COMPLEX; 
#endif  
#endif

  if (n <= 32) {
    info = TRTI2(n, a, lda, buffer);
    return info;
  }

  blocking = GEMM_Q;
  if (n <= 4 * GEMM_Q) blocking = n / 4;

  for (i = 0; i < n; i += blocking) {
    
    bk = MIN(blocking, n - i);

    trsm_thread( mode | BLAS_RSIDE,
		 i, bk, ZERO, ZERO,
		 a + (i + i * lda) * COMPSIZE, lda, 
		 a + (    i * lda) * COMPSIZE, lda, 
		 (void *)inner_thread1,
		 buffer);
    
    gemm_thread_n( mode,
		 i, n - i - bk, bk, dp1, ZERO,
		 a + (     i       * lda) * COMPSIZE, lda,
		 a + (i + (i + bk) * lda) * COMPSIZE, lda,
		 a + (    (i + bk) * lda) * COMPSIZE, lda, 
		 (void *)inner_thread2,
		 buffer);

    CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, buffer);
  }
  
  return 0;
}
