/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"
	
#define STACK	 8
#define ARGS     8

#define V	 0 + STACK(%esp)

#define STACK_M		 4 + STACK + ARGS(%esp)
#define STACK_X		 8 + STACK + ARGS(%esp)
#define STACK_INCX	12 + STACK + ARGS(%esp)

#define I	%eax
#define M	%ecx
#define X	%esi
#define INCX	%ebx

#ifndef HAVE_SSE2
#define movsd	movlps
#define clear(a)	xorps a, a
#else
#define clear(a)
#endif

	PROLOGUE

	subl	$ARGS, %esp
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	STACK_M, M
	movl	STACK_X, X
	movl	STACK_INCX, INCX

#ifdef F_INTERFACE
	movl	(M), M
	movl	(INCX), INCX
#endif

#ifdef HAVE_SSE2
	pxor	%xmm0, %xmm0
	pxor	%xmm7, %xmm7
#else
	xorps	%xmm0, %xmm0
	xorps	%xmm7, %xmm7
#endif

	testl	M, M
	jle	.L999
	testl	INCX, INCX
	jle	.L999

#ifdef HAVE_SSE2
	pxor	%xmm1, %xmm1
	cmpeqps	%xmm7, %xmm7
	psrld	$1, %xmm7
#else
	xorps	%xmm1, %xmm1
	movl	$0x7fffffff, V
	movss	V, %xmm7
	shufps	$0, %xmm7, %xmm7
#endif

	leal	(, INCX, SIZE), INCX
	cmpl	$SIZE, INCX
	jne	.L100		# incx != 1

	testl	$3, X
	jne	.L50		# Unaligned Mode

	cmpl	$15, M
	jle	.L50

	testl	$4, X
	je	.L05

	movss	0 * SIZE(X), %xmm0
	addl	$SIZE, X
	andps	%xmm7, %xmm0
	decl	M
	jle	.L999
	ALIGN_3

	
.L05:
	testl	$8, X
	je	.L10

	subl	$2, M
	movsd	0 * SIZE(X), %xmm0
	addl	$2 * SIZE, X
	andps	%xmm7, %xmm0
	ALIGN_3

/* aligned mode */

.L10:
	movl	M,  I
	sarl	$6, I
	jle	.L15

	movaps	 0 * SIZE(X), %xmm2
	movaps	16 * SIZE(X), %xmm3
	movaps	32 * SIZE(X), %xmm4
	movaps	48 * SIZE(X), %xmm5
	decl	I
	jle	.L12
	ALIGN_4
	
#ifdef PENTIUM3
#define PREFETCHSIZE	216
#endif

#ifdef PENTIUM4
#define PREFETCHSIZE	216
#endif

#ifdef PENTIUMM
#define PREFETCHSIZE	216
#endif

#ifdef CORE2
#define PREFETCHSIZE	216
#endif

#ifdef OPTERON
#define PREFETCHSIZE	344
#endif

.L11:
#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
#endif

	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	 4 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

	movaps	 8 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	12 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE + 16) * SIZE(X)
#endif

	movaps	 64 * SIZE(X), %xmm2
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	20 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1

	movaps	24 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	28 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1

#ifdef PENTIUM4
       prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
#endif
#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE + 32) * SIZE(X)
#endif
	
	movaps	 80 * SIZE(X), %xmm3
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	36 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movaps	40 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	44 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE + 48) * SIZE(X)
#endif

	movaps	 96 * SIZE(X), %xmm4
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm0

	movaps	52 * SIZE(X), %xmm5
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm1

	movaps	56 * SIZE(X), %xmm5
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm0

	movaps	60 * SIZE(X), %xmm5
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm1

	movaps	112 * SIZE(X), %xmm5

#ifdef PENTIUM4
	prefetcht0	(PREFETCHSIZE + 32) * SIZE(X)
#endif

	addl	$64 * SIZE, X
	decl	I
	jg	.L11
	ALIGN_4

.L12:
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	 4 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

	movaps	 8 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	12 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	20 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1

	movaps	24 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	28 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1

	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	36 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movaps	40 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	44 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movaps	 96 * SIZE(X), %xmm4
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm0

	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm1

	movaps	56 * SIZE(X), %xmm5
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm0

	movaps	60 * SIZE(X), %xmm5
	andps	%xmm7, %xmm5
	addps	%xmm5, %xmm1
	addl	$64 * SIZE, X
	ALIGN_4

.L15:
	testl	$63,  M
	je	.L998

	testl	$32, M
	je	.L16

	movaps	 0 * SIZE(X), %xmm2
	movaps	16 * SIZE(X), %xmm3

	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	 4 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

	movaps	 8 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm0

	movaps	12 * SIZE(X), %xmm2
	andps	%xmm7, %xmm2
	addps	%xmm2, %xmm1

	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	20 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1

	movaps	24 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm0

	movaps	28 * SIZE(X), %xmm3
	andps	%xmm7, %xmm3
	addps	%xmm3, %xmm1
	addl	$32 * SIZE, X
	ALIGN_3

.L16:
	testl	$16, M
	je	.L17

	movaps	 0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	 4 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movaps	 8 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	12 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	addl	$16 * SIZE, X
	ALIGN_3

.L17:
	testl	$8, M
	je	.L18

	movaps	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movaps	4 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	addl	$8 * SIZE, X
	ALIGN_3

.L18:
	testl	$4, M
	je	.L19

	movaps	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0
	addl	$4 * SIZE, X
	ALIGN_3

.L19:
	testl	$2, M
	je	.L20

	clear	(%xmm4)
	movsd	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1
	addl	$2 * SIZE, X
	ALIGN_3

.L20:
	testl	$1, M
	je	.L998

	movss	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0
	jmp	.L998
	ALIGN_4


/* unaligned mode */

.L50:
	movl	M,  I
	sarl	$4, I
	jle	.L55
	ALIGN_4
	
.L51:
	movsd	 0 * SIZE(X), %xmm4
	movhps	 2 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	prefetcht0	PREFETCHSIZE * SIZE(X)

	movsd	 4 * SIZE(X), %xmm4
	movhps	 6 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movsd	 8 * SIZE(X), %xmm4
	movhps	10 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movsd	12 * SIZE(X), %xmm4
	movhps	14 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	addl	$16 * SIZE, X
	decl	I
	jg	.L51
	ALIGN_4

.L55:
	clear	(%xmm4)
	testl	$15,  M
	je	.L998

	testl	$8, M
	je	.L56

	movsd	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movsd	2 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	movsd	4 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movsd	6 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	addl	$8 * SIZE, X
	ALIGN_3

.L56:
	testl	$4, M
	je	.L57

	movsd	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0

	movsd	2 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1

	addl	$4 * SIZE, X
	ALIGN_3

.L57:
	testl	$2, M
	je	.L58

	movsd	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm1
	addl	$2 * SIZE, X
	ALIGN_3

.L58:
	testl	$1, M
	je	.L998

	movss	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0
	jmp	.L998
	ALIGN_4

/* incx != 1 */

.L100:
	movl	M,  I
	sarl	$3, I
	jle	.L105
	ALIGN_4
	
.L101:
	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm0

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm1

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm0

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm1

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm0

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm1

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm0

	movss	0 * SIZE(X), %xmm4
	addl	INCX, X
	andps	%xmm7, %xmm4
	addss	%xmm4, %xmm1

	decl	I
	jg	.L101
	ALIGN_4

.L105:
	andl	$7,  M
	jle	.L998
	ALIGN_4

.L106:
	movss	0 * SIZE(X), %xmm4
	andps	%xmm7, %xmm4
	addps	%xmm4, %xmm0
	addl	INCX, X
	decl	M
	jg	.L106
	ALIGN_4

.L998:
	addps	%xmm1, %xmm0

	movhlps	%xmm0, %xmm1
	addps	%xmm1, %xmm0
	
	movaps	%xmm0, %xmm1
	shufps  $1, %xmm0, %xmm0
	addss	 %xmm1, %xmm0
	
	ALIGN_4

.L999:
	movss	%xmm0, V
	flds	V

	popl	%ebx
	popl	%esi
	addl	$ARGS, %esp

	ret

	EPILOGUE
