/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"
	
#define STACK	 8
#define ARGS     8

#define V	 0 + STACK(%esp)

#define STACK_M		 4 + STACK + ARGS(%esp)
#define STACK_X		 8 + STACK + ARGS(%esp)
#define STACK_INCX	12 + STACK + ARGS(%esp)

#define I	%eax
#define M	%ecx
#define X	%esi
#define INCX	%ebx

#ifdef CORE2
#define PREFETCHSIZE	160
#endif

#ifdef PENTIUM4
#define PREFETCHSIZE	160
#endif

#ifdef PENTIUMM
#define PREFETCHSIZE	128
#endif

#ifdef OPTERON
#define PREFETCHSIZE	168
#endif

	PROLOGUE

	subl	$ARGS, %esp
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	STACK_M, M
	movl	STACK_X, X
	movl	STACK_INCX, INCX

#ifdef F_INTERFACE
	movl	(M), M
	movl	(INCX), INCX
#endif

	pxor	%xmm0, %xmm0
	pxor	%xmm7, %xmm7
	testl	M, M
	jle	.L999
	testl	INCX, INCX
	jle	.L999

	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

	cmpeqpd	%xmm7, %xmm7
	psrlq	$1, %xmm7

	leal	(, INCX, SIZE), INCX
	cmpl	$SIZE, INCX
	jne	.L40

	testl	$SIZE, X
	je	.L05

	movsd	0 * SIZE(X), %xmm0
	andpd	%xmm7, %xmm0
	addl	$SIZE, X
	decl	M
	jle	.L999
	ALIGN_3

.L05:
	movl	M,  I
	sarl	$5, I
	jle	.L15
	ALIGN_3
	
	movapd	0 * SIZE(X), %xmm2
	movapd	2 * SIZE(X), %xmm3
	movapd	4 * SIZE(X), %xmm4
	movapd	6 * SIZE(X), %xmm5
	decl	I
	jle	.L13
	ALIGN_3

.L12:
#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
#endif

	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0
	movapd	 8 * SIZE(X), %xmm2

	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1
	movapd	10 * SIZE(X), %xmm3

	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0
	movapd	12 * SIZE(X), %xmm4

	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1
	movapd	14 * SIZE(X), %xmm5

#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE +  8) * SIZE(X)
#endif

	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0
	movapd	16 * SIZE(X), %xmm2

	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1
	movapd	18 * SIZE(X), %xmm3

	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0
	movapd	20 * SIZE(X), %xmm4

	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1
	movapd	22 * SIZE(X), %xmm5

#ifdef PENTIUM4
       prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
#endif
#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE + 16) * SIZE(X)
#endif

	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0
	movapd	24 * SIZE(X), %xmm2

	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1
	movapd	26 * SIZE(X), %xmm3

	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0
	movapd	28 * SIZE(X), %xmm4

	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1
	movapd	30 * SIZE(X), %xmm5

#ifdef OPTERON
       prefetcht0	(PREFETCHSIZE + 24) * SIZE(X)
#endif

	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0
	movapd	32 * SIZE(X), %xmm2

	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1
	movapd	34 * SIZE(X), %xmm3

	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0
	movapd	36 * SIZE(X), %xmm4

	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1
	movapd	38 * SIZE(X), %xmm5

#ifdef PENTIUM4
	prefetcht0	(PREFETCHSIZE + 16) * SIZE(X)
#endif

	addl	$32 * SIZE, X
	subl	$1, I
	jg	.L12
	ALIGN_3

.L13:
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	movapd	 8 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	10 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	12 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	14 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	movapd	16 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	18 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	20 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	22 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	movapd	24 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	26 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	28 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	30 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	addl	$32 * SIZE, X
	ALIGN_3

.L15:
	andl	$31,  M
	jle	.L998

	testl	$16, M
	je	.L16

	movapd	0 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	2 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	4 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	6 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	movapd	8 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	10 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	12 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	14 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	addl	$16 * SIZE, X
	ALIGN_3

.L16:
	testl	$8, M
	je	.L17

	movapd	0 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	2 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1

	movapd	4 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movapd	6 * SIZE(X), %xmm5
	andpd	%xmm7, %xmm5
	addpd	%xmm5, %xmm1

	addl	$8 * SIZE, X
	ALIGN_3

.L17:
	testl	$4, M
	je	.L18

	movapd	0 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0

	movapd	2 * SIZE(X), %xmm3
	andpd	%xmm7, %xmm3
	addpd	%xmm3, %xmm1
	addl	$4 * SIZE, X
	ALIGN_3	

.L18:
	testl	$2, M
	je	.L19

	movapd	0 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm0
	addl	$2 * SIZE, X
	
.L19:
	testl	$1, M
	je	.L998

	MOVSD	0 * SIZE(X), %xmm2
	andpd	%xmm7, %xmm2
	addpd	%xmm2, %xmm1
	jmp	.L998
	ALIGN_3


.L40:
	movl	M,  I
	sarl	$3, I
	jle	.L60
	ALIGN_3
	
.L50:
	movsd	0 * SIZE(X), %xmm4
	addl	INCX, X
	movhpd	0 * SIZE(X), %xmm4
	addl	INCX, X
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movsd	0 * SIZE(X), %xmm4
	addl	INCX, X
	movhpd	0 * SIZE(X), %xmm4
	addl	INCX, X
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm1

	movsd	0 * SIZE(X), %xmm4
	addl	INCX, X
	movhpd	0 * SIZE(X), %xmm4
	addl	INCX, X
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0

	movsd	0 * SIZE(X), %xmm4
	addl	INCX, X
	movhpd	0 * SIZE(X), %xmm4
	addl	INCX, X
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm1

	decl	I
	jg	.L50
	ALIGN_3

.L60:
	andl	$7,  M
	jle	.L998
	ALIGN_3

.L61:
	movsd	0 * SIZE(X), %xmm4
	andpd	%xmm7, %xmm4
	addpd	%xmm4, %xmm0
	addl	INCX, X
	decl	M
	jg	.L61
	ALIGN_3

.L998:
	addpd	%xmm1, %xmm0
	movapd	%xmm0, %xmm1
	unpckhpd %xmm0, %xmm0
	addsd	 %xmm1, %xmm0
	ALIGN_3

.L999:
	movsd	%xmm0, V
	fldl	V
	popl	%ebx
	popl	%esi
	addl	$ARGS, %esp

	ret

	EPILOGUE

