/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#ifdef DOUBLE
#define PREFETCH_SIZE (32 * 8)
#else
#define PREFETCH_SIZE (64 * 8)
#endif

#define SP	r12

#define N	r32
#define X	r37
#define INCX	r38
#define Y	r39
#define INCY	r36

#define PRE1	r2
#define PRE2	r3

#define I	r14
#define J	r15
#define Y1	r16
#define Y2	r17
#define X2	r18
#define XX	r19
#define YY	r20
#define INCXM1	r21
#define INCYM1	r22
#define INCX3M1	r23
#define INCY3M1	r24
#define INCX8	r25
#define INCY8	r26
#define PR	r30
#define ARLC	r31

#define ALPHA_R	f8
#define ALPHA_I	f9

	PROLOGUE
	.prologue
	PROFCODE
	{ .mfi
	adds	r14 = 16, SP
	fcmp.eq	p13, p0 = f8, f0
	.save ar.lc, ARLC
	mov	ARLC = ar.lc
	}
	{ .mib
	cmp.gt	p15, p0 = r0, N
	(p15) br.ret.sptk.many b0
	}
	;;
	{ .mmi
	ld8	INCY   = [r14]
	shladd	INCX = INCX, ZBASE_SHIFT, r0
	mov	PR = pr
	}
	{ .mmi
	mov	Y1 = Y
	}
	;;
	.body
	{ .mmi
	shladd	INCY = INCY, ZBASE_SHIFT, r0
	shladd	INCX3M1 = INCX,   1, INCX
	mov	pr.rot= 0
	}
	{ .mmi
	shladd	INCX8  = INCX, 3, r0
	shr	I =  N, 3
	}
	;;
	{ .mmi
	shladd	INCY3M1 = INCY,   1, INCY
	shladd	INCY8  = INCY, 3, r0
	}
	{ .mmi
	adds	INCXM1  = -SIZE, INCX
	adds	INCYM1  = -SIZE, INCY
	and	J =  7, N
	}
	;;
	{ .mmi
	adds	INCX3M1 = -SIZE, INCX3M1
	adds	INCY3M1 = -SIZE, INCY3M1
	adds	I = -1, I
	}
	{ .mmi
	cmp.eq	p12, p0 = r0, J
	cmp.eq	p16, p0 = r0, r0
	}
	;;
	sub	XX = Y, X
	;;
	shr.u  XX = XX, BASE_SHIFT
	;;
#ifdef DOUBLE
	adds	XX =  2, XX
#else
	adds	YY = -44, XX
	adds	XX = -20, XX
#endif
	;;
#ifdef DOUBLE
	and	XX = 31, XX
#else
	and	XX = 63, XX
	and	YY = 63, YY
#endif
	;;
#ifdef DOUBLE
	cmp.le	p11, p0 = 14, XX
#else
	cmp.gt	p11, p0 = 16, XX
	cmp.gt	p10, p0 = 16, YY
#endif
	;;
	{ .mfi
	(p13) fcmp.eq.unc p14, p0 = f9, f0
	mov	ar.ec = 4
	}
	{ .mbb
#ifndef DOUBLE
	(p10) br.cond.dpnt .L20
#endif
	(p11) br.cond.dpnt .L20
	}
	;;
	{ .mmi
#ifdef DOUBLE
	adds	PRE1 = (PREFETCH_SIZE + 28) * SIZE, X
	adds	PRE2 = (PREFETCH_SIZE + 24) * SIZE, Y
#else
	adds	PRE1 = (PREFETCH_SIZE -  8) * SIZE, X
	adds	PRE2 = (PREFETCH_SIZE + 20) * SIZE, Y
#endif
	mov	ar.lc = I
	}
	{ .mbb
	cmp.eq	p11 ,p0 = -1, I
	(p13) br.cond.dptk .L999
	(p11) br.cond.dpnt  .L15
	}
	;;
	.align 32

.L12:
	{ .mmf
	(p19) STFD	[Y1] = f11
	(p17) LDFD	f33  = [X], 1 * SIZE
	(p19) FMA	f15 = ALPHA_I, f77, f15
	}
	{ .mmf
	(p16) LDFD	f80  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_R, f43, f91
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f12,  1 * SIZE
	(p16) lfetch.nt1 [PRE1], INCX8
	(p18) FMA	f12 = ALPHA_R, f46, f94
	}
	{ .mmf
	(p16) LDFD	f83  = [Y], INCYM1
	(p17) LDFD	f36  = [X], INCXM1
	(p18) FNMA	f6  = ALPHA_I, f37, f6
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f13
	(p17) LDFD	f39  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_R, f49, f97
	}
	{ .mmf
	(p16) LDFD	f86  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_I, f34, f7
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f14,  1 * SIZE
	(p16) lfetch.excl.nt1 [PRE2], INCY8
	(p18) FMA	f14 = ALPHA_R, f52, f100
	}
	{ .mmf
	(p16) LDFD	f89  = [Y], INCYM1
	(p17) LDFD	f42  = [X], INCXM1
	(p18) FNMA	f10 = ALPHA_I, f43, f10
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f15
	(p17) LDFD	f45  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_R, f55, f103
	}
	{ .mmf
	(p16) LDFD	f92  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_I, f40, f11
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f6,   1 * SIZE
	(p17) LDFD	f48  = [X], INCXM1
	(p18) FNMA	f12 = ALPHA_I, f49, f12
	}
	{ .mmf
	(p16) LDFD	f95  = [Y], INCYM1
	nop.m 0
	(p18) FMA	f6  = ALPHA_R, f58, f106
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f7
	(p17) LDFD	f51  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_I, f46, f13
	}
	{ .mmf
	(p16) LDFD	f98  = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_R, f61, f109
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f10,  1 * SIZE
	(p17) LDFD	f54  = [X], INCXM1
	(p18) FNMA	f14 = ALPHA_I, f55, f14
	}
	{ .mmf
	(p16) LDFD	f101 = [Y], INCYM1
	nop.m 0
	(p18) FMA	f10 = ALPHA_R, f64, f112
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f11
	(p17) LDFD	f57  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_I, f52, f15
	}
	{ .mmf
	(p16) LDFD	f104 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_R, f67, f115
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f12,  1 * SIZE
	(p17) LDFD	f60  = [X], INCXM1
	(p18) FMA	f12 = ALPHA_R, f70, f118
	}
	{ .mmf
	(p16) LDFD	f107 = [Y], INCYM1
	nop.m 0
	(p18) FNMA	f6  = ALPHA_I, f61, f6
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f13
	(p17) LDFD	f63  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_R, f73, f121
	}
	{ .mmf
	(p16) LDFD	f110 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_I, f58, f7
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f14,  1 * SIZE
	(p17) LDFD	f66  = [X], INCXM1
	(p18) FMA	f14 = ALPHA_R, f76, f124
	}
	{ .mmf
	(p16) LDFD	f113 = [Y], INCYM1
	nop.m 0
	(p18) FNMA	f10 = ALPHA_I, f67, f10
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f15
	(p17) LDFD	f69  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_R, f79, f127
	}
	{ .mmf
	(p16) LDFD	f116 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_I, f64, f11
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f6,   1 * SIZE
	(p17) LDFD	f72  = [X], INCXM1
	(p18) FNMA	f12 = ALPHA_I, f73, f12
	}
	{ .mmf
	(p16) LDFD	f119 = [Y], INCYM1
	nop.m 0
	(p17) FMA	f6  = ALPHA_R, f33, f81
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f7
	(p17) LDFD	f75  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_I, f70, f13
	}
	{ .mmf
	(p16) LDFD	f122 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p17) FMA	f7  = ALPHA_R, f36, f84
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f10,  1 * SIZE
	(p17) LDFD	f78  = [X], INCXM1
	(p18) FNMA	f14 = ALPHA_I, f79, f14
	}
	{ .mfb
	(p16) LDFD	f125 = [Y], INCYM1
	(p17) FMA	f10 = ALPHA_R, f39, f87
	br.ctop.sptk.few .L12
	}
	;;
	.align 32

.L15:
	{ .mmi
	mov	XX = X
	mov	YY = Y1
	mov	ar.lc = ARLC
	}
	{ .mmi
	shladd	X2 = INCX, 1, X
	shladd	Y2 = INCY, 1, Y1
	tbit.z	p0, p13 = N, 2
	}
	;;
	{ .mmi
	(p13) LDFD	f32 = [X],  1 * SIZE
	(p13) LDFD	f80 = [Y],  1 * SIZE
	mov	pr = PR, -65474
	}
	{ .mmb
	(p13) LDFD	f36 = [X2], 1 * SIZE
	(p13) LDFD	f84 = [Y2], 1 * SIZE
	(p12) br.ret.sptk.many b0
	}
	;;
	{ .mmi
	(p13) LDFD	f33 = [X],  INCXM1
	(p13) LDFD	f81 = [Y],  INCYM1
	tbit.z	p0, p14 = N, 1
	}
	{ .mmi
	(p13) LDFD	f37 = [X2], INCXM1
	(p13) LDFD	f85 = [Y2], INCYM1
	tbit.z	p0, p15 = N, 0
	}
	;;
	{ .mmi
	(p13) LDFD	f34 = [X],  1 * SIZE
	(p13) LDFD	f82 = [Y],  1 * SIZE
	(p13) shladd XX = INCX, 2, XX
	}
	{ .mmi
	(p13) LDFD	f38 = [X2], 1 * SIZE
	(p13) LDFD	f86 = [Y2], 1 * SIZE
	(p13) shladd YY = INCY, 2, YY
	}
	;;
	{ .mmi
	(p13) LDFD	f35 = [X],  INCX3M1
	(p13) LDFD	f83 = [Y],  INCY3M1
	(p14) shladd XX = INCX, 1, XX
	}
	{ .mmi
	(p13) LDFD	f39 = [X2], INCX3M1
	(p13) LDFD	f87 = [Y2], INCY3M1
	(p14) shladd YY = INCY, 1, YY
	}
	;;
	(p14) LDFD	f40 = [X],  1 * SIZE
	(p14) LDFD	f88 = [Y],  1 * SIZE
	(p15) LDFD	f44 = [XX], 1 * SIZE
	(p15) LDFD	f92 = [YY], 1 * SIZE
	;;
	(p14) LDFD	f41 = [X], INCXM1
	(p14) LDFD	f89 = [Y], INCYM1
	(p15) LDFD	f45 = [XX]
	(p15) LDFD	f93 = [YY]
	;;
	(p14) LDFD	f42 = [X], 1 * SIZE
	(p13) FMA	f100 = ALPHA_R, f32, f80
	(p14) LDFD	f90 = [Y], 1 * SIZE
	(p13) FMA	f104 = ALPHA_R, f36, f84
	;;
	{ .mmf
	(p14) LDFD	f43 = [X]
	(p13) FMA	f101 = ALPHA_R, f33, f81
	}
	{ .mmf
	(p14) LDFD	f91 = [Y]
	mov	YY = Y1
	(p13) FMA	f105 = ALPHA_R, f37, f85
	}

	(p13) FMA	f102 = ALPHA_R, f34, f82
	(p13) FMA	f106 = ALPHA_R, f38, f86
	(p13) FMA	f103 = ALPHA_R, f35, f83
	(p13) FMA	f107 = ALPHA_R, f39, f87
	;;
	(p14) FMA	f108 = ALPHA_R, f40, f88
	(p14) FMA	f109 = ALPHA_R, f41, f89
	(p14) FMA	f110 = ALPHA_R, f42, f90
	(p14) FMA	f111 = ALPHA_R, f43, f91
	(p15) FMA	f112 = ALPHA_R, f44, f92
	(p15) FMA	f113 = ALPHA_R, f45, f93
	;;
	(p13) FNMA	f100 = ALPHA_I, f33, f100
	(p13) FNMA	f104 = ALPHA_I, f37, f104
	(p13) FMA	f101 = ALPHA_I, f32, f101
	(p13) FMA	f105 = ALPHA_I, f36, f105
	;;
	(p13) FNMA	f102 = ALPHA_I, f35, f102
	(p13) FNMA	f106 = ALPHA_I, f39, f106
	(p13) shladd YY = INCY, 2, YY
	;;
	(p14) shladd YY = INCY, 1, YY
	(p13) FMA	f103 = ALPHA_I, f34, f103
	shladd	Y2 = INCY,  1, Y1
	(p13) FMA	f107 = ALPHA_I, f38, f107
	;;
	(p13) STFD	[Y1] = f100,   1 * SIZE
	(p14) FNMA	f108 = ALPHA_I, f41, f108
	(p13) STFD	[Y2] = f104,   1 * SIZE
	(p15) FNMA	f112 = ALPHA_I, f45, f112
	;;
	(p13) STFD	[Y1] = f101
	(p13) add	Y1 = Y1, INCYM1
	(p14) FMA	f109 = ALPHA_I, f40, f109
	(p13) STFD	[Y2] = f105
	(p13) add	Y2 = Y2, INCYM1
	(p15) FMA	f113 = ALPHA_I, f44, f113
	;;
	(p13) STFD	[Y1] = f102,   1 * SIZE
	(p13) STFD	[Y2] = f106,   1 * SIZE
	(p14) FNMA	f110 = ALPHA_I, f43, f110
	;;
	(p13) STFD	[Y1] = f103
	(p13) add	Y1 = Y1, INCY3M1
	(p13) STFD	[Y2] = f107
	(p13) add	Y2 = Y2, INCY3M1
	(p14) FMA	f111 = ALPHA_I, f42, f111
	;;
	(p14) STFD	[Y1] = f108,   1 * SIZE
	(p15) STFD	[YY] = f112,   1 * SIZE
	;;
	(p14) STFD	[Y1] = f109
	(p15) STFD	[YY] = f113
	(p14) add	Y1 = Y1, INCYM1
	;;
	(p14) STFD	[Y1] = f110,   1 * SIZE
	;;
	(p14) STFD	[Y1] = f111
	br.ret.sptk.many b0
	;;
	.align 32

.L20:
	{ .mmi
#ifdef DOUBLE
	adds	PRE1 = (PREFETCH_SIZE +  4) * SIZE, X
	adds	PRE2 = (PREFETCH_SIZE -  6) * SIZE, Y
#else
	adds	PRE1 = (PREFETCH_SIZE + 40) * SIZE, X
	adds	PRE2 = (PREFETCH_SIZE -  4) * SIZE, Y
#endif
	mov	ar.lc = I
	}
	{ .mbb
	cmp.eq	p11 ,p0 = -1, I
	(p13) br.cond.dptk .L999
	(p11) br.cond.dpnt  .L25
	}
	;;
	.align 32

.L22:
	{ .mmf
	(p19) STFD	[Y1] = f11
	(p16) LDFD	f32  = [X], 1 * SIZE
	(p19) FMA	f15 = ALPHA_I, f77, f15
	}
	{ .mmf
	(p16) LDFD	f80  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_R, f43, f91
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f12,  1 * SIZE
	(p16) lfetch.nt1 [PRE1], INCX8
	(p18) FMA	f12 = ALPHA_R, f46, f94
	}
	{ .mmf
	(p16) LDFD	f35  = [X], INCXM1
	(p16) LDFD	f83  = [Y], INCYM1
	(p18) FNMA	f6  = ALPHA_I, f37, f6
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f13
	(p16) LDFD	f38  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_R, f49, f97
	}
	{ .mmf
	(p16) LDFD	f86  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_I, f34, f7
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f14,  1 * SIZE
	(p16) lfetch.excl.nt1 [PRE2], INCY8
	(p18) FMA	f14 = ALPHA_R, f52, f100
	}
	{ .mmf
	(p16) LDFD	f41  = [X], INCXM1
	(p16) LDFD	f89  = [Y], INCYM1
	(p18) FNMA	f10 = ALPHA_I, f43, f10
	}
	;;
	{ .mmf
	(p19) STFD	[Y1] = f15
	(p16) LDFD	f44  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_R, f55, f103
	}
	{ .mmf
	(p16) LDFD	f92  = [Y], 1 * SIZE
	(p19) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_I, f40, f11
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f6,   1 * SIZE
	(p16) LDFD	f47  = [X], INCXM1
	(p18) FNMA	f12 = ALPHA_I, f49, f12
	}
	{ .mmf
	(p16) LDFD	f95  = [Y], INCYM1
	nop.m 0
	(p18) FMA	f6  = ALPHA_R, f58, f106
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f7
	(p16) LDFD	f50  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_I, f46, f13
	}
	{ .mmf
	(p16) LDFD	f98  = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_R, f61, f109
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f10,  1 * SIZE
	(p16) LDFD	f53  = [X], INCXM1
	(p18) FNMA	f14 = ALPHA_I, f55, f14
	}
	{ .mmf
	(p16) LDFD	f101 = [Y], INCYM1
	nop.m 0
	(p18) FMA	f10 = ALPHA_R, f64, f112
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f11
	(p16) LDFD	f56  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_I, f52, f15
	}
	{ .mmf
	(p16) LDFD	f104 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_R, f67, f115
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f12,  1 * SIZE
	(p16) LDFD	f59  = [X], INCXM1
	(p18) FMA	f12 = ALPHA_R, f70, f118
	}
	{ .mmf
	(p16) LDFD	f107 = [Y], INCYM1
	nop.m 0
	(p18) FNMA	f6  = ALPHA_I, f61, f6
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f13
	(p16) LDFD	f62  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_R, f73, f121
	}
	{ .mmf
	(p16) LDFD	f110 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f7  = ALPHA_I, f58, f7
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f14,  1 * SIZE
	(p16) LDFD	f65  = [X], INCXM1
	(p18) FMA	f14 = ALPHA_R, f76, f124
	}
	{ .mmf
	(p16) LDFD	f113 = [Y], INCYM1
	nop.m 0
	(p18) FNMA	f10 = ALPHA_I, f67, f10
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f15
	(p16) LDFD	f68  = [X], 1 * SIZE
	(p18) FMA	f15 = ALPHA_R, f79, f127
	}
	{ .mmf
	(p16) LDFD	f116 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p18) FMA	f11 = ALPHA_I, f64, f11
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f6,   1 * SIZE
	(p16) LDFD	f71  = [X], INCXM1
	(p18) FNMA	f12 = ALPHA_I, f73, f12
	}
	{ .mmf
	(p16) LDFD	f119 = [Y], INCYM1
	nop.m 0
	(p17) FMA	f6  = ALPHA_R, f33, f81
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f7
	(p16) LDFD	f74  = [X], 1 * SIZE
	(p18) FMA	f13 = ALPHA_I, f70, f13
	}
	{ .mmf
	(p16) LDFD	f122 = [Y], 1 * SIZE
	(p18) add	Y1 = Y1, INCYM1
	(p17) FMA	f7  = ALPHA_R, f36, f84
	}
	;;
	{ .mmf
	(p18) STFD	[Y1] = f10,  1 * SIZE
	(p16) LDFD	f77  = [X], INCXM1
	(p18) FNMA	f14 = ALPHA_I, f79, f14
	}
	{ .mfb
	(p16) LDFD	f125 = [Y], INCYM1
	(p17) FMA	f10 = ALPHA_R, f39, f87
	br.ctop.sptk.few .L22
	}
	;;
	.align 32

.L25:
	{ .mmi
	mov	XX = X
	mov	YY = Y1
	mov	ar.lc = ARLC
	}
	{ .mmi
	shladd	X2 = INCX, 1, X
	shladd	Y2 = INCY, 1, Y1
	tbit.z	p0, p13 = N, 2
	}
	;;
	{ .mmi
	(p13) LDFD	f32 = [X],  1 * SIZE
	(p13) LDFD	f80 = [Y],  1 * SIZE
	mov	pr = PR, -65474
	}
	{ .mmb
	(p13) LDFD	f36 = [X2], 1 * SIZE
	(p13) LDFD	f84 = [Y2], 1 * SIZE
	(p12) br.ret.sptk.many b0
	}
	;;
	{ .mmi
	(p13) LDFD	f33 = [X],  INCXM1
	(p13) LDFD	f81 = [Y],  INCYM1
	tbit.z	p0, p14 = N, 1
	}
	{ .mmi
	(p13) LDFD	f37 = [X2], INCXM1
	(p13) LDFD	f85 = [Y2], INCYM1
	tbit.z	p0, p15 = N, 0
	}
	;;
	{ .mmi
	(p13) LDFD	f34 = [X],  1 * SIZE
	(p13) LDFD	f82 = [Y],  1 * SIZE
	(p13) shladd XX = INCX, 2, XX
	}
	{ .mmi
	(p13) LDFD	f38 = [X2], 1 * SIZE
	(p13) LDFD	f86 = [Y2], 1 * SIZE
	(p13) shladd YY = INCY, 2, YY
	}
	;;
	{ .mmi
	(p13) LDFD	f35 = [X],  INCX3M1
	(p13) LDFD	f83 = [Y],  INCY3M1
	(p14) shladd XX = INCX, 1, XX
	}
	{ .mmi
	(p13) LDFD	f39 = [X2], INCX3M1
	(p13) LDFD	f87 = [Y2], INCY3M1
	(p14) shladd YY = INCY, 1, YY
	}
	;;
	(p14) LDFD	f40 = [X],  1 * SIZE
	(p14) LDFD	f88 = [Y],  1 * SIZE
	(p15) LDFD	f44 = [XX], 1 * SIZE
	(p15) LDFD	f92 = [YY], 1 * SIZE
	;;
	(p14) LDFD	f41 = [X], INCXM1
	(p14) LDFD	f89 = [Y], INCYM1
	(p15) LDFD	f45 = [XX]
	(p15) LDFD	f93 = [YY]
	;;
	(p14) LDFD	f42 = [X], 1 * SIZE
	(p13) FMA	f100 = ALPHA_R, f32, f80
	(p14) LDFD	f90 = [Y], 1 * SIZE
	(p13) FMA	f104 = ALPHA_R, f36, f84
	;;
	{ .mmf
	(p14) LDFD	f43 = [X]
	(p13) FMA	f101 = ALPHA_R, f33, f81
	}
	{ .mmf
	(p14) LDFD	f91 = [Y]
	mov	YY = Y1
	(p13) FMA	f105 = ALPHA_R, f37, f85
	}

	(p13) FMA	f102 = ALPHA_R, f34, f82
	(p13) FMA	f106 = ALPHA_R, f38, f86
	(p13) FMA	f103 = ALPHA_R, f35, f83
	(p13) FMA	f107 = ALPHA_R, f39, f87
	;;
	(p14) FMA	f108 = ALPHA_R, f40, f88
	(p14) FMA	f109 = ALPHA_R, f41, f89
	(p14) FMA	f110 = ALPHA_R, f42, f90
	(p14) FMA	f111 = ALPHA_R, f43, f91
	(p15) FMA	f112 = ALPHA_R, f44, f92
	(p15) FMA	f113 = ALPHA_R, f45, f93
	;;
	(p13) FNMA	f100 = ALPHA_I, f33, f100
	(p13) FNMA	f104 = ALPHA_I, f37, f104
	(p13) FMA	f101 = ALPHA_I, f32, f101
	(p13) FMA	f105 = ALPHA_I, f36, f105
	;;
	(p13) FNMA	f102 = ALPHA_I, f35, f102
	(p13) FNMA	f106 = ALPHA_I, f39, f106
	(p13) shladd YY = INCY, 2, YY
	;;
	(p14) shladd YY = INCY, 1, YY
	(p13) FMA	f103 = ALPHA_I, f34, f103
	shladd	Y2 = INCY,  1, Y1
	(p13) FMA	f107 = ALPHA_I, f38, f107
	;;
	(p13) STFD	[Y1] = f100,   1 * SIZE
	(p14) FNMA	f108 = ALPHA_I, f41, f108
	(p13) STFD	[Y2] = f104,   1 * SIZE
	(p15) FNMA	f112 = ALPHA_I, f45, f112
	;;
	(p13) STFD	[Y1] = f101
	(p13) add	Y1 = Y1, INCYM1
	(p14) FMA	f109 = ALPHA_I, f40, f109
	(p13) STFD	[Y2] = f105
	(p13) add	Y2 = Y2, INCYM1
	(p15) FMA	f113 = ALPHA_I, f44, f113
	;;
	(p13) STFD	[Y1] = f102,   1 * SIZE
	(p13) STFD	[Y2] = f106,   1 * SIZE
	(p14) FNMA	f110 = ALPHA_I, f43, f110
	;;
	(p13) STFD	[Y1] = f103
	(p13) add	Y1 = Y1, INCY3M1
	(p13) STFD	[Y2] = f107
	(p13) add	Y2 = Y2, INCY3M1
	(p14) FMA	f111 = ALPHA_I, f42, f111
	;;
	(p14) STFD	[Y1] = f108,   1 * SIZE
	(p15) STFD	[YY] = f112,   1 * SIZE
	;;
	(p14) STFD	[Y1] = f109
	(p15) STFD	[YY] = f113
	(p14) add	Y1 = Y1, INCYM1
	;;
	(p14) STFD	[Y1] = f110,   1 * SIZE
	;;
	(p14) STFD	[Y1] = f111
	br.ret.sptk.many b0
	;;
	.align 32

.L999:
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	EPILOGUE
