/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define M	ARG1	/* rdi */
#define X	ARG4	/* rsi */
#define INCX	ARG5	/* rdx */
#define Y	ARG6	/* rcx */
#define INCY	ARG2	/* r8  */

#define	YY	%r10
#define ALPHA	%xmm15

#define PREFETCH	prefetch
#define PREFETCHW	prefetchw
#define PREFETCHSIZE	64

	PROLOGUE
	PROFCODE

	movq	8(%rsp), INCY

	movapd	%xmm0,  ALPHA
	unpcklpd ALPHA, ALPHA

	leaq	(, INCX, SIZE), INCX
	leaq	(, INCY, SIZE), INCY

	cmpq	$SIZE, INCX
	jne	.L40
	cmpq	$SIZE, INCY
	jne	.L40

	subq	$-16 * SIZE, X
	subq	$-16 * SIZE, Y

	testq	$SIZE, Y
	je	.L10

	movsd	-16 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	-16 * SIZE(Y), %xmm0
	movlpd	%xmm0, -16 * SIZE(Y)
	addq	$1 * SIZE, X
	addq	$1 * SIZE, Y
	decq	M
	jle	.L19
	ALIGN_3

.L10:
	testq	$SIZE, X
	jne	.L20

	movq	M,  %rax
	sarq	$4, %rax
	jle	.L13

	movapd	-16 * SIZE(X), %xmm0
	movapd	-14 * SIZE(X), %xmm1
	movapd	-12 * SIZE(X), %xmm2
	movapd	-10 * SIZE(X), %xmm3
	movapd	 -8 * SIZE(X), %xmm8
	movapd	 -6 * SIZE(X), %xmm9
	movapd	 -4 * SIZE(X), %xmm10
	movapd	 -2 * SIZE(X), %xmm11

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	decq	%rax
	jle .L12
	ALIGN_3

.L11:
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)

	addpd	-16 * SIZE(Y), %xmm0
	mulpd	ALPHA, %xmm8
	addpd	-14 * SIZE(Y), %xmm1
	mulpd	ALPHA, %xmm9
	addpd	-12 * SIZE(Y), %xmm2
	mulpd	ALPHA, %xmm10
	addpd	-10 * SIZE(Y), %xmm3
	mulpd	ALPHA, %xmm11

	PREFETCHW	(PREFETCHSIZE + 0) * SIZE(Y)

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	movapd	 0 * SIZE(X), %xmm0
	movapd	 2 * SIZE(X), %xmm1
	movapd	 4 * SIZE(X), %xmm2
	movapd	 6 * SIZE(X), %xmm3

	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)

	addpd	-8 * SIZE(Y), %xmm8
	mulpd	ALPHA, %xmm0
	addpd	-6 * SIZE(Y), %xmm9
	mulpd	ALPHA, %xmm1
	addpd	-4 * SIZE(Y), %xmm10
	mulpd	ALPHA, %xmm2
	addpd	-2 * SIZE(Y), %xmm11
	mulpd	ALPHA, %xmm3

	PREFETCHW	(PREFETCHSIZE + 8) * SIZE(Y)

	movapd	%xmm8,  -8 * SIZE(Y)
	movapd	%xmm9,  -6 * SIZE(Y)
	movapd	%xmm10, -4 * SIZE(Y)
	movapd	%xmm11, -2 * SIZE(Y)

	movapd	 8 * SIZE(X), %xmm8
	movapd	10 * SIZE(X), %xmm9
	movapd	12 * SIZE(X), %xmm10
	movapd	14 * SIZE(X), %xmm11

	subq	$-16 * SIZE, X
	subq	$-16 * SIZE, Y
	decq	%rax
	jg	.L11
	ALIGN_3

.L12:
	addpd	-16 * SIZE(Y), %xmm0
	mulpd	ALPHA, %xmm8
	addpd	-14 * SIZE(Y), %xmm1
	mulpd	ALPHA, %xmm9
	addpd	-12 * SIZE(Y), %xmm2
	mulpd	ALPHA, %xmm10
	addpd	-10 * SIZE(Y), %xmm3
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	addpd	-8 * SIZE(Y), %xmm8
	addpd	-6 * SIZE(Y), %xmm9
	addpd	-4 * SIZE(Y), %xmm10
	addpd	-2 * SIZE(Y), %xmm11

	movapd	%xmm8,  -8 * SIZE(Y)
	movapd	%xmm9,  -6 * SIZE(Y)
	movapd	%xmm10, -4 * SIZE(Y)
	movapd	%xmm11, -2 * SIZE(Y)

	subq	$-16 * SIZE, X
	subq	$-16 * SIZE, Y
	ALIGN_3

.L13:
	testq	$15, M
	jle	.L19

	testq	$8, M
	jle	.L14

	movapd	-16 * SIZE(X), %xmm0
	movapd	-14 * SIZE(X), %xmm1
	movapd	-12 * SIZE(X), %xmm2
	movapd	-10 * SIZE(X), %xmm3

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	addpd	-16 * SIZE(Y), %xmm0
	addpd	-14 * SIZE(Y), %xmm1
	addpd	-12 * SIZE(Y), %xmm2
	addpd	-10 * SIZE(Y), %xmm3

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	ALIGN_3

.L14:
	testq	$4, M
	jle	.L15

	movapd	-16 * SIZE(X), %xmm0
	movapd	-14 * SIZE(X), %xmm1

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1

	addpd	-16 * SIZE(Y), %xmm0
	addpd	-14 * SIZE(Y), %xmm1

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	ALIGN_3

.L15:
	testq	$2, M
	jle	.L16

	movapd	-16 * SIZE(X), %xmm0

	mulpd	ALPHA, %xmm0

	addpd	-16 * SIZE(Y), %xmm0

	movapd	%xmm0, -16 * SIZE(Y)

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	ALIGN_3

.L16:
	testq	$1, M
	jle	.L19

	movlpd	-16 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	-16 * SIZE(Y), %xmm0

	movlpd	%xmm0, 	-16 * SIZE(Y)
	addq	$SIZE, Y
	ALIGN_3

.L19:
	xorq	%rax,%rax
	ret
	ALIGN_3

.L20:
	movq	M,  %rax
	sarq	$4, %rax
	jle	.L23

	movlpd	-16 * SIZE(X), %xmm0
	movhpd	-15 * SIZE(X), %xmm0
	movlpd	-14 * SIZE(X), %xmm1
	movhpd	-13 * SIZE(X), %xmm1
	movlpd	-12 * SIZE(X), %xmm2
	movhpd	-11 * SIZE(X), %xmm2
	movlpd	-10 * SIZE(X), %xmm3
	movhpd	 -9 * SIZE(X), %xmm3
	movlpd	 -8 * SIZE(X), %xmm8
	movhpd	 -7 * SIZE(X), %xmm8
	movlpd	 -6 * SIZE(X), %xmm9
	movhpd	 -5 * SIZE(X), %xmm9
	movlpd	 -4 * SIZE(X), %xmm10
	movhpd	 -3 * SIZE(X), %xmm10
	movlpd	 -2 * SIZE(X), %xmm11
	movhpd	 -1 * SIZE(X), %xmm11

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	decq	%rax
	jle .L22
	ALIGN_3

.L21:
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)

	addpd	-16 * SIZE(Y), %xmm0
	mulpd	ALPHA, %xmm8
	addpd	-14 * SIZE(Y), %xmm1
	mulpd	ALPHA, %xmm9
	addpd	-12 * SIZE(Y), %xmm2
	mulpd	ALPHA, %xmm10
	addpd	-10 * SIZE(Y), %xmm3
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	PREFETCHW	(PREFETCHSIZE + 0) * SIZE(Y)

	movlpd	 0 * SIZE(X), %xmm0
	movhpd	 1 * SIZE(X), %xmm0
	movlpd	 2 * SIZE(X), %xmm1
	movhpd	 3 * SIZE(X), %xmm1
	movlpd	 4 * SIZE(X), %xmm2
	movhpd	 5 * SIZE(X), %xmm2
	movlpd	 6 * SIZE(X), %xmm3
	movhpd	 7 * SIZE(X), %xmm3

	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)

	addpd	-8 * SIZE(Y), %xmm8
	mulpd	ALPHA, %xmm0
	addpd	-6 * SIZE(Y), %xmm9
	mulpd	ALPHA, %xmm1
	addpd	-4 * SIZE(Y), %xmm10
	mulpd	ALPHA, %xmm2
	addpd	-2 * SIZE(Y), %xmm11
	mulpd	ALPHA, %xmm3

	movapd	%xmm8,  -8 * SIZE(Y)
	movapd	%xmm9,  -6 * SIZE(Y)
	movapd	%xmm10, -4 * SIZE(Y)
	movapd	%xmm11, -2 * SIZE(Y)

	PREFETCHW	(PREFETCHSIZE + 8) * SIZE(Y)

	movlpd	 8 * SIZE(X), %xmm8
	movhpd	 9 * SIZE(X), %xmm8
	movlpd	10 * SIZE(X), %xmm9
	movhpd	11 * SIZE(X), %xmm9
	movlpd	12 * SIZE(X), %xmm10
	movhpd	13 * SIZE(X), %xmm10
	movlpd	14 * SIZE(X), %xmm11
	movhpd	15 * SIZE(X), %xmm11

	addq	$16 * SIZE, Y
	addq	$16 * SIZE, X
	decq	%rax
	jg	.L21
	ALIGN_3

.L22:
	addpd	-16 * SIZE(Y), %xmm0
	mulpd	ALPHA, %xmm8
	addpd	-14 * SIZE(Y), %xmm1
	mulpd	ALPHA, %xmm9
	addpd	-12 * SIZE(Y), %xmm2
	mulpd	ALPHA, %xmm10
	addpd	-10 * SIZE(Y), %xmm3
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	addpd	-8 * SIZE(Y), %xmm8
	addpd	-6 * SIZE(Y), %xmm9
	addpd	-4 * SIZE(Y), %xmm10
	addpd	-2 * SIZE(Y), %xmm11

	movapd	%xmm8,  -8 * SIZE(Y)
	movapd	%xmm9,  -6 * SIZE(Y)
	movapd	%xmm10, -4 * SIZE(Y)
	movapd	%xmm11, -2 * SIZE(Y)

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	ALIGN_3

.L23:
	movq	M,  %rax
	andq	$8, %rax
	jle	.L24
	ALIGN_3

	movlpd	-16 * SIZE(X), %xmm0
	movhpd	-15 * SIZE(X), %xmm0
	movlpd	-14 * SIZE(X), %xmm1
	movhpd	-13 * SIZE(X), %xmm1
	movlpd	-12 * SIZE(X), %xmm2
	movhpd	-11 * SIZE(X), %xmm2
	movlpd	-10 * SIZE(X), %xmm3
	movhpd	 -9 * SIZE(X), %xmm3

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	addpd	-16 * SIZE(Y), %xmm0
	addpd	-14 * SIZE(Y), %xmm1
	addpd	-12 * SIZE(Y), %xmm2
	addpd	-10 * SIZE(Y), %xmm3

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)
	movapd	%xmm2, -12 * SIZE(Y)
	movapd	%xmm3, -10 * SIZE(Y)

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	ALIGN_3

.L24:
	movq	M,  %rax
	andq	$4, %rax
	jle	.L25
	ALIGN_3

	movlpd	-16 * SIZE(X), %xmm0
	movhpd	-15 * SIZE(X), %xmm0
	movlpd	-14 * SIZE(X), %xmm1
	movhpd	-13 * SIZE(X), %xmm1

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1

	addpd	-16 * SIZE(Y), %xmm0
	addpd	-14 * SIZE(Y), %xmm1

	movapd	%xmm0, -16 * SIZE(Y)
	movapd	%xmm1, -14 * SIZE(Y)

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	ALIGN_3

.L25:
	movq	M,  %rax
	andq	$2, %rax
	jle	.L26
	ALIGN_3

	movlpd	-16 * SIZE(X), %xmm0
	movhpd	-15 * SIZE(X), %xmm0

	mulpd	ALPHA, %xmm0

	addpd	-16 * SIZE(Y), %xmm0

	movapd	%xmm0, -16 * SIZE(Y)

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	ALIGN_3

.L26:
	movq	M,  %rax
	andq	$1, %rax
	jle	.L29
	ALIGN_3

	movlpd	-16 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	-16 * SIZE(Y), %xmm0

	movlpd	%xmm0, 	-16 * SIZE(Y)
	addq	$SIZE, Y
	ALIGN_3

.L29:
	xorq	%rax,%rax
	ret
	ALIGN_3


.L40:
	movq	Y, YY
	movq	M,  %rax
	sarq	$3, %rax
	jle	.L45
	ALIGN_3

.L41:
	movlpd	0 * SIZE(X), %xmm0
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm0
	addq	INCX, X
	mulpd	ALPHA, %xmm0

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm0

	movlpd	0 * SIZE(X), %xmm1
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm1
	addq	INCX, X
	mulpd	ALPHA, %xmm1

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm1

	movlpd	0 * SIZE(X), %xmm2
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulpd	ALPHA, %xmm2

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm2

	movlpd	0 * SIZE(X), %xmm3
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm3
	addq	INCX, X
	mulpd	ALPHA, %xmm3

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm3

	movlpd	%xmm0, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm0, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm1, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm1, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm2, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm2, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm3, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm3, 0 * SIZE(Y)
	addq	INCY, Y

	decq	%rax
	jg	.L41
	ALIGN_3

.L45:
	movq	M,  %rax
	andq	$7, %rax
	jle	.L47
	ALIGN_3

.L46:
	movlpd	(X), %xmm0
	addq	INCX, X
	mulsd	%xmm15, %xmm0
	addsd	(Y), %xmm0
	movlpd	%xmm0, (Y)
	addq	INCY, Y
	decq	%rax
	jg	.L46
	ALIGN_3

.L47:
	xorq	%rax, %rax
	ret

	EPILOGUE
