/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
#define RESULT	ARG1	/* rdi */
#define N	ARG2	/* rsi */
#define X	ARG3	/* rdx */
#define INCX	ARG4	/* rcx */
#define Y	ARG5	/* r8  */
#define INCY	ARG6	/* r9  */
#else
#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */
#endif

#define PREFETCHSIZE 112

#define MOVDDUP(a, b, c)	movlpd	a(b), c;unpcklpd	c, c

	PROLOGUE
	PROFCODE

#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	salq	$ZBASE_SHIFT, INCX
	salq	$ZBASE_SHIFT, INCY

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

	cmpq	$0, N
	jle	.L999

	cmpq	$2 * SIZE, INCX
	jne	.L50
	cmpq	$2 * SIZE, INCY
	jne	.L50

	movq	N,  %rax
	sarq	$3, %rax
	jle	.L25

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5
	movlpd	 4 * SIZE(X), %xmm6
	movhpd	 5 * SIZE(X), %xmm6
	movlpd	 6 * SIZE(X), %xmm7
	movhpd	 7 * SIZE(X), %xmm7

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	MOVDDUP( 2 * SIZE, Y, %xmm10)
	MOVDDUP( 3 * SIZE, Y, %xmm11)

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 4 * SIZE, Y, %xmm12)
	MOVDDUP( 5 * SIZE, Y, %xmm13)
	MOVDDUP( 6 * SIZE, Y, %xmm14)
	MOVDDUP( 7 * SIZE, Y, %xmm15)

	movlpd	 8 * SIZE(X), %xmm4
	movhpd	 9 * SIZE(X), %xmm4
	movlpd	10 * SIZE(X), %xmm5
	movhpd	11 * SIZE(X), %xmm5
	decq	%rax
	jle	.L22
	ALIGN_3

.L21:
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	movlpd	12 * SIZE(X), %xmm6
	movhpd	13 * SIZE(X), %xmm6
	MOVDDUP( 8 * SIZE, Y, %xmm8)
	MOVDDUP( 9 * SIZE, Y, %xmm9)

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	14 * SIZE(X), %xmm7
	movhpd	15 * SIZE(X), %xmm7
	MOVDDUP(10 * SIZE, Y, %xmm10)
	MOVDDUP(11 * SIZE, Y, %xmm11)

	PREFETCH	(PREFETCHSIZE + 0) * SIZE(Y)

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9

	movlpd	16 * SIZE(X), %xmm4
	movhpd	17 * SIZE(X), %xmm4
	MOVDDUP(12 * SIZE, Y, %xmm12)
	MOVDDUP(13 * SIZE, Y, %xmm13)

	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	movlpd	18 * SIZE(X), %xmm5
	movhpd	19 * SIZE(X), %xmm5
	MOVDDUP(14 * SIZE, Y, %xmm14)
	MOVDDUP(15 * SIZE, Y, %xmm15)

	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	movlpd	20 * SIZE(X), %xmm6
	movhpd	21 * SIZE(X), %xmm6
	MOVDDUP(16 * SIZE, Y, %xmm8)
	MOVDDUP(17 * SIZE, Y, %xmm9)

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	22 * SIZE(X), %xmm7
	movhpd	23 * SIZE(X), %xmm7

	MOVDDUP(18 * SIZE, Y, %xmm10)
	MOVDDUP(19 * SIZE, Y, %xmm11)

	PREFETCH	(PREFETCHSIZE + 8) * SIZE(Y)

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9

	movlpd	24 * SIZE(X), %xmm4
	movhpd	25 * SIZE(X), %xmm4
	MOVDDUP(20 * SIZE, Y, %xmm12)
	MOVDDUP(21 * SIZE, Y, %xmm13)

	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	movlpd	26 * SIZE(X), %xmm5
	movhpd	27 * SIZE(X), %xmm5
	MOVDDUP(22 * SIZE, Y, %xmm14)
	MOVDDUP(23 * SIZE, Y, %xmm15)

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y

	decq	%rax
	jg	.L21
	ALIGN_3

.L22:
	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	movlpd	12 * SIZE(X), %xmm6
	movhpd	13 * SIZE(X), %xmm6
	MOVDDUP( 8 * SIZE, Y, %xmm8)
	MOVDDUP( 9 * SIZE, Y, %xmm9)

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	14 * SIZE(X), %xmm7
	movhpd	15 * SIZE(X), %xmm7
	MOVDDUP(10 * SIZE, Y, %xmm10)
	MOVDDUP(11 * SIZE, Y, %xmm11)

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9

	MOVDDUP(12 * SIZE, Y, %xmm12)
	MOVDDUP(13 * SIZE, Y, %xmm13)

	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	MOVDDUP(14 * SIZE, Y, %xmm14)
	MOVDDUP(15 * SIZE, Y, %xmm15)

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	addpd	 %xmm12, %xmm0
	addpd	 %xmm13, %xmm1
	addpd	 %xmm14, %xmm2
	addpd	 %xmm15, %xmm3

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	ALIGN_3

.L25:
	testq	$7, N
	je	.L999

	movq	N,  %rax
	andq	$4, %rax
	jle	.L26

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5
	movlpd	 4 * SIZE(X), %xmm6
	movhpd	 5 * SIZE(X), %xmm6
	movlpd	 6 * SIZE(X), %xmm7
	movhpd	 7 * SIZE(X), %xmm7

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	MOVDDUP( 2 * SIZE, Y, %xmm10)
	MOVDDUP( 3 * SIZE, Y, %xmm11)

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 4 * SIZE, Y, %xmm12)
	MOVDDUP( 5 * SIZE, Y, %xmm13)
	MOVDDUP( 6 * SIZE, Y, %xmm14)
	MOVDDUP( 7 * SIZE, Y, %xmm15)

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	addpd	 %xmm12, %xmm0
	addpd	 %xmm13, %xmm1
	addpd	 %xmm14, %xmm2
	addpd	 %xmm15, %xmm3

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	ALIGN_3

.L26:
	movq	N,  %rax
	andq	$2, %rax
	jle	.L27

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	MOVDDUP( 2 * SIZE, Y, %xmm10)
	MOVDDUP( 3 * SIZE, Y, %xmm11)

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	addpd	 %xmm8, %xmm0
	addpd	 %xmm9, %xmm1
	addpd	 %xmm10, %xmm2
	addpd	 %xmm11, %xmm3

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	ALIGN_3

.L27:
	movq	N,  %rax
	andq	$1, %rax
	jle	.L999

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	addpd	 %xmm8, %xmm0
	addpd	 %xmm9, %xmm1
	jmp	 .L999
	ALIGN_3


.L50:
#ifdef F_INTERFACE
	testq	INCX, INCX		# if (incx < 0)
	jge	.L51

	movq	N, %rax		# n
	decq	%rax			# n - 1
	imulq	INCX, %rax		# (n - 1) * incx
	subq	%rax, X
	ALIGN_3

.L51:
	testq	INCY, INCY		# if (incy < 0)
	jge	.L52

	movq	N, %rax
	decq	%rax			# (n - 1)
	imulq	INCY, %rax		# (n - 1) * incy
	subq	%rax, Y
	ALIGN_3
.L52:
#endif

	movq	N,  %rax
	sarq	$3, %rax
	jle	.L55

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm6
	movhpd	 1 * SIZE(X), %xmm6
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm7
	movhpd	 1 * SIZE(X), %xmm7
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 0 * SIZE, Y, %xmm12)
	MOVDDUP( 1 * SIZE, Y, %xmm13)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm14)
	MOVDDUP( 1 * SIZE, Y, %xmm15)
	addq	 INCY, Y

	decq	%rax
	jle	.L54
	ALIGN_3

.L53:
	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm6
	movhpd	 1 * SIZE(X), %xmm6
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm7
	movhpd	 1 * SIZE(X), %xmm7
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9

	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 0 * SIZE, Y, %xmm12)
	MOVDDUP( 1 * SIZE, Y, %xmm13)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm14)
	MOVDDUP( 1 * SIZE, Y, %xmm15)
	addq	 INCY, Y

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm6
	movhpd	 1 * SIZE(X), %xmm6
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm7
	movhpd	 1 * SIZE(X), %xmm7
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9
	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 0 * SIZE, Y, %xmm12)
	MOVDDUP( 1 * SIZE, Y, %xmm13)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm14)
	MOVDDUP( 1 * SIZE, Y, %xmm15)
	addq	 INCY, Y

	decq	%rax
	jg	.L53
	ALIGN_3

.L54:
	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm6
	movhpd	 1 * SIZE(X), %xmm6
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm7
	movhpd	 1 * SIZE(X), %xmm7
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9

	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 0 * SIZE, Y, %xmm12)
	MOVDDUP( 1 * SIZE, Y, %xmm13)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm14)
	MOVDDUP( 1 * SIZE, Y, %xmm15)
	addq	 INCY, Y

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	addpd	 %xmm12, %xmm0
	mulpd	 %xmm4, %xmm8
	addpd	 %xmm13, %xmm1
	mulpd	 %xmm4, %xmm9
	addpd	 %xmm14, %xmm2
	mulpd	 %xmm5, %xmm10
	addpd	 %xmm15, %xmm3
	mulpd	 %xmm5, %xmm11
	ALIGN_3

.L55:
	testq	$7, N
	je	.L999

	movq	N,  %rax
	andq	$4, %rax
	jle	.L56

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm6
	movhpd	 1 * SIZE(X), %xmm6
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm7
	movhpd	 1 * SIZE(X), %xmm7
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	MOVDDUP( 0 * SIZE, Y, %xmm12)
	MOVDDUP( 1 * SIZE, Y, %xmm13)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm14)
	MOVDDUP( 1 * SIZE, Y, %xmm15)
	addq	 INCY, Y

	addpd	 %xmm8, %xmm0
	mulpd	 %xmm6, %xmm12
	addpd	 %xmm9, %xmm1
	mulpd	 %xmm6, %xmm13

	addpd	 %xmm10, %xmm2
	mulpd	 %xmm7, %xmm14
	addpd	 %xmm11, %xmm3
	mulpd	 %xmm7, %xmm15

	addpd	 %xmm12, %xmm0
	addpd	 %xmm13, %xmm1
	addpd	 %xmm14, %xmm2
	addpd	 %xmm15, %xmm3
	ALIGN_3

.L56:
	movq	N,  %rax
	andq	$2, %rax
	jle	.L57

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X
	movlpd	 0 * SIZE(X), %xmm5
	movhpd	 1 * SIZE(X), %xmm5
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)
	addq	 INCY, Y
	MOVDDUP( 0 * SIZE, Y, %xmm10)
	MOVDDUP( 1 * SIZE, Y, %xmm11)
	addq	 INCY, Y

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	mulpd	 %xmm5, %xmm10
	mulpd	 %xmm5, %xmm11

	addpd	 %xmm8, %xmm0
	addpd	 %xmm9, %xmm1
	addpd	 %xmm10, %xmm2
	addpd	 %xmm11, %xmm3
	ALIGN_3

.L57:
	movq	N,  %rax
	andq	$1, %rax
	jle	.L999

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	addq	 INCX, X

	MOVDDUP( 0 * SIZE, Y, %xmm8)
	MOVDDUP( 1 * SIZE, Y, %xmm9)

	mulpd	 %xmm4, %xmm8
	mulpd	 %xmm4, %xmm9
	addpd	 %xmm8, %xmm0
	addpd	 %xmm9, %xmm1
	jmp	 .L999
	ALIGN_3

.L999:
	addpd	%xmm0, %xmm2
	addpd	%xmm1, %xmm3

	movapd	%xmm2,  %xmm0
	unpckhpd %xmm2, %xmm2

	movapd	%xmm3, %xmm1
	unpckhpd %xmm3, %xmm3

#ifndef CONJ
	subsd	 %xmm3, %xmm0
	addsd	 %xmm2, %xmm1
#else
	addsd	 %xmm3, %xmm0
	subsd	 %xmm2, %xmm1
#endif

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
	movlpd	%xmm0, 0 * SIZE(RESULT)
	movlpd	%xmm1, 1 * SIZE(RESULT)
#endif
	ret

	EPILOGUE
