/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define PREFETCH_SIZE 144

#ifdef USEMAX
#define FMAX	fmax
#endif

#ifdef USEMIN
#define FMAX	fmin
#endif

#ifdef USEAMAX
#define FMAX	famax
#endif

#ifdef USEAMIN
#define FMAX	famin
#endif

#define RET	r8

#define PRE1	r2

#define N	r14
#define DX	r15
#define INCX	r16

#define J	r17
#define K	r18
#define X1	r19
#define X2	r20
#define X3	r21
#define INCX5	r22
#define INCX16	r23
#define I	r24
#define IVAL	r25

#define DMAX1	f8
#define DMAX2	f9
#define DMAX3	f10
#define DMAX4	f11
#define DMAX5	f12
#define DMAX6	f13
#define DMAX7	f14
#define DMAX8	f15

#define PR	r30
#define ARLC	r31

	PROLOGUE
	.prologue
	PROFCODE

	{ .mmi
	alloc	r16 = ar.pfs, 8, 72, 0, 80
	mov	RET = 0
	.save	ar.lc, ARLC
	mov	ARLC = ar.lc
	}
	;;
	.body

#ifdef F_INTERFACE
	{ .mmi
	LDINT	N = [r32]
	LDINT	INCX = [r34]
	mov	DX   = r33
	}
	;;
#ifndef USE64BITINT
	{ .mii
	nop.m 0
	sxt4	N = N
	sxt4	INCX = INCX
	}
	;;
#endif
#else
	{ .mmi
	mov	N    = r32
	mov	DX   = r33
	mov	INCX = r34
	}
	;;
#endif

	{ .mii
	mov	X1 = DX
	mov	PR = pr
	cmp.ge	p6, p0 = 0, INCX
	}
	{ .mbb
	cmp.ge	p8, p0 = 0, N
	(p8) br.ret.sptk.many b0
	(p6) br.ret.sptk.many b0
	}
	;;
	{ .mmi
	LDFD	DMAX1 = [X1]
	shladd	INCX = INCX, BASE_SHIFT, r0
	mov	pr.rot= 0
	}
	;;
	{ .mmf
	add	X1 = X1, INCX
	adds	K = -1, N
	mov	DMAX2 = DMAX1
	}
	;;
	{ .mmf
	shladd	X2    = INCX, 2, X1
	mov	DMAX5 = DMAX1
	}
	{ .mfi
	cmp.eq	p16, p0 = r0, r0
	mov	DMAX6 = DMAX1
	shr	J =  K, 4
	}
	;;
	{ .mfi
	shladd	INCX5 = INCX, 2, INCX
	mov	DMAX3 = DMAX1
	mov	ar.ec= 4
	}
	{ .mmf
	shladd	INCX16= INCX, 4, r0
	adds	J = -1, J
	mov	DMAX7 = DMAX1
	}
	;;
	{ .mfi
	adds	PRE1 = PREFETCH_SIZE * SIZE, X1
	mov	DMAX4 = DMAX1
	mov	ar.lc = J
	}
	{ .mfb
	cmp.eq	p7 ,p0  =  -1, J
	mov	DMAX8 = DMAX1
	(p7) br.cond.dpnt  .L15
	}
	.align 32
	;;
.L10:
	{ .mmf
	(p16) lfetch.nt1  [PRE1], INCX16
	(p16) LDFD	f32 = [X1], INCX
	(p19) FMAX DMAX1 = f35, DMAX1
	}
	{ .mmf
	(p16) LDFD	f48 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX5 = f51, DMAX5
	}
	;;
	{ .mmf
	(p16) LDFD	f36 = [X1], INCX
	nop.m 0
	(p19) FMAX DMAX2 = f39, DMAX2
	}
	{ .mmf
	(p16) LDFD	f52 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX6 = f55, DMAX6
	}
	;;
	{ .mmf
	(p16) LDFD	f40 = [X1], INCX
	nop.m 0
	(p19) FMAX DMAX3 = f43, DMAX3
	}
	{ .mmf
	(p16) LDFD	f56 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX7 = f59, DMAX7
	}
	;;
	{ .mmf
	(p16) LDFD	f44 = [X1], INCX5
	nop.m 0
	(p19) FMAX DMAX4 = f47, DMAX4
	}
	{ .mmf
	(p16) LDFD	f60 = [X2], INCX5
	nop.m 0
	(p19) FMAX DMAX8 = f63, DMAX8
	}
	;;
	{ .mmf
	(p16) LDFD	f64 = [X1], INCX
	nop.m 0
	(p19) FMAX DMAX1 = f67, DMAX1
	}
	{ .mmf
	(p16) LDFD	f80 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX5 = f83, DMAX5
	}
	;;
	{ .mmf
	(p16) LDFD	f68 = [X1], INCX
	nop.m 0
	(p19) FMAX DMAX2 = f71, DMAX2
	}
	{ .mmf
	(p16) LDFD	f84 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX6 = f87, DMAX6
	}
	;;
	{ .mmf
	(p16) LDFD	f72 = [X1], INCX
	nop.m 0
	(p19) FMAX DMAX3 = f75, DMAX3
	}
	{ .mmf
	(p16) LDFD	f88 = [X2], INCX
	nop.m 0
	(p19) FMAX DMAX7 = f91, DMAX7
	}
	;;
	{ .mmf
	(p16) LDFD	f76 = [X1], INCX5
	nop.m 0
	(p19) FMAX DMAX4 = f79, DMAX4
	}
	{ .mfb
	(p16) LDFD	f92 = [X2], INCX5
	(p19) FMAX DMAX8 = f95, DMAX8
	br.ctop.sptk.few .L10
	}
	.align 32
	;;
.L15:
	{ .mmi
	and	J =  15, K
	mov	X3 = X1
	tbit.z	p0, p12 = K, 3
	}
	;;
	{ .mmi
	(p12) LDFD	f32 = [X1], INCX
	(p12) LDFD	f36 = [X2], INCX
	tbit.z	p0, p13 = K, 2
	}
	{ .mib
	cmp.eq	p8 ,p0  =   r0, J
	tbit.z	p0, p14 = K, 1
	(p8) br.cond.dpnt  .L30
	}
	;;
	{ .mmi
	(p12) LDFD	f33 = [X1], INCX
	(p12) LDFD	f37 = [X2], INCX
	tbit.z	p0, p15 = K, 0
	}
	;;
	{ .mmi
	(p12) LDFD	f34 = [X1], INCX
	(p12) LDFD	f38 = [X2], INCX
	(p12) shladd X3 = INCX, 3, X3
	}
	;;
	{ .mmi
	(p12) LDFD	f35 = [X1], INCX5
	(p12) LDFD	f39 = [X2], INCX5
	(p13) shladd X3 = INCX, 2, X3
	}
	;;
	{ .mmi
	(p13) LDFD	f40 = [X1], INCX
	(p14) LDFD	f44 = [X3], INCX
	nop.i 0
	}
	;;
	{ .mmi
	(p13) LDFD	f41 = [X1], INCX
	(p14) LDFD	f45 = [X3], INCX
	nop.i 0
	}
	;;
	{ .mmf
	(p13) LDFD	f42 = [X1], INCX
	nop.m 0 
	(p12) FMAX DMAX1 = f32, DMAX1
	}
	{ .mmf
	(p15) LDFD	f46 = [X3], INCX
	nop.m 0 
	(p12) FMAX DMAX5 = f36, DMAX5
	}
	;;
	{ .mmf
	(p13) LDFD	f43 = [X1], INCX
	nop.m 0 
	(p12) FMAX DMAX2 = f33, DMAX2
	}
	(p12) FMAX DMAX6 = f37, DMAX6
	(p12) FMAX DMAX3 = f34, DMAX3
	(p12) FMAX DMAX7 = f38, DMAX7
	(p12) FMAX DMAX4 = f35, DMAX4
	(p12) FMAX DMAX8 = f39, DMAX8
	;;
	(p13) FMAX DMAX1 = f40, DMAX1
	(p14) FMAX DMAX5 = f44, DMAX5
	(p13) FMAX DMAX2 = f41, DMAX2
	(p14) FMAX DMAX6 = f45, DMAX6
	(p13) FMAX DMAX3 = f42, DMAX3
	(p15) FMAX DMAX7 = f46, DMAX7
	(p13) FMAX DMAX4 = f43, DMAX4
	;;
	.align 32
	
.L30:
	{ .mfi
	mov	X1 = DX
	FMAX	DMAX1 = DMAX5, DMAX1
	mov	pr.rot = 0
	}
	{ .mmf
	FMAX	DMAX2 = DMAX6, DMAX2
	}
	;;
	{ .mmf
	cmp.eq	p16, p0 = r0, r0
	FMAX	DMAX3 = DMAX7, DMAX3
	}
	{ .mmf
	shladd	X2    = INCX, 2, DX
	FMAX	DMAX4 = DMAX8, DMAX4
	}
	;;
	{ .mfi
 	FMAX	DMAX1 = DMAX2, DMAX1
	shr	J =  N, 4
	}
	{ .mfb
	FMAX	DMAX3 = DMAX4, DMAX3
	}
	;;
	{ .mfi
 	FMAX	DMAX1 = DMAX3, DMAX1
	mov	ar.ec= 4
	}
	;;
	{ .mmi
#ifdef DOUBLE
	getf.d	IVAL = DMAX1
#else
	getf.s	IVAL = DMAX1
#endif
	adds	J = -1, J
	adds	PRE1 = PREFETCH_SIZE * SIZE, X1
	}
	;;
	{ .mib
	cmp.eq	p7 ,p0  =  -1, J
	mov	ar.lc = J
	(p7) br.cond.dpnt  .L70
	}
	.align 32
	;;
.L40:
	{ .mmi
	(p16) lfetch.nt1  [PRE1], INCX16
	nop.m 0
	(p19) cmp.eq.unc p6,  p0 = r35, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7,  p0 = r39, IVAL
	(p6 ) br.cond.dpnt .L50
	(p7 ) br.cond.dpnt .L51
	}
	;;
	{ .mmi
	(p16) LD8	r32 = [X1], INCX
	(p16) LD8	r48 = [X2], INCX
	(p19) cmp.eq.unc p6,  p0 = r43, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7,  p0 = r47, IVAL
	(p6 ) br.cond.dpnt .L52
	(p7 ) br.cond.dpnt .L53
	}
	;;
	{ .mmi
	(p16) LD8	r36 = [X1], INCX
	(p16) LD8	r52 = [X2], INCX
	(p19) cmp.eq.unc p6,  p0 = r51, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7,  p0 = r55, IVAL
	(p6 ) br.cond.dpnt .L54
	(p7 ) br.cond.dpnt .L55
	}
	;;
	{ .mmi
	(p16) LD8	r40 = [X1], INCX
	(p16) LD8	r56 = [X2], INCX
	(p19) cmp.eq.unc p6, p0 = r59, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7, p0 = r63, IVAL
	(p6) br.cond.dpnt .L56
	(p7) br.cond.dpnt .L57
	}
	;;
	{ .mmi
	(p16) LD8	r44 = [X1], INCX5
	(p16) LD8	r60 = [X2], INCX5
	(p19) cmp.eq.unc p6,  p0 = r67, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7,  p0 = r71, IVAL
	(p6 ) br.cond.dpnt .L58
	(p7 ) br.cond.dpnt .L59
	}
	;;
	{ .mmi
	(p16) LD8	r64 = [X1], INCX
	(p16) LD8	r80 = [X2], INCX
	(p19) cmp.eq.unc p6,  p0 = r75, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7,  p0 = r79, IVAL
	(p6 ) br.cond.dpnt .L60
	(p7 ) br.cond.dpnt .L61
	}
	;;
	{ .mmi
	(p16) LD8	r68 = [X1], INCX
	(p16) LD8	r84 = [X2], INCX
	(p19) cmp.eq.unc p6, p0 = r83, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7, p0 = r87, IVAL
	(p6) br.cond.dpnt .L62
	(p7) br.cond.dpnt .L63
	}
	;;
	{ .mmi
	(p16) LD8	r72 = [X1], INCX
	(p16) LD8	r88 = [X2], INCX
	(p19) cmp.eq.unc p6, p0 = r91, IVAL
	}
	{ .mbb
	(p19) cmp.eq.unc p7, p0 = r95, IVAL
	(p6) br.cond.dpnt .L64
	(p7) br.cond.dpnt .L65
	}
	;;
	{ .mmi
	(p16) LD8	r76 = [X1], INCX5
	(p16) LD8	r92 = [X2], INCX5
	nop.i 0
	}
	{ .mbb
	(p19) adds  RET = 16, RET
	nop.b 0
	br.ctop.sptk.few .L40
	}
	;;
	{ .mmb
	nop.m 0
	nop.m 0
	br .L70
	}
	.align 32
	;;
.L50:
	adds	RET = 1, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L51:
	adds	RET = 2, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L52:
	adds	RET = 3, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L53:
	adds	RET = 4, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L54:
	adds	RET = 5, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L55:
	adds	RET = 6, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L56:
	adds	RET = 7, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L57:
	adds	RET = 8, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L58:
	adds	RET = 9, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L59:
	adds	RET = 10, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L60:
	adds	RET = 11, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L61:
	adds	RET = 12, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L62:
	adds	RET = 13, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L63:
	adds	RET = 14, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L64:
	adds	RET = 15, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L65:
	adds	RET = 16, RET
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	br.ret.sptk.many b0
	;;
	.align 32	

.L70:
	and	J =  15, N
	tbit.z	p0, p12 = N, 3
	mov	X3 = X1
	mov	ar.lc = ARLC
	mov	pr = PR, -65474
	;;
	{ .mmi
	(p12) LD8	r32 = [X1], INCX
	(p12) LD8	r36 = [X2], INCX
	tbit.z	p0, p13 = N, 2
	}
	{ .mib
	tbit.z	p0, p14 = N, 1
	}
	;;
	{ .mmi
	(p12) LD8	r33 = [X1], INCX
	(p12) LD8	r37 = [X2], INCX
	tbit.z	p0, p15 = N, 0
	}
	;;
	{ .mmi
	(p12) LD8	r34 = [X1], INCX
	(p12) LD8	r38 = [X2], INCX
	(p12) shladd X3 = INCX, 3, X3
	}
	;;
	{ .mmi
	(p12) LD8	r35 = [X1], INCX5
	(p12) LD8	r39 = [X2], INCX5
	(p13) shladd X3 = INCX, 2, X3
	}
	;;
	{ .mmi
	(p13) LD8	r40 = [X1], INCX
	(p14) LD8	r44 = [X3], INCX
	nop.i 0
	}
	;;
	{ .mmi
	(p13) LD8	r41 = [X1], INCX
	(p14) LD8	r45 = [X3], INCX
	nop.i 0
	}
	;;
	{ .mbb
	(p13) LD8	r42 = [X1], INCX
	}
	{ .mmi
	(p15) LD8	r46 = [X3], INCX
	}
	;;
	{ .mbb
	(p13) LD8	r43 = [X1], INCX
	}
	;;

	(p12) cmp.eq.unc p6, p0 = r32, IVAL
	(p12) cmp.eq.unc p7, p0 = r33, IVAL
	;;
	(p12) cmp.eq.unc p8, p0 = r34, IVAL
	(p12) adds  RET = 1, RET
	(p6) br.ret.sptk.many b0
	;;
	(p12) cmp.eq.unc p9, p0 = r35, IVAL
	(p12) adds  RET = 1, RET
	(p7) br.ret.sptk.many b0
	;;
	(p12) cmp.eq.unc p6, p0 = r36, IVAL
	(p12) adds  RET = 1, RET
	(p8) br.ret.sptk.many b0
	;;
	(p12) cmp.eq.unc p7, p0 = r37, IVAL
	(p12) adds  RET = 1, RET
	(p9) br.ret.sptk.many b0
	;;
	(p12) cmp.eq.unc p8, p0 = r38, IVAL
	(p12) adds  RET = 1, RET
	(p6) br.ret.sptk.many b0
	;;
	(p12) cmp.eq.unc p9, p0 = r39, IVAL
	(p12) adds  RET = 1, RET
	(p7) br.ret.sptk.many b0
	;;
	(p13) cmp.eq.unc p6, p0 = r40, IVAL
	(p12) adds  RET = 1, RET
	(p8) br.ret.sptk.many b0
	;;
	(p13) cmp.eq.unc p7, p0 = r41, IVAL
	(p12) adds  RET = 1, RET
	(p9) br.ret.sptk.many b0
	;;
	(p13) cmp.eq.unc p8, p0 = r42, IVAL
	(p13) adds  RET = 1, RET
	(p6) br.ret.sptk.many b0
	;;
	(p13) cmp.eq.unc p9, p0 = r43, IVAL
	(p13) adds  RET = 1, RET
	(p7) br.ret.sptk.many b0
	;;
	(p14) cmp.eq.unc p6, p0 = r44, IVAL
	(p13) adds  RET = 1, RET
	(p8) br.ret.sptk.many b0
	;;
	(p14) cmp.eq.unc p7, p0 = r45, IVAL
	(p13) adds  RET = 1, RET
	(p9) br.ret.sptk.many b0
	;;
	(p14) adds  RET = 1, RET
	(p6) br.ret.sptk.many b0
	;;
	(p14) adds  RET = 1, RET
	(p7) br.ret.sptk.many b0
	;;
	adds  RET = 1, RET
	br.ret.sptk.many b0
	;;
	EPILOGUE
	
