/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifndef SWITCHING
#define SWITCHING	(GEMM_P *  2)
#define GEMM_PP		(GEMM_P *  1)
#endif

#if defined(NT) || defined(TT) || defined(CT) || defined(RT) || \
    defined(NC) || defined(TC) || defined(CC) || defined(RC)
#undef SWITCHING
#endif

#define RRR  GEMM_UNROLL_N

GEMMRETTYPE CNAME(long m, long n, long k,
		  FLOAT alpha_r,
#ifdef COMPLEX
		  FLOAT alpha_i,
#endif
		  FLOAT *a, long lda,
		  FLOAT *b, long ldb,
		  FLOAT *c, long ldc, FLOAT *buffer
){
  long ls, is, js;
  long min_l, min_i, min_j;

#ifndef PPC440
  FLOAT *sa, *sb;

  sa = (FLOAT *)((void *)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)((void *)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_BUFFER_A_OFFSET * COMPSIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
#else
  extern FLOAT *sa, *sb;
#endif

#if defined(SWITCHING)
  if (m >= SWITCHING) {
#endif
  for(ls = 0; ls < k; ls += min_l){
    min_l = k - ls;
    if (min_l >= GEMM_Q * 2) {
      min_l = GEMM_Q;
    } else 
      if (min_l > GEMM_Q) {
	min_l = (min_l / 2) & ~(GEMM_UNROLL_N - 1);
      }
    
    for(js = 0; js < n; js += GEMM_R){
      min_j = n - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
#if defined(NN) || defined(TN) || defined(CN) || defined(RN) || \
    defined(NR) || defined(TR) || defined(CR) || defined(RR)
      GEMM_ONCOPY(min_l, min_j, b + (ls + js * ldb) * COMPSIZE, ldb, sb);
#else
      GEMM_OTCOPY(min_l, min_j, b + (js + ls * ldb) * COMPSIZE, ldb, sb);
#endif
      
      for(is = 0; is < m; is += min_i){
	min_i = m - is;
	if (min_i >= GEMM_P * 2) {
	  min_i = GEMM_P;
	} else 
	  if (min_i > GEMM_P) {
	    min_i = (min_i / 2) & ~(GEMM_UNROLL_M - 1);
	  }

#if defined(NN) || defined(NT) || defined(NC) || defined(NR) || \
    defined(RN) || defined(RT) || defined(RC) || defined(RR)
	GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
	GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
	
#if defined(NN) || defined(NT) || defined(TN) || defined(TT)
	GEMM_KERNEL_N
#endif
#if defined(CN) || defined(CT) || defined(RN) || defined(RT)
        GEMM_KERNEL_L
#endif
#if defined(NC) || defined(TC) || defined(NR) || defined(TR)
        GEMM_KERNEL_R
#endif
#if defined(CC) || defined(CR) || defined(RC) || defined(RR)
        GEMM_KERNEL_B
#endif
	             (min_i, min_j, min_l,
		      alpha_r,
#ifdef COMPLEX
		      alpha_i,
#endif
		      sa, sb, c + (is + js * ldc) * COMPSIZE, ldc); 

      } /* end of ls */
    } /* end of is */
  } /* end of js */

#if defined(SWITCHING)

  } else {
    for(ls = 0; ls < k; ls += min_l){
      min_l = k - ls;
    if (min_l >= GEMM_Q * 2) {
      min_l = GEMM_Q;
    } else 
      if (min_l > GEMM_Q) {
	min_l = (min_l / 2) & ~(GEMM_UNROLL_N - 1);
      }
      
      for(is = 0; is < m; is += min_i){
	min_i = m - is;
	if (min_i >= GEMM_P * 2) {
	  min_i = GEMM_P;
	} else 
	  if (min_i > GEMM_P) {
	    min_i = (min_i / 2) & ~(GEMM_UNROLL_M - 1);
	  }
	
#if defined(NN) || defined(NT) || defined(NC) || defined(NR) || \
    defined(RN) || defined(RT) || defined(RC) || defined(RR)
	GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
	GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
	
	for(js = 0; js < n; js += RRR){
	  min_j = n - js;
	  if (min_j > RRR) min_j = RRR;
	  
#if defined(NN) || defined(TN) || defined(CN) || defined(RN) || \
    defined(NR) || defined(TR) || defined(CR) || defined(RR)
      GEMM_ONCOPY(min_l, min_j, b + (ls + js * ldb) * COMPSIZE, ldb, sb);
#else
      GEMM_OTCOPY(min_l, min_j, b + (js + ls * ldb) * COMPSIZE, ldb, sb);
#endif
	  
#if defined(NN) || defined(NT) || defined(TN) || defined(TT)
	GEMM_KERNEL_N
#endif
#if defined(CN) || defined(CT) || defined(RN) || defined(RT)
        GEMM_KERNEL_L
#endif
#if defined(NC) || defined(TC) || defined(NR) || defined(TR)
        GEMM_KERNEL_R
#endif
#if defined(CC) || defined(CR) || defined(RC) || defined(RR)
        GEMM_KERNEL_B
#endif
	             (min_i, min_j, min_l,
		      alpha_r,
#ifdef COMPLEX
		      alpha_i,
#endif
		      sa, sb, c + (is + js * ldc) * COMPSIZE, ldc); 
	  
	} /* end of ls */
      } /* end of is */
    } /* end of js */
  }
#endif

  return 0;
}
