/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define RPREFETCHSIZE (48 + 4)
#define WPREFETCHSIZE (16 + 4)

#define M	ARG1	/* rdi */
#define N	ARG2	/* rsi */
#define A	ARG3	/* rdx */
#define LDA	ARG4	/* rcx */
#define B	ARG5	/* r8  */

#define I	%r9
#define J	%r10
#define AO1	%r11
#define AO2	%r12
#define AO3	%r13
#define AO4	%rax

	PROLOGUE
	PROFCODE
	
	pushq	%r13
	pushq	%r12

	EMMS

	leaq	(,LDA, SIZE), LDA		# Scaling

	movq	N,  J
	sarq	$2, J
	jle	.L20
	ALIGN_4

.L11:
#if 0
	movq	A, AO1
	leaq	(A,   LDA, 1), AO2
	leaq	(A,   LDA, 2), AO3
	leaq	(AO2, LDA, 2), AO4

	movq	M,  I
	sarq	$3, I
	jle	.L13
	ALIGN_4

.L12:
	MOVQ	0 * SIZE(AO1), %mm0
	addq	$8 * SIZE, AO1
	MOVQ	0 * SIZE(AO2), %mm1
	addq	$8 * SIZE, AO2
	MOVQ	0 * SIZE(AO3), %mm2
	addq	$8 * SIZE, AO3
	MOVQ	0 * SIZE(AO4), %mm3
	addq	$8 * SIZE, AO4

	decq	I
	jg	.L12
	ALIGN_4

.L13:
#endif

	movq	A, AO1
	leaq	(A,   LDA), AO2
	leaq	(A,   LDA, 2), AO3
	leaq	(AO2, LDA, 2), AO4
	leaq	(A, LDA, 4), A

	movq	M,  I
	sarq	$2, I
	jle	.L15
	ALIGN_4

.L14:
	prefetchw	(WPREFETCHSIZE +  0) * SIZE(B)
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	0 * SIZE(AO2), %mm1
	MOVQ	0 * SIZE(AO3), %mm2
	MOVQ	0 * SIZE(AO4), %mm3

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO1)
	MOVQ	1 * SIZE(AO1), %mm4
	MOVQ	1 * SIZE(AO2), %mm5
	MOVQ	1 * SIZE(AO3), %mm6
	MOVQ	1 * SIZE(AO4), %mm7

	MOVQ	%mm0,   0 * SIZE(B)
	MOVQ	%mm1,   1 * SIZE(B)
	MOVQ	%mm2,   2 * SIZE(B)
	MOVQ	%mm3,   3 * SIZE(B)

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO2)
	MOVQ	%mm4,   4 * SIZE(B)
	MOVQ	%mm5,   5 * SIZE(B)
	MOVQ	%mm6,   6 * SIZE(B)
	MOVQ	%mm7,   7 * SIZE(B)

	prefetchw	(WPREFETCHSIZE +  8) * SIZE(B)
	MOVQ	2 * SIZE(AO1), %mm0
	MOVQ	2 * SIZE(AO2), %mm1
	MOVQ	2 * SIZE(AO3), %mm2
	MOVQ	2 * SIZE(AO4), %mm3

	MOVQ	3 * SIZE(AO1), %mm4
	MOVQ	3 * SIZE(AO2), %mm5
	MOVQ	3 * SIZE(AO3), %mm6
	MOVQ	3 * SIZE(AO4), %mm7

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO3)
	MOVQ	%mm0,   8 * SIZE(B)
	MOVQ	%mm1,   9 * SIZE(B)
 	MOVQ	%mm2,  10 * SIZE(B)
 	MOVQ	%mm3,  11 * SIZE(B)

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO4)
	MOVQ	%mm4,  12 * SIZE(B)
	MOVQ	%mm5,  13 * SIZE(B)
	MOVQ	%mm6,  14 * SIZE(B)
	MOVQ	%mm7,  15 * SIZE(B)

	addq	$4 * SIZE, AO1
	addq	$4 * SIZE, AO2
	addq	$4 * SIZE, AO3
	addq	$4 * SIZE, AO4

	subq	$-16 * SIZE, B
	decq	I
	jg	.L14
	ALIGN_4

.L15:
	movq	M,  I
	andq	$3, I
	jle	.L17
	ALIGN_4

.L16:
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	0 * SIZE(AO2), %mm1
	MOVQ	0 * SIZE(AO3), %mm2
	MOVQ	0 * SIZE(AO4), %mm3

	MOVQ	%mm0,   0 * SIZE(B)
	MOVQ	%mm1,   1 * SIZE(B)
	MOVQ	%mm2,   2 * SIZE(B)
	MOVQ	%mm3,   3 * SIZE(B)

	addq	$SIZE, AO1
	addq	$SIZE, AO2
	addq	$SIZE, AO3
	addq	$SIZE, AO4
	addq	$4 * SIZE, B
	decq	I
	jg	.L16
	ALIGN_4

.L17:
	decq	J
	jg	.L11
	ALIGN_4

.L20:
	testq	$2, N
	jle	.L30

	movq	A, AO1
	leaq	(A,   LDA), AO2
	leaq	(A, LDA, 2), A

	movq	M,  I
	sarq	$2, I
	jle	.L24
	ALIGN_4

.L23:
	prefetch	(RPREFETCHSIZE) * SIZE(AO1)
	MOVQ	0 * SIZE(AO1), %mm0
	prefetch	(RPREFETCHSIZE) * SIZE(AO2)
	MOVQ	0 * SIZE(AO2), %mm1
	MOVQ	1 * SIZE(AO1), %mm2
	MOVQ	1 * SIZE(AO2), %mm3
	MOVQ	2 * SIZE(AO1), %mm4
	MOVQ	2 * SIZE(AO2), %mm5
	MOVQ	3 * SIZE(AO1), %mm6
	MOVQ	3 * SIZE(AO2), %mm7

	prefetchw	(WPREFETCHSIZE +  0) * SIZE(B)

	MOVQ	%mm0,   0 * SIZE(B)
	MOVQ	%mm1,   1 * SIZE(B)
	MOVQ	%mm2,   2 * SIZE(B)
	MOVQ	%mm3,   3 * SIZE(B)
	MOVQ	%mm4,   4 * SIZE(B)
	MOVQ	%mm5,   5 * SIZE(B)
	MOVQ	%mm6,   6 * SIZE(B)
	MOVQ	%mm7,   7 * SIZE(B)

	addq	$4 * SIZE, AO1
	addq	$4 * SIZE, AO2
	subq	$-8 * SIZE, B
	decq	I
	jg	.L23
	ALIGN_4

.L24:
	movq	M,  I
	andq	$3, I
	jle	.L30
	ALIGN_4

.L25:
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	0 * SIZE(AO2), %mm1

	MOVQ	%mm0,   0 * SIZE(B)
	MOVQ	%mm1,   1 * SIZE(B)

	addq	$SIZE, AO1
	addq	$SIZE, AO2
	addq	$2 * SIZE, B
	decq	I
	jg	.L25
	ALIGN_4

.L30:
	testq	$1, N
	jle	.L999

	movq	A, AO1

	movq	M,  I
	sarq	$2, I
	jle	.L34
	ALIGN_4

.L33:
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	2 * SIZE(AO1), %mm2
	MOVQ	3 * SIZE(AO1), %mm3

	MOVQ	%mm0,   0 * SIZE(B)
	MOVQ	%mm1,   1 * SIZE(B)
	MOVQ	%mm2,   2 * SIZE(B)
	MOVQ	%mm3,   3 * SIZE(B)

	addq	$4 * SIZE, AO1
	subq	$-4 * SIZE, B
	decq	I
	jg	.L33
	ALIGN_4

.L34:
	movq	M,  I
	andq	$3, I
	jle	.L999
	ALIGN_4

.L35:
	MOVQ	0 * SIZE(AO1), %mm0
	addq	$SIZE, AO1

	MOVQ	%mm0,   0 * SIZE(B)
	addq	$1 * SIZE, B
	decq	I
	jg	.L35
	ALIGN_4


.L999:
	EMMS

	popq	%r12
	popq	%r13
	ret

	EPILOGUE
