/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define RPREFETCHSIZE (16 + 4)
#define WPREFETCHSIZE (16 + 4)

#define M	ARG1	/* rdi */
#define N	ARG2	/* rsi */
#define A	ARG3	/* rdx */
#define LDA	ARG4	/* rcx */
#define B	ARG5	/* r8  */

#define I	%r10
#define J	%rbp

#define AO1	%r9
#define AO2	%r15
#define AO3	%r11
#define AO4	%r14
#define BO1	%r13
#define BO2	%r12
#define M8	%rbx
#define BO	%rax

	PROLOGUE
	PROFCODE
	
	pushq	%r15
	pushq	%r14
	pushq	%r13
	pushq	%r12
	pushq	%rbp
	pushq	%rbx

	movq	N,    %rax
	movq	N,    %rbx
	andq	$-4,  %rax
	andq	$-2,  %rbx
	imulq	M,    %rax
	imulq	M,    %rbx

	EMMS

	leaq	(B, %rax, SIZE), BO1
	leaq	(B, %rbx, SIZE), BO2

	leaq	(,   LDA, SIZE), LDA
	leaq	(,   M,   SIZE), M8
	movq	M,  J
	sarq	$2, J
	jle	.L20
	ALIGN_4

.L11:
	movq	A, AO1
	leaq	(A,   LDA   ), AO2
	leaq	(A,   LDA, 2), AO3
	leaq	(AO2, LDA, 2), AO4
	leaq	(A,   LDA, 4), A

	movq	B, BO
	addq	$16 * SIZE, B

	movq	N,  I
	sarq	$2, I
	jle	.L15
	ALIGN_4

.L14:
	prefetchnta	(RPREFETCHSIZE) * SIZE(AO1)
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	2 * SIZE(AO1), %mm2
	MOVQ	3 * SIZE(AO1), %mm3

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO2)
	MOVQ	0 * SIZE(AO2), %mm4
	MOVQ	1 * SIZE(AO2), %mm5
	MOVQ	2 * SIZE(AO2), %mm6
	MOVQ	3 * SIZE(AO2), %mm7

	prefetchw	(WPREFETCHSIZE +  0) * SIZE(B)
	MOVQ	%mm0,   0 * SIZE(BO)
	MOVQ	%mm1,   1 * SIZE(BO)
	MOVQ	%mm2,   2 * SIZE(BO)
	MOVQ	%mm3,   3 * SIZE(BO)
	MOVQ	%mm4,   4 * SIZE(BO)
	MOVQ	%mm5,   5 * SIZE(BO)
	MOVQ	%mm6,   6 * SIZE(BO)
	MOVQ	%mm7,   7 * SIZE(BO)

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO3)
	MOVQ	0 * SIZE(AO3), %mm0
	MOVQ	1 * SIZE(AO3), %mm1
	MOVQ	2 * SIZE(AO3), %mm2
	MOVQ	3 * SIZE(AO3), %mm3

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO4)
	MOVQ	0 * SIZE(AO4), %mm4
	MOVQ	1 * SIZE(AO4), %mm5
	MOVQ	2 * SIZE(AO4), %mm6
	MOVQ	3 * SIZE(AO4), %mm7

	prefetchw	(WPREFETCHSIZE +  8) * SIZE(B)
	MOVQ	%mm0,   8 * SIZE(BO)
	MOVQ	%mm1,   9 * SIZE(BO)
 	MOVQ	%mm2,  10 * SIZE(BO)
 	MOVQ	%mm3,  11 * SIZE(BO)
	MOVQ	%mm4,  12 * SIZE(BO)
	MOVQ	%mm5,  13 * SIZE(BO)
	MOVQ	%mm6,  14 * SIZE(BO)
	MOVQ	%mm7,  15 * SIZE(BO)

	addq	$4 * SIZE, AO1
	addq	$4 * SIZE, AO2
	addq	$4 * SIZE, AO3
	addq	$4 * SIZE, AO4

	leaq	(BO, M8, 4), BO
	decq	I
	jg	.L14
	ALIGN_4

.L15:
	testq	$2, N
	jle	.L16

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	0 * SIZE(AO2), %mm2
	MOVQ	1 * SIZE(AO2), %mm3

	MOVQ	0 * SIZE(AO3), %mm4
	MOVQ	1 * SIZE(AO3), %mm5
	MOVQ	0 * SIZE(AO4), %mm6
	MOVQ	1 * SIZE(AO4), %mm7

	MOVQ	%mm0,   0 * SIZE(BO1)
	MOVQ	%mm1,   1 * SIZE(BO1)
	MOVQ	%mm2,   2 * SIZE(BO1)
	MOVQ	%mm3,   3 * SIZE(BO1)
	MOVQ	%mm4,   4 * SIZE(BO1)
	MOVQ	%mm5,   5 * SIZE(BO1)
	MOVQ	%mm6,   6 * SIZE(BO1)
	MOVQ	%mm7,   7 * SIZE(BO1)

	addq	$2 * SIZE, AO1
	addq	$2 * SIZE, AO2
	addq	$2 * SIZE, AO3
	addq	$2 * SIZE, AO4
	addq	$8 * SIZE, BO1
	ALIGN_4

.L16:
	testq	$1, N
	jle	.L19

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	0 * SIZE(AO2), %mm1
	MOVQ	0 * SIZE(AO3), %mm2
	MOVQ	0 * SIZE(AO4), %mm3

	MOVQ	%mm0,   0 * SIZE(BO2)
	MOVQ	%mm1,   1 * SIZE(BO2)
	MOVQ	%mm2,   2 * SIZE(BO2)
	MOVQ	%mm3,   3 * SIZE(BO2)

	addq	$4 * SIZE, BO2
	ALIGN_4

.L19:
	decq	J
	jg	.L11
	ALIGN_4

.L20:
	testq	$2, M
	jle	.L30
	ALIGN_4

.L21:
	movq	A, AO1
	leaq	(A,   LDA   ), AO2
	leaq	(A,   LDA, 2), A

	movq	B, BO
	addq	$8 * SIZE, B

	movq	N,  I
	sarq	$2, I
	jle	.L23
	ALIGN_4

.L22:
	prefetchnta	(RPREFETCHSIZE) * SIZE(AO1)
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	2 * SIZE(AO1), %mm2
	MOVQ	3 * SIZE(AO1), %mm3

	prefetchnta	(RPREFETCHSIZE) * SIZE(AO2)
	MOVQ	0 * SIZE(AO2), %mm4
	MOVQ	1 * SIZE(AO2), %mm5
	MOVQ	2 * SIZE(AO2), %mm6
	MOVQ	3 * SIZE(AO2), %mm7

	prefetchw	(WPREFETCHSIZE +  0) * SIZE(B)
	MOVQ	%mm0,   0 * SIZE(BO)
	MOVQ	%mm1,   1 * SIZE(BO)
	MOVQ	%mm2,   2 * SIZE(BO)
	MOVQ	%mm3,   3 * SIZE(BO)
	MOVQ	%mm4,   4 * SIZE(BO)
	MOVQ	%mm5,   5 * SIZE(BO)
	MOVQ	%mm6,   6 * SIZE(BO)
	MOVQ	%mm7,   7 * SIZE(BO)

	addq	$4 * SIZE, AO1
	addq	$4 * SIZE, AO2
	leaq	(BO, M8, 4), BO
	decq	I
	jg	.L22
	ALIGN_4

.L23:
	testq	$2, N
	jle	.L24

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	0 * SIZE(AO2), %mm2
	MOVQ	1 * SIZE(AO2), %mm3

	MOVQ	%mm0,   0 * SIZE(BO1)
	MOVQ	%mm1,   1 * SIZE(BO1)
	MOVQ	%mm2,   2 * SIZE(BO1)
	MOVQ	%mm3,   3 * SIZE(BO1)

	addq	$2 * SIZE, AO1
	addq	$2 * SIZE, AO2
	addq	$4 * SIZE, BO1
	ALIGN_4

.L24:
	testq	$1, N
	jle	.L30

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	0 * SIZE(AO2), %mm1

	MOVQ	%mm0,   0 * SIZE(BO2)
	MOVQ	%mm1,   1 * SIZE(BO2)

	addq	$2 * SIZE, BO2
	ALIGN_4

.L30:
	testq	$1, M
	jle	.L999
	ALIGN_4

.L31:
	movq	A, AO1
	movq	B, BO

	movq	N,  I
	sarq	$2, I
	jle	.L33
	ALIGN_4

.L32:
	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1
	MOVQ	2 * SIZE(AO1), %mm2
	MOVQ	3 * SIZE(AO1), %mm3

	MOVQ	%mm0,   0 * SIZE(BO)
	MOVQ	%mm1,   1 * SIZE(BO)
	MOVQ	%mm2,   2 * SIZE(BO)
	MOVQ	%mm3,   3 * SIZE(BO)

	addq	$4 * SIZE, AO1
	leaq	(BO, M8, 4), BO
	decq	I
	jg	.L32
	ALIGN_4

.L33:
	testq	$2, N
	jle	.L34

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	1 * SIZE(AO1), %mm1

	MOVQ	%mm0,   0 * SIZE(BO1)
	MOVQ	%mm1,   1 * SIZE(BO1)

	addq	$2 * SIZE, AO1
	addq	$2 * SIZE, BO1
	ALIGN_4

.L34:
	testq	$1, N
	jle	.L999

	MOVQ	0 * SIZE(AO1), %mm0
	MOVQ	%mm0,   0 * SIZE(BO2)

	addq	$1 * SIZE, BO2
	ALIGN_4

.L999:
	EMMS

	popq	%rbx
	popq	%rbp
	popq	%r12
	popq	%r13
	popq	%r14
	popq	%r15
	ret

	EPILOGUE
