/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

void CNAME(long m, long n, FLOAT *a, long lda, long posX, long posY, FLOAT *b){

  long i, js, ii;
  long X;

  FLOAT *a01, *a02, *a03 ,*a04, *a05, *a06, *a07, *a08;

  lda *= 2;

  js = (n >> 3);
  if (js > 0){
    do {
      X = posX;

      if (posX <= posY) {
	a01 = a + posY * 2 + (posX +  0) * lda;
	a02 = a + posY * 2 + (posX +  1) * lda;
	a03 = a + posY * 2 + (posX +  2) * lda;
	a04 = a + posY * 2 + (posX +  3) * lda;
	a05 = a + posY * 2 + (posX +  4) * lda;
	a06 = a + posY * 2 + (posX +  5) * lda;
	a07 = a + posY * 2 + (posX +  6) * lda;
	a08 = a + posY * 2 + (posX +  7) * lda;
      } else {
	a01 = a + posX * 2 + (posY +  0) * lda;
	a02 = a + posX * 2 + (posY +  1) * lda;
	a03 = a + posX * 2 + (posY +  2) * lda;
	a04 = a + posX * 2 + (posY +  3) * lda;
	a05 = a + posX * 2 + (posY +  4) * lda;
	a06 = a + posX * 2 + (posY +  5) * lda;
	a07 = a + posX * 2 + (posY +  6) * lda;
	a08 = a + posX * 2 + (posY +  7) * lda;
      }

      i = (m >> 3);
      if (i > 0) {
	do {
	  if (X > posY) {
	    
	    for (ii = 0; ii < 8; ii++){
	      b[ 0] = *(a01 + 0);
	      b[ 1] = -*(a01 + 1);
	      b[ 2] = *(a02 + 0);
	      b[ 3] = -*(a02 + 1);
	      b[ 4] = *(a03 + 0);
	      b[ 5] = -*(a03 + 1);
	      b[ 6] = *(a04 + 0);
	      b[ 7] = -*(a04 + 1);
	      b[ 8] = *(a05 + 0);
	      b[ 9] = -*(a05 + 1);
	      b[10] = *(a06 + 0);
	      b[11] = -*(a06 + 1);
	      b[12] = *(a07 + 0);
	      b[13] = -*(a07 + 1);
	      b[14] = *(a08 + 0);
	      b[15] = -*(a08 + 1);
	      
	      a01 += 2;
	      a02 += 2;
	      a03 += 2;
	      a04 += 2;
	      a05 += 2;
	      a06 += 2;
	      a07 += 2;
	      a08 += 2;
	      b += 16;
	    }

	  } else 
	    if (X < posY) {

	      for (ii = 0; ii < 8; ii++){
		
		b[  0] = *(a01 +  0);
		b[  1] = *(a01 +  1);
		b[  2] = *(a01 +  2);
		b[  3] = *(a01 +  3);
		b[  4] = *(a01 +  4);
		b[  5] = *(a01 +  5);
		b[  6] = *(a01 +  6);
		b[  7] = *(a01 +  7);
		
		b[  8] = *(a01 +  8);
		b[  9] = *(a01 +  9);
		b[ 10] = *(a01 + 10);
		b[ 11] = *(a01 + 11);
		b[ 12] = *(a01 + 12);
		b[ 13] = *(a01 + 13);
		b[ 14] = *(a01 + 14);
		b[ 15] = *(a01 + 15);
		
		a01 += lda;
		b += 16;
	      }

	      a02 += 8 * lda;
	      a03 += 8 * lda;
	      a04 += 8 * lda;
	      a05 += 8 * lda;
	      a06 += 8 * lda;
	      a07 += 8 * lda;
	      a08 += 8 * lda;
		
	    } else {
	      b[  0] = *(a01 +  0);
	      b[  1] = ZERO;
	      b[  2] = *(a01 +  2);
	      b[  3] = *(a01 +  3);
	      b[  4] = *(a01 +  4);
	      b[  5] = *(a01 +  5);
	      b[  6] = *(a01 +  6);
	      b[  7] = *(a01 +  7);
	      
	      b[  8] = *(a01 +  8);
	      b[  9] = *(a01 +  9);
	      b[ 10] = *(a01 + 10);
	      b[ 11] = *(a01 + 11);
	      b[ 12] = *(a01 + 12);
	      b[ 13] = *(a01 + 13);
	      b[ 14] = *(a01 + 14);
	      b[ 15] = *(a01 + 15);
      
	      b[ 16] = *(a01 +  2);
	      b[ 17] = -*(a01 +  3);
	      b[ 18] = *(a02 +  2);
	      b[ 19] = ZERO;
	      b[ 20] = *(a02 +  4);
	      b[ 21] = *(a02 +  5);
	      b[ 22] = *(a02 +  6);
	      b[ 23] = *(a02 +  7);
	      
	      b[ 24] = *(a02 +  8);
	      b[ 25] = *(a02 +  9);
	      b[ 26] = *(a02 + 10);
	      b[ 27] = *(a02 + 11);
	      b[ 28] = *(a02 + 12);
	      b[ 29] = *(a02 + 13);
	      b[ 30] = *(a02 + 14);
	      b[ 31] = *(a02 + 15);

	      b[ 32] = *(a01 +  4);
	      b[ 33] = -*(a01 +  5);
	      b[ 34] = *(a02 +  4);
	      b[ 35] = -*(a02 +  5);
	      b[ 36] = *(a03 +  4);
	      b[ 37] = ZERO;
	      b[ 38] = *(a03 +  6);
	      b[ 39] = *(a03 +  7);
	      
	      b[ 40] = *(a03 +  8);
	      b[ 41] = *(a03 +  9);
	      b[ 42] = *(a03 + 10);
	      b[ 43] = *(a03 + 11);
	      b[ 44] = *(a03 + 12);
	      b[ 45] = *(a03 + 13);
	      b[ 46] = *(a03 + 14);
	      b[ 47] = *(a03 + 15);

	      b[ 48] = *(a01 +  6);
	      b[ 49] = -*(a01 +  7);
	      b[ 50] = *(a02 +  6);
	      b[ 51] = -*(a02 +  7);
	      b[ 52] = *(a03 +  6);
	      b[ 53] = -*(a03 +  7);
	      b[ 54] = *(a04 +  6);
	      b[ 55] = ZERO;
	      
	      b[ 56] = *(a04 +  8);
	      b[ 57] = *(a04 +  9);
	      b[ 58] = *(a04 + 10);
	      b[ 59] = *(a04 + 11);
	      b[ 60] = *(a04 + 12);
	      b[ 61] = *(a04 + 13);
	      b[ 62] = *(a04 + 14);
	      b[ 63] = *(a04 + 15);

	      b[ 64] = *(a01 +  8);
	      b[ 65] = -*(a01 +  9);
	      b[ 66] = *(a02 +  8);
	      b[ 67] = -*(a02 +  9);
	      b[ 68] = *(a03 +  8);
	      b[ 69] = -*(a03 +  9);
	      b[ 70] = *(a04 +  8);
	      b[ 71] = -*(a04 +  9);
	      
	      b[ 72] = *(a05 +  8);
	      b[ 73] = ZERO;
	      b[ 74] = *(a05 + 10);
	      b[ 75] = *(a05 + 11);
	      b[ 76] = *(a05 + 12);
	      b[ 77] = *(a05 + 13);
	      b[ 78] = *(a05 + 14);
	      b[ 79] = *(a05 + 15);

	      b[ 80] = *(a01 + 10);
	      b[ 81] = -*(a01 + 11);
	      b[ 82] = *(a02 + 10);
	      b[ 83] = -*(a02 + 11);
	      b[ 84] = *(a03 + 10);
	      b[ 85] = -*(a03 + 11);
	      b[ 86] = *(a04 + 10);
	      b[ 87] = -*(a04 + 11);
	      
	      b[ 88] = *(a05 + 10);
	      b[ 89] = -*(a05 + 11);
	      b[ 90] = *(a06 + 10);
	      b[ 91] = ZERO;
	      b[ 92] = *(a06 + 12);
	      b[ 93] = *(a06 + 13);
	      b[ 94] = *(a06 + 14);
	      b[ 95] = *(a06 + 15);

	      b[ 96] = *(a01 + 12);
	      b[ 97] = -*(a01 + 13);
	      b[ 98] = *(a02 + 12);
	      b[ 99] = -*(a02 + 13);
	      b[100] = *(a03 + 12);
	      b[101] = -*(a03 + 13);
	      b[102] = *(a04 + 12);
	      b[103] = -*(a04 + 13);
	      
	      b[104] = *(a05 + 12);
	      b[105] = -*(a05 + 13);
	      b[106] = *(a06 + 12);
	      b[107] = -*(a06 + 13);
	      b[108] = *(a07 + 12);
	      b[109] = ZERO;
	      b[110] = *(a07 + 14);
	      b[111] = *(a07 + 15);

	      b[112] = *(a01 + 14);
	      b[113] = -*(a01 + 15);
	      b[114] = *(a02 + 14);
	      b[115] = -*(a02 + 15);
	      b[116] = *(a03 + 14);
	      b[117] = -*(a03 + 15);
	      b[118] = *(a04 + 14);
	      b[119] = -*(a04 + 15);
	      
	      b[120] = *(a05 + 14);
	      b[121] = -*(a05 + 15);
	      b[122] = *(a06 + 14);
	      b[123] = -*(a06 + 15);
	      b[124] = *(a07 + 14);
	      b[125] = -*(a07 + 15);
	      b[126] = *(a08 + 14);
	      b[127] = ZERO;

	      a01 += 16;
	      a02 += 16;
	      a03 += 16;
	      a04 += 16;
	      a05 += 16;
	      a06 += 16;
	      a07 += 16;
	      a08 += 16;
	      b += 128;
	    }

	  X += 8;
	  i --;
	} while (i > 0);
      }
      
      i = (m & 7);
      if (i > 0) {
	if (X > posY) {
	  for (ii = 0; ii < i; ii++){
	    b[ 0] = *(a01 + 0);
	    b[ 1] = -*(a01 + 1);
	    b[ 2] = *(a02 + 0);
	    b[ 3] = -*(a02 + 1);
	    b[ 4] = *(a03 + 0);
	    b[ 5] = -*(a03 + 1);
	    b[ 6] = *(a04 + 0);
	    b[ 7] = -*(a04 + 1);
	    b[ 8] = *(a05 + 0);
	    b[ 9] = -*(a05 + 1);
	    b[10] = *(a06 + 0);
	    b[11] = -*(a06 + 1);
	    b[12] = *(a07 + 0);
	    b[13] = -*(a07 + 1);
	    b[14] = *(a08 + 0);
	    b[15] = -*(a08 + 1);
	    
	    a01 += 2;
	    a02 += 2;
	    a03 += 2;
	    a04 += 2;
	    a05 += 2;
	    a06 += 2;
	    a07 += 2;
	    a08 += 2;
	    b += 16;
	  }
	} else 
	  if (X < posY) {
	    for (ii = 0; ii < i; ii++){
	      b[  0] = *(a01 +  0);
	      b[  1] = *(a01 +  1);
	      b[  2] = *(a01 +  2);
	      b[  3] = *(a01 +  3);
	      b[  4] = *(a01 +  4);
	      b[  5] = *(a01 +  5);
	      b[  6] = *(a01 +  6);
	      b[  7] = *(a01 +  7);
	      
	      b[  8] = *(a01 +  8);
	      b[  9] = *(a01 +  9);
	      b[ 10] = *(a01 + 10);
	      b[ 11] = *(a01 + 11);
	      b[ 12] = *(a01 + 12);
	      b[ 13] = *(a01 + 13);
	      b[ 14] = *(a01 + 14);
	      b[ 15] = *(a01 + 15);
	      
	      a01 += lda;
	      a02 += lda;
	      a03 += lda;
	      a04 += lda;
	      a05 += lda;
	      a06 += lda;
	      a07 += lda;
	      a08 += lda;
	      b += 16;
	    }
	  } else {
	    b[  0] = *(a01 +  0);
	    b[  1] = ZERO;
	    b[  2] = *(a01 +  2);
	    b[  3] = *(a01 +  3);
	    b[  4] = *(a01 +  4);
	    b[  5] = *(a01 +  5);
	    b[  6] = *(a01 +  6);
	    b[  7] = *(a01 +  7);
	    
	    b[  8] = *(a01 +  8);
	    b[  9] = *(a01 +  9);
	    b[ 10] = *(a01 + 10);
	    b[ 11] = *(a01 + 11);
	    b[ 12] = *(a01 + 12);
	    b[ 13] = *(a01 + 13);
	    b[ 14] = *(a01 + 14);
	    b[ 15] = *(a01 + 15);
	    b += 16;

	    if (i >= 2) {
	      b[ 0] = *(a01 +  2);
	      b[ 1] = -*(a01 +  3);
	      b[ 2] = *(a02 +  2);
	      b[ 3] = ZERO;
	      b[ 4] = *(a02 +  4);
	      b[ 5] = *(a02 +  5);
	      b[ 6] = *(a02 +  6);
	      b[ 7] = *(a02 +  7);
	      
	      b[ 8] = *(a02 +  8);
	      b[ 9] = *(a02 +  9);
	      b[10] = *(a02 + 10);
	      b[11] = *(a02 + 11);
	      b[12] = *(a02 + 12);
	      b[13] = *(a02 + 13);
	      b[14] = *(a02 + 14);
	      b[15] = *(a02 + 15);
	      b += 16;
	    }

	    if (i >= 3) {
	      b[ 0] = *(a01 +  4);
	      b[ 1] = -*(a01 +  5);
	      b[ 2] = *(a02 +  4);
	      b[ 3] = -*(a02 +  5);
	      b[ 4] = *(a03 +  4);
	      b[ 5] = ZERO;
	      b[ 6] = *(a03 +  6);
	      b[ 7] = *(a03 +  7);
	      
	      b[ 8] = *(a03 +  8);
	      b[ 9] = *(a03 +  9);
	      b[10] = *(a03 + 10);
	      b[11] = *(a03 + 11);
	      b[12] = *(a03 + 12);
	      b[13] = *(a03 + 13);
	      b[14] = *(a03 + 14);
	      b[15] = *(a03 + 15);
	      b += 16;
	    }

	    if (i >= 4) {
	      b[ 0] = *(a01 +  6);
	      b[ 1] = -*(a01 +  7);
	      b[ 2] = *(a02 +  6);
	      b[ 3] = -*(a02 +  7);
	      b[ 4] = *(a03 +  6);
	      b[ 5] = -*(a03 +  7);
	      b[ 6] = *(a04 +  6);
	      b[ 7] = ZERO;
	      
	      b[ 8] = *(a04 +  8);
	      b[ 9] = *(a04 +  9);
	      b[10] = *(a04 + 10);
	      b[11] = *(a04 + 11);
	      b[12] = *(a04 + 12);
	      b[13] = *(a04 + 13);
	      b[14] = *(a04 + 14);
	      b[15] = *(a04 + 15);
	      b += 16;
	    }

	    if (i >= 5) {
	      b[ 0] = *(a01 +  8);
	      b[ 1] = -*(a01 +  9);
	      b[ 2] = *(a02 +  8);
	      b[ 3] = -*(a02 +  9);
	      b[ 4] = *(a03 +  8);
	      b[ 5] = -*(a03 +  9);
	      b[ 6] = *(a04 +  8);
	      b[ 7] = -*(a04 +  9);
	      
	      b[ 8] = *(a05 +  8);
	      b[ 9] = ZERO;
	      b[10] = *(a05 + 10);
	      b[11] = *(a05 + 11);
	      b[12] = *(a05 + 12);
	      b[13] = *(a05 + 13);
	      b[14] = *(a05 + 14);
	      b[15] = *(a05 + 15);
	      b += 16;
	    }

	    if (i >= 6) {
	      b[ 0] = *(a01 + 10);
	      b[ 1] = -*(a01 + 11);
	      b[ 2] = *(a02 + 10);
	      b[ 3] = -*(a02 + 11);
	      b[ 4] = *(a03 + 10);
	      b[ 5] = -*(a03 + 11);
	      b[ 6] = *(a04 + 10);
	      b[ 7] = -*(a04 + 11);
	      
	      b[ 8] = *(a05 + 10);
	      b[ 9] = -*(a05 + 11);
	      b[10] = *(a06 + 10);
	      b[11] = ZERO;
	      b[12] = *(a06 + 12);
	      b[13] = *(a06 + 13);
	      b[14] = *(a06 + 14);
	      b[15] = *(a06 + 15);
	      b += 16;
	    }

	    if (i >= 7) {
	      b[ 0] = *(a01 + 12);
	      b[ 1] = -*(a01 + 13);
	      b[ 2] = *(a02 + 12);
	      b[ 3] = -*(a02 + 13);
	      b[ 4] = *(a03 + 12);
	      b[ 5] = -*(a03 + 13);
	      b[ 6] = *(a04 + 12);
	      b[ 7] = -*(a04 + 13);
	      
	      b[ 8] = *(a05 + 12);
	      b[ 9] = -*(a05 + 13);
	      b[10] = *(a06 + 12);
	      b[11] = -*(a06 + 13);
	      b[12] = *(a07 + 12);
	      b[13] = ZERO;
	      b[14] = *(a07 + 14);
	      b[15] = *(a07 + 15);
	      b += 16;
	    }
	  }	    
      }

      posY += 8;
      js --;
    } while (js > 0);
  } /* End of main loop */


  if (n & 4){
    X = posX;
    
    if (posX <= posY) {
      a01 = a + posY * 2 + (posX +  0) * lda;
      a02 = a + posY * 2 + (posX +  1) * lda;
      a03 = a + posY * 2 + (posX +  2) * lda;
      a04 = a + posY * 2 + (posX +  3) * lda;
    } else {
      a01 = a + posX * 2 + (posY +  0) * lda;
      a02 = a + posX * 2 + (posY +  1) * lda;
      a03 = a + posX * 2 + (posY +  2) * lda;
      a04 = a + posX * 2 + (posY +  3) * lda;
    }
    
    i = (m >> 2);
    if (i > 0) {
      do {
	if (X > posY) {
	  for (ii = 0; ii < 4; ii++){
	    b[ 0] = *(a01 + 0);
	    b[ 1] = -*(a01 + 1);
	    b[ 2] = *(a02 + 0);
	    b[ 3] = -*(a02 + 1);
	    b[ 4] = *(a03 + 0);
	    b[ 5] = -*(a03 + 1);
	    b[ 6] = *(a04 + 0);
	    b[ 7] = -*(a04 + 1);

	    a01 += 2;
	    a02 += 2;
	    a03 += 2;
	    a04 += 2;
	    b += 8;
	  }
	} else 
	  if (X < posY) {
	    for (ii = 0; ii < 4; ii++){
	      b[  0] = *(a01 +  0);
	      b[  1] = *(a01 +  1);
	      b[  2] = *(a01 +  2);
	      b[  3] = *(a01 +  3);
	      b[  4] = *(a01 +  4);
	      b[  5] = *(a01 +  5);
	      b[  6] = *(a01 +  6);
	      b[  7] = *(a01 +  7);
	      
	      a01 += lda;
	      b += 8;
	      }

	    a02 += 4 * lda;
	    a03 += 4 * lda;
	    a04 += 4 * lda;
	    } else {
	      b[  0] = *(a01 +  0);
	      b[  1] = ZERO;
	      b[  2] = *(a01 +  2);
	      b[  3] = *(a01 +  3);
	      b[  4] = *(a01 +  4);
	      b[  5] = *(a01 +  5);
	      b[  6] = *(a01 +  6);
	      b[  7] = *(a01 +  7);
	      
	      b[  8] = *(a01 +  2);
	      b[  9] = -*(a01 +  3);
	      b[ 10] = *(a02 +  2);
	      b[ 11] = ZERO;
	      b[ 12] = *(a02 +  4);
	      b[ 13] = *(a02 +  5);
	      b[ 14] = *(a02 +  6);
	      b[ 15] = *(a02 +  7);

	      b[ 16] = *(a01 +  4);
	      b[ 17] = -*(a01 +  5);
	      b[ 18] = *(a02 +  4);
	      b[ 19] = -*(a02 +  5);
	      b[ 20] = *(a03 +  4);
	      b[ 21] = ZERO;
	      b[ 22] = *(a03 +  6);
	      b[ 23] = *(a03 +  7);
	      
	      b[ 24] = *(a01 +  6);
	      b[ 25] = -*(a01 +  7);
	      b[ 26] = *(a02 +  6);
	      b[ 27] = -*(a02 +  7);
	      b[ 28] = *(a03 +  6);
	      b[ 29] = -*(a03 +  7);
	      b[ 30] = *(a04 +  6);
	      b[ 31] = ZERO;
	      
	      a01 += 8;
	      a02 += 8;
	      a03 += 8;
	      a04 += 8;
	      b += 32;
	    }
	
	X += 4;
	i --;
      } while (i > 0);
    }
    
    i = (m & 3);
    if (i > 0) {
      if (X > posY) {
	for (ii = 0; ii < i; ii++){
	  b[ 0] = *(a01 + 0);
	  b[ 1] = -*(a01 + 1);
	  b[ 2] = *(a02 + 0);
	  b[ 3] = -*(a02 + 1);
	  b[ 4] = *(a03 + 0);
	  b[ 5] = -*(a03 + 1);
	  b[ 6] = *(a04 + 0);
	  b[ 7] = -*(a04 + 1);
	  
	  a01 += 2;
	  a02 += 2;
	  a03 += 2;
	  a04 += 2;
	  b += 8;
	}
      } else 
	if (X < posY) {
	  for (ii = 0; ii < i; ii++){
	    b[  0] = *(a01 +  0);
	    b[  1] = *(a01 +  1);
	    b[  2] = *(a01 +  2);
	    b[  3] = *(a01 +  3);
	    b[  4] = *(a01 +  4);
	    b[  5] = *(a01 +  5);
	    b[  6] = *(a01 +  6);
	    b[  7] = *(a01 +  7);
	    
	    a01 += lda;
	    a02 += lda;
	    a03 += lda;
	    a04 += lda;
	    b += 8;
	  }
	} else {
	  b[  0] = *(a01 +  0);
	  b[  1] = ZERO;
	  b[  2] = *(a01 +  2);
	  b[  3] = *(a01 +  3);
	  b[  4] = *(a01 +  4);
	  b[  5] = *(a01 +  5);
	  b[  6] = *(a01 +  6);
	  b[  7] = *(a01 +  7);
	  b += 8;

	  if (i >= 2) {
	    b[ 0] = *(a01 +  2);
	    b[ 1] = -*(a01 +  3);
	    b[ 2] = *(a02 +  2);
	    b[ 3] = ZERO;
	    b[ 4] = *(a02 +  4);
	    b[ 5] = *(a02 +  5);
	    b[ 6] = *(a02 +  6);
	    b[ 7] = *(a02 +  7);
	    b += 8;
	  }

	  if (i >= 3) {
	    b[ 0] = *(a01 +  4);
	    b[ 1] = -*(a01 +  5);
	    b[ 2] = *(a02 +  4);
	    b[ 3] = -*(a02 +  5);
	    b[ 4] = *(a03 +  4);
	    b[ 5] = ZERO;
	    b[ 6] = *(a03 +  6);
	    b[ 7] = *(a03 +  7);
	    b += 8;
	  }
	}
    }
    posY += 4;
  }

  if (n & 2){
    X = posX;
    
    if (posX <= posY) {
      a01 = a + posY * 2 + (posX +  0) * lda;
      a02 = a + posY * 2 + (posX +  1) * lda;
    } else {
      a01 = a + posX * 2 + (posY +  0) * lda;
      a02 = a + posX * 2 + (posY +  1) * lda;
    }
    
    i = (m >> 1);
    if (i > 0) {
      do {
	if (X > posY) {
	  b[0] = *(a01 + 0);
	  b[1] = -*(a01 + 1);
	  b[2] = *(a02 + 0);
	  b[3] = -*(a02 + 1);
	  b[4] = *(a01 + 2);
	  b[5] = -*(a01 + 3);
	  b[6] = *(a02 + 2);
	  b[7] = -*(a02 + 3);
	  a01 += 4;
	  a02 += 4;
	  b += 8;
	} else 
	  if (X < posY) {
	    b[0] = *(a01 +  0);
	    b[1] = *(a01 +  1);
	    b[2] = *(a01 +  2);
	    b[3] = *(a01 +  3);
	    b[4] = *(a02 +  0);
	    b[5] = *(a02 +  1);
	    b[6] = *(a02 +  2);
	    b[7] = *(a02 +  3);
	    a01 += 2 * lda;
	    a02 += 2 * lda;
	    b += 8;
	  } else {
	    b[0] = *(a01 +  0);
	    b[1] = ZERO;
	    b[2] = *(a01 +  2);
	    b[3] = *(a01 +  3);
	    
	    b[4] = *(a01 +  2);
	    b[5] = -*(a01 +  3);
	    b[6] = *(a02 +  2);
	    b[7] = ZERO;
	    
	    a01 += 4;
	    a02 += 4;
	    b += 8;
	  }
	
	X += 2;
	i --;
      } while (i > 0);
    }
    
    i = (m & 1);
    if (i > 0) {
      if (X > posY) {
	b[ 0] = *(a01 + 0);
	b[ 1] = -*(a01 + 1);
	b[ 2] = *(a02 + 0);
	b[ 3] = -*(a02 + 1);
	
	a01 += 2;
	a02 += 2;
	b += 4;
      } else 
	if (X < posY) {
	  b[  0] = *(a01 +  0);
	  b[  1] = *(a01 +  1);
	  b[  2] = *(a01 +  2);
	  b[  3] = *(a01 +  3);
	  
	  a01 += lda;
	  a02 += lda;
	  b += 4;
	} else {
	  b[  0] = *(a01 +  0);
	  b[  1] = ZERO;
	  b[  2] = *(a01 +  2);
	  b[  3] = *(a01 +  3);
	  b += 4;
	}
    }
    posY += 2;
  }

  if (n & 1){
    X = posX;
    
    if (posX <= posY) {
      a01 = a + posY * 2 + (posX +  0) * lda;
    } else {
      a01 = a + posX * 2 + (posY +  0) * lda;
    }
    
    i = m;
    if (i > 0) {
      do {
	if (X > posY) {
	  b[0] = *(a01 + 0);
	  b[1] = -*(a01 + 1);
	  a01 += 2;
	  b += 2;
	} else 
	  if (X < posY) {
	    b[0] = *(a01 + 0);
	    b[1] = *(a01 + 1);
	    a01 += lda;
	    b += 2;
	  } else {
	    b[0] = *(a01 + 0);
	    b[1] = ZERO;
	    a01 += 2;
	    b += 2;
	  }
	
	X += 1;
	i --;
      } while (i > 0);
    }
    posY += 1;
  }

}


