/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

const static FLOAT dp1 = 1.;

#ifdef CONJ
#define GEMM_KERNEL   GEMM_KERNEL_R
#define TRMM_KERNEL_N TRMM_KERNEL_RR
#define TRMM_KERNEL_T TRMM_KERNEL_RC
#else
#define GEMM_KERNEL   GEMM_KERNEL_N
#define TRMM_KERNEL_N TRMM_KERNEL_RN
#define TRMM_KERNEL_T TRMM_KERNEL_RT
#endif

int CNAME(long m, long n, long dummy1,
	  FLOAT dummyalpha1, 
#ifdef COMPLEX
	  FLOAT dummyalpha2,
#endif
	  FLOAT *a, long lda,
	  FLOAT *dummy3, long dummy4, 
	  FLOAT *b, long ldb, FLOAT *buffer){

  long ls, is, js;
  long min_l, min_i, min_j;
  long offset;
#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
  int start_ls, start_js;
#endif
#ifndef PPC440
  FLOAT *sa, *sb;

  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)((long)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_BUFFER_A_OFFSET * COMPSIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
#else
  extern FLOAT *sa, *sb;
#endif

#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
  start_ls = 0;
  while (start_ls < n) start_ls += GEMM_Q;
  start_ls -= GEMM_Q;

  start_js = 0;
  while (start_js < n) start_js += GEMM_R;
  start_js -= GEMM_R;
#endif

#if (!defined(TRANSA) && !defined(UPPER)) || (defined(TRANSA) && defined(UPPER))
  for(ls = 0; ls < n; ls += GEMM_Q){
#else
  for(ls = start_ls; ls >= 0; ls -= GEMM_Q){
#endif

    min_l = n - ls;
    if (min_l > GEMM_Q) min_l = GEMM_Q;
    
#if (!defined(TRANSA) && !defined(UPPER)) || (defined(TRANSA) && defined(UPPER))
    for(js = 0; js < n; js += GEMM_R){
#else
    for(js = start_js; js >= 0; js -= GEMM_R){
#endif
      min_j = n - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
#if (defined(UPPER) && !defined(TRANSA)) || (!defined(UPPER) && defined(TRANSA))
      
      if (js + GEMM_R > ls) {
	if (ls + GEMM_Q < js) {

#ifndef TRANSA
	  GEMM_ONCOPY(min_l, min_j, a + (ls + js * lda) * COMPSIZE, lda, sb);
#else
	  GEMM_OTCOPY(min_l, min_j, a + (js + ls * lda) * COMPSIZE, lda, sb);
#endif
	  
	  for(is = 0; is < m; is += GEMM_P){
	    min_i = m - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	    GEMM_ITCOPY(min_l, min_i, b + (is + ls * ldb) * COMPSIZE, ldb, sa);
	    
	    GEMM_KERNEL(min_i, min_j, min_l, dp1, 
#ifdef COMPLEX
			  ZERO,
#endif
			  sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	  }
	} else {
#ifndef TRANSA
	  TRMM_OUNCOPY(min_l, min_j, a, lda, ls, js, sb);
#else
	  TRMM_OLTCOPY(min_l, min_j, a, lda, ls, js, sb);
#endif
	  
	  for(is = 0; is < m; is += GEMM_P){
	    min_i = m - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	    GEMM_ITCOPY(min_l, min_i, b + (is + ls * ldb) * COMPSIZE, ldb, sa);
	    
	    offset = ls - js;

	    if (min_j > offset) {
	      FLOAT *aa = sa;
	      FLOAT *bb = sb;
	      FLOAT *cc = b + (is + js * ldb) * COMPSIZE;
	      long nn = min_j;
	      
	      if (offset > 0) {
		bb += offset * min_l * COMPSIZE;
		cc += offset * ldb   * COMPSIZE;
		nn -= offset;
		offset = 0;
	      }

	      if (nn > min_l + offset) {

		GEMM_KERNEL(min_i, nn - min_l - offset, min_l, dp1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    aa,
			    bb + (min_l + offset) * min_l * COMPSIZE,
			    cc + (min_l + offset) * ldb   * COMPSIZE, ldb); 
		
		nn = min_l + offset;
	      }
	      
	      TRMM_KERNEL_N(min_i, nn, min_l, dp1,
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb, offset); 
	    }


	  }
	}
      }
#else
     
      if (ls + GEMM_Q > js) {
	if (js + GEMM_R < ls) {

#ifndef TRANSA
	  GEMM_ONCOPY(min_l, min_j, a + (ls + js * lda) * COMPSIZE, lda, sb);
#else
	  GEMM_OTCOPY(min_l, min_j, a + (js + ls * lda) * COMPSIZE, lda, sb);
#endif

	  for(is = 0; is < m; is += GEMM_P){
	    min_i = m - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	    GEMM_ITCOPY(min_l, min_i, b + (is + ls * ldb) * COMPSIZE, ldb, sa);
	    
	    GEMM_KERNEL(min_i, min_j, min_l, dp1, 
#ifdef COMPLEX
			  ZERO,
#endif
			  sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	  } /* end of ls */
	} else {

#ifndef TRANSA
	  TRMM_OLNCOPY(min_l, min_j, a, lda, ls, js, sb);
#else
	  TRMM_OUTCOPY(min_l, min_j, a, lda, ls, js, sb);
#endif
	  
	  for(is = 0; is < m; is += GEMM_P){
	    min_i = m - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	    GEMM_ITCOPY(min_l, min_i, b + (is + ls * ldb) * COMPSIZE, ldb, sa);
	    
	    offset = ls - js;

	    if (min_l > -offset) {
	      FLOAT *aa = sa;
	      FLOAT *bb = sb;
	      FLOAT *cc = b + (is + js * ldb) * COMPSIZE;
	      long nn = min_j;

	      if (offset > 0) {

		GEMM_KERNEL(min_i, offset, min_l, dp1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb); 

		bb += offset * min_l * COMPSIZE;
		cc += offset * ldb   * COMPSIZE;
		nn -= offset;
		offset = 0;
	      }

	      TRMM_KERNEL_T(min_i, MIN(nn, min_l + offset), min_l, dp1,
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb, offset); 
	    }
	      
	  } /* end of ls */
	}

      }
#endif

    } /* end of is */
  } /* end of js */
  return 0;
}
