/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

const static FLOAT dm1 = -1.;

#ifdef CONJ
#define GEMM_KERNEL   GEMM_KERNEL_L
#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
#define TRSM_KERNEL   TRSM_KERNEL_LR
#else
#define TRSM_KERNEL   TRSM_KERNEL_LC
#endif
#else
#define GEMM_KERNEL   GEMM_KERNEL_N
#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
#define TRSM_KERNEL   TRSM_KERNEL_LN
#else
#define TRSM_KERNEL   TRSM_KERNEL_LT
#endif
#endif

int CNAME(long m, long n, long dummy1,
	  FLOAT dummyalpha1, 
#ifdef COMPLEX
	  FLOAT dummyalpha2,
#endif
	  FLOAT *a, long lda,
	  FLOAT *dummy3, long dummy4, 
	  FLOAT *b, long ldb, FLOAT *buffer){

  long ls, is, js;
  long min_l, min_i, min_j;
  long offset;
#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
  int start_ls, start_is;
#endif

#ifndef PPC440
  FLOAT *sa, *sb;
  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)((long)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_BUFFER_A_OFFSET * COMPSIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
#else
  extern FLOAT *sa, *sb;
#endif

#if (defined(TRANSA) && defined(UPPER)) || (!defined(TRANSA) && !defined(UPPER))
  for(ls = 0; ls < m; ls += GEMM_Q){
#else
  start_is = 0;
  while (start_is < m) start_is += GEMM_P;
  start_is -= GEMM_P;

  start_ls = 0;
  while (start_ls < m) start_ls += GEMM_Q;
  start_ls -= GEMM_Q;

  for(ls = start_ls; ls >= 0; ls -= GEMM_Q){
#endif

    min_l = m - ls;
    if (min_l > GEMM_Q) min_l = GEMM_Q;
    
    for(js = 0; js < n; js += GEMM_R){
      min_j = n - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
      GEMM_ONCOPY(min_l, min_j, b + (ls + js * ldb) * COMPSIZE, ldb, sb);

#if (defined(TRANSA) && defined(UPPER)) || (!defined(TRANSA) && !defined(UPPER))
	for(is = 0; is < m; is += GEMM_P){
#else
        for(is = start_is; is >= 0; is -= GEMM_P){
#endif
	  min_i = m - is;
	  if (min_i > GEMM_P) min_i = GEMM_P;

#if (defined(TRANSA) && defined(UPPER)) || (!defined(TRANSA) && !defined(UPPER))

	  /* Left Lower */
	  if (is + GEMM_P > ls) {

#ifndef TRANSA
	    TRSM_ILTCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, is - ls, sa);
#else
	    TRSM_IUNCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, is - ls, sa);
#endif

	    if (ls + GEMM_Q > is) {
	      offset = is - ls;

	      if (offset < 0) {
		TRSM_KERNEL(MIN(min_i + offset, min_l), min_j, min_l, dm1,
#ifdef COMPLEX
			    ZERO,
#endif
			    sa - min_l * offset * COMPSIZE, sb, 
			    b + (is - offset + js * ldb) * COMPSIZE, ldb, 0);
		
		if (min_i >  min_l - offset) {
		  GEMM_KERNEL(min_i - min_l + offset, min_j, min_l, dm1, 
#ifdef COMPLEX
				ZERO,
#endif
				sa + min_l * ( min_l - offset) * COMPSIZE, sb,
				b + (is - offset + min_l + js * ldb) * COMPSIZE, ldb); 
		}
	      } else {
		TRSM_KERNEL(MIN(min_i, min_l - offset), min_j, min_l, dm1,
#ifdef COMPLEX
			    ZERO,
#endif
			    sa, sb,
			    b + (is + js * ldb) * COMPSIZE, ldb, offset);
		
		if (min_i >  min_l - offset) {
		  GEMM_KERNEL(min_i - min_l + offset, min_j, min_l, dm1, 
#ifdef COMPLEX
				ZERO,
#endif
				sa + min_l * (min_l - offset) * COMPSIZE, sb,
				b  + ( min_l - offset + is + js * ldb) * COMPSIZE, ldb); 
		}
	      }
	    } else {
	      GEMM_KERNEL(min_i, min_j, min_l, dm1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	    }
	  }
#else

	  /* Left Upper */
	  if (ls + GEMM_Q > is) {

#ifndef TRANSA
	    TRSM_IUTCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, is - ls, sa);
#else
	    TRSM_ILNCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, is - ls, sa);
#endif

	    if (is + GEMM_P > ls) {
	      /* Partial calculation */
	      offset = is - ls;

	      if (offset >= 0) {
		TRSM_KERNEL(MIN(min_i, min_l - offset), min_j, min_l, dm1,
#ifdef COMPLEX
			    ZERO,
#endif
			    sa, sb, 
			    b +  (is + js * ldb) * COMPSIZE, ldb, offset);
	      } else {
		TRSM_KERNEL(MIN(min_i + offset, min_l), min_j, min_l, dm1,
#ifdef COMPLEX
			    ZERO,
#endif
			    sa - min_l * offset * COMPSIZE, sb, 
			    b + (is - offset + js * ldb) * COMPSIZE, ldb, 0);
		
		GEMM_KERNEL(-offset, min_j, min_l, dm1, 
#ifdef COMPLEX
			      ZERO,
#endif
			      sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 

	      }

	    } else {
	      GEMM_KERNEL(min_i, min_j, min_l, dm1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	    }
	  }
#endif
	}
    } 
  }

  return 0;
}
