/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"
 
#define M	ARG1
#define N	ARG2
#define K	ARG3
#define A	ARG4
#define B	ARG5
#define C	ARG6
#define LDC	%r10
	
#define I	%r11
#define J	%r12
#define AO	%r13
#define BO	%r14
#define	CO1	%r15
#define CO2	%rbp
/* rbp is reserved */
	
#ifdef OPTERON
#define movsd movlps
#endif

#ifdef OPTERON
#define PREFETCH     prefetch
#define PREFETCHSIZE (16 * 11 + 8)
#endif

#define LOCAL_BUFFER_SIZE  GEMM_Q * GEMM_UNROLL_N * COMPSIZE * 16

#define STACKSIZE 64
	
#define ALPHA	  0(%rsp)
#define OFFSET	 16(%rsp)
#define KK	 24(%rsp)
#define KKK	 32(%rsp)
#define AORIG	 40(%rsp)
#define BORIG	 48(%rsp)
#define BUFFER	128(%rsp)

#define KERNEL(xx) \
	mulps	%xmm8, %xmm9 ;\
	addps	%xmm9, %xmm0 ;\
	movaps	 0 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulps	%xmm8, %xmm11 ;\
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE + 1 * (xx) * SIZE(AO) ;\
	addps	%xmm11, %xmm1 ;\
	movaps	 4 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm8, %xmm13 ;\
	mulps	12 * SIZE + 2 * (xx) * SIZE(BO), %xmm8 ;\
	addps	%xmm13, %xmm2 ;\
	movaps	 8 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm8, %xmm3 ;\
	movaps	16 * SIZE + 1 * (xx) * SIZE(AO), %xmm8 ;\
	mulps	%xmm10, %xmm9 ;\
	addps	%xmm9, %xmm4 ;\
	movaps	32 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulps	%xmm10, %xmm11 ;\
	addps	%xmm11, %xmm5 ;\
	movaps	20 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm10, %xmm13 ;\
	mulps	12 * SIZE + 2 * (xx) * SIZE(BO), %xmm10 ;\
	addps	%xmm13, %xmm6 ;\
	movaps	24 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm10, %xmm7 ;\
	movaps	20 * SIZE + 1 * (xx) * SIZE(AO), %xmm10 ;\
	mulps	%xmm12, %xmm15 ;\
	addps	%xmm15, %xmm0 ;\
	movaps	16 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulps	%xmm12, %xmm11 ;\
	addps	%xmm11, %xmm1 ;\
	movaps	20 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm12, %xmm13 ;\
	mulps	28 * SIZE + 2 * (xx) * SIZE(BO), %xmm12 ;\
	addps	%xmm13, %xmm2 ;\
	movaps	24 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm12, %xmm3 ;\
	movaps	24 * SIZE + 1 * (xx) * SIZE(AO), %xmm12 ;\
	mulps	%xmm14, %xmm15 ;\
	addps	%xmm15, %xmm4 ;\
	movaps	48 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulps	%xmm14, %xmm11 ;\
	addps	%xmm11, %xmm5 ;\
	movaps	36 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm14, %xmm13 ;\
	mulps	28 * SIZE + 2 * (xx) * SIZE(BO), %xmm14 ;\
	addps	%xmm13, %xmm6 ;\
	movaps	40 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm14, %xmm7 ;\
	movaps	28 * SIZE + 1 * (xx) * SIZE(AO), %xmm14 ;\
	mulps	%xmm8, %xmm9 ;\
	addps	%xmm9, %xmm0 ;\
	movaps	32 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulps	%xmm8, %xmm11 ;\
	PREFETCH	(PREFETCHSIZE     +  8) * SIZE + 1 * (xx) * SIZE(AO) ;\
	addps	%xmm11, %xmm1 ;\
	movaps	36 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm8, %xmm13 ;\
	mulps	44 * SIZE + 2 * (xx) * SIZE(BO), %xmm8 ;\
	addps	%xmm13, %xmm2 ;\
	movaps	40 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm8, %xmm3 ;\
	movaps	32 * SIZE + 1 * (xx) * SIZE(AO), %xmm8 ;\
	mulps	%xmm10, %xmm9 ;\
	addps	%xmm9, %xmm4 ;\
	movaps	64 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulps	%xmm10, %xmm11 ;\
	addps	%xmm11, %xmm5 ;\
	movaps	52 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm10, %xmm13 ;\
	mulps	44 * SIZE + 2 * (xx) * SIZE(BO), %xmm10 ;\
	addps	%xmm13, %xmm6 ;\
	movaps	56 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm10, %xmm7 ;\
	movaps	36 * SIZE + 1 * (xx) * SIZE(AO), %xmm10 ;\
	mulps	%xmm12, %xmm15 ;\
	addps	%xmm15, %xmm0 ;\
	movaps	48 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulps	%xmm12, %xmm11 ;\
	addps	%xmm11, %xmm1 ;\
	movaps	52 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm12, %xmm13 ;\
	mulps	60 * SIZE + 2 * (xx) * SIZE(BO), %xmm12 ;\
	addps	%xmm13, %xmm2 ;\
	movaps	56 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm12, %xmm3 ;\
	movaps	40 * SIZE + 1 * (xx) * SIZE(AO), %xmm12 ;\
	mulps	%xmm14, %xmm15 ;\
	addps	%xmm15, %xmm4 ;\
	movaps	80 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulps	%xmm14, %xmm11 ;\
	addps	%xmm11, %xmm5 ;\
	movaps	68 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulps	%xmm14, %xmm13 ;\
	mulps	60 * SIZE + 2 * (xx) * SIZE(BO), %xmm14 ;\
	addps	%xmm13, %xmm6 ;\
	movaps	72 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addps	%xmm14, %xmm7 ;\
	movaps	44 * SIZE + 1 * (xx) * SIZE(AO), %xmm14

	PROLOGUE
	PROFCODE
	
	subq	$STACKSIZE, %rsp

	EMMS
	
	movq	%rbx,  0(%rsp)
	movq	%rbp,  8(%rsp)
	movq	%r12, 16(%rsp)
	movq	%r13, 24(%rsp)
	movq	%r14, 32(%rsp)
	movq	%r15, 40(%rsp)

	movq	72(%rsp), LDC
	movsd	80(%rsp), %xmm4

	movq	%rsp, %rbx	# save old stack
	subq	$128 + LOCAL_BUFFER_SIZE, %rsp
	andq	$-1024, %rsp	# align stack

	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0, ALPHA
	
	movsd	%xmm4, OFFSET
	movsd	%xmm4, KK

	leaq	(, LDC, SIZE), LDC

#ifdef LN
       leaq	(, M, SIZE), %rax
       addq	%rax, C
       imulq	K, %rax
       addq	%rax, A
#endif

#ifdef RT
       leaq	(, N, SIZE), %rax
       imulq	K, %rax
       addq	%rax, B
       movq	N, %rax
       imulq	LDC, %rax
       addq	%rax, C
#endif

#ifdef RN
	negq	KK
#endif	

#ifdef RT
       movq	N, %rax
       subq	OFFSET, %rax
       movq	%rax, KK
#endif

	movq	N,  J
	sarq	$2, J		# j = (n >> 2)
	jle	.L50

.L01:
/* Copying to Sub Buffer */

#ifdef LN
	movq	OFFSET, %rax
	addq	M, %rax
	movq	%rax, KK
#endif	

	leaq	BUFFER, BO
	
#ifdef RT
       movq	K, %rax
       salq	$2 + BASE_SHIFT, %rax
       subq	%rax, B
#endif

#if defined(LN) || defined(RT)
	movq	KK, %rax
	movq	B, BORIG
	leaq	(, %rax, SIZE), %rax
	leaq	(B,  %rax, 4), B
	leaq	(BO, %rax, 8), BO
#endif	

#ifdef LT
	movq	OFFSET, %rax
	movq	%rax, KK
#endif

#if defined(LT) || defined(RN)
	movq	KK, %rax
#else
	movq	K, %rax
	subq	KK, %rax
#endif
	sarq	$2, %rax
	jle	.L03

	addq	%rax, %rax
	ALIGN_4
	
.L02:
	movd	 0 * SIZE(B), %mm0
	movd	 1 * SIZE(B), %mm1
	movd	 2 * SIZE(B), %mm2
	movd	 3 * SIZE(B), %mm3
	movd	 4 * SIZE(B), %mm4
	movd	 5 * SIZE(B), %mm5
	movd	 6 * SIZE(B), %mm6
	movd	 7 * SIZE(B), %mm7

	prefetchnta	 56 * SIZE(B)

#define WPREFETCHSIZE 24

	prefetchw	 (WPREFETCHSIZE + 0) * SIZE(BO)
	prefetchw	 (WPREFETCHSIZE + 8) * SIZE(BO)

	punpckldq %mm0, %mm0
	punpckldq %mm1, %mm1
	punpckldq %mm2, %mm2
	punpckldq %mm3, %mm3
	punpckldq %mm4, %mm4
	punpckldq %mm5, %mm5
	punpckldq %mm6, %mm6
	punpckldq %mm7, %mm7
	
	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
	movq	%mm1,  4 * SIZE(BO)
	movq	%mm1,  6 * SIZE(BO)
	movq	%mm2,  8 * SIZE(BO)
	movq	%mm2, 10 * SIZE(BO)
	movq	%mm3, 12 * SIZE(BO)
	movq	%mm3, 14 * SIZE(BO)
	movq	%mm4, 16 * SIZE(BO)
	movq	%mm4, 18 * SIZE(BO)
	movq	%mm5, 20 * SIZE(BO)
	movq	%mm5, 22 * SIZE(BO)
	movq	%mm6, 24 * SIZE(BO)
	movq	%mm6, 26 * SIZE(BO)
	movq	%mm7, 28 * SIZE(BO)
	movq	%mm7, 30 * SIZE(BO)

	addq	$ 8 * SIZE, B
	addq	$32 * SIZE, BO

	decq	%rax
	jne	.L02
	ALIGN_4

.L03:
#if defined(LT) || defined(RN)
	movq	KK, %rax
#else
	movq	K, %rax
	subq	KK, %rax
#endif
	andq	$3, %rax
	BRANCH
	jle	.L10
	ALIGN_4

.L04:
	movd	 0 * SIZE(B), %mm0
	movd	 1 * SIZE(B), %mm1
	movd	 2 * SIZE(B), %mm2
	movd	 3 * SIZE(B), %mm3

	punpckldq %mm0, %mm0
	punpckldq %mm1, %mm1
	punpckldq %mm2, %mm2
	punpckldq %mm3, %mm3

	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
	movq	%mm1,  4 * SIZE(BO)
	movq	%mm1,  6 * SIZE(BO)
	movq	%mm2,  8 * SIZE(BO)
	movq	%mm2, 10 * SIZE(BO)
	movq	%mm3, 12 * SIZE(BO)
	movq	%mm3, 14 * SIZE(BO)

	addq	$ 4 * SIZE, B
	addq	$16 * SIZE, BO
	decq	%rax
	jne	.L04
	ALIGN_4
	
.L10:
#if defined(LT) || defined(RN)
	movq	A, AO
#else
	movq	A, AORIG
#endif

#ifdef RT
       leaq	(, LDC, 4), %rax
       subq	%rax, C
#endif

	movq	C, CO1			# coffset1 = c
	leaq	(C, LDC, 1), CO2	# coffset2 = c + ldc
#ifndef RT
	leaq	(C, LDC, 4), C
#endif

	movq	M,  I
	sarq	$3, I	# i = (m >> 3)
	jle	.L20
	ALIGN_4

.L11:
#ifdef LN
       movq	K, %rax
       salq	$2 + BASE_SHIFT, %rax
       subq	%rax, AORIG
#endif

#if defined(LN) || defined(RT)
	movq	KK, %rax
	movq	AORIG, AO
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
#endif

	leaq	BUFFER, BO

#if defined(LN) || defined(RT)
	movq	KK, %rax
	salq	$2 + BASE_SHIFT, %rax
	leaq	(BO, %rax, 2), BO
#endif	

	movaps	 0 * SIZE(BO), %xmm9
	movaps	 4 * SIZE(BO), %xmm11
	movaps	 8 * SIZE(BO), %xmm13
	movaps	16 * SIZE(BO), %xmm15

	movaps	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movaps	 4 * SIZE(AO), %xmm10
	pxor	%xmm1, %xmm1
	movaps	 8 * SIZE(AO), %xmm12
	pxor	%xmm2, %xmm2
	movaps	12 * SIZE(AO), %xmm14
	pxor	%xmm3, %xmm3

	prefetchw      8 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	prefetchw      8 * SIZE(CO2)
	pxor	%xmm5, %xmm5
	prefetchw      8 * SIZE(CO1, LDC, 2)
	pxor	%xmm6, %xmm6
	prefetchw      8 * SIZE(CO2, LDC, 2)
	pxor	%xmm7, %xmm7

#if defined(LT) || defined(RN)
	movq	KK, %rax
#else
	movq	K, %rax
	subq	KK, %rax
#endif
	andq	$-8, %rax
	salq	$4, %rax
	je	.L15
.L1X:	
	KERNEL(32 *  0)		# 0
	KERNEL(32 *  1)		# 1
	cmpq	$64 *  2, %rax
	jle	.L12
	KERNEL(32 *  2)		# 2
	KERNEL(32 *  3)		# 3
	cmpq	$64 *  4, %rax
	jle	.L12
	KERNEL(32 *  4)		# 4
	KERNEL(32 *  5)		# 5
	cmpq	$64 *  6, %rax
	jle	.L12
	KERNEL(32 *  6)		# 6
	KERNEL(32 *  7)		# 7
	cmpq	$64 *  8, %rax
	jle	.L12
	KERNEL(32 *  8)		# 8
	KERNEL(32 *  9)		# 9
	cmpq	$64 * 10, %rax
	jle	.L12
	KERNEL(32 * 10)		# 10
	KERNEL(32 * 11)		# 11
	cmpq	$64 * 12, %rax
	jle	.L12
	KERNEL(32 * 12)		# 12
	KERNEL(32 * 13)		# 13
	cmpq	$64 * 14, %rax
	jle	.L12
	KERNEL(32 * 14)		# 14
	KERNEL(32 * 15)		# 15
	addq	$ 64 * 8  * SIZE, AO
	addq	$128 * 8  * SIZE, BO
	subq	$128 * 8, %rax
	jg	.L1X
	jmp	.L15

.L12:
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 4), BO
	ALIGN_4

.L15:
#if defined(LT) || defined(RN)
	movq	KK, %rax
#else
	movq	K, %rax
	subq	KK, %rax
#endif
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L18
	ALIGN_4
.L16:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movaps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movaps	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm2
	movaps	 0 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm3
	movaps	 8 * SIZE(AO), %xmm8
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm4
	movaps	 4 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm5
	movaps	 8 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	mulps	12 * SIZE(BO), %xmm10
	addps	%xmm9, %xmm6
	movaps	16 * SIZE(BO), %xmm9
	addps	%xmm10, %xmm7
	movaps	12 * SIZE(AO), %xmm10

	addq	$8  * SIZE, AO		# aoffset  += 4
	addq	$16 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L16
	ALIGN_4

.L18:
#if defined(LN) || defined(RT)
	movq	KK, %rax
#ifdef LN
	subq	$4, %rax
#else
	subq	$4, %rax
#endif

	movq	AORIG, AO
	movq	BORIG, B
	leaq	BUFFER, BO

	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(B,  %rax, 4), B
	leaq	(BO, %rax, 8), BO
#endif

#if defined(LN) || defined(LT)
	movaps	%xmm0, %xmm8
	movaps	%xmm2, %xmm9
	movaps	%xmm1, %xmm10
	movaps	%xmm3, %xmm11

	unpcklps %xmm4, %xmm0
	unpcklps %xmm6, %xmm2
	unpcklps %xmm5, %xmm1
	unpcklps %xmm7, %xmm3

	unpckhps %xmm4, %xmm8
	unpckhps %xmm6, %xmm9
	unpckhps %xmm5, %xmm10
	unpckhps %xmm7, %xmm11

	movaps	 %xmm0, %xmm6
	movaps	 %xmm8, %xmm4
	movaps	 %xmm1, %xmm7
	movaps	 %xmm10, %xmm5

	unpcklps %xmm2, %xmm0
	unpcklps %xmm9, %xmm8
	unpcklps %xmm3, %xmm1
	unpcklps %xmm11, %xmm10

	unpckhps %xmm2, %xmm6
	unpckhps %xmm9, %xmm4
	unpckhps %xmm3, %xmm7
	unpckhps %xmm11, %xmm5

	movaps	 0 * SIZE(B), %xmm2
	movaps	 4 * SIZE(B), %xmm3
	movaps	 8 * SIZE(B), %xmm9
	movaps	12 * SIZE(B), %xmm11
	movaps	16 * SIZE(B), %xmm12
	movaps	20 * SIZE(B), %xmm13
	movaps	24 * SIZE(B), %xmm14
	movaps	28 * SIZE(B), %xmm15

	subps	%xmm0,  %xmm2
	subps	%xmm6,  %xmm3
	subps	%xmm8,  %xmm9
	subps	%xmm4,  %xmm11
	subps	%xmm1,  %xmm12
	subps	%xmm7,  %xmm13
	subps	%xmm10, %xmm14
	subps	%xmm5,  %xmm15
#else
	movaps	 0 * SIZE(AO), %xmm8
	movaps	 2 * SIZE(AO), %xmm9
	movaps	 4 * SIZE(AO), %xmm10
	movaps	 6 * SIZE(AO), %xmm11

	movaps	 8 * SIZE(AO), %xmm12
	movaps	10 * SIZE(AO), %xmm13
	movaps	12 * SIZE(AO), %xmm14
	movaps	14 * SIZE(AO), %xmm15

	subps	%xmm0, %xmm8
	subps	%xmm4, %xmm9
	subps	%xmm1, %xmm10
	subps	%xmm5, %xmm11
	subps	%xmm2, %xmm12
	subps	%xmm6, %xmm13
	subps	%xmm3, %xmm14
	subps	%xmm7, %xmm15
#endif

#ifdef LN
	movss	15 * SIZE(AO), %xmm0
	shufps	$0, %xmm0, %xmm0
	mulps	 %xmm0, %xmm13
	mulps	 %xmm0, %xmm15

	movss	14 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm13, %xmm2
	subps	 %xmm2, %xmm9
	movss	14 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm15, %xmm2
	subps	 %xmm2, %xmm11

	movss	13 * SIZE(AO), %xmm4
	shufps	$0, %xmm4, %xmm4
	mulps	 %xmm13, %xmm4
	subps	 %xmm4, %xmm5
	movss	13 * SIZE(AO), %xmm4
	shufps	$0, %xmm4, %xmm4
	mulps	 %xmm15, %xmm4
	subps	 %xmm4, %xmm7

	movss	12 * SIZE(AO), %xmm6
	shufps	$0, %xmm6, %xmm6
	mulps	 %xmm13, %xmm6
	subps	 %xmm6, %xmm1
	movss	12 * SIZE(AO), %xmm6
	shufps	$0, %xmm6, %xmm6
	mulps	 %xmm15, %xmm6
	subps	 %xmm6, %xmm3

	movss	10 * SIZE(AO), %xmm0
	shufps	$0, %xmm0, %xmm0
	mulps	 %xmm0, %xmm9
	mulps	 %xmm0, %xmm11

	movss	 9 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm9, %xmm2
	subps	 %xmm2, %xmm5
	movss	 9 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm11, %xmm2
	subps	 %xmm2, %xmm7

	movss	 8 * SIZE(AO), %xmm4
	shufps	$0, %xmm4, %xmm4
	mulps	 %xmm9, %xmm4
	subps	 %xmm4, %xmm1
	movss	 8 * SIZE(AO), %xmm4
	shufps	$0, %xmm4, %xmm4
	mulps	 %xmm11, %xmm4
	subps	 %xmm4, %xmm3

	movss	 5 * SIZE(AO), %xmm0
	shufps	$0, %xmm0, %xmm0
	mulps	 %xmm0, %xmm5
	mulps	 %xmm0, %xmm7

	movss	 4 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm5, %xmm2
	subps	 %xmm2, %xmm1
	movss	 4 * SIZE(AO), %xmm2
	shufps	$0, %xmm2, %xmm2
	mulps	 %xmm7, %xmm2
	subps	 %xmm2, %xmm3

	movss	 0 * SIZE(AO), %xmm0
	shufps	$0, %xmm0, %xmm0
	mulps	 %xmm0, %xmm1
	mulps	 %xmm0, %xmm3
#endif

#ifdef LT
	movss	 0 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm2

	movss	 1 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm3

	movss	 2 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm9

	movss	 3 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm11

	movss	 4 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm12

	movss	 5 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm13

	movss	 6 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm14

	movss	 7 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm2, %xmm8
	subps	 %xmm8, %xmm15


	movss	 9 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm3

	movss	10 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm9

	movss	11 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm11

	movss	12 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm12

	movss	13 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm13

	movss	14 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm14

	movss	15 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm3, %xmm8
	subps	 %xmm8, %xmm15


	movss	18 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm9

	movss	19 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm9, %xmm8
	subps	 %xmm8, %xmm11

	movss	20 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm9, %xmm8
	subps	 %xmm8, %xmm12

	movss	21 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm9, %xmm8
	subps	 %xmm8, %xmm13

	movss	22 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm9, %xmm8
	subps	 %xmm8, %xmm14

	movss	23 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm9, %xmm8
	subps	 %xmm8, %xmm15


	movss	27 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm11

	movss	28 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm11, %xmm8
	subps	 %xmm8, %xmm12

	movss	29 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm11, %xmm8
	subps	 %xmm8, %xmm13

	movss	30 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm11, %xmm8
	subps	 %xmm8, %xmm14

	movss	31 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm11, %xmm8
	subps	 %xmm8, %xmm15


	movss	36 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm12

	movss	37 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm12, %xmm8
	subps	 %xmm8, %xmm13

	movss	38 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm12, %xmm8
	subps	 %xmm8, %xmm14

	movss	39 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm12, %xmm8
	subps	 %xmm8, %xmm15

	movss	45 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm13

	movss	46 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm13, %xmm8
	subps	 %xmm8, %xmm14

	movss	47 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm13, %xmm8
	subps	 %xmm8, %xmm15

	movss	54 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm14

	movss	55 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm14, %xmm8
	subps	 %xmm8, %xmm15

	movss	63 * SIZE(AO), %xmm8
	shufps	 $0, %xmm8, %xmm8
	mulps	 %xmm8, %xmm15
#endif

#ifdef RN
	movlps	 0 * SIZE(B), %xmm0
	movhps	 0 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm8
	mulps	 %xmm0, %xmm9

	movlps	 2 * SIZE(B), %xmm1
	movhps	 2 * SIZE(B), %xmm1
	mulps	 %xmm8, %xmm1
	subps	 %xmm1, %xmm10
	movlps	 2 * SIZE(B), %xmm1
	movhps	 2 * SIZE(B), %xmm1
	mulps	 %xmm9, %xmm1
	subps	 %xmm1, %xmm11

	movlps	 4 * SIZE(B), %xmm2
	movhps	 4 * SIZE(B), %xmm2
	mulps	 %xmm8, %xmm2
	subps	 %xmm2, %xmm12
	movlps	 4 * SIZE(B), %xmm2
	movhps	 4 * SIZE(B), %xmm2
	mulps	 %xmm9, %xmm2
	subps	 %xmm2, %xmm13

	movlps	 6 * SIZE(B), %xmm3
	movhps	 6 * SIZE(B), %xmm3
	mulps	 %xmm8, %xmm3
	subps	 %xmm3, %xmm14
	movlps	 6 * SIZE(B), %xmm3
	movhps	 6 * SIZE(B), %xmm3
	mulps	 %xmm9, %xmm3
	subps	 %xmm3, %xmm15

	movlps	10 * SIZE(B), %xmm0
	movhps	10 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm10
	mulps	 %xmm0, %xmm11

	movlps	12 * SIZE(B), %xmm1
	movhps	12 * SIZE(B), %xmm1
	mulps	 %xmm10, %xmm1
	subps	 %xmm1, %xmm12
	movlps	12 * SIZE(B), %xmm1
	movhps	12 * SIZE(B), %xmm1
	mulps	 %xmm11, %xmm1
	subps	 %xmm1, %xmm13

	movlps	14 * SIZE(B), %xmm2
	movhps	14 * SIZE(B), %xmm2
	mulps	 %xmm10, %xmm2
	subps	 %xmm2, %xmm14
	movlps	14 * SIZE(B), %xmm2
	movhps	14 * SIZE(B), %xmm2
	mulps	 %xmm11, %xmm2
	subps	 %xmm2, %xmm15

	movlps	20 * SIZE(B), %xmm0
	movhps	20 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm12
	mulps	 %xmm0, %xmm13

	movlps	22 * SIZE(B), %xmm1
	movhps	22 * SIZE(B), %xmm1
	mulps	 %xmm12, %xmm1
	subps	 %xmm1, %xmm14
	movlps	22 * SIZE(B), %xmm1
	movhps	22 * SIZE(B), %xmm1
	mulps	 %xmm13, %xmm1
	subps	 %xmm1, %xmm15

	movlps	30 * SIZE(B), %xmm0
	movhps	30 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm14
	mulps	 %xmm0, %xmm15
#endif

#ifdef RT
	movlps	15 * SIZE(B), %xmm0
	movhps	15 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm14
	mulps	 %xmm0, %xmm15

	movlps	14 * SIZE(B), %xmm1
	movhps	14 * SIZE(B), %xmm1
	mulps	 %xmm14, %xmm1
	subps	 %xmm1, %xmm12
	movlps	14 * SIZE(B), %xmm1
	movhps	14 * SIZE(B), %xmm1
	mulps	 %xmm15, %xmm1
	subps	 %xmm1, %xmm13

	movlps	13 * SIZE(B), %xmm2
	movhps	13 * SIZE(B), %xmm2
	mulps	 %xmm14, %xmm2
	subps	 %xmm2, %xmm10
	movlps	13 * SIZE(B), %xmm2
	movhps	13 * SIZE(B), %xmm2
	mulps	 %xmm15, %xmm2
	subps	 %xmm2, %xmm11

	movlps	12 * SIZE(B), %xmm3
	movhps	12 * SIZE(B), %xmm3
	mulps	 %xmm14, %xmm3
	subps	 %xmm3, %xmm8
	movlps	12 * SIZE(B), %xmm3
	movhps	12 * SIZE(B), %xmm3
	mulps	 %xmm15, %xmm3
	subps	 %xmm3, %xmm9

	movlps	10 * SIZE(B), %xmm0
	movhps	10 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm12
	mulps	 %xmm0, %xmm13

	movlps	 9 * SIZE(B), %xmm1
	movhps	 9 * SIZE(B), %xmm1
	mulps	 %xmm12, %xmm1
	subps	 %xmm1, %xmm10
	movlps	 9 * SIZE(B), %xmm1
	movhps	 9 * SIZE(B), %xmm1
	mulps	 %xmm13, %xmm1
	subps	 %xmm1, %xmm11

	movlps	 8 * SIZE(B), %xmm2
	movhps	 8 * SIZE(B), %xmm2
	mulps	 %xmm12, %xmm2
	subps	 %xmm2, %xmm8
	movlps	 8 * SIZE(B), %xmm2
	movhps	 8 * SIZE(B), %xmm2
	mulps	 %xmm13, %xmm2
	subps	 %xmm2, %xmm9

	movlps	 5 * SIZE(B), %xmm0
	movhps	 5 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm10
	mulps	 %xmm0, %xmm11

	movlps	 4 * SIZE(B), %xmm1
	movhps	 4 * SIZE(B), %xmm1
	mulps	 %xmm10, %xmm1
	subps	 %xmm1, %xmm8
	movlps	 4 * SIZE(B), %xmm1
	movhps	 4 * SIZE(B), %xmm1
	mulps	 %xmm11, %xmm1
	subps	 %xmm1, %xmm9

	movlps	 0 * SIZE(B), %xmm0
	movhps	 0 * SIZE(B), %xmm0
	mulps	 %xmm0, %xmm8
	mulps	 %xmm0, %xmm9
#endif

#ifdef LN
	subq	$8 * SIZE, CO1
	subq	$8 * SIZE, CO2
#endif

#if defined(LN) || defined(LT)
	movaps	%xmm2,   0 * SIZE(B)
	movaps	%xmm12,  4 * SIZE(B)
	movaps	%xmm9,   8 * SIZE(B)
	movaps	%xmm14, 12 * SIZE(B)
	movaps	%xmm3,  16 * SIZE(B)
	movaps	%xmm13, 20 * SIZE(B)
	movaps	%xmm11, 24 * SIZE(B)
	movaps	%xmm15, 28 * SIZE(B)

	movss	 0 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,   0 * SIZE(BO)
	movss	 1 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,   4 * SIZE(BO)
	movss	 2 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,   8 * SIZE(BO)
	movss	 3 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  12 * SIZE(BO)
	movss	 4 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  16 * SIZE(BO)
	movss	 5 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  20 * SIZE(BO)
	movss	 6 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  24 * SIZE(BO)
	movss	 7 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  28 * SIZE(BO)

	movss	 8 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  32 * SIZE(BO)
	movss	 9 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  36 * SIZE(BO)
	movss	10 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  40 * SIZE(BO)
	movss	11 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  44 * SIZE(BO)
	movss	12 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  48 * SIZE(BO)
	movss	13 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  52 * SIZE(BO)
	movss	14 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  56 * SIZE(BO)
	movss	15 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  60 * SIZE(BO)

	movss	16 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  64 * SIZE(BO)
	movss	17 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  68 * SIZE(BO)
	movss	18 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  72 * SIZE(BO)
	movss	19 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  76 * SIZE(BO)
	movss	20 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  80 * SIZE(BO)
	movss	21 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  84 * SIZE(BO)
	movss	22 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  88 * SIZE(BO)
	movss	23 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  92 * SIZE(BO)

	movss	24 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  96 * SIZE(BO)
	movss	25 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  100 * SIZE(BO)
	movss	26 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  104 * SIZE(BO)
	movss	27 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  108 * SIZE(BO)
	movss	28 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  112 * SIZE(BO)
	movss	29 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  116 * SIZE(BO)
	movss	30 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  120 * SIZE(BO)
	movss	31 * SIZE(B), %xmm0
	shufps	$0, %xmm0, %xmm0
	movaps	%xmm0,  124 * SIZE(BO)


#if 0
	movlps	%xmm1,   1 * SIZE(BO)
	movhps	%xmm1,   2 * SIZE(BO)
	movhps	%xmm1,   3 * SIZE(BO)
	movlps	%xmm3,   4 * SIZE(BO)
	movlps	%xmm3,   5 * SIZE(BO)
	movhps	%xmm3,   6 * SIZE(BO)
	movhps	%xmm3,   7 * SIZE(BO)
	movlps	%xmm5,   8 * SIZE(BO)
	movlps	%xmm5,   9 * SIZE(BO)
	movhps	%xmm5,  10 * SIZE(BO)
	movhps	%xmm5,  11 * SIZE(BO)
	movlps	%xmm7,  12 * SIZE(BO)
	movlps	%xmm7,  13 * SIZE(BO)
	movhps	%xmm7,  14 * SIZE(BO)
	movhps	%xmm7,  15 * SIZE(BO)
	movlps	%xmm9,  16 * SIZE(BO)
	movlps	%xmm9,  17 * SIZE(BO)
	movhps	%xmm9,  18 * SIZE(BO)
	movhps	%xmm9,  19 * SIZE(BO)
	movlps	%xmm11, 20 * SIZE(BO)
	movlps	%xmm11, 21 * SIZE(BO)
	movhps	%xmm11, 22 * SIZE(BO)
	movhps	%xmm11, 23 * SIZE(BO)
	movlps	%xmm13, 24 * SIZE(BO)
	movlps	%xmm13, 25 * SIZE(BO)
	movhps	%xmm13, 26 * SIZE(BO)
	movhps	%xmm13, 27 * SIZE(BO)
	movlps	%xmm15, 28 * SIZE(BO)
	movlps	%xmm15, 29 * SIZE(BO)
	movhps	%xmm15, 30 * SIZE(BO)
	movhps	%xmm15, 31 * SIZE(BO)
#endif

#else
	movaps	%xmm8,   0 * SIZE(AO)
	movaps	%xmm9,   2 * SIZE(AO)
	movaps	%xmm10,  4 * SIZE(AO)
	movaps	%xmm11,  6 * SIZE(AO)
	movaps	%xmm12,  8 * SIZE(AO)
	movaps	%xmm13, 10 * SIZE(AO)
	movaps	%xmm14, 12 * SIZE(AO)
	movaps	%xmm15, 14 * SIZE(AO)
#endif

#if defined(LN) || defined(LT)
	movaps	 %xmm2,  %xmm0
	movaps	 %xmm3,  %xmm1
	movaps	 %xmm12, %xmm4
	movaps	 %xmm13, %xmm5

	unpcklps %xmm9,  %xmm2
	unpcklps %xmm11, %xmm3
	unpcklps %xmm14, %xmm12
	unpcklps %xmm15, %xmm13

	unpckhps %xmm9,  %xmm0
	unpckhps %xmm11, %xmm1
	unpckhps %xmm14, %xmm4
	unpckhps %xmm15, %xmm5

	movaps	 %xmm2,  %xmm11
	movaps	 %xmm0,  %xmm9
	movaps	 %xmm12, %xmm15
	movaps	 %xmm4,  %xmm14

	unpcklps %xmm3,  %xmm2
	unpcklps %xmm1,  %xmm0
	unpcklps %xmm13, %xmm12
	unpcklps %xmm5,  %xmm4

	unpckhps %xmm3,  %xmm11
	unpckhps %xmm1,  %xmm9
	unpckhps %xmm13, %xmm15
	unpckhps %xmm5,  %xmm14

	movsd	%xmm2,  0 * SIZE(CO1)
	movhps	%xmm2,  2 * SIZE(CO1)
	movsd	%xmm12, 4 * SIZE(CO1)
	movhps	%xmm12, 6 * SIZE(CO1)

	movsd	%xmm11, 0 * SIZE(CO2)
	movhps	%xmm11, 2 * SIZE(CO2)
	movsd	%xmm15, 4 * SIZE(CO2)
	movhps	%xmm15, 6 * SIZE(CO2)

	movsd	%xmm0,  0 * SIZE(CO1, LDC, 2)
	movhps	%xmm0,  2 * SIZE(CO1, LDC, 2)
	movsd	%xmm4,  4 * SIZE(CO1, LDC, 2)
	movhps	%xmm4,  6 * SIZE(CO1, LDC, 2)

	movsd	%xmm9,  0 * SIZE(CO2, LDC, 2)
	movhps	%xmm9,  2 * SIZE(CO2, LDC, 2)
	movsd	%xmm14, 4 * SIZE(CO2, LDC, 2)
	movhps	%xmm14, 6 * SIZE(CO2, LDC, 2)
#else
	movsd	%xmm8,  0 * SIZE(CO1)
	movhps	%xmm8,  1 * SIZE(CO1)
	movsd	%xmm9,  2 * SIZE(CO1)
	movhps	%xmm9,  3 * SIZE(CO1)

	movsd	%xmm10,  0 * SIZE(CO2)
	movhps	%xmm10,  1 * SIZE(CO2)
	movsd	%xmm11,  2 * SIZE(CO2)
	movhps	%xmm11,  3 * SIZE(CO2)

	movsd	%xmm12,  0 * SIZE(CO1, LDC, 2)
	movhps	%xmm12,  1 * SIZE(CO1, LDC, 2)
	movsd	%xmm13,  2 * SIZE(CO1, LDC, 2)
	movhps	%xmm13,  3 * SIZE(CO1, LDC, 2)

	movsd	%xmm14,  0 * SIZE(CO2, LDC, 2)
	movhps	%xmm14,  1 * SIZE(CO2, LDC, 2)
	movsd	%xmm15,  2 * SIZE(CO2, LDC, 2)
	movhps	%xmm15,  3 * SIZE(CO2, LDC, 2)
#endif

#ifndef LN
	addq	$8 * SIZE, CO1
	addq	$8 * SIZE, CO2
#endif

#if defined(LT) || defined(RN)
	movq	K,  %rax
	subq	KK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO			###
#ifdef LT
	addq	$32 * SIZE, B
#endif
#endif

#ifdef LN
	subq	$8, KK
	movq	BORIG, B
#endif

#ifdef LT
	addq	$8, KK
#endif

#ifdef RT
	movq	K, %rax
	movq	BORIG, B
	salq	$2 + BASE_SHIFT, %rax
	addq	%rax, AORIG
#endif

	decq	I			# i --
	jg	.L11
	ALIGN_4	

.L20:
	testq	$4, M
	je	.L30

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	16 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L25
	ALIGN_4

.L22:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movaps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movaps	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm2
	movaps	64 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm3
	movaps	 4 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm0
	movaps	20 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm1
	movaps	24 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	mulps	28 * SIZE(BO), %xmm8
	addps	%xmm11, %xmm2
	movaps	80 * SIZE(BO), %xmm11
	addps	%xmm8, %xmm3
	movaps	 8 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm13
	addps	%xmm13, %xmm0
	movaps	36 * SIZE(BO), %xmm13
	mulps	%xmm8, %xmm13
	addps	%xmm13, %xmm1
	movaps	40 * SIZE(BO), %xmm13
	mulps	%xmm8, %xmm13
	mulps	44 * SIZE(BO), %xmm8
	addps	%xmm13, %xmm2
	movaps	96 * SIZE(BO), %xmm13
	addps	%xmm8, %xmm3
	movaps	12 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm15
	addps	%xmm15, %xmm0
	movaps	52 * SIZE(BO), %xmm15
	mulps	%xmm8, %xmm15
	addps	%xmm15, %xmm1
	movaps	56 * SIZE(BO), %xmm15
	mulps	%xmm8, %xmm15
	mulps	60 * SIZE(BO), %xmm8
	addps	%xmm15, %xmm2
	movaps	112 * SIZE(BO), %xmm15
	addps	%xmm8, %xmm3
	movaps	32 * SIZE(AO), %xmm8

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 16) * SIZE(AO)
#endif
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm0
	movaps	68 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm1
	movaps	72 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	mulps	76 * SIZE(BO), %xmm10
	addps	%xmm9, %xmm2
	movaps	128 * SIZE(BO), %xmm9
	addps	%xmm10, %xmm3
	movaps	20 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm11
	addps	%xmm11, %xmm0
	movaps	84 * SIZE(BO), %xmm11
	mulps	%xmm10, %xmm11
	addps	%xmm11, %xmm1
	movaps	88 * SIZE(BO), %xmm11
	mulps	%xmm10, %xmm11
	mulps	92 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm2
	movaps	144 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm3
	movaps	24 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm0
	movaps	100 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm1
	movaps	104 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	mulps	108 * SIZE(BO), %xmm10
	addps	%xmm13, %xmm2
	movaps	160 * SIZE(BO), %xmm13
	addps	%xmm10, %xmm3
	movaps	28 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm0
	movaps	116 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm1
	movaps	120 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	mulps	124 * SIZE(BO), %xmm10
	addps	%xmm15, %xmm2
	movaps	176 * SIZE(BO), %xmm15
	addps	%xmm10, %xmm3
	movaps	48 * SIZE(AO), %xmm10

	addq   $ 32 * SIZE, AO
	addq   $128 * SIZE, BO
	decq   %rax
	jne    .L22
	ALIGN_4

.L25:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L28
	ALIGN_4

.L26:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movaps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movaps	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm2
	movaps	16 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm3
	movaps	 4 * SIZE(AO), %xmm8

	addq	$ 4 * SIZE, AO		# aoffset  += 4
	addq	$16 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L26
	ALIGN_4

.L28:
	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm1
 	mulps	%xmm15, %xmm2
	mulps	%xmm15, %xmm3

	movlps	0 * SIZE(CO1), %xmm8
	movhps	2 * SIZE(CO1), %xmm8
	movlps	0 * SIZE(CO2), %xmm10
	movhps	2 * SIZE(CO2), %xmm10

	movlps	0 * SIZE(CO1, LDC, 2), %xmm12
	movhps	2 * SIZE(CO1, LDC, 2), %xmm12
	movlps	0 * SIZE(CO2, LDC, 2), %xmm14
	movhps	2 * SIZE(CO2, LDC, 2), %xmm14

	addps	%xmm8,  %xmm0
	addps	%xmm10, %xmm1
	addps	%xmm12, %xmm2
	addps	%xmm14, %xmm3

	movlps	%xmm0, 0 * SIZE(CO1)
	movhps	%xmm0, 2 * SIZE(CO1)
	movlps	%xmm1, 0 * SIZE(CO2)
	movhps	%xmm1, 2 * SIZE(CO2)

	movlps	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movhps	%xmm2, 2 * SIZE(CO1, LDC, 2)
	movlps	%xmm3, 0 * SIZE(CO2, LDC, 2)
	movhps	%xmm3, 2 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	addq	$4 * SIZE, CO2		# coffset += 4
	ALIGN_4	

.L30:
	testq	$2, M
	je	.L40

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	 8 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L35
	ALIGN_4

.L32:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movlps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movlps	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm2
	movlps	12 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm3
	movlps	64 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm0
	movlps	20 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm1
	movlps	24 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm2
	movlps	28 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	movlps	 4 * SIZE(AO), %xmm8
	addps	%xmm11, %xmm3
	movlps	80 * SIZE(BO), %xmm11

	mulps	%xmm8, %xmm13
	addps	%xmm13, %xmm0
	movlps	36 * SIZE(BO), %xmm13
	mulps	%xmm8, %xmm13
	addps	%xmm13, %xmm1
	movlps	40 * SIZE(BO), %xmm13
	mulps	%xmm8, %xmm13
	addps	%xmm13, %xmm2
	movlps	44 * SIZE(BO), %xmm13
	mulps	%xmm8, %xmm13
	movlps	 6 * SIZE(AO), %xmm8
	addps	%xmm13, %xmm3
	movlps	96 * SIZE(BO), %xmm13

	mulps	%xmm8, %xmm15
	addps	%xmm15, %xmm0
	movlps	52 * SIZE(BO), %xmm15
	mulps	%xmm8, %xmm15
	addps	%xmm15, %xmm1
	movlps	56 * SIZE(BO), %xmm15
	mulps	%xmm8, %xmm15
	addps	%xmm15, %xmm2
	movlps	60 * SIZE(BO), %xmm15
	mulps	%xmm8, %xmm15
	movlps	16 * SIZE(AO), %xmm8
	addps	%xmm15, %xmm3
	movlps	112 * SIZE(BO), %xmm15

	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm0
	movlps	68 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm1
	movlps	72 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	addps	%xmm9, %xmm2
	movlps	76 * SIZE(BO), %xmm9
	mulps	%xmm10, %xmm9
	movlps	10 * SIZE(AO), %xmm10
	addps	%xmm9, %xmm3
	movlps	128 * SIZE(BO), %xmm9

	mulps	%xmm10, %xmm11
	addps	%xmm11, %xmm0
	movlps	84 * SIZE(BO), %xmm11
	mulps	%xmm10, %xmm11
	addps	%xmm11, %xmm1
	movlps	88 * SIZE(BO), %xmm11
	mulps	%xmm10, %xmm11
	addps	%xmm11, %xmm2
	movlps	92 * SIZE(BO), %xmm11
	mulps	%xmm10, %xmm11
	movlps	12 * SIZE(AO), %xmm10
	addps	%xmm11, %xmm3
	movlps	144 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm0
	movlps	100 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm1
	movlps	104 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm2
	movlps	108 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	movlps	14 * SIZE(AO), %xmm10
	addps	%xmm13, %xmm3
	movlps	160 * SIZE(BO), %xmm13

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm0
	movlps	116 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm1
	movlps	120 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm2
	movlps	124 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	movlps	24 * SIZE(AO), %xmm10
	addps	%xmm15, %xmm3
	movlps	176 * SIZE(BO), %xmm15

	addq   $ 16 * SIZE, AO
	addq   $128 * SIZE, BO
	decq   %rax
	jne    .L32
	ALIGN_4

.L35:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L38
	ALIGN_4

.L36:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movlps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movlps	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm2
	movlps	12 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm3
	movlps	16 * SIZE(BO), %xmm9

	addq	$ 2 * SIZE, AO		# aoffset  += 4
	addq	$16 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L36
	ALIGN_4

.L38:
	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm1
 	mulps	%xmm15, %xmm2
	mulps	%xmm15, %xmm3

	movlps	0 * SIZE(CO1), %xmm8
	movlps	0 * SIZE(CO2), %xmm10
	movlps	0 * SIZE(CO1, LDC, 2), %xmm12
	movlps	0 * SIZE(CO2, LDC, 2), %xmm14

	addps	%xmm8,  %xmm0
	addps	%xmm10, %xmm1
	addps	%xmm12, %xmm2
	addps	%xmm14, %xmm3

	movlps	%xmm0, 0 * SIZE(CO1)
	movlps	%xmm1, 0 * SIZE(CO2)
	movlps	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movlps	%xmm3, 0 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif

	addq	$2 * SIZE, CO1		# coffset += 4
	addq	$2 * SIZE, CO2		# coffset += 4
	ALIGN_4	

.L40:
	testq	$1, M
	je	.L49

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   4), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
	leaq	(BO, %rax, 8), BO
#endif	

	movss	 0 * SIZE(AO), %xmm8
	movss	 4 * SIZE(AO), %xmm10

	movss	 0 * SIZE(BO), %xmm9
	movss	16 * SIZE(BO), %xmm11
	movss	32 * SIZE(BO), %xmm13
	movss	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L45
	ALIGN_4

.L42:
	mulss	%xmm8, %xmm9
	addss	%xmm9, %xmm0
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movss	 4 * SIZE(BO), %xmm9
	mulss	%xmm8, %xmm9
	addss	%xmm9, %xmm1
	movss	 8 * SIZE(BO), %xmm9
	mulss	%xmm8, %xmm9
	addss	%xmm9, %xmm2
	movss	12 * SIZE(BO), %xmm9
	mulss	%xmm8, %xmm9
	movss	 1 * SIZE(AO), %xmm8
	addss	%xmm9, %xmm3
	movss	64 * SIZE(BO), %xmm9

	mulss	%xmm8, %xmm11
	addss	%xmm11, %xmm0
	movss	20 * SIZE(BO), %xmm11
	mulss	%xmm8, %xmm11
	addss	%xmm11, %xmm1
	movss	24 * SIZE(BO), %xmm11
	mulss	%xmm8, %xmm11
	addss	%xmm11, %xmm2
	movss	28 * SIZE(BO), %xmm11
	mulss	%xmm8, %xmm11
	movss	 2 * SIZE(AO), %xmm8
	addss	%xmm11, %xmm3
	movss	80 * SIZE(BO), %xmm11

	mulss	%xmm8, %xmm13
	addss	%xmm13, %xmm0
	movss	36 * SIZE(BO), %xmm13
	mulss	%xmm8, %xmm13
	addss	%xmm13, %xmm1
	movss	40 * SIZE(BO), %xmm13
	mulss	%xmm8, %xmm13
	addss	%xmm13, %xmm2
	movss	44 * SIZE(BO), %xmm13
	mulss	%xmm8, %xmm13
	movss	 3 * SIZE(AO), %xmm8
	addss	%xmm13, %xmm3
	movss	96 * SIZE(BO), %xmm13

	mulss	%xmm8, %xmm15
	addss	%xmm15, %xmm0
	movss	52 * SIZE(BO), %xmm15
	mulss	%xmm8, %xmm15
	addss	%xmm15, %xmm1
	movss	56 * SIZE(BO), %xmm15
	mulss	%xmm8, %xmm15
	addss	%xmm15, %xmm2
	movss	60 * SIZE(BO), %xmm15
	mulss	%xmm8, %xmm15
	movss	 8 * SIZE(AO), %xmm8
	addss	%xmm15, %xmm3
	movss	112 * SIZE(BO), %xmm15

	mulss	%xmm10, %xmm9
	addss	%xmm9, %xmm0
	movss	68 * SIZE(BO), %xmm9
	mulss	%xmm10, %xmm9
	addss	%xmm9, %xmm1
	movss	72 * SIZE(BO), %xmm9
	mulss	%xmm10, %xmm9
	addss	%xmm9, %xmm2
	movss	76 * SIZE(BO), %xmm9
	mulss	%xmm10, %xmm9
	movss	 5 * SIZE(AO), %xmm10
	addss	%xmm9, %xmm3
	movss	128 * SIZE(BO), %xmm9

	mulss	%xmm10, %xmm11
	addss	%xmm11, %xmm0
	movss	84 * SIZE(BO), %xmm11
	mulss	%xmm10, %xmm11
	addss	%xmm11, %xmm1
	movss	88 * SIZE(BO), %xmm11
	mulss	%xmm10, %xmm11
	addss	%xmm11, %xmm2
	movss	92 * SIZE(BO), %xmm11
	mulss	%xmm10, %xmm11
	movss	 6 * SIZE(AO), %xmm10
	addss	%xmm11, %xmm3
	movss	144 * SIZE(BO), %xmm11

	mulss	%xmm10, %xmm13
	addss	%xmm13, %xmm0
	movss	100 * SIZE(BO), %xmm13
	mulss	%xmm10, %xmm13
	addss	%xmm13, %xmm1
	movss	104 * SIZE(BO), %xmm13
	mulss	%xmm10, %xmm13
	addss	%xmm13, %xmm2
	movss	108 * SIZE(BO), %xmm13
	mulss	%xmm10, %xmm13
	movss	 7 * SIZE(AO), %xmm10
	addss	%xmm13, %xmm3
	movss	160 * SIZE(BO), %xmm13

	mulss	%xmm10, %xmm15
	addss	%xmm15, %xmm0
	movss	116 * SIZE(BO), %xmm15
	mulss	%xmm10, %xmm15
	addss	%xmm15, %xmm1
	movss	120 * SIZE(BO), %xmm15
	mulss	%xmm10, %xmm15
	addss	%xmm15, %xmm2
	movss	124 * SIZE(BO), %xmm15
	mulss	%xmm10, %xmm15
	movss	 12 * SIZE(AO), %xmm10
	addss	%xmm15, %xmm3
	movss	176 * SIZE(BO), %xmm15

	addq   $  8 * SIZE, AO
	addq   $128 * SIZE, BO
	decq   %rax
	jne    .L42
	ALIGN_4

.L45:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L48
	ALIGN_4

.L46:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movss	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm1
	movss	 8 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm2
	movss	12 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movss	 1 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm3
	movss	16 * SIZE(BO), %xmm9

	addq	$ 1 * SIZE, AO		# aoffset  += 4
	addq	$16 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L46
	ALIGN_4

.L48:
	mulss	%xmm15, %xmm0
	mulss	%xmm15, %xmm1
 	mulss	%xmm15, %xmm2
	mulss	%xmm15, %xmm3

	movss	0 * SIZE(CO1), %xmm8
	movss	0 * SIZE(CO2), %xmm10
	movss	0 * SIZE(CO1, LDC, 2), %xmm12
	movss	0 * SIZE(CO2, LDC, 2), %xmm14

	addss	%xmm8,  %xmm0
	addss	%xmm10, %xmm1
	addss	%xmm12, %xmm2
	addss	%xmm14, %xmm3

	movss	%xmm0, 0 * SIZE(CO1)
	movss	%xmm1, 0 * SIZE(CO2)
	movss	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movss	%xmm3, 0 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    4), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$1, KK
#endif
	ALIGN_4	
	
.L49:
#if defined(TRMMKERNEL) && !defined(LEFT)
	addl	$4, KK
#endif
	leaq	(C, LDC, 4), C		# c += 4 * ldc
	decq	J			# j --
	jg	.L01

.L50:
	testq	$2, N
	je	.L100

.L51:
#if defined(TRMMKERNEL) && defined(LEFT)
	movq	OFFSET, %rax
	movq	%rax, KK
#endif	

/* Copying to Sub Buffer */
	leaq	BUFFER, BO
	
	movq	K, %rax
	sarq	$2, %rax
	jle	.L53
	ALIGN_4
	
.L52:
#ifdef PENTIUM4
	movss	 0 * SIZE(B), %xmm0
	movss	 1 * SIZE(B), %xmm1
	movss	 2 * SIZE(B), %xmm2
	movss	 3 * SIZE(B), %xmm3
	movss	 4 * SIZE(B), %xmm0
	movss	 5 * SIZE(B), %xmm5
	movss	 6 * SIZE(B), %xmm6
	movss	 7 * SIZE(B), %xmm7

	prefetchnta	 32 * SIZE(B)

	shufps	 $0, %xmm0, %xmm0
	shufps	 $0, %xmm1, %xmm1
	shufps	 $0, %xmm2, %xmm2
	shufps	 $0, %xmm3, %xmm3
	shufps	 $0, %xmm0, %xmm0
	shufps	 $0, %xmm5, %xmm5
	shufps	 $0, %xmm6, %xmm6
	shufps	 $0, %xmm7, %xmm7

	movaps	%xmm0,  0 * SIZE(BO)
	movaps	%xmm1,  4 * SIZE(BO)
	movaps	%xmm2,  8 * SIZE(BO)
	movaps	%xmm3, 12 * SIZE(BO)
	movaps	%xmm4, 16 * SIZE(BO)
	movaps	%xmm5, 20 * SIZE(BO)
	movaps	%xmm6, 24 * SIZE(BO)
	movaps	%xmm7, 28 * SIZE(BO)

	addq	$ 8 * SIZE, B
	addq	$32 * SIZE, BO
#endif

#ifdef OPTERON
	prefetchnta	 32 * SIZE(B)

	movd	 0 * SIZE(B), %mm0
	movd	 1 * SIZE(B), %mm1
	movd	 2 * SIZE(B), %mm2
	movd	 3 * SIZE(B), %mm3
	movd	 4 * SIZE(B), %mm4
	movd	 5 * SIZE(B), %mm5
	movd	 6 * SIZE(B), %mm6
	movd	 7 * SIZE(B), %mm7

	punpckldq %mm0, %mm0
	punpckldq %mm1, %mm1
	punpckldq %mm2, %mm2
	punpckldq %mm3, %mm3
	punpckldq %mm4, %mm4
	punpckldq %mm5, %mm5
	punpckldq %mm6, %mm6
	punpckldq %mm7, %mm7
	
	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
	movq	%mm1,  4 * SIZE(BO)
	movq	%mm1,  6 * SIZE(BO)
	movq	%mm2,  8 * SIZE(BO)
	movq	%mm2, 10 * SIZE(BO)
	movq	%mm3, 12 * SIZE(BO)
	movq	%mm3, 14 * SIZE(BO)
	movq	%mm4, 16 * SIZE(BO)
	movq	%mm4, 18 * SIZE(BO)
	movq	%mm5, 20 * SIZE(BO)
	movq	%mm5, 22 * SIZE(BO)
	movq	%mm6, 24 * SIZE(BO)
	movq	%mm6, 26 * SIZE(BO)
	movq	%mm7, 28 * SIZE(BO)
	movq	%mm7, 30 * SIZE(BO)

	addq	$ 8 * SIZE, B
	addq	$32 * SIZE, BO
#endif

	decq	%rax
	jne	.L52
	ALIGN_4

.L53:
	movq	K, %rax
	andq	$3, %rax
	BRANCH
	jle	.L60
	ALIGN_4

.L54:
#ifdef PENTIUM4
	movss	 0 * SIZE(B), %xmm0
	movss	 1 * SIZE(B), %xmm1

	shufps	 $0, %xmm0, %xmm0
	shufps	 $0, %xmm1, %xmm1

	movaps	%xmm0,  0 * SIZE(BO)
	movaps	%xmm1,  4 * SIZE(BO)
#endif

#ifdef OPTERON
	movd	 0 * SIZE(B), %mm0
	movd	 1 * SIZE(B), %mm1

	punpckldq %mm0, %mm0
	punpckldq %mm1, %mm1

	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
	movq	%mm1,  4 * SIZE(BO)
	movq	%mm1,  6 * SIZE(BO)
#endif

	addq	$ 2 * SIZE, B
	addq	$ 8 * SIZE, BO
	decq	%rax
	jne	.L54
	ALIGN_4
	
.L60:
	movq	C, CO1			# coffset1 = c
	leaq	(C, LDC, 1), CO2	# coffset2 = c + ldc
	movq	A, AO		# aoffset = a

	movq	M,  I
	sarq	$3, I	# i = (m >> 3)
	jle	.L70
	ALIGN_4

.L61:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	16 * SIZE(AO), %xmm10
	movaps	32 * SIZE(AO), %xmm12
	movaps	48 * SIZE(AO), %xmm14

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1

#ifdef HAVE_3DNOW
	prefetchw      4 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	prefetchw      4 * SIZE(CO2)
	pxor	%xmm5, %xmm5
#else
	prefetcht2      4 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	prefetcht2     4 * SIZE(CO2)
	pxor	%xmm5, %xmm5
#endif

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$8, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L65
	ALIGN_4

.L62:
	mulps	%xmm8, %xmm9
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 0 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	 4 * SIZE(AO), %xmm8
	mulps	%xmm8, %xmm9
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm4
	movaps	 8 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm5
	movaps	 8 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 8 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	12 * SIZE(AO), %xmm8
	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm4
	movaps	64 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm5
	movaps	64 * SIZE(AO), %xmm8

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 16) * SIZE(AO)
#endif
	mulps	%xmm10, %xmm11
	mulps	20 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm0
	movaps	16 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm1
	movaps	20 * SIZE(AO), %xmm10
	mulps	%xmm10, %xmm11
	mulps	20 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm4
	movaps	24 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm5
	movaps	24 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm11
	mulps	28 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm0
	movaps	24 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm1
	movaps	28 * SIZE(AO), %xmm10
	mulps	%xmm10, %xmm11
	mulps	28 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm4
	movaps	80 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm5
	movaps	80 * SIZE(AO), %xmm10

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 32) * SIZE(AO)
#endif
	mulps	%xmm12, %xmm13
	mulps	36 * SIZE(BO), %xmm12
	addps	%xmm13, %xmm0
	movaps	32 * SIZE(BO), %xmm13
	addps	%xmm12, %xmm1
	movaps	36 * SIZE(AO), %xmm12
	mulps	%xmm12, %xmm13
	mulps	36 * SIZE(BO), %xmm12
	addps	%xmm13, %xmm4
	movaps	40 * SIZE(BO), %xmm13
	addps	%xmm12, %xmm5
	movaps	40 * SIZE(AO), %xmm12

	mulps	%xmm12, %xmm13
	mulps	44 * SIZE(BO), %xmm12
	addps	%xmm13, %xmm0
	movaps	40 * SIZE(BO), %xmm13
	addps	%xmm12, %xmm1
	movaps	44 * SIZE(AO), %xmm12
	mulps	%xmm12, %xmm13
	mulps	44 * SIZE(BO), %xmm12
	addps	%xmm13, %xmm4
	addps	%xmm12, %xmm5
	movaps	96 * SIZE(BO), %xmm13
	movaps	96 * SIZE(AO), %xmm12

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 48) * SIZE(AO)
#endif
	mulps	%xmm14, %xmm15
	mulps	52 * SIZE(BO), %xmm14
	addps	%xmm15, %xmm0
	movaps	48 * SIZE(BO), %xmm15
	addps	%xmm14, %xmm1
	movaps	52 * SIZE(AO), %xmm14
	mulps	%xmm14, %xmm15
	mulps	52 * SIZE(BO), %xmm14
	addps	%xmm15, %xmm4
	movaps	56 * SIZE(BO), %xmm15
	addps	%xmm14, %xmm5
	movaps	56 * SIZE(AO), %xmm14

	mulps	%xmm14, %xmm15
	mulps	60 * SIZE(BO), %xmm14
	addps	%xmm15, %xmm0
	movaps	56 * SIZE(BO), %xmm15
	addps	%xmm14, %xmm1
	movaps	60 * SIZE(AO), %xmm14
	mulps	%xmm14, %xmm15
	mulps	60 * SIZE(BO), %xmm14
	addps	%xmm15, %xmm4
	movaps	112 * SIZE(BO), %xmm15
	addps	%xmm14, %xmm5
	movaps	112 * SIZE(AO), %xmm14

	addq   $64 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L62
	ALIGN_4

.L65:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L68
	ALIGN_4

.L66:
	mulps	%xmm8, %xmm9
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 0 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	 4 * SIZE(AO), %xmm8
	mulps	%xmm8, %xmm9
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm4
	movaps	 8 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm5
	movaps	 8 * SIZE(AO), %xmm8

	addq	$8 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L66
	ALIGN_4

.L68:
	movlps	0 * SIZE(CO1), %xmm8
	movhps	2 * SIZE(CO1), %xmm8
	movlps	4 * SIZE(CO1), %xmm9
	movhps	6 * SIZE(CO1), %xmm9

	movlps	0 * SIZE(CO2), %xmm10
	movhps	2 * SIZE(CO2), %xmm10
	movlps	4 * SIZE(CO2), %xmm11
	movhps	6 * SIZE(CO2), %xmm11

	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm4
	mulps	%xmm15, %xmm1
	mulps	%xmm15, %xmm5

	addps	%xmm8,  %xmm0
	addps	%xmm9,  %xmm4
	addps	%xmm10, %xmm1
	addps	%xmm11, %xmm5

	movlps	%xmm0, 0 * SIZE(CO1)
	movhps	%xmm0, 2 * SIZE(CO1)
	movlps	%xmm4, 4 * SIZE(CO1)
	movhps	%xmm4, 6 * SIZE(CO1)

	movlps	%xmm1, 0 * SIZE(CO2)
	movhps	%xmm1, 2 * SIZE(CO2)
	movlps	%xmm5, 4 * SIZE(CO2)
	movhps	%xmm5, 6 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$8, KK
#endif

	addq	$8 * SIZE, CO1		# coffset += 4
	addq	$8 * SIZE, CO2		# coffset += 4
	decq	I			# i --
	jg	.L61
	ALIGN_4	

.L70:
	testq	$4, M
	je	.L80


#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	16 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L75
	ALIGN_4

.L72:
	mulps	%xmm8, %xmm9
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif

	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 8 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	 4 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm9
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm2
	movaps	64 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm3
	movaps	 8 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm11
	mulps	20 * SIZE(BO), %xmm8
	addps	%xmm11, %xmm0
	movaps	24 * SIZE(BO), %xmm11
	addps	%xmm8, %xmm1
	movaps	12 * SIZE(AO), %xmm8

	mulps	%xmm8, %xmm11
	mulps	28 * SIZE(BO), %xmm8
	addps	%xmm11, %xmm2
	movaps	80 * SIZE(BO), %xmm11
	addps	%xmm8, %xmm3
	movaps	32 * SIZE(AO), %xmm8

	mulps	%xmm10, %xmm13
	mulps	36 * SIZE(BO), %xmm10
	addps	%xmm13, %xmm0
	movaps	40 * SIZE(BO), %xmm13
	addps	%xmm10, %xmm1
	movaps	20 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm13
	mulps	44 * SIZE(BO), %xmm10
	addps	%xmm13, %xmm2
	movaps	96 * SIZE(BO), %xmm13
	addps	%xmm10, %xmm3
	movaps	24 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm15
	mulps	52 * SIZE(BO), %xmm10
	addps	%xmm15, %xmm0
	movaps	56 * SIZE(BO), %xmm15
	addps	%xmm10, %xmm1
	movaps	28 * SIZE(AO), %xmm10

	mulps	%xmm10, %xmm15
	mulps	60 * SIZE(BO), %xmm10
	addps	%xmm15, %xmm2
	movaps	112 * SIZE(BO), %xmm15
	addps	%xmm10, %xmm3
	movaps	48 * SIZE(AO), %xmm10

	addq   $32 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L72
	ALIGN_4

.L75:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L78
	ALIGN_4

.L76:
	mulps	%xmm8, %xmm9
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 8 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	 4 * SIZE(AO), %xmm8

	addq	$4 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L76
	ALIGN_4

.L78:
	movlps	0 * SIZE(CO1), %xmm8
	movhps	2 * SIZE(CO1), %xmm8
	movlps	0 * SIZE(CO2), %xmm10
	movhps	2 * SIZE(CO2), %xmm10

	addps	%xmm2, %xmm0
	addps	%xmm3, %xmm1

	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm1

	addps	%xmm8,  %xmm0
	addps	%xmm10, %xmm1

	movlps	%xmm0, 0 * SIZE(CO1)
	movhps	%xmm0, 2 * SIZE(CO1)
	movlps	%xmm1, 0 * SIZE(CO2)
	movhps	%xmm1, 2 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	addq	$4 * SIZE, CO2		# coffset += 4
	ALIGN_4	

.L80:
	testq	$2, M
	je	.L90

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	 8 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L85
	ALIGN_4

.L82:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movlps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movlps	 8 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm2
	movlps	12 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 4 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm3
	movlps	64 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm0
	movlps	20 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	movlps	 6 * SIZE(AO), %xmm8
	addps	%xmm11, %xmm1
	movlps	24 * SIZE(BO), %xmm11

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm2
	movlps	28 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	movlps	16 * SIZE(AO), %xmm8
	addps	%xmm11, %xmm3
	movlps	 80 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm0
	movlps	36 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	movlps	10 * SIZE(AO), %xmm10
	addps	%xmm13, %xmm1
	movlps	40 * SIZE(BO), %xmm13

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm2
	movlps	44 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	movlps	12 * SIZE(AO), %xmm10
	addps	%xmm13, %xmm3
	movlps	 96 * SIZE(BO), %xmm13

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm0
	movlps	52 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	movlps	14 * SIZE(AO), %xmm10
	addps	%xmm15, %xmm1
	movlps	56 * SIZE(BO), %xmm15

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm2
	movlps	60 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	movlps	24 * SIZE(AO), %xmm10
	addps	%xmm15, %xmm3
	movlps	112 * SIZE(BO), %xmm15

	addq   $16 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L82
	ALIGN_4

.L85:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L88
	ALIGN_4

.L86:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movlps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movlps	 8 * SIZE(BO), %xmm9

	addq	$2 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L86
	ALIGN_4

.L88:
	movlps	0 * SIZE(CO1), %xmm8
	movlps	0 * SIZE(CO2), %xmm10

	addps	%xmm2, %xmm0
	addps	%xmm3, %xmm1

	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm1
	addps	%xmm8,  %xmm0
	addps	%xmm10, %xmm1

	movlps	%xmm0, 0 * SIZE(CO1)
	movlps	%xmm1, 0 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif

	addq	$2 * SIZE, CO1		# coffset += 4
	addq	$2 * SIZE, CO2		# coffset += 4
	ALIGN_4	

.L90:
	testq	$1, M
	je	.L99

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   4), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movss	 0 * SIZE(AO), %xmm8
	movss	 4 * SIZE(AO), %xmm10

	movss	 0 * SIZE(BO), %xmm9
	movss	16 * SIZE(BO), %xmm11
	movss	32 * SIZE(BO), %xmm13
	movss	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L95
	ALIGN_4

.L92:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movss	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movss	 1 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movss	 8 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm2
	movss	12 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movss	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm3
	movss	64 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm0
	movss	20 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	movss	 3 * SIZE(AO), %xmm8
	addps	%xmm11, %xmm1
	movss	24 * SIZE(BO), %xmm11

	mulps	%xmm8, %xmm11
	addps	%xmm11, %xmm2
	movss	28 * SIZE(BO), %xmm11
	mulps	%xmm8, %xmm11
	movss	 8 * SIZE(AO), %xmm8
	addps	%xmm11, %xmm3
	movss	 80 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm0
	movss	36 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	movss	 5 * SIZE(AO), %xmm10
	addps	%xmm13, %xmm1
	movss	40 * SIZE(BO), %xmm13

	mulps	%xmm10, %xmm13
	addps	%xmm13, %xmm2
	movss	44 * SIZE(BO), %xmm13
	mulps	%xmm10, %xmm13
	movss	 6 * SIZE(AO), %xmm10
	addps	%xmm13, %xmm3
	movss	 96 * SIZE(BO), %xmm13

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm0
	movss	52 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	movss	 7 * SIZE(AO), %xmm10
	addps	%xmm15, %xmm1
	movss	56 * SIZE(BO), %xmm15

	mulps	%xmm10, %xmm15
	addps	%xmm15, %xmm2
	movss	60 * SIZE(BO), %xmm15
	mulps	%xmm10, %xmm15
	movss	12 * SIZE(AO), %xmm10
	addps	%xmm15, %xmm3
	movss	112 * SIZE(BO), %xmm15

	addq   $ 8 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L92
	ALIGN_4

.L95:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L98
	ALIGN_4

.L96:
	mulps	%xmm8, %xmm9
	addps	%xmm9, %xmm0
	movss	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movss	 1 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movss	 8 * SIZE(BO), %xmm9

	addq	$1 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L96
	ALIGN_4

.L98:
	movss	0 * SIZE(CO1), %xmm8
	movss	0 * SIZE(CO2), %xmm10
	addss	%xmm2, %xmm0
	addss	%xmm3, %xmm1
	mulss	%xmm15, %xmm0
	mulss	%xmm15, %xmm1
	addss	%xmm8,  %xmm0
	addss	%xmm10, %xmm1
	movss	%xmm0, 0 * SIZE(CO1)
	movss	%xmm1, 0 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    4), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$1, KK
#endif
	ALIGN_4	
	
.L99:
#if defined(TRMMKERNEL) && !defined(LEFT)
	addl	$2, KK
#endif
	leaq	(C, LDC, 2), C		# c += 4 * ldc
	ALIGN_4


.L100:
	testq	$1, N
	je	.L999

.L101:
#if defined(TRMMKERNEL) && defined(LEFT)
	movq	OFFSET, %rax
	movq	%rax, KK
#endif	

/* Copying to Sub Buffer */
	leaq	BUFFER, BO
	
	movq	K, %rax
	sarq	$3, %rax
	jle	.L103
	ALIGN_4
	

.L102:
#ifdef PENTIUM4
	movss	 0 * SIZE(B), %xmm0
	movss	 1 * SIZE(B), %xmm1
	movss	 2 * SIZE(B), %xmm2
	movss	 3 * SIZE(B), %xmm3
	movss	 4 * SIZE(B), %xmm4
	movss	 5 * SIZE(B), %xmm5
	movss	 6 * SIZE(B), %xmm6
	movss	 7 * SIZE(B), %xmm7

	prefetchnta	 32 * SIZE(B)

	shufps	 $0, %xmm0, %xmm0
	shufps	 $0, %xmm1, %xmm1
	shufps	 $0, %xmm2, %xmm2
	shufps	 $0, %xmm3, %xmm3
	shufps	 $0, %xmm4, %xmm4
	shufps	 $0, %xmm5, %xmm5
	shufps	 $0, %xmm6, %xmm6
	shufps	 $0, %xmm7, %xmm7

	movaps	%xmm0,  0 * SIZE(BO)
	movaps	%xmm1,  4 * SIZE(BO)
	movaps	%xmm2,  8 * SIZE(BO)
	movaps	%xmm3, 12 * SIZE(BO)
	movaps	%xmm4, 16 * SIZE(BO)
	movaps	%xmm5, 20 * SIZE(BO)
	movaps	%xmm6, 24 * SIZE(BO)
	movaps	%xmm7, 28 * SIZE(BO)

	addq	$ 8 * SIZE, B
	addq	$32 * SIZE, BO
#endif

#ifdef OPTERON
	prefetchnta	 32 * SIZE(B)

	movd	 0 * SIZE(B), %mm0
	movd	 1 * SIZE(B), %mm1
	movd	 2 * SIZE(B), %mm2
	movd	 3 * SIZE(B), %mm3
	movd	 4 * SIZE(B), %mm4
	movd	 5 * SIZE(B), %mm5
	movd	 6 * SIZE(B), %mm6
	movd	 7 * SIZE(B), %mm7

	punpckldq %mm0, %mm0
	punpckldq %mm1, %mm1
	punpckldq %mm2, %mm2
	punpckldq %mm3, %mm3
	punpckldq %mm4, %mm4
	punpckldq %mm5, %mm5
	punpckldq %mm6, %mm6
	punpckldq %mm7, %mm7
	
	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
	movq	%mm1,  4 * SIZE(BO)
	movq	%mm1,  6 * SIZE(BO)
	movq	%mm2,  8 * SIZE(BO)
	movq	%mm2, 10 * SIZE(BO)
	movq	%mm3, 12 * SIZE(BO)
	movq	%mm3, 14 * SIZE(BO)
	movq	%mm4, 16 * SIZE(BO)
	movq	%mm4, 18 * SIZE(BO)
	movq	%mm5, 20 * SIZE(BO)
	movq	%mm5, 22 * SIZE(BO)
	movq	%mm6, 24 * SIZE(BO)
	movq	%mm6, 26 * SIZE(BO)
	movq	%mm7, 28 * SIZE(BO)
	movq	%mm7, 30 * SIZE(BO)

	addq	$ 8 * SIZE, B
	addq	$32 * SIZE, BO
#endif

	decq	%rax
	jne	.L102
	ALIGN_4

.L103:
	movq	K, %rax
	andq	$7, %rax
	BRANCH
	jle	.L110
	ALIGN_4

.L104:
#ifdef PENTIUM4
	movss	 0 * SIZE(B), %xmm0
	shufps	 $0, %xmm0, %xmm0
	movaps	%xmm0,  0 * SIZE(BO)
#endif

#ifdef OPTERON
	movd	 0 * SIZE(B), %mm0
	punpckldq %mm0, %mm0
	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  2 * SIZE(BO)
#endif

	addq	$ 1 * SIZE, B
	addq	$ 4 * SIZE, BO
	decq	%rax
	jne	.L104
	ALIGN_4
	
.L110:
	movq	C, CO1			# coffset1 = c
	movq	A, AO		# aoffset = a

	movq	M,  I
	sarq	$3, I	# i = (m >> 3)
	jle	.L120
	ALIGN_4

.L111:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	16 * SIZE(AO), %xmm10
	movaps	32 * SIZE(AO), %xmm12
	movaps	48 * SIZE(AO), %xmm14

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11
	movaps	32 * SIZE(BO), %xmm13
	movaps	48 * SIZE(BO), %xmm15

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1

#ifdef HAVE_3DNOW
	prefetchw      4 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	pxor	%xmm5, %xmm5
#else
	prefetcht2      4 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	pxor	%xmm5, %xmm5
#endif

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$8, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L115
	ALIGN_4

.L112:
	mulps	%xmm9, %xmm8
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif

	mulps	 4 * SIZE(AO), %xmm9
	addps	%xmm8, %xmm0
	movaps	 8 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm4
	movaps	 4 * SIZE(BO), %xmm9

	mulps	%xmm9, %xmm8
	mulps	12 * SIZE(AO), %xmm9
	addps	%xmm8, %xmm0
	movaps	64 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm4
	movaps	 8 * SIZE(BO), %xmm9

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 16) * SIZE(AO)
#endif
	mulps	%xmm9, %xmm10
	mulps	20 * SIZE(AO), %xmm9
	addps	%xmm10, %xmm0
	movaps	24 * SIZE(AO), %xmm10
	addps	%xmm9, %xmm4
	movaps	12 * SIZE(BO), %xmm9

	mulps	%xmm9, %xmm10
	mulps	28 * SIZE(AO), %xmm9
	addps	%xmm10, %xmm0
	movaps	80 * SIZE(AO), %xmm10
	addps	%xmm9, %xmm4
	movaps	32 * SIZE(BO), %xmm9

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 32) * SIZE(AO)
#endif
	mulps	%xmm11, %xmm12
	mulps	36 * SIZE(AO), %xmm11
	addps	%xmm12, %xmm0
	movaps	40 * SIZE(AO), %xmm12
	addps	%xmm11, %xmm4
	movaps	20 * SIZE(BO), %xmm11

	mulps	%xmm11, %xmm12
	mulps	44 * SIZE(AO), %xmm11
	addps	%xmm12, %xmm0
	movaps	96 * SIZE(AO), %xmm12
	addps	%xmm11, %xmm4
	movaps	24 * SIZE(BO), %xmm11

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 48) * SIZE(AO)
#endif
	mulps	%xmm11, %xmm14
	mulps	52 * SIZE(AO), %xmm11
	addps	%xmm14, %xmm0
	movaps	56 * SIZE(AO), %xmm14
	addps	%xmm11, %xmm4
	movaps	28 * SIZE(BO), %xmm11

	mulps	%xmm11, %xmm14
	mulps	60 * SIZE(AO), %xmm11
	addps	%xmm14, %xmm0
	movaps	112 * SIZE(AO), %xmm14
	addps	%xmm11, %xmm4
	movaps	48 * SIZE(BO), %xmm11

	addq   $64 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L112
	ALIGN_4

.L115:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L118
	ALIGN_4

.L116:
	mulps	%xmm9, %xmm8
	mulps	 4 * SIZE(AO), %xmm9
	addps	%xmm8, %xmm0
	movaps	 8 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm4
	movaps	 4 * SIZE(BO), %xmm9

	addq	$8 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L116
	ALIGN_4

.L118:
	movlps	0 * SIZE(CO1), %xmm8
	movhps	2 * SIZE(CO1), %xmm8
	movlps	4 * SIZE(CO1), %xmm9
	movhps	6 * SIZE(CO1), %xmm9

	mulps	%xmm15, %xmm0
	mulps	%xmm15, %xmm4
	addps	%xmm8,  %xmm0
	addps	%xmm9,  %xmm4

	movlps	%xmm0, 0 * SIZE(CO1)
	movhps	%xmm0, 2 * SIZE(CO1)
	movlps	%xmm4, 4 * SIZE(CO1)
	movhps	%xmm4, 6 * SIZE(CO1)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 2), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$8, KK
#endif

	addq	$8 * SIZE, CO1		# coffset += 4
	decq	I			# i --
	jg	.L111
	ALIGN_4	

.L120:
	testq	$4, M
	je	.L130


#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	16 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L125
	ALIGN_4

.L122:
	mulps	%xmm8, %xmm9
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movaps	 4 * SIZE(AO), %xmm8
	mulps	 4 * SIZE(BO), %xmm8
	addps	%xmm9, %xmm0
	movaps	32 * SIZE(BO), %xmm9
	addps	%xmm8, %xmm1
	movaps	 8 * SIZE(AO), %xmm8
	mulps	 8 * SIZE(BO), %xmm8
	addps	%xmm8, %xmm2
	movaps	12 * SIZE(AO), %xmm8
	mulps	12 * SIZE(BO), %xmm8
	addps	%xmm8, %xmm3
	movaps	32 * SIZE(AO), %xmm8

#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE + 16) * SIZE(AO)
#endif
	mulps	%xmm10, %xmm11
	movaps	20 * SIZE(AO), %xmm10
	mulps	20 * SIZE(BO), %xmm10
	addps	%xmm11, %xmm0
	movaps	48 * SIZE(BO), %xmm11
	addps	%xmm10, %xmm1
	movaps	24 * SIZE(AO), %xmm10
	mulps	24 * SIZE(BO), %xmm10
	addps	%xmm10, %xmm2
	movaps	28 * SIZE(AO), %xmm10
	mulps	28 * SIZE(BO), %xmm10
	addps	%xmm10, %xmm3
	movaps	48 * SIZE(AO), %xmm10

	addq   $32 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L122
	ALIGN_4

.L125:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L128
	ALIGN_4

.L126:
	mulps	%xmm8, %xmm9
	movaps	 4 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm0
	movaps	 4 * SIZE(BO), %xmm9

	addq	$4 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L126
	ALIGN_4

.L128:
	movlps	0 * SIZE(CO1), %xmm8
	movhps	2 * SIZE(CO1), %xmm8
	addps	%xmm1, %xmm0
	addps	%xmm3, %xmm2
	addps	%xmm2, %xmm0

	mulps	%xmm15, %xmm0
	addps	%xmm8,  %xmm0
	movlps	%xmm0, 0 * SIZE(CO1)
	movhps	%xmm0, 2 * SIZE(CO1)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 2), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	ALIGN_4	

.L130:
	testq	$2, M
	je	.L140

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movaps	 0 * SIZE(AO), %xmm8
	movaps	 8 * SIZE(AO), %xmm10

	movaps	 0 * SIZE(BO), %xmm9
	movaps	16 * SIZE(BO), %xmm11

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L135
	ALIGN_4

.L132:
	mulps	%xmm8, %xmm9
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm0
	movlps	 4 * SIZE(BO), %xmm9
	mulps	%xmm8, %xmm9
	movlps	 4 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movlps	 8 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm9
	movlps	 6 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm0
	movlps	12 * SIZE(BO), %xmm9

	mulps	%xmm8, %xmm9
	movlps	16 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm1
	movlps	32 * SIZE(BO), %xmm9

	mulps	%xmm10, %xmm11
	movlps	10 * SIZE(AO), %xmm10
	addps	%xmm11, %xmm0
	movlps	20 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm11
	movlps	12 * SIZE(AO), %xmm10
	addps	%xmm11, %xmm1
	movlps	24 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm11
	movlps	14 * SIZE(AO), %xmm10
	addps	%xmm11, %xmm0
	movlps	28 * SIZE(BO), %xmm11

	mulps	%xmm10, %xmm11
	movlps	24 * SIZE(AO), %xmm10
	addps	%xmm11, %xmm1
	movlps	48 * SIZE(BO), %xmm11

	addq   $16 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L132
	ALIGN_4

.L135:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movaps	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L138
	ALIGN_4

.L136:
	mulps	%xmm8, %xmm9
	movlps	 2 * SIZE(AO), %xmm8
	addps	%xmm9, %xmm0
	movlps	 4 * SIZE(BO), %xmm9

	addq	$2 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L136
	ALIGN_4

.L138:
	movlps	0 * SIZE(CO1), %xmm8
	addps	%xmm1,  %xmm0
	mulps	%xmm15, %xmm0
	addps	%xmm8,  %xmm0
	movlps	%xmm0, 0 * SIZE(CO1)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax,    8), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 2), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif

	addq	$2 * SIZE, CO1		# coffset += 4
	ALIGN_4	

.L140:
	testq	$1, M
	je	.L999

#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax,   4), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movss	 0 * SIZE(AO), %xmm8
	movss	 4 * SIZE(AO), %xmm10

	movss	 0 * SIZE(BO), %xmm9
	movss	16 * SIZE(BO), %xmm11

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L145
	ALIGN_4

.L142:
	mulss	%xmm8, %xmm9
#if defined(OPTERON) && defined(HAVE_PREFETCH)
	PREFETCH	(PREFETCHSIZE +  0) * SIZE(AO)
#endif
	movss	 1 * SIZE(AO), %xmm8
	mulss	 4 * SIZE(BO), %xmm8
	addss	%xmm9, %xmm0
	movss	32 * SIZE(BO), %xmm9
	addss	%xmm8, %xmm1
	movss	 2 * SIZE(AO), %xmm8
	mulss	 8 * SIZE(BO), %xmm8
	addss	%xmm8, %xmm2
	movss	 3 * SIZE(AO), %xmm8
	mulss	12 * SIZE(BO), %xmm8
	addss	%xmm8, %xmm3
	movss	 8 * SIZE(AO), %xmm8
	mulss	%xmm10, %xmm11
	movss	 5 * SIZE(AO), %xmm10
	mulss	20 * SIZE(BO), %xmm10
	addss	%xmm11, %xmm0
	movss	48 * SIZE(BO), %xmm11
	addss	%xmm10, %xmm1
	movss	 6 * SIZE(AO), %xmm10
	mulss	24 * SIZE(BO), %xmm10
	addss	%xmm10, %xmm2
	movss	 7 * SIZE(AO), %xmm10
	mulss	28 * SIZE(BO), %xmm10
	addss	%xmm10, %xmm3
	movss	12 * SIZE(AO), %xmm10

	addq   $ 8 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L142
	ALIGN_4

.L145:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movss	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L148
	ALIGN_4

.L146:
	mulss	%xmm8, %xmm9
	movss	 1 * SIZE(AO), %xmm8
	addss	%xmm9, %xmm0
	movss	 4 * SIZE(BO), %xmm9

	addq	$1 * SIZE, AO
	addq	$4 * SIZE, BO
	decq	%rax
	jg	.L146
	ALIGN_4

.L148:
	movss	0 * SIZE(CO1), %xmm8
	addss	%xmm1, %xmm0
	addss	%xmm3, %xmm2
	addss	%xmm2, %xmm0

	mulss	%xmm15, %xmm0
	addss	%xmm8,  %xmm0
	movss	%xmm0, 0 * SIZE(CO1)
	ALIGN_4	
	
.L999:
	movq	%rbx, %rsp
	EMMS
	movq	  0(%rsp), %rbx
	movq	  8(%rsp), %rbp
	movq	 16(%rsp), %r12
	movq	 24(%rsp), %r13
	movq	 32(%rsp), %r14
	movq	 40(%rsp), %r15
	addq	$STACKSIZE, %rsp
	ret

	EPILOGUE
