/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LEVEL3
#define COMMON_LEVEL3

#ifndef ASSEMBLER

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
extern long sgemm_p;
extern long sgemm_q;
extern long sgemm_r;
extern long dgemm_p;
extern long dgemm_q;
extern long dgemm_r;
extern long qgemm_p;
extern long qgemm_q;
extern long qgemm_r;
extern long cgemm_p;
extern long cgemm_q;
extern long cgemm_r;
extern long zgemm_p;
extern long zgemm_q;
extern long zgemm_r;
extern long xgemm_p;
extern long xgemm_q;
extern long xgemm_r;
#endif

int sgemm_beta(long, long, long, float, 
	       float  *, long, float   *, long, float  *, long);
int dgemm_beta(long, long, long, double, 
	       double *, long, double  *, long, double *, long);
int qgemm_beta(long, long, long, xdouble, 
	       xdouble *, long, xdouble  *, long, xdouble *, long);
int cgemm_beta(long, long, long, float,  float,
	       float  *, long, float   *, long, float  *, long);
int zgemm_beta(long, long, long, double, double,
	       double *, long, double  *, long, double *, long);
int xgemm_beta(long, long, long, xdouble, xdouble,
	       xdouble *, long, xdouble  *, long, xdouble *, long);

int ssyrk_beta_U(long, long, long, float, 
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int ssyrk_beta_L(long, long, long, float, 
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int dsyrk_beta_U(long, long, long, double, 
		 double *, long, double  *, long, double *, long, double *, long, long);
int dsyrk_beta_L(long, long, long, double, 
		 double *, long, double  *, long, double *, long, double *, long, long);
int qsyrk_beta_U(long, long, long, xdouble, 
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);
int qsyrk_beta_L(long, long, long, xdouble, 
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);
int csyrk_beta_U(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int csyrk_beta_L(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int zsyrk_beta_U(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int zsyrk_beta_L(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int xsyrk_beta_U(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);
int xsyrk_beta_L(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);

int cherk_beta_U(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int cherk_beta_L(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int zherk_beta_U(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int zherk_beta_L(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int xherk_beta_U(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);
int xherk_beta_L(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble  *, long, xdouble *, long, xdouble *, long, long);

void sgemm_incopy(long m, long n, float *a, long lda, float *b);
void sgemm_itcopy(long m, long n, float *a, long lda, float *b);
void sgemm_oncopy(long m, long n, float *a, long lda, float *b);
void sgemm_otcopy(long m, long n, float *a, long lda, float *b);
void dgemm_incopy(long m, long n, double *a, long lda, double *b);
void dgemm_itcopy(long m, long n, double *a, long lda, double *b);
void dgemm_oncopy(long m, long n, double *a, long lda, double *b);
void dgemm_otcopy(long m, long n, double *a, long lda, double *b);
void qgemm_incopy(long m, long n, xdouble *a, long lda, xdouble *b);
void qgemm_itcopy(long m, long n, xdouble *a, long lda, xdouble *b);
void qgemm_oncopy(long m, long n, xdouble *a, long lda, xdouble *b);
void qgemm_otcopy(long m, long n, xdouble *a, long lda, xdouble *b);
void cgemm_incopy(long m, long n, float *a, long lda, float *b);
void cgemm_itcopy(long m, long n, float *a, long lda, float *b);
void cgemm_oncopy(long m, long n, float *a, long lda, float *b);
void cgemm_otcopy(long m, long n, float *a, long lda, float *b);
void zgemm_incopy(long m, long n, double *a, long lda, double *b);
void zgemm_itcopy(long m, long n, double *a, long lda, double *b);
void zgemm_oncopy(long m, long n, double *a, long lda, double *b);
void zgemm_otcopy(long m, long n, double *a, long lda, double *b);
void xgemm_incopy(long m, long n, xdouble *a, long lda, xdouble *b);
void xgemm_itcopy(long m, long n, xdouble *a, long lda, xdouble *b);
void xgemm_oncopy(long m, long n, xdouble *a, long lda, xdouble *b);
void xgemm_otcopy(long m, long n, xdouble *a, long lda, xdouble *b);

void strsm_kernel_LN(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_LT(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_RN(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_RT(long, long, long, float, float *, float *, float *, long, long);
void dtrsm_kernel_LN(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_LT(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_RN(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_RT(long, long, long, double, double *, double *, double *, long, long);

void qtrsm_kernel_LN(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrsm_kernel_LT(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrsm_kernel_RN(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrsm_kernel_RT(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);

void ctrsm_kernel_LN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LC(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RC(long, long, long, float, float, float *, float *, float *, long, long);

void ztrsm_kernel_LN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LC(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RC(long, long, long, double, double, double *, double *, double *, long, long);

void xtrsm_kernel_LN(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_LT(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_LR(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_LC(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_RN(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_RT(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_RR(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrsm_kernel_RC(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);

void strmm_kernel_RN(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_RT(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_LN(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_LT(long, long, long, float, float *, float *, float *, long, long);

void dtrmm_kernel_RN(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_RT(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_LN(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_LT(long, long, long, double, double *, double *, double *, long, long);

void qtrmm_kernel_RN(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrmm_kernel_RT(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrmm_kernel_LN(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void qtrmm_kernel_LT(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long);

void ctrmm_kernel_RN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RC(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LC(long, long, long, float, float, float *, float *, float *, long, long);

void ztrmm_kernel_RN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RC(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LC(long, long, long, double, double, double *, double *, double *, long, long);

void xtrmm_kernel_RN(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_RT(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_RR(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_RC(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_LN(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_LT(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_LR(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);
void xtrmm_kernel_LC(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long, long);

void strmm_iunucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iunncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iutucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iutncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ounucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ounncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_outucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_outncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ilnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ilnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_olnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_olnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_oltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_oltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);

void dtrmm_iunucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iunncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iutucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iutncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ounucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ounncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_outucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_outncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ilnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ilnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_olnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_olnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_oltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_oltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void qtrmm_iunucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_iunncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_iutucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_iutncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_ounucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_ounncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_outucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_outncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_ilnucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_ilnncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_iltucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_iltncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_olnucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_olnncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_oltucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qtrmm_oltncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);

void ctrmm_iunucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iunncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iutucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iutncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ounucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ounncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_outucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_outncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ilnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ilnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_olnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_olnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_oltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_oltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);

void ztrmm_iunucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iunncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iutucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iutncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ounucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ounncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_outucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_outncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ilnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ilnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_olnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_olnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_oltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_oltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void xtrmm_iunucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_iunncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_iutucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_iutncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_ounucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_ounncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_outucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_outncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_ilnucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_ilnncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_iltucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_iltncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_olnucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_olnncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_oltucopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xtrmm_oltncopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);

void strsm_iunucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iunncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iutucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iutncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ounucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ounncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_outucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_outncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ilnucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ilnncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iltucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iltncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_olnucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_olnncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_oltucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_oltncopy(long m, long n, float *a, long lda, long offset, float *b);

void dtrsm_iunucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iunncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iutucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iutncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ounucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ounncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_outucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_outncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ilnucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ilnncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iltucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iltncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_olnucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_olnncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_oltucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_oltncopy(long m, long n, double *a, long lda, long offset, double *b);

void qtrsm_iunucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_iunncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_iutucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_iutncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_ounucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_ounncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_outucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_outncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_ilnucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_ilnncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_iltucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_iltncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_olnucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_olnncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_oltucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void qtrsm_oltncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);

void ctrsm_iunucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iunncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iutucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iutncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ounucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ounncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_outucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_outncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ilnucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ilnncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iltucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iltncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_olnucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_olnncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_oltucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_oltncopy(long m, long n, float *a, long lda, long offset, float *b);

void ztrsm_iunucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iunncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iutucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iutncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ounucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ounncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_outucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_outncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ilnucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ilnncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iltucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iltncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_olnucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_olnncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_oltucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_oltncopy(long m, long n, double *a, long lda, long offset, double *b);

void xtrsm_iunucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_iunncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_iutucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_iutncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_ounucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_ounncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_outucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_outncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_ilnucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_ilnncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_iltucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_iltncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_olnucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_olnncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_oltucopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);
void xtrsm_oltncopy(long m, long n, xdouble *a, long lda, long offset, xdouble *b);

void ssymm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void dsymm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void qsymm_iutcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qsymm_outcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qsymm_iltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void qsymm_oltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void csymm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void zsymm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void xsymm_iutcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xsymm_outcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xsymm_iltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xsymm_oltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);

void chemm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void zhemm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void xhemm_iutcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xhemm_outcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xhemm_iltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);
void xhemm_oltcopy(long m, long n, xdouble *a, long lda, long posX, long posY, xdouble *b);

void ssyrk_kernel_U(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void ssyrk_kernel_L(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);

void dsyrk_kernel_U(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void dsyrk_kernel_L(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void qsyrk_kernel_U(long m, long n, long k, xdouble alpha, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);
void qsyrk_kernel_L(long m, long n, long k, xdouble alpha, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);

void csyrk_kernel_U(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void csyrk_kernel_L(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void zsyrk_kernel_U(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zsyrk_kernel_L(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void cher2m_kernel_UN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_UL(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_UR(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_UB(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_LN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_LL(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_LR(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cher2m_kernel_LB(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);

void zher2m_kernel_UN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_UL(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_UR(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_UB(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_LN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_LL(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_LR(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zher2m_kernel_LB(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void ssyr2k_kernel_U(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void ssyr2k_kernel_L(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void dsyr2k_kernel_U(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void dsyr2k_kernel_L(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void qsyr2k_kernel_U(long m, long n, long k, xdouble alpha, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);
void qsyr2k_kernel_L(long m, long n, long k, xdouble alpha, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);

void csyr2k_kernel_U(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void csyr2k_kernel_L(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void zsyr2k_kernel_U(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zsyr2k_kernel_L(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);

void xsyr2k_kernel_U(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);
void xsyr2k_kernel_L(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);

void cherk_kernel_UN(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_UC(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_LN(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_LC(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);

void zherk_kernel_UN(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_UC(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_LN(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_LC(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void xherk_kernel_UN(long m, long n, long k, xdouble alpha_r, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);
void xherk_kernel_UC(long m, long n, long k, xdouble alpha_r, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);
void xherk_kernel_LN(long m, long n, long k, xdouble alpha_r, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);
void xherk_kernel_LC(long m, long n, long k, xdouble alpha_r, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer);

void cher2k_kernel_UN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_UC(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_LN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_LC(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);

void zher2k_kernel_UN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_UC(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_LN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_LC(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);

void xher2k_kernel_UN(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);
void xher2k_kernel_UC(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);
void xher2k_kernel_LN(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);
void xher2k_kernel_LC(long m, long n, long k, xdouble alpha_r, xdouble alpha_i, xdouble *a, xdouble *b, xdouble *c, long ldc, long offset, xdouble *subbuffer, int flag);

#ifndef PREFETCHTEST
unsigned int sgemm_kernel(long, long, long, float,  float  *, float  *, float  *, long);
unsigned int dgemm_kernel(long, long, long, double, double *, double *, double *, long);
unsigned int qgemm_kernel(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int sgemm_kernel_b0(long, long, long, float,  float  *, float  *, float  *, long);
unsigned int dgemm_kernel_b0(long, long, long, double, double *, double *, double *, long);
unsigned int qgemm_kernel_b0(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long);
#else
unsigned int sgemm_kernel(long, long, long, float,  float  *, float  *, float  *, long, long, long, long);
unsigned int dgemm_kernel(long, long, long, double, double *, double *, double *, long, long, long, long);
unsigned int qgemm_kernel(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long, long, long);
unsigned int sgemm_kernel_b0(long, long, long, float,  float  *, float  *, float  *, long, long, long, long);
unsigned int dgemm_kernel_b0(long, long, long, double, double *, double *, double *, long, long, long, long);
unsigned int qgemm_kernel_b0(long, long, long, xdouble, xdouble *, xdouble *, xdouble *, long, long, long, long);
#endif

unsigned int cgemm_kernel_n(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_l(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_r(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_b(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_n_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_l_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_r_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_b_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);

unsigned int zgemm_kernel_n(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_l(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_r(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_b(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_n_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_l_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_r_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_b_b0(long, long, long, double, double, double *, double *, double *, long);

unsigned int xgemm_kernel_n(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_l(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_r(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_b(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_n_b0(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_l_b0(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_r_b0(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);
unsigned int xgemm_kernel_b_b0(long, long, long, xdouble, xdouble, xdouble *, xdouble *, xdouble *, long);

GEMMRETTYPE sgemm_nn(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE sgemm_nt(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE sgemm_tn(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE sgemm_tt(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE dgemm_nn(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE dgemm_nt(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE dgemm_tn(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE dgemm_tt(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE qgemm_nn(long, long, long, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE qgemm_nt(long, long, long, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE qgemm_tn(long, long, long, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE qgemm_tt(long, long, long, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE cgemm_nn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_ct(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE zgemm_nn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_ct(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE xgemm_nn(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_nt(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_nr(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_nc(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_tn(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_tt(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_tr(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_tc(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_rn(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_rt(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_rr(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_rc(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_cn(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_ct(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_cr(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE xgemm_cc(long, long, long, xdouble, xdouble, xdouble *, long,
		     xdouble *, long, xdouble *, long, xdouble *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

int cher2m_LNN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LNT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LNR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LNC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LTN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LTT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LTR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LTC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LRN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LRT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LRR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LRC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LCN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LCT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LCR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_LCC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UNN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UNT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UNR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UNC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UTN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UTT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UTR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UTC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_URN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_URT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_URR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_URC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UCN(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UCT(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UCR(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);
int cher2m_UCC(long, long, long, float, float, float *, long,
	       float *, long, float *, long, float *, long, long);

int zher2m_LNN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LNT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LNR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LNC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LTN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LTT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LTR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LTC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LRN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LRT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LRR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LRC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LCN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LCT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LCR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_LCC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UNN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UNT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UNR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UNC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UTN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UTT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UTR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UTC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_URN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_URT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_URR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_URC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UCN(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UCT(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UCR(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);
int zher2m_UCC(long, long, long, double, double, double *, long,
	       double *, long, double *, long, double *, long, long);

int strsm_LNUU(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);
int strsm_LNUN(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);
int strsm_LNLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LNLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTLN(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);

int dtrsm_LNUU(long, long, long, double, 
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTLN(long, long, long, double, 
		   double *, long, double *, long, double *, long, double *);

int qtrsm_LNUU(long, long, long, xdouble, 
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LNUN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LNLU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LNLN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LTUU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LTUN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LTLU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_LTLN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RNUU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RNUN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RNLU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RNLN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RTUU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RTUN(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RTLU(long, long, long, xdouble,
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrsm_RTLN(long, long, long, xdouble, 
		   xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);

int ctrsm_LNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);

int ztrsm_LNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);

int xtrsm_LNUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LNUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LNLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LNLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LTUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LTUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LTLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LTLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LRUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LRUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LRLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LRLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LCUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LCUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LCLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_LCLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RNUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RNUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RNLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RNLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RTUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RTUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RTLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RTLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RRUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RRUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RRLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RRLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RCUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RCUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RCLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrsm_RCLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);

int strmm_LNUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTLN(long, long, long, float, float *, long, float *, long, float *, long, float *);

int dtrmm_LNUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTLN(long, long, long, double, double *, long, double *, long, double *, long, double *);

int qtrmm_LNUU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LNUN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LNLU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LNLN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LTUU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LTUN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LTLU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_LTLN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RNUU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RNUN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RNLU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RNLN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RTUU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RTUN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RTLU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int qtrmm_RTLN(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);

int ctrmm_LNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);

int ztrmm_LNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);

int xtrmm_LNUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LNUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LNLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LNLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LTUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LTUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LTLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LTLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LRUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LRUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LRLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LRLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LCUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LCUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LCLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_LCLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RNUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RNUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RNLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RNLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RTUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RTUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RTLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RTLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RRUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RRUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RRLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RRLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RCUU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RCUN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RCLU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);
int xtrmm_RCLN(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *, long, xdouble *);

int ssymm_LU(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_LL(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_RU(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_RL(long, long, long, float , float  *, long, float  *, long, float *, long, float *);

int dsymm_LU(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_LL(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_RU(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_RL(long, long, long, double, double *, long, double *, long, double *,long, double *);

int qsymm_LU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int qsymm_LL(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int qsymm_RU(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int qsymm_RL(long, long, long, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);

int csymm_LU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_LL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_RU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_RL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int zsymm_LU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_LL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_RU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_RL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int xsymm_LU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xsymm_LL(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xsymm_RU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xsymm_RL(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);

int ssyrk_UN(long, long, long, float , 
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_UT(long, long, long, float ,
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_LN(long, long, long, float , 
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_LT(long, long, long, float ,
		 float  *, long, float *, long, float *, long, float *, long, long);

int dsyrk_UN(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_UT(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_LN(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_LT(long, long, long, double, 
		 double *, long, double*, long, double *, long, double *, long, long);

int qsyrk_UN(long, long, long, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int qsyrk_UT(long, long, long, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int qsyrk_LN(long, long, long, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int qsyrk_LT(long, long, long, xdouble, 
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);

int csyrk_UN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_UT(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_LN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_LT(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);

int zsyrk_UN(long, long, long, double, double, 
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_UT(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_LN(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_LT(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);

int xsyrk_UN(long, long, long, xdouble, xdouble, 
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int xsyrk_UT(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int xsyrk_LN(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);
int xsyrk_LT(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble*, long, xdouble *, long, xdouble *, long, long);

int ssyr2k_UN(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_UT(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_LN(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_LT(long, long, long, float ,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int dsyr2k_UN(long, long, long, double, 
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_UT(long, long, long, double, 
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_LN(long, long, long, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_LT(long, long, long, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int qsyr2k_UN(long, long, long, xdouble, 
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int qsyr2k_UT(long, long, long, xdouble, 
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int qsyr2k_LN(long, long, long, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int qsyr2k_LT(long, long, long, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);

int csyr2k_UN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_UT(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_LN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_LT(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int zsyr2k_UN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_UT(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_LN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_LT(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int xsyr2k_UN(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xsyr2k_UT(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xsyr2k_LN(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xsyr2k_LT(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);

int chemm_LU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_LL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_RU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_RL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);

int zhemm_LU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_LL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_RU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_RL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);

int xhemm_LU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xhemm_LL(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xhemm_RU(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);
int xhemm_RL(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, xdouble *,long, xdouble *);

int cherk_UN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_UC(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_LN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_LC(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);

int zherk_UN(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_UC(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_LN(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_LC(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);

int xherk_UN(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble *, long, xdouble *, long, xdouble *, long, long);
int xherk_UC(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble *, long, xdouble *, long, xdouble *, long, long);
int xherk_LN(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble *, long, xdouble *, long, xdouble *, long, long);
int xherk_LC(long, long, long, xdouble, xdouble,
		 xdouble *, long, xdouble *, long, xdouble *, long, xdouble *, long, long);

int cher2k_UN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_UC(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_LN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_LC(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int zher2k_UN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_UC(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_LN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_LC(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int xher2k_UN(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xher2k_UC(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xher2k_LN(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);
int xher2k_LC(long, long, long, xdouble, xdouble,
		  xdouble *, long, xdouble *, long, xdouble*, long, xdouble *, long, long);

int sgemt_n(long, long, float,  float   *, long, float    *, int);
int sgemt_t(long, long, float,  float   *, long, float    *, int);
int dgemt_n(long, long, double, double  *, long, double   *, int);
int dgemt_t(long, long, double, double  *, long, double   *, int);

int cgemt_n(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_t(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_r(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_c(long, long, float,  float,  float   *, long, float    *, int);
int zgemt_n(long, long, double, double, double  *, long, double   *, int);
int zgemt_t(long, long, double, double, double  *, long, double   *, int);
int zgemt_r(long, long, double, double, double  *, long, double   *, int);
int zgemt_c(long, long, double, double, double  *, long, double   *, int);

int sgema_n(long, long, float,  float   *, long, float    *, long);
int sgema_t(long, long, float,  float   *, long, float    *, long);
int dgema_n(long, long, double, double  *, long, double   *, long);
int dgema_t(long, long, double, double  *, long, double   *, long);

int cgema_n(long, long, float,  float,  float   *, long, float    *, long);
int cgema_t(long, long, float,  float,  float   *, long, float    *, long);
int cgema_r(long, long, float,  float,  float   *, long, float    *, long);
int cgema_c(long, long, float,  float,  float   *, long, float    *, long);
int zgema_n(long, long, double, double, double  *, long, double   *, long);
int zgema_t(long, long, double, double, double  *, long, double   *, long);
int zgema_r(long, long, double, double, double  *, long, double   *, long);
int zgema_c(long, long, double, double, double  *, long, double   *, long);

#endif

#ifndef COMPLEX
#ifdef XDOUBLE
/* REAL && EXPRECISION */
#define   GEMM_BETA   qgemm_beta
#define   GEMM_ONCOPY qgemm_oncopy
#define   GEMM_OTCOPY qgemm_otcopy
#if QGEMM_UNROLL_M != QGEMM_UNROLL_N
#define   GEMM_INCOPY qgemm_incopy
#define   GEMM_ITCOPY qgemm_itcopy
#else
#define   GEMM_INCOPY qgemm_oncopy
#define   GEMM_ITCOPY qgemm_otcopy
#endif
#define   GEMM_KERNEL_N qgemm_kernel
#define   GEMM_KERNEL_L qgemm_kernel
#define   GEMM_KERNEL_R qgemm_kernel
#define   GEMM_KERNEL_B qgemm_kernel
#define   GEMM_KERNEL_N_B0 qgemm_kernel_b0
#define   GEMM_KERNEL_L_B0 qgemm_kernel_b0
#define   GEMM_KERNEL_R_B0 qgemm_kernel_b0
#define   GEMM_KERNEL_B_B0 qgemm_kernel_b0
#define   GEMM_NN     qgemm_nn
#define   GEMM_TN     qgemm_tn
#define   GEMM_RN     qgemm_nn
#define   GEMM_CN     qgemm_tn
#define   GEMM_NT     qgemm_nt
#define   GEMM_TT     qgemm_tt
#define   GEMM_RT     qgemm_nt
#define   GEMM_CT     qgemm_tt
#define   GEMM_NR     qgemm_nn
#define   GEMM_TR     qgemm_tn
#define   GEMM_RR     qgemm_nn
#define   GEMM_CR     qgemm_tn
#define   GEMM_NC     qgemm_nt
#define   GEMM_TC     qgemm_tt
#define   GEMM_RC     qgemm_nt
#define   GEMM_CC     qgemm_tt

#define TRMM_KERNEL_LN  qtrmm_kernel_LN
#define TRMM_KERNEL_LT  qtrmm_kernel_LT
#define TRMM_KERNEL_LR  qtrmm_kernel_LN
#define TRMM_KERNEL_LC  qtrmm_kernel_LT
#define TRMM_KERNEL_RN  qtrmm_kernel_RN
#define TRMM_KERNEL_RT  qtrmm_kernel_RT
#define TRMM_KERNEL_RR  qtrmm_kernel_RN
#define TRMM_KERNEL_RC  qtrmm_kernel_RT

#define TRSM_KERNEL_LN  qtrsm_kernel_LN
#define TRSM_KERNEL_LT  qtrsm_kernel_LT
#define TRSM_KERNEL_LR  qtrsm_kernel_LN
#define TRSM_KERNEL_LC  qtrsm_kernel_LT
#define TRSM_KERNEL_RN  qtrsm_kernel_RN
#define TRSM_KERNEL_RT  qtrsm_kernel_RT
#define TRSM_KERNEL_RR  qtrsm_kernel_RN
#define TRSM_KERNEL_RC  qtrsm_kernel_RT

#define TRMM_LNUU qtrmm_LNUU
#define TRMM_LNUN qtrmm_LNUN
#define TRMM_LNLU qtrmm_LNLU
#define TRMM_LNLN qtrmm_LNLN
#define TRMM_LTUU qtrmm_LTUU
#define TRMM_LTUN qtrmm_LTUN
#define TRMM_LTLU qtrmm_LTLU
#define TRMM_LTLN qtrmm_LTLN
#define TRMM_LRUU qtrmm_LNUU
#define TRMM_LRUN qtrmm_LNUN
#define TRMM_LRLU qtrmm_LNLU
#define TRMM_LRLN qtrmm_LNLN
#define TRMM_LCUU qtrmm_LTUU
#define TRMM_LCUN qtrmm_LTUN
#define TRMM_LCLU qtrmm_LTLU
#define TRMM_LCLN qtrmm_LTLN

#define TRMM_RNUU qtrmm_RNUU
#define TRMM_RNUN qtrmm_RNUN
#define TRMM_RNLU qtrmm_RNLU
#define TRMM_RNLN qtrmm_RNLN
#define TRMM_RTUU qtrmm_RTUU
#define TRMM_RTUN qtrmm_RTUN
#define TRMM_RTLU qtrmm_RTLU
#define TRMM_RTLN qtrmm_RTLN
#define TRMM_RRUU qtrmm_RNUU
#define TRMM_RRUN qtrmm_RNUN
#define TRMM_RRLU qtrmm_RNLU
#define TRMM_RRLN qtrmm_RNLN
#define TRMM_RCUU qtrmm_RTUU
#define TRMM_RCUN qtrmm_RTUN
#define TRMM_RCLU qtrmm_RTLU
#define TRMM_RCLN qtrmm_RTLN

#define TRSM_LNUU qtrsm_LNUU
#define TRSM_LNUN qtrsm_LNUN
#define TRSM_LNLU qtrsm_LNLU
#define TRSM_LNLN qtrsm_LNLN
#define TRSM_LTUU qtrsm_LTUU
#define TRSM_LTUN qtrsm_LTUN
#define TRSM_LTLU qtrsm_LTLU
#define TRSM_LTLN qtrsm_LTLN
#define TRSM_LRUU qtrsm_LNUU
#define TRSM_LRUN qtrsm_LNUN
#define TRSM_LRLU qtrsm_LNLU
#define TRSM_LRLN qtrsm_LNLN
#define TRSM_LCUU qtrsm_LTUU
#define TRSM_LCUN qtrsm_LTUN
#define TRSM_LCLU qtrsm_LTLU
#define TRSM_LCLN qtrsm_LTLN

#define TRSM_RNUU qtrsm_RNUU
#define TRSM_RNUN qtrsm_RNUN
#define TRSM_RNLU qtrsm_RNLU
#define TRSM_RNLN qtrsm_RNLN
#define TRSM_RTUU qtrsm_RTUU
#define TRSM_RTUN qtrsm_RTUN
#define TRSM_RTLU qtrsm_RTLU
#define TRSM_RTLN qtrsm_RTLN
#define TRSM_RRUU qtrsm_RNUU
#define TRSM_RRUN qtrsm_RNUN
#define TRSM_RRLU qtrsm_RNLU
#define TRSM_RRLN qtrsm_RNLN
#define TRSM_RCUU qtrsm_RTUU
#define TRSM_RCUN qtrsm_RTUN
#define TRSM_RCLU qtrsm_RTLU
#define TRSM_RCLN qtrsm_RTLN

#define  SYRK_UN   qsyrk_UN
#define  SYRK_UT   qsyrk_UT
#define  SYRK_LN   qsyrk_LN
#define  SYRK_LT   qsyrk_LT
#define  SYRK_UR   qsyrk_UN
#define  SYRK_UC   qsyrk_UT
#define  SYRK_LR   qsyrk_LN
#define  SYRK_LC   qsyrk_LT

#define  SYR2K_UN  qsyr2k_UN
#define  SYR2K_UT  qsyr2k_UT
#define  SYR2K_LN  qsyr2k_LN
#define  SYR2K_LT  qsyr2k_LT
#define  SYR2K_UR  qsyr2k_UN
#define  SYR2K_UC  qsyr2k_UT
#define  SYR2K_LR  qsyr2k_LN
#define  SYR2K_LC  qsyr2k_LT

#define  HERK_UN   qsyrk_UN
#define  HERK_UT   qsyrk_UT
#define  HERK_LN   qsyrk_LN
#define  HERK_LT   qsyrk_LT
#define  HERK_UR   qsyrk_UN
#define  HERK_UC   qsyrk_UT
#define  HERK_LR   qsyrk_LN
#define  HERK_LC   qsyrk_LT

#define SYRK_KERNEL_U qsyrk_kernel_U
#define SYRK_KERNEL_L qsyrk_kernel_L

#define  SYMCOPY_L qsymcopy_L
#define  SYMCOPY_U qsymcopy_U
#define  HEMCOPY_L qsymcopy_L
#define  HEMCOPY_U qsymcopy_U

#elif defined(DOUBLE)

/* REAL && DOUBLE */
#define   GEMM_BETA   dgemm_beta
#define   GEMM_ONCOPY dgemm_oncopy
#define   GEMM_OTCOPY dgemm_otcopy
#if DGEMM_UNROLL_M != DGEMM_UNROLL_N
#define   GEMM_INCOPY dgemm_incopy
#define   GEMM_ITCOPY dgemm_itcopy
#else
#define   GEMM_INCOPY dgemm_oncopy
#define   GEMM_ITCOPY dgemm_otcopy
#endif
#define   GEMM_KERNEL_N dgemm_kernel
#define   GEMM_KERNEL_L dgemm_kernel
#define   GEMM_KERNEL_R dgemm_kernel
#define   GEMM_KERNEL_B dgemm_kernel
#define   GEMM_KERNEL_N_B0 dgemm_kernel_b0
#define   GEMM_KERNEL_L_B0 dgemm_kernel_b0
#define   GEMM_KERNEL_R_B0 dgemm_kernel_b0
#define   GEMM_KERNEL_B_B0 dgemm_kernel_b0
#define   GEMM_NN     dgemm_nn
#define   GEMM_TN     dgemm_tn
#define   GEMM_RN     dgemm_nn
#define   GEMM_CN     dgemm_tn
#define   GEMM_NT     dgemm_nt
#define   GEMM_TT     dgemm_tt
#define   GEMM_RT     dgemm_nt
#define   GEMM_CT     dgemm_tt
#define   GEMM_NR     dgemm_nn
#define   GEMM_TR     dgemm_tn
#define   GEMM_RR     dgemm_nn
#define   GEMM_CR     dgemm_tn
#define   GEMM_NC     dgemm_nt
#define   GEMM_TC     dgemm_tt
#define   GEMM_RC     dgemm_nt
#define   GEMM_CC     dgemm_tt

#define TRMM_KERNEL_LN  dtrmm_kernel_LN
#define TRMM_KERNEL_LT  dtrmm_kernel_LT
#define TRMM_KERNEL_LR  dtrmm_kernel_LN
#define TRMM_KERNEL_LC  dtrmm_kernel_LT
#define TRMM_KERNEL_RN  dtrmm_kernel_RN
#define TRMM_KERNEL_RT  dtrmm_kernel_RT
#define TRMM_KERNEL_RR  dtrmm_kernel_RN
#define TRMM_KERNEL_RC  dtrmm_kernel_RT

#define TRSM_KERNEL_LN  dtrsm_kernel_LN
#define TRSM_KERNEL_LT  dtrsm_kernel_LT
#define TRSM_KERNEL_LR  dtrsm_kernel_LN
#define TRSM_KERNEL_LC  dtrsm_kernel_LT
#define TRSM_KERNEL_RN  dtrsm_kernel_RN
#define TRSM_KERNEL_RT  dtrsm_kernel_RT
#define TRSM_KERNEL_RR  dtrsm_kernel_RN
#define TRSM_KERNEL_RC  dtrsm_kernel_RT

#define TRMM_LNUU dtrmm_LNUU
#define TRMM_LNUN dtrmm_LNUN
#define TRMM_LNLU dtrmm_LNLU
#define TRMM_LNLN dtrmm_LNLN
#define TRMM_LTUU dtrmm_LTUU
#define TRMM_LTUN dtrmm_LTUN
#define TRMM_LTLU dtrmm_LTLU
#define TRMM_LTLN dtrmm_LTLN
#define TRMM_LRUU dtrmm_LNUU
#define TRMM_LRUN dtrmm_LNUN
#define TRMM_LRLU dtrmm_LNLU
#define TRMM_LRLN dtrmm_LNLN
#define TRMM_LCUU dtrmm_LTUU
#define TRMM_LCUN dtrmm_LTUN
#define TRMM_LCLU dtrmm_LTLU
#define TRMM_LCLN dtrmm_LTLN

#define TRMM_RNUU dtrmm_RNUU
#define TRMM_RNUN dtrmm_RNUN
#define TRMM_RNLU dtrmm_RNLU
#define TRMM_RNLN dtrmm_RNLN
#define TRMM_RTUU dtrmm_RTUU
#define TRMM_RTUN dtrmm_RTUN
#define TRMM_RTLU dtrmm_RTLU
#define TRMM_RTLN dtrmm_RTLN
#define TRMM_RRUU dtrmm_RNUU
#define TRMM_RRUN dtrmm_RNUN
#define TRMM_RRLU dtrmm_RNLU
#define TRMM_RRLN dtrmm_RNLN
#define TRMM_RCUU dtrmm_RTUU
#define TRMM_RCUN dtrmm_RTUN
#define TRMM_RCLU dtrmm_RTLU
#define TRMM_RCLN dtrmm_RTLN

#define TRSM_LNUU dtrsm_LNUU
#define TRSM_LNUN dtrsm_LNUN
#define TRSM_LNLU dtrsm_LNLU
#define TRSM_LNLN dtrsm_LNLN
#define TRSM_LTUU dtrsm_LTUU
#define TRSM_LTUN dtrsm_LTUN
#define TRSM_LTLU dtrsm_LTLU
#define TRSM_LTLN dtrsm_LTLN
#define TRSM_LRUU dtrsm_LNUU
#define TRSM_LRUN dtrsm_LNUN
#define TRSM_LRLU dtrsm_LNLU
#define TRSM_LRLN dtrsm_LNLN
#define TRSM_LCUU dtrsm_LTUU
#define TRSM_LCUN dtrsm_LTUN
#define TRSM_LCLU dtrsm_LTLU
#define TRSM_LCLN dtrsm_LTLN

#define TRSM_RNUU dtrsm_RNUU
#define TRSM_RNUN dtrsm_RNUN
#define TRSM_RNLU dtrsm_RNLU
#define TRSM_RNLN dtrsm_RNLN
#define TRSM_RTUU dtrsm_RTUU
#define TRSM_RTUN dtrsm_RTUN
#define TRSM_RTLU dtrsm_RTLU
#define TRSM_RTLN dtrsm_RTLN
#define TRSM_RRUU dtrsm_RNUU
#define TRSM_RRUN dtrsm_RNUN
#define TRSM_RRLU dtrsm_RNLU
#define TRSM_RRLN dtrsm_RNLN
#define TRSM_RCUU dtrsm_RTUU
#define TRSM_RCUN dtrsm_RTUN
#define TRSM_RCLU dtrsm_RTLU
#define TRSM_RCLN dtrsm_RTLN

#define  SYRK_UN   dsyrk_UN
#define  SYRK_UT   dsyrk_UT
#define  SYRK_LN   dsyrk_LN
#define  SYRK_LT   dsyrk_LT
#define  SYRK_UR   dsyrk_UN
#define  SYRK_UC   dsyrk_UT
#define  SYRK_LR   dsyrk_LN
#define  SYRK_LC   dsyrk_LT

#define  HERK_UN   dsyrk_UN
#define  HERK_UT   dsyrk_UT
#define  HERK_LN   dsyrk_LN
#define  HERK_LT   dsyrk_LT
#define  HERK_UR   dsyrk_UN
#define  HERK_UC   dsyrk_UT
#define  HERK_LR   dsyrk_LN
#define  HERK_LC   dsyrk_LT

#define SYRK_KERNEL_U dsyrk_kernel_U
#define SYRK_KERNEL_L dsyrk_kernel_L

#define  SYR2K_UN  dsyr2k_UN
#define  SYR2K_UT  dsyr2k_UT
#define  SYR2K_LN  dsyr2k_LN
#define  SYR2K_LT  dsyr2k_LT
#define  SYR2K_UR  dsyr2k_UN
#define  SYR2K_UC  dsyr2k_UT
#define  SYR2K_LR  dsyr2k_LN
#define  SYR2K_LC  dsyr2k_LT

#define  SYMCOPY_L dsymcopy_L
#define  SYMCOPY_U dsymcopy_U
#define  HEMCOPY_L dsymcopy_L
#define  HEMCOPY_U dsymcopy_U

#else
/* REAL && SINGLE */

#define   GEMM_BETA   sgemm_beta
#define   GEMM_ONCOPY sgemm_oncopy
#define   GEMM_OTCOPY sgemm_otcopy
#if SGEMM_UNROLL_M != SGEMM_UNROLL_N
#define   GEMM_INCOPY sgemm_incopy
#define   GEMM_ITCOPY sgemm_itcopy
#else
#define   GEMM_INCOPY sgemm_oncopy
#define   GEMM_ITCOPY sgemm_otcopy
#endif
#define   GEMM_KERNEL_N sgemm_kernel
#define   GEMM_KERNEL_L sgemm_kernel
#define   GEMM_KERNEL_R sgemm_kernel
#define   GEMM_KERNEL_B sgemm_kernel
#define   GEMM_KERNEL_N_B0 sgemm_kernel_b0
#define   GEMM_KERNEL_L_B0 sgemm_kernel_b0
#define   GEMM_KERNEL_R_B0 sgemm_kernel_b0
#define   GEMM_KERNEL_B_B0 sgemm_kernel_b0
#define   GEMM_NN     sgemm_nn
#define   GEMM_TN     sgemm_tn
#define   GEMM_RN     sgemm_nn
#define   GEMM_CN     sgemm_tn
#define   GEMM_NT     sgemm_nt
#define   GEMM_TT     sgemm_tt
#define   GEMM_RT     sgemm_nt
#define   GEMM_CT     sgemm_tt
#define   GEMM_NR     sgemm_nn
#define   GEMM_TR     sgemm_tn
#define   GEMM_RR     sgemm_nn
#define   GEMM_CR     sgemm_tn
#define   GEMM_NC     sgemm_nt
#define   GEMM_TC     sgemm_tt
#define   GEMM_RC     sgemm_nt
#define   GEMM_CC     sgemm_tt

#define TRMM_KERNEL_LN  strmm_kernel_LN
#define TRMM_KERNEL_LT  strmm_kernel_LT
#define TRMM_KERNEL_LR  strmm_kernel_LN
#define TRMM_KERNEL_LC  strmm_kernel_LT
#define TRMM_KERNEL_RN  strmm_kernel_RN
#define TRMM_KERNEL_RT  strmm_kernel_RT
#define TRMM_KERNEL_RR  strmm_kernel_RN
#define TRMM_KERNEL_RC  strmm_kernel_RT

#define TRSM_KERNEL_LN  strsm_kernel_LN
#define TRSM_KERNEL_LT  strsm_kernel_LT
#define TRSM_KERNEL_LR  strsm_kernel_LN
#define TRSM_KERNEL_LC  strsm_kernel_LT
#define TRSM_KERNEL_RN  strsm_kernel_RN
#define TRSM_KERNEL_RT  strsm_kernel_RT
#define TRSM_KERNEL_RR  strsm_kernel_RN
#define TRSM_KERNEL_RC  strsm_kernel_RT

#define TRMM_LNUU strmm_LNUU
#define TRMM_LNUN strmm_LNUN
#define TRMM_LNLU strmm_LNLU
#define TRMM_LNLN strmm_LNLN
#define TRMM_LTUU strmm_LTUU
#define TRMM_LTUN strmm_LTUN
#define TRMM_LTLU strmm_LTLU
#define TRMM_LTLN strmm_LTLN
#define TRMM_LRUU strmm_LNUU
#define TRMM_LRUN strmm_LNUN
#define TRMM_LRLU strmm_LNLU
#define TRMM_LRLN strmm_LNLN
#define TRMM_LCUU strmm_LTUU
#define TRMM_LCUN strmm_LTUN
#define TRMM_LCLU strmm_LTLU
#define TRMM_LCLN strmm_LTLN

#define TRMM_RNUU strmm_RNUU
#define TRMM_RNUN strmm_RNUN
#define TRMM_RNLU strmm_RNLU
#define TRMM_RNLN strmm_RNLN
#define TRMM_RTUU strmm_RTUU
#define TRMM_RTUN strmm_RTUN
#define TRMM_RTLU strmm_RTLU
#define TRMM_RTLN strmm_RTLN
#define TRMM_RRUU strmm_RNUU
#define TRMM_RRUN strmm_RNUN
#define TRMM_RRLU strmm_RNLU
#define TRMM_RRLN strmm_RNLN
#define TRMM_RCUU strmm_RTUU
#define TRMM_RCUN strmm_RTUN
#define TRMM_RCLU strmm_RTLU
#define TRMM_RCLN strmm_RTLN

#define TRSM_LNUU strsm_LNUU
#define TRSM_LNUN strsm_LNUN
#define TRSM_LNLU strsm_LNLU
#define TRSM_LNLN strsm_LNLN
#define TRSM_LTUU strsm_LTUU
#define TRSM_LTUN strsm_LTUN
#define TRSM_LTLU strsm_LTLU
#define TRSM_LTLN strsm_LTLN
#define TRSM_LRUU strsm_LNUU
#define TRSM_LRUN strsm_LNUN
#define TRSM_LRLU strsm_LNLU
#define TRSM_LRLN strsm_LNLN
#define TRSM_LCUU strsm_LTUU
#define TRSM_LCUN strsm_LTUN
#define TRSM_LCLU strsm_LTLU
#define TRSM_LCLN strsm_LTLN

#define TRSM_RNUU strsm_RNUU
#define TRSM_RNUN strsm_RNUN
#define TRSM_RNLU strsm_RNLU
#define TRSM_RNLN strsm_RNLN
#define TRSM_RTUU strsm_RTUU
#define TRSM_RTUN strsm_RTUN
#define TRSM_RTLU strsm_RTLU
#define TRSM_RTLN strsm_RTLN
#define TRSM_RRUU strsm_RNUU
#define TRSM_RRUN strsm_RNUN
#define TRSM_RRLU strsm_RNLU
#define TRSM_RRLN strsm_RNLN
#define TRSM_RCUU strsm_RTUU
#define TRSM_RCUN strsm_RTUN
#define TRSM_RCLU strsm_RTLU
#define TRSM_RCLN strsm_RTLN

#define  SYRK_UN   ssyrk_UN
#define  SYRK_UT   ssyrk_UT
#define  SYRK_LN   ssyrk_LN
#define  SYRK_LT   ssyrk_LT
#define  SYRK_UR   ssyrk_UN
#define  SYRK_UC   ssyrk_UT
#define  SYRK_LR   ssyrk_LN
#define  SYRK_LC   ssyrk_LT

#define  SYR2K_UN  ssyr2k_UN
#define  SYR2K_UT  ssyr2k_UT
#define  SYR2K_LN  ssyr2k_LN
#define  SYR2K_LT  ssyr2k_LT
#define  SYR2K_UR  ssyr2k_UN
#define  SYR2K_UC  ssyr2k_UT
#define  SYR2K_LR  ssyr2k_LN
#define  SYR2K_LC  ssyr2k_LT

#define  HERK_UN   ssyrk_UN
#define  HERK_UT   ssyrk_UT
#define  HERK_LN   ssyrk_LN
#define  HERK_LT   ssyrk_LT
#define  HERK_UR   ssyrk_UN
#define  HERK_UC   ssyrk_UT
#define  HERK_LR   ssyrk_LN
#define  HERK_LC   ssyrk_LT

#define SYRK_KERNEL_U ssyrk_kernel_U
#define SYRK_KERNEL_L ssyrk_kernel_L

#define  SYMCOPY_L ssymcopy_L
#define  SYMCOPY_U ssymcopy_U
#define  HEMCOPY_L ssymcopy_L
#define  HEMCOPY_U ssymcopy_U
#endif

#else

#ifdef XDOUBLE
/* COMPLEX && EXTENDED */
#define   GEMM_BETA   xgemm_beta
#define   GEMM_ONCOPY xgemm_oncopy
#define   GEMM_OTCOPY xgemm_otcopy
#if XGEMM_UNROLL_M != XGEMM_UNROLL_N
#define   GEMM_INCOPY xgemm_incopy
#define   GEMM_ITCOPY xgemm_itcopy
#else
#define   GEMM_INCOPY xgemm_oncopy
#define   GEMM_ITCOPY xgemm_otcopy
#endif
#define   GEMM_KERNEL_N xgemm_kernel_n
#define   GEMM_KERNEL_L xgemm_kernel_l
#define   GEMM_KERNEL_R xgemm_kernel_r
#define   GEMM_KERNEL_B xgemm_kernel_b
#define   GEMM_KERNEL_N_B0 xgemm_kernel_n_b0
#define   GEMM_KERNEL_L_B0 xgemm_kernel_l_b0
#define   GEMM_KERNEL_R_B0 xgemm_kernel_r_b0
#define   GEMM_KERNEL_B_B0 xgemm_kernel_b_b0
#define   GEMM_NN     xgemm_nn
#define   GEMM_CN     xgemm_cn
#define   GEMM_TN     xgemm_tn
#define   GEMM_NC     xgemm_nc
#define   GEMM_NT     xgemm_nt
#define   GEMM_CC     xgemm_cc
#define   GEMM_CT     xgemm_ct
#define   GEMM_TC     xgemm_tc
#define   GEMM_TT     xgemm_tt
#define   GEMM_NR     xgemm_nr
#define   GEMM_TR     xgemm_tr
#define   GEMM_CR     xgemm_cr
#define   GEMM_RN     xgemm_rn
#define   GEMM_RT     xgemm_rt
#define   GEMM_RC     xgemm_rc
#define   GEMM_RR     xgemm_rr

#define TRMM_KERNEL_LN  xtrmm_kernel_LN
#define TRMM_KERNEL_LT  xtrmm_kernel_LT
#define TRMM_KERNEL_LR  xtrmm_kernel_LR
#define TRMM_KERNEL_LC  xtrmm_kernel_LC
#define TRMM_KERNEL_RN  xtrmm_kernel_RN
#define TRMM_KERNEL_RT  xtrmm_kernel_RT
#define TRMM_KERNEL_RR  xtrmm_kernel_RR
#define TRMM_KERNEL_RC  xtrmm_kernel_RC

#define TRSM_KERNEL_LN  xtrsm_kernel_LN
#define TRSM_KERNEL_LT  xtrsm_kernel_LT
#define TRSM_KERNEL_LR  xtrsm_kernel_LR
#define TRSM_KERNEL_LC  xtrsm_kernel_LC
#define TRSM_KERNEL_RN  xtrsm_kernel_RN
#define TRSM_KERNEL_RT  xtrsm_kernel_RT
#define TRSM_KERNEL_RR  xtrsm_kernel_RR
#define TRSM_KERNEL_RC  xtrsm_kernel_RC

#define TRMM_LNUU xtrmm_LNUU
#define TRMM_LNUN xtrmm_LNUN
#define TRMM_LNLU xtrmm_LNLU
#define TRMM_LNLN xtrmm_LNLN
#define TRMM_LTUU xtrmm_LTUU
#define TRMM_LTUN xtrmm_LTUN
#define TRMM_LTLU xtrmm_LTLU
#define TRMM_LTLN xtrmm_LTLN
#define TRMM_LRUU xtrmm_LRUU
#define TRMM_LRUN xtrmm_LRUN
#define TRMM_LRLU xtrmm_LRLU
#define TRMM_LRLN xtrmm_LRLN
#define TRMM_LCUU xtrmm_LCUU
#define TRMM_LCUN xtrmm_LCUN
#define TRMM_LCLU xtrmm_LCLU
#define TRMM_LCLN xtrmm_LCLN

#define TRMM_RNUU xtrmm_RNUU
#define TRMM_RNUN xtrmm_RNUN
#define TRMM_RNLU xtrmm_RNLU
#define TRMM_RNLN xtrmm_RNLN
#define TRMM_RTUU xtrmm_RTUU
#define TRMM_RTUN xtrmm_RTUN
#define TRMM_RTLU xtrmm_RTLU
#define TRMM_RTLN xtrmm_RTLN
#define TRMM_RRUU xtrmm_RRUU
#define TRMM_RRUN xtrmm_RRUN
#define TRMM_RRLU xtrmm_RRLU
#define TRMM_RRLN xtrmm_RRLN
#define TRMM_RCUU xtrmm_RCUU
#define TRMM_RCUN xtrmm_RCUN
#define TRMM_RCLU xtrmm_RCLU
#define TRMM_RCLN xtrmm_RCLN

#define TRSM_LNUU xtrsm_LNUU
#define TRSM_LNUN xtrsm_LNUN
#define TRSM_LNLU xtrsm_LNLU
#define TRSM_LNLN xtrsm_LNLN
#define TRSM_LTUU xtrsm_LTUU
#define TRSM_LTUN xtrsm_LTUN
#define TRSM_LTLU xtrsm_LTLU
#define TRSM_LTLN xtrsm_LTLN
#define TRSM_LRUU xtrsm_LRUU
#define TRSM_LRUN xtrsm_LRUN
#define TRSM_LRLU xtrsm_LRLU
#define TRSM_LRLN xtrsm_LRLN
#define TRSM_LCUU xtrsm_LCUU
#define TRSM_LCUN xtrsm_LCUN
#define TRSM_LCLU xtrsm_LCLU
#define TRSM_LCLN xtrsm_LCLN

#define TRSM_RNUU xtrsm_RNUU
#define TRSM_RNUN xtrsm_RNUN
#define TRSM_RNLU xtrsm_RNLU
#define TRSM_RNLN xtrsm_RNLN
#define TRSM_RTUU xtrsm_RTUU
#define TRSM_RTUN xtrsm_RTUN
#define TRSM_RTLU xtrsm_RTLU
#define TRSM_RTLN xtrsm_RTLN
#define TRSM_RRUU xtrsm_RRUU
#define TRSM_RRUN xtrsm_RRUN
#define TRSM_RRLU xtrsm_RRLU
#define TRSM_RRLN xtrsm_RRLN
#define TRSM_RCUU xtrsm_RCUU
#define TRSM_RCUN xtrsm_RCUN
#define TRSM_RCLU xtrsm_RCLU
#define TRSM_RCLN xtrsm_RCLN

#define  SYRK_UN   xsyrk_UN
#define  SYRK_UT   xsyrk_UT
#define  SYRK_LN   xsyrk_LN
#define  SYRK_LT   xsyrk_LT
#define  SYRK_UR   xsyrk_UR
#define  SYRK_UC   xsyrk_UC
#define  SYRK_LR   xsyrk_LR
#define  SYRK_LC   xsyrk_LC

#define  HERK_UN   xherk_UN
#define  HERK_UT   xherk_UT
#define  HERK_LN   xherk_LN
#define  HERK_LT   xherk_LT
#define  HERK_UR   xherk_UR
#define  HERK_UC   xherk_UC
#define  HERK_LR   xherk_LR
#define  HERK_LC   xherk_LC

#define SYRK_KERNEL_U xsyrk_kernel_U
#define SYRK_KERNEL_L xsyrk_kernel_L

#define  SYR2K_UN  xsyr2k_UN
#define  SYR2K_UT  xsyr2k_UT
#define  SYR2K_LN  xsyr2k_LN
#define  SYR2K_LT  xsyr2k_LT
#define  SYR2K_UR  xsyr2k_UR
#define  SYR2K_UC  xsyr2k_UC
#define  SYR2K_LR  xsyr2k_LR
#define  SYR2K_LC  xsyr2k_LC

#define  SYMCOPY_L xsymcopy_L
#define  SYMCOPY_U xsymcopy_U
#define  HEMCOPY_L xhemcopy_L
#define  HEMCOPY_U xhemcopy_U

#elif defined(DOUBLE)
/* COMPLEX && DOUBLE */
#define   GEMM_BETA   zgemm_beta
#define   GEMM_ONCOPY zgemm_oncopy
#define   GEMM_OTCOPY zgemm_otcopy
#if ZGEMM_UNROLL_M != ZGEMM_UNROLL_N
#define   GEMM_INCOPY zgemm_incopy
#define   GEMM_ITCOPY zgemm_itcopy
#else
#define   GEMM_INCOPY zgemm_oncopy
#define   GEMM_ITCOPY zgemm_otcopy
#endif
#define   GEMM_KERNEL_N zgemm_kernel_n
#define   GEMM_KERNEL_L zgemm_kernel_l
#define   GEMM_KERNEL_R zgemm_kernel_r
#define   GEMM_KERNEL_B zgemm_kernel_b
#define   GEMM_KERNEL_N_B0 zgemm_kernel_n_b0
#define   GEMM_KERNEL_L_B0 zgemm_kernel_l_b0
#define   GEMM_KERNEL_R_B0 zgemm_kernel_r_b0
#define   GEMM_KERNEL_B_B0 zgemm_kernel_b_b0
#define   GEMM_NN     zgemm_nn
#define   GEMM_CN     zgemm_cn
#define   GEMM_TN     zgemm_tn
#define   GEMM_NC     zgemm_nc
#define   GEMM_NT     zgemm_nt
#define   GEMM_CC     zgemm_cc
#define   GEMM_CT     zgemm_ct
#define   GEMM_TC     zgemm_tc
#define   GEMM_TT     zgemm_tt
#define   GEMM_NR     zgemm_nr
#define   GEMM_TR     zgemm_tr
#define   GEMM_CR     zgemm_cr
#define   GEMM_RN     zgemm_rn
#define   GEMM_RT     zgemm_rt
#define   GEMM_RC     zgemm_rc
#define   GEMM_RR     zgemm_rr

#define TRMM_KERNEL_LN  ztrmm_kernel_LN
#define TRMM_KERNEL_LT  ztrmm_kernel_LT
#define TRMM_KERNEL_LR  ztrmm_kernel_LR
#define TRMM_KERNEL_LC  ztrmm_kernel_LC
#define TRMM_KERNEL_RN  ztrmm_kernel_RN
#define TRMM_KERNEL_RT  ztrmm_kernel_RT
#define TRMM_KERNEL_RR  ztrmm_kernel_RR
#define TRMM_KERNEL_RC  ztrmm_kernel_RC

#define TRSM_KERNEL_LN  ztrsm_kernel_LN
#define TRSM_KERNEL_LT  ztrsm_kernel_LT
#define TRSM_KERNEL_LR  ztrsm_kernel_LR
#define TRSM_KERNEL_LC  ztrsm_kernel_LC
#define TRSM_KERNEL_RN  ztrsm_kernel_RN
#define TRSM_KERNEL_RT  ztrsm_kernel_RT
#define TRSM_KERNEL_RR  ztrsm_kernel_RR
#define TRSM_KERNEL_RC  ztrsm_kernel_RC

#define TRMM_LNUU ztrmm_LNUU
#define TRMM_LNUN ztrmm_LNUN
#define TRMM_LNLU ztrmm_LNLU
#define TRMM_LNLN ztrmm_LNLN
#define TRMM_LTUU ztrmm_LTUU
#define TRMM_LTUN ztrmm_LTUN
#define TRMM_LTLU ztrmm_LTLU
#define TRMM_LTLN ztrmm_LTLN
#define TRMM_LRUU ztrmm_LRUU
#define TRMM_LRUN ztrmm_LRUN
#define TRMM_LRLU ztrmm_LRLU
#define TRMM_LRLN ztrmm_LRLN
#define TRMM_LCUU ztrmm_LCUU
#define TRMM_LCUN ztrmm_LCUN
#define TRMM_LCLU ztrmm_LCLU
#define TRMM_LCLN ztrmm_LCLN

#define TRMM_RNUU ztrmm_RNUU
#define TRMM_RNUN ztrmm_RNUN
#define TRMM_RNLU ztrmm_RNLU
#define TRMM_RNLN ztrmm_RNLN
#define TRMM_RTUU ztrmm_RTUU
#define TRMM_RTUN ztrmm_RTUN
#define TRMM_RTLU ztrmm_RTLU
#define TRMM_RTLN ztrmm_RTLN
#define TRMM_RRUU ztrmm_RRUU
#define TRMM_RRUN ztrmm_RRUN
#define TRMM_RRLU ztrmm_RRLU
#define TRMM_RRLN ztrmm_RRLN
#define TRMM_RCUU ztrmm_RCUU
#define TRMM_RCUN ztrmm_RCUN
#define TRMM_RCLU ztrmm_RCLU
#define TRMM_RCLN ztrmm_RCLN

#define TRSM_LNUU ztrsm_LNUU
#define TRSM_LNUN ztrsm_LNUN
#define TRSM_LNLU ztrsm_LNLU
#define TRSM_LNLN ztrsm_LNLN
#define TRSM_LTUU ztrsm_LTUU
#define TRSM_LTUN ztrsm_LTUN
#define TRSM_LTLU ztrsm_LTLU
#define TRSM_LTLN ztrsm_LTLN
#define TRSM_LRUU ztrsm_LRUU
#define TRSM_LRUN ztrsm_LRUN
#define TRSM_LRLU ztrsm_LRLU
#define TRSM_LRLN ztrsm_LRLN
#define TRSM_LCUU ztrsm_LCUU
#define TRSM_LCUN ztrsm_LCUN
#define TRSM_LCLU ztrsm_LCLU
#define TRSM_LCLN ztrsm_LCLN

#define TRSM_RNUU ztrsm_RNUU
#define TRSM_RNUN ztrsm_RNUN
#define TRSM_RNLU ztrsm_RNLU
#define TRSM_RNLN ztrsm_RNLN
#define TRSM_RTUU ztrsm_RTUU
#define TRSM_RTUN ztrsm_RTUN
#define TRSM_RTLU ztrsm_RTLU
#define TRSM_RTLN ztrsm_RTLN
#define TRSM_RRUU ztrsm_RRUU
#define TRSM_RRUN ztrsm_RRUN
#define TRSM_RRLU ztrsm_RRLU
#define TRSM_RRLN ztrsm_RRLN
#define TRSM_RCUU ztrsm_RCUU
#define TRSM_RCUN ztrsm_RCUN
#define TRSM_RCLU ztrsm_RCLU
#define TRSM_RCLN ztrsm_RCLN

#define  SYRK_UN   zsyrk_UN
#define  SYRK_UT   zsyrk_UT
#define  SYRK_LN   zsyrk_LN
#define  SYRK_LT   zsyrk_LT
#define  SYRK_UR   zsyrk_UR
#define  SYRK_UC   zsyrk_UC
#define  SYRK_LR   zsyrk_LR
#define  SYRK_LC   zsyrk_LC

#define  HERK_UN   zherk_UN
#define  HERK_UT   zherk_UT
#define  HERK_LN   zherk_LN
#define  HERK_LT   zherk_LT
#define  HERK_UR   zherk_UR
#define  HERK_UC   zherk_UC
#define  HERK_LR   zherk_LR
#define  HERK_LC   zherk_LC

#define SYRK_KERNEL_U zsyrk_kernel_U
#define SYRK_KERNEL_L zsyrk_kernel_L

#define  SYR2K_UN  zsyr2k_UN
#define  SYR2K_UT  zsyr2k_UT
#define  SYR2K_LN  zsyr2k_LN
#define  SYR2K_LT  zsyr2k_LT
#define  SYR2K_UR  zsyr2k_UR
#define  SYR2K_UC  zsyr2k_UC
#define  SYR2K_LR  zsyr2k_LR
#define  SYR2K_LC  zsyr2k_LC

#define  SYMCOPY_L zsymcopy_L
#define  SYMCOPY_U zsymcopy_U
#define  HEMCOPY_L zhemcopy_L
#define  HEMCOPY_U zhemcopy_U

#else

/* COMPLEX && SINGLE */
#define   GEMM_BETA   cgemm_beta
#define   GEMM_ONCOPY cgemm_oncopy
#define   GEMM_OTCOPY cgemm_otcopy
#if CGEMM_UNROLL_M != CGEMM_UNROLL_N
#define   GEMM_INCOPY cgemm_incopy
#define   GEMM_ITCOPY cgemm_itcopy
#else
#define   GEMM_INCOPY cgemm_oncopy
#define   GEMM_ITCOPY cgemm_otcopy
#endif
#define   GEMM_KERNEL_N cgemm_kernel_n
#define   GEMM_KERNEL_L cgemm_kernel_l
#define   GEMM_KERNEL_R cgemm_kernel_r
#define   GEMM_KERNEL_B cgemm_kernel_b
#define   GEMM_KERNEL_N_B0 cgemm_kernel_n_b0
#define   GEMM_KERNEL_L_B0 cgemm_kernel_l_b0
#define   GEMM_KERNEL_R_B0 cgemm_kernel_r_b0
#define   GEMM_KERNEL_B_B0 cgemm_kernel_b_b0
#define   GEMM_NN     cgemm_nn
#define   GEMM_CN     cgemm_cn
#define   GEMM_TN     cgemm_tn
#define   GEMM_NC     cgemm_nc
#define   GEMM_NT     cgemm_nt
#define   GEMM_CC     cgemm_cc
#define   GEMM_CT     cgemm_ct
#define   GEMM_TC     cgemm_tc
#define   GEMM_TT     cgemm_tt
#define   GEMM_NR     cgemm_nr
#define   GEMM_TR     cgemm_tr
#define   GEMM_CR     cgemm_cr
#define   GEMM_RN     cgemm_rn
#define   GEMM_RT     cgemm_rt
#define   GEMM_RC     cgemm_rc
#define   GEMM_RR     cgemm_rr

#define TRMM_KERNEL_LN  ctrmm_kernel_LN
#define TRMM_KERNEL_LT  ctrmm_kernel_LT
#define TRMM_KERNEL_LR  ctrmm_kernel_LR
#define TRMM_KERNEL_LC  ctrmm_kernel_LC
#define TRMM_KERNEL_RN  ctrmm_kernel_RN
#define TRMM_KERNEL_RT  ctrmm_kernel_RT
#define TRMM_KERNEL_RR  ctrmm_kernel_RR
#define TRMM_KERNEL_RC  ctrmm_kernel_RC

#define TRSM_KERNEL_LN  ctrsm_kernel_LN
#define TRSM_KERNEL_LT  ctrsm_kernel_LT
#define TRSM_KERNEL_LR  ctrsm_kernel_LR
#define TRSM_KERNEL_LC  ctrsm_kernel_LC
#define TRSM_KERNEL_RN  ctrsm_kernel_RN
#define TRSM_KERNEL_RT  ctrsm_kernel_RT
#define TRSM_KERNEL_RR  ctrsm_kernel_RR
#define TRSM_KERNEL_RC  ctrsm_kernel_RC

#define TRMM_LNUU ctrmm_LNUU
#define TRMM_LNUN ctrmm_LNUN
#define TRMM_LNLU ctrmm_LNLU
#define TRMM_LNLN ctrmm_LNLN
#define TRMM_LTUU ctrmm_LTUU
#define TRMM_LTUN ctrmm_LTUN
#define TRMM_LTLU ctrmm_LTLU
#define TRMM_LTLN ctrmm_LTLN
#define TRMM_LRUU ctrmm_LRUU
#define TRMM_LRUN ctrmm_LRUN
#define TRMM_LRLU ctrmm_LRLU
#define TRMM_LRLN ctrmm_LRLN
#define TRMM_LCUU ctrmm_LCUU
#define TRMM_LCUN ctrmm_LCUN
#define TRMM_LCLU ctrmm_LCLU
#define TRMM_LCLN ctrmm_LCLN

#define TRMM_RNUU ctrmm_RNUU
#define TRMM_RNUN ctrmm_RNUN
#define TRMM_RNLU ctrmm_RNLU
#define TRMM_RNLN ctrmm_RNLN
#define TRMM_RTUU ctrmm_RTUU
#define TRMM_RTUN ctrmm_RTUN
#define TRMM_RTLU ctrmm_RTLU
#define TRMM_RTLN ctrmm_RTLN
#define TRMM_RRUU ctrmm_RRUU
#define TRMM_RRUN ctrmm_RRUN
#define TRMM_RRLU ctrmm_RRLU
#define TRMM_RRLN ctrmm_RRLN
#define TRMM_RCUU ctrmm_RCUU
#define TRMM_RCUN ctrmm_RCUN
#define TRMM_RCLU ctrmm_RCLU
#define TRMM_RCLN ctrmm_RCLN

#define TRSM_LNUU ctrsm_LNUU
#define TRSM_LNUN ctrsm_LNUN
#define TRSM_LNLU ctrsm_LNLU
#define TRSM_LNLN ctrsm_LNLN
#define TRSM_LTUU ctrsm_LTUU
#define TRSM_LTUN ctrsm_LTUN
#define TRSM_LTLU ctrsm_LTLU
#define TRSM_LTLN ctrsm_LTLN
#define TRSM_LRUU ctrsm_LRUU
#define TRSM_LRUN ctrsm_LRUN
#define TRSM_LRLU ctrsm_LRLU
#define TRSM_LRLN ctrsm_LRLN
#define TRSM_LCUU ctrsm_LCUU
#define TRSM_LCUN ctrsm_LCUN
#define TRSM_LCLU ctrsm_LCLU
#define TRSM_LCLN ctrsm_LCLN

#define TRSM_RNUU ctrsm_RNUU
#define TRSM_RNUN ctrsm_RNUN
#define TRSM_RNLU ctrsm_RNLU
#define TRSM_RNLN ctrsm_RNLN
#define TRSM_RTUU ctrsm_RTUU
#define TRSM_RTUN ctrsm_RTUN
#define TRSM_RTLU ctrsm_RTLU
#define TRSM_RTLN ctrsm_RTLN
#define TRSM_RRUU ctrsm_RRUU
#define TRSM_RRUN ctrsm_RRUN
#define TRSM_RRLU ctrsm_RRLU
#define TRSM_RRLN ctrsm_RRLN
#define TRSM_RCUU ctrsm_RCUU
#define TRSM_RCUN ctrsm_RCUN
#define TRSM_RCLU ctrsm_RCLU
#define TRSM_RCLN ctrsm_RCLN

#define  SYRK_UN   csyrk_UN
#define  SYRK_UT   csyrk_UT
#define  SYRK_LN   csyrk_LN
#define  SYRK_LT   csyrk_LT
#define  SYRK_UR   csyrk_UR
#define  SYRK_UC   csyrk_UC
#define  SYRK_LR   csyrk_LR
#define  SYRK_LC   csyrk_LC

#define  HERK_UN   cherk_UN
#define  HERK_UT   cherk_UT
#define  HERK_LN   cherk_LN
#define  HERK_LT   cherk_LT
#define  HERK_UR   cherk_UR
#define  HERK_UC   cherk_UC
#define  HERK_LR   cherk_LR
#define  HERK_LC   cherk_LC

#define SYRK_KERNEL_U csyrk_kernel_U
#define SYRK_KERNEL_L csyrk_kernel_L

#define  SYR2K_UN  csyr2k_UN
#define  SYR2K_UT  csyr2k_UT
#define  SYR2K_LN  csyr2k_LN
#define  SYR2K_LT  csyr2k_LT
#define  SYR2K_UR  csyr2k_UR
#define  SYR2K_UC  csyr2k_UC
#define  SYR2K_LR  csyr2k_LR
#define  SYR2K_LC  csyr2k_LC

#define  SYMCOPY_L csymcopy_L
#define  SYMCOPY_U csymcopy_U
#define  HEMCOPY_L chemcopy_L
#define  HEMCOPY_U chemcopy_U
#endif
#endif

#ifdef UNIT
#ifndef COMPLEX
#ifdef XDOUBLE
#define TRMM_IUNCOPY qtrmm_iunucopy
#define TRMM_IUTCOPY qtrmm_iutucopy
#define TRMM_ILNCOPY qtrmm_ilnucopy
#define TRMM_ILTCOPY qtrmm_iltucopy
#define TRMM_OUNCOPY qtrmm_ounucopy
#define TRMM_OUTCOPY qtrmm_outucopy
#define TRMM_OLNCOPY qtrmm_olnucopy
#define TRMM_OLTCOPY qtrmm_oltucopy

#define TRSM_IUNCOPY qtrsm_iunucopy
#define TRSM_IUTCOPY qtrsm_iutucopy
#define TRSM_ILNCOPY qtrsm_ilnucopy
#define TRSM_ILTCOPY qtrsm_iltucopy
#define TRSM_OUNCOPY qtrsm_ounucopy
#define TRSM_OUTCOPY qtrsm_outucopy
#define TRSM_OLNCOPY qtrsm_olnucopy
#define TRSM_OLTCOPY qtrsm_oltucopy
#elif defined(DOUBLE)
#define TRMM_IUNCOPY dtrmm_iunucopy
#define TRMM_IUTCOPY dtrmm_iutucopy
#define TRMM_ILNCOPY dtrmm_ilnucopy
#define TRMM_ILTCOPY dtrmm_iltucopy
#define TRMM_OUNCOPY dtrmm_ounucopy
#define TRMM_OUTCOPY dtrmm_outucopy
#define TRMM_OLNCOPY dtrmm_olnucopy
#define TRMM_OLTCOPY dtrmm_oltucopy

#define TRSM_IUNCOPY dtrsm_iunucopy
#define TRSM_IUTCOPY dtrsm_iutucopy
#define TRSM_ILNCOPY dtrsm_ilnucopy
#define TRSM_ILTCOPY dtrsm_iltucopy
#define TRSM_OUNCOPY dtrsm_ounucopy
#define TRSM_OUTCOPY dtrsm_outucopy
#define TRSM_OLNCOPY dtrsm_olnucopy
#define TRSM_OLTCOPY dtrsm_oltucopy
#else
#define TRMM_IUNCOPY strmm_iunucopy
#define TRMM_IUTCOPY strmm_iutucopy
#define TRMM_ILNCOPY strmm_ilnucopy
#define TRMM_ILTCOPY strmm_iltucopy
#define TRMM_OUNCOPY strmm_ounucopy
#define TRMM_OUTCOPY strmm_outucopy
#define TRMM_OLNCOPY strmm_olnucopy
#define TRMM_OLTCOPY strmm_oltucopy

#define TRSM_IUNCOPY strsm_iunucopy
#define TRSM_IUTCOPY strsm_iutucopy
#define TRSM_ILNCOPY strsm_ilnucopy
#define TRSM_ILTCOPY strsm_iltucopy
#define TRSM_OUNCOPY strsm_ounucopy
#define TRSM_OUTCOPY strsm_outucopy
#define TRSM_OLNCOPY strsm_olnucopy
#define TRSM_OLTCOPY strsm_oltucopy
#endif
#else
#ifdef XDOUBLE
#define TRMM_IUNCOPY xtrmm_iunucopy
#define TRMM_IUTCOPY xtrmm_iutucopy
#define TRMM_ILNCOPY xtrmm_ilnucopy
#define TRMM_ILTCOPY xtrmm_iltucopy
#define TRMM_OUNCOPY xtrmm_ounucopy
#define TRMM_OUTCOPY xtrmm_outucopy
#define TRMM_OLNCOPY xtrmm_olnucopy
#define TRMM_OLTCOPY xtrmm_oltucopy

#define TRSM_IUNCOPY xtrsm_iunucopy
#define TRSM_IUTCOPY xtrsm_iutucopy
#define TRSM_ILNCOPY xtrsm_ilnucopy
#define TRSM_ILTCOPY xtrsm_iltucopy
#define TRSM_OUNCOPY xtrsm_ounucopy
#define TRSM_OUTCOPY xtrsm_outucopy
#define TRSM_OLNCOPY xtrsm_olnucopy
#define TRSM_OLTCOPY xtrsm_oltucopy
#elif defined(DOUBLE)
#define TRMM_IUNCOPY ztrmm_iunucopy
#define TRMM_IUTCOPY ztrmm_iutucopy
#define TRMM_ILNCOPY ztrmm_ilnucopy
#define TRMM_ILTCOPY ztrmm_iltucopy
#define TRMM_OUNCOPY ztrmm_ounucopy
#define TRMM_OUTCOPY ztrmm_outucopy
#define TRMM_OLNCOPY ztrmm_olnucopy
#define TRMM_OLTCOPY ztrmm_oltucopy

#define TRSM_IUNCOPY ztrsm_iunucopy
#define TRSM_IUTCOPY ztrsm_iutucopy
#define TRSM_ILNCOPY ztrsm_ilnucopy
#define TRSM_ILTCOPY ztrsm_iltucopy
#define TRSM_OUNCOPY ztrsm_ounucopy
#define TRSM_OUTCOPY ztrsm_outucopy
#define TRSM_OLNCOPY ztrsm_olnucopy
#define TRSM_OLTCOPY ztrsm_oltucopy
#else
#define TRMM_IUNCOPY ctrmm_iunucopy
#define TRMM_IUTCOPY ctrmm_iutucopy
#define TRMM_ILNCOPY ctrmm_ilnucopy
#define TRMM_ILTCOPY ctrmm_iltucopy
#define TRMM_OUNCOPY ctrmm_ounucopy
#define TRMM_OUTCOPY ctrmm_outucopy
#define TRMM_OLNCOPY ctrmm_olnucopy
#define TRMM_OLTCOPY ctrmm_oltucopy

#define TRSM_IUNCOPY ctrsm_iunucopy
#define TRSM_IUTCOPY ctrsm_iutucopy
#define TRSM_ILNCOPY ctrsm_ilnucopy
#define TRSM_ILTCOPY ctrsm_iltucopy
#define TRSM_OUNCOPY ctrsm_ounucopy
#define TRSM_OUTCOPY ctrsm_outucopy
#define TRSM_OLNCOPY ctrsm_olnucopy
#define TRSM_OLTCOPY ctrsm_oltucopy
#endif
#endif
#else
#ifndef COMPLEX
#ifdef XDOUBLE
#define TRMM_IUNCOPY qtrmm_iunncopy
#define TRMM_IUTCOPY qtrmm_iutncopy
#define TRMM_ILNCOPY qtrmm_ilnncopy
#define TRMM_ILTCOPY qtrmm_iltncopy
#define TRMM_OUNCOPY qtrmm_ounncopy
#define TRMM_OUTCOPY qtrmm_outncopy
#define TRMM_OLNCOPY qtrmm_olnncopy
#define TRMM_OLTCOPY qtrmm_oltncopy

#define TRSM_IUNCOPY qtrsm_iunncopy
#define TRSM_IUTCOPY qtrsm_iutncopy
#define TRSM_ILNCOPY qtrsm_ilnncopy
#define TRSM_ILTCOPY qtrsm_iltncopy
#define TRSM_OUNCOPY qtrsm_ounncopy
#define TRSM_OUTCOPY qtrsm_outncopy
#define TRSM_OLNCOPY qtrsm_olnncopy
#define TRSM_OLTCOPY qtrsm_oltncopy
#elif defined(DOUBLE)
#define TRMM_IUNCOPY dtrmm_iunncopy
#define TRMM_IUTCOPY dtrmm_iutncopy
#define TRMM_ILNCOPY dtrmm_ilnncopy
#define TRMM_ILTCOPY dtrmm_iltncopy
#define TRMM_OUNCOPY dtrmm_ounncopy
#define TRMM_OUTCOPY dtrmm_outncopy
#define TRMM_OLNCOPY dtrmm_olnncopy
#define TRMM_OLTCOPY dtrmm_oltncopy

#define TRSM_IUNCOPY dtrsm_iunncopy
#define TRSM_IUTCOPY dtrsm_iutncopy
#define TRSM_ILNCOPY dtrsm_ilnncopy
#define TRSM_ILTCOPY dtrsm_iltncopy
#define TRSM_OUNCOPY dtrsm_ounncopy
#define TRSM_OUTCOPY dtrsm_outncopy
#define TRSM_OLNCOPY dtrsm_olnncopy
#define TRSM_OLTCOPY dtrsm_oltncopy
#else
#define TRMM_IUNCOPY strmm_iunncopy
#define TRMM_IUTCOPY strmm_iutncopy
#define TRMM_ILNCOPY strmm_ilnncopy
#define TRMM_ILTCOPY strmm_iltncopy
#define TRMM_OUNCOPY strmm_ounncopy
#define TRMM_OUTCOPY strmm_outncopy
#define TRMM_OLNCOPY strmm_olnncopy
#define TRMM_OLTCOPY strmm_oltncopy

#define TRSM_IUNCOPY strsm_iunncopy
#define TRSM_IUTCOPY strsm_iutncopy
#define TRSM_ILNCOPY strsm_ilnncopy
#define TRSM_ILTCOPY strsm_iltncopy
#define TRSM_OUNCOPY strsm_ounncopy
#define TRSM_OUTCOPY strsm_outncopy
#define TRSM_OLNCOPY strsm_olnncopy
#define TRSM_OLTCOPY strsm_oltncopy
#endif
#else
#ifdef XDOUBLE
#define TRMM_IUNCOPY xtrmm_iunncopy
#define TRMM_IUTCOPY xtrmm_iutncopy
#define TRMM_ILNCOPY xtrmm_ilnncopy
#define TRMM_ILTCOPY xtrmm_iltncopy
#define TRMM_OUNCOPY xtrmm_ounncopy
#define TRMM_OUTCOPY xtrmm_outncopy
#define TRMM_OLNCOPY xtrmm_olnncopy
#define TRMM_OLTCOPY xtrmm_oltncopy

#define TRSM_IUNCOPY xtrsm_iunncopy
#define TRSM_IUTCOPY xtrsm_iutncopy
#define TRSM_ILNCOPY xtrsm_ilnncopy
#define TRSM_ILTCOPY xtrsm_iltncopy
#define TRSM_OUNCOPY xtrsm_ounncopy
#define TRSM_OUTCOPY xtrsm_outncopy
#define TRSM_OLNCOPY xtrsm_olnncopy
#define TRSM_OLTCOPY xtrsm_oltncopy
#elif defined(DOUBLE)
#define TRMM_IUNCOPY ztrmm_iunncopy
#define TRMM_IUTCOPY ztrmm_iutncopy
#define TRMM_ILNCOPY ztrmm_ilnncopy
#define TRMM_ILTCOPY ztrmm_iltncopy
#define TRMM_OUNCOPY ztrmm_ounncopy
#define TRMM_OUTCOPY ztrmm_outncopy
#define TRMM_OLNCOPY ztrmm_olnncopy
#define TRMM_OLTCOPY ztrmm_oltncopy

#define TRSM_IUNCOPY ztrsm_iunncopy
#define TRSM_IUTCOPY ztrsm_iutncopy
#define TRSM_ILNCOPY ztrsm_ilnncopy
#define TRSM_ILTCOPY ztrsm_iltncopy
#define TRSM_OUNCOPY ztrsm_ounncopy
#define TRSM_OUTCOPY ztrsm_outncopy
#define TRSM_OLNCOPY ztrsm_olnncopy
#define TRSM_OLTCOPY ztrsm_oltncopy
#else
#define TRMM_IUNCOPY ctrmm_iunncopy
#define TRMM_IUTCOPY ctrmm_iutncopy
#define TRMM_ILNCOPY ctrmm_ilnncopy
#define TRMM_ILTCOPY ctrmm_iltncopy
#define TRMM_OUNCOPY ctrmm_ounncopy
#define TRMM_OUTCOPY ctrmm_outncopy
#define TRMM_OLNCOPY ctrmm_olnncopy
#define TRMM_OLTCOPY ctrmm_oltncopy

#define TRSM_IUNCOPY ctrsm_iunncopy
#define TRSM_IUTCOPY ctrsm_iutncopy
#define TRSM_ILNCOPY ctrsm_ilnncopy
#define TRSM_ILTCOPY ctrsm_iltncopy
#define TRSM_OUNCOPY ctrsm_ounncopy
#define TRSM_OUTCOPY ctrsm_outncopy
#define TRSM_OLNCOPY ctrsm_olnncopy
#define TRSM_OLTCOPY ctrsm_oltncopy
#endif
#endif
#endif

#if GEMM_UNROLL_M == GEMM_UNROLL_N
#undef  TRMM_IUNCOPY
#undef  TRMM_IUTCOPY
#undef  TRMM_ILNCOPY
#undef  TRMM_ILTCOPY
#define TRMM_IUNCOPY TRMM_OUNCOPY
#define TRMM_IUTCOPY TRMM_OUTCOPY
#define TRMM_ILNCOPY TRMM_OLNCOPY
#define TRMM_ILTCOPY TRMM_OLTCOPY

#undef  TRSM_IUNCOPY
#undef  TRSM_IUTCOPY
#undef  TRSM_ILNCOPY
#undef  TRSM_ILTCOPY
#define TRSM_IUNCOPY TRSM_OUNCOPY
#define TRSM_IUTCOPY TRSM_OUTCOPY
#define TRSM_ILNCOPY TRSM_OLNCOPY
#define TRSM_ILTCOPY TRSM_OLTCOPY
#endif

#endif
