/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#ifdef DOUBLE
#define PREFETCHSIZE  (16 *  8)
#else
#define PREFETCHSIZE  (32 *  4)
#endif

#define CPREFETCHSIZE 8
#define CPREFETCH     lfetch.excl.nt1

#define M	r32
#define N	r33
#define K	r34
#define A	r37
#define B	r38
#define C	r39
#define LDC	r35

#define I	r15
#define J	r16
#define AOFFSET	r17
#define BOFFSET	r18
#define TEMP	r19
#define L	r20

#define C1	r21
#define C2	r22
#define C3	r23
#define C4	r24
#define C5	r25
#define C6	r26
#define C7	r27
#define C8	r28

#define PREA	r8
#define PREB	r9
#define PREC	r10
#define SP	r12
#define ARLC	r29
#define PR	r30
#define ARPFS	r31

#define ALPHA_R	f8
#define ALPHA_I	f9

#define AORIG	loc0
#define KK	loc1
#define KK8	loc2
#define OFFSET	loc3

#if   defined(NN) || defined(NT) || defined(TN) || defined(TT) || \
      defined(CC) || defined(CR) || defined(RC) || defined(RR)
#define FCALC_A	  FSUB
#define FCALC_B	  FADD
#else
#define FCALC_A	  FADD
#define FCALC_B	  FSUB
#endif

#if   defined(NN) || defined(NT) || defined(TN) || defined(TT) || \
      defined(NC) || defined(TC) || defined(NR) || defined(TR)
#define FCALC_C	  FMA
#define FCALC_D	  FNMA
#else
#define FCALC_C	  FNMA
#define FCALC_D	  FMA
#endif

	PROLOGUE
	.prologue
	PROFCODE

#ifdef TRMMKERNEL
	.save	ar.pfs, ARPFS
	alloc	ARPFS = ar.pfs, 8, 8, 0, 0
#endif

	adds	r14 = 16, SP
	adds	r15 = 24, SP
	adds	r8 = -16 * 16, SP
	adds	r9 = -15 * 16, SP
	adds	SP = -16 * 16, SP
	;;
	ld8	LDC    = [r14]
#ifdef TRMMKERNEL
	ld8	OFFSET = [r15]
#endif
	mov	PR = pr
	;;
	stf.spill  [r8] = f16, 32
	stf.spill  [r9] = f17, 32
	;;	
	stf.spill  [r8] = f18, 32
	stf.spill  [r9] = f19, 32
	;;	
	stf.spill  [r8] = f20, 32
	stf.spill  [r9] = f21, 32
	;;
	stf.spill  [r8] = f22, 32
	stf.spill  [r9] = f23, 32
	;;	
	stf.spill  [r8] = f24, 32
	stf.spill  [r9] = f25, 32
	;;	
	stf.spill  [r8] = f26, 32
	stf.spill  [r9] = f27, 32
	;;	
	stf.spill  [r8] = f28, 32
	stf.spill  [r9] = f29, 32
       ;;	
	stf.spill  [r8] = f30
	stf.spill  [r9] = f31
	mov	ARLC  = ar.lc
	;;
	{ .mmi
	shladd	LDC = LDC, ZBASE_SHIFT, r0
#if defined(TRMMKERNEL) && !defined(LEFT)
	sub	KK  = r0, OFFSET
#endif
	shr	J = N, 2
	}
	;;
	.body
	{ .mmb
	mov	AOFFSET = A
	cmp.ge	p6, p0  = 0, J
	(p6)	br.cond.dpnt .L050
	}
	;;
	.align 16

.L010:
	{ .mfi
	adds	J = -1, J
	mov	f64  = f0
	shr	I  = M, 2
	}
	{ .mfi
	mov	C1 = C			// coffset1 = c + 0 * ldc
	mov	f72  = f0
	}
	;;
	{ .mmf
	cmp.eq	p6, p7 = 0, I
#if defined(TRMMKERNEL) &&  defined(LEFT)
	mov	KK  = OFFSET
#else
	nop	__LINE__
#endif
	mov	f80  = f0
	} 
	{ .mmf
	add	C2 = LDC, C		// coffset2 = c + 1 * ldc
	shladd	C3 = LDC, 1, C		// coffset3 = c + 2 * ldc
	mov	f88  = f0
	}
	;;
	{ .mfi
	shladd	C = LDC, 2, C		// coffset += 8 * ldc
	mov	f96  = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mmf
	shladd	C4 = LDC, 1, C2		// coffset4 = c + 3 * ldc
	nop	__LINE__
	mov	f104 = f0
	}
	;;
	{ .mfi
	nop	__LINE__
	mov	f112 = f0
	nop	__LINE__
	}{ .mfb
	nop	__LINE__
	mov	f120 = f0
	(p6)	br.cond.dpnt .L020
	}
	;;
	.align 16

.L011:
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f65  = f0
	nop	__LINE__
	}
	{ .mfb
	adds	BOFFSET = 2 * SIZE, B
	mov	f73  = f0
	nop	__LINE__
	}
	;;
#else
	{ .mfi
	shladd	BOFFSET = KK8, 2, B
	mov	f65  = f0
	shladd	AOFFSET = KK8, 2, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f73  = f0
	nop	__LINE__
	}
	;;
#endif
	{ .mfb
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	mov	f81  = f0
	nop	__LINE__
	}
	{ .mfb
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	mov	f89  = f0
	nop	__LINE__
	}
	;;
	{ .mmf
	LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	setf.d	f97  = r0
	mov	f105 = f0
	}
	{ .mfb
	setf.d	f113 = r0
	mov	f121 = f0
	nop	__LINE__
	}
	;;
	{ .mmf
	LDFPD	f54, f55 = [BOFFSET], 2 * SIZE
	setf.d	f66  = r0
	mov	f74  = f0
	}
	{ .mfb
	setf.d	f82  = r0
	mov	f90  = f0
	nop	__LINE__
	}
	;;
	{ .mmf
	LDFPD	f34, f35  = [AOFFSET], 2 * SIZE
	setf.d	f98  = r0
	mov	f106 = f0
	}
	{ .mfb
	setf.d	f114 = r0
	mov	f122 = f0
	nop	__LINE__
	}
	;;
	{ .mmf
	LDFPD	f36, f37  = [AOFFSET], 2 * SIZE
	setf.d	f67  = r0
	mov	f75  = f0
	}
	{ .mfi
	setf.d	f83  = r0
	mov	f91  = f0
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 4, KK
#else
	adds	L = 4, KK
#endif
#endif
	}
	;;
	{ .mmf
	LDFPD	f38, f39  = [AOFFSET], 2 * SIZE
	setf.d	f99  = r0
	mov	f107 = f0
	}
	{ .mfi
	setf.d	f115 = r0
	mov	f123 = f0
	adds	PREC = CPREFETCHSIZE * SIZE, C1
	}
	;;
	{ .mmf
	CPREFETCH [PREC], LDC
	setf.d	f68  = r0
	mov	f76  = f0
	}
	{ .mfi
	setf.d	f84  = r0
	mov	f92  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
	{ .mmf
	CPREFETCH [PREC], LDC
	setf.d	f100 = r0
	mov	f108 = f0
	}
	{ .mfi
	setf.d	f116 = r0
	mov	f124 = f0
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	}
	;;
	{ .mmf
	CPREFETCH [PREC], LDC
	setf.d	f69  = r0
	mov	f77  = f0
	}
	{ .mfi
	setf.d	f85  = r0
	mov	f93  = f0
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	}
	;;
	{ .mmf
	CPREFETCH [PREC]
	setf.d	f101 = r0
	mov	f109 = f0
	}
	{ .mfi
	setf.d	f117 = r0
	mov	f125 = f0
	tbit.z	p12, p0 = L, 0
	}
	;;
	{ .mmf
	adds	C5 = 4 * SIZE, C1
	setf.d	f70  = r0
	mov	f78  = f0
	}
	{ .mfi
	setf.d	f86  = r0
	mov	f94  = f0
	shr	L = L, 1
	}
	;;
	{ .mmf
	adds	C6 = 4 * SIZE, C2
	setf.d	f102 = r0
	mov	f110 = f0
	}
	{ .mfi
	setf.d	f118 = r0
	mov	f126 = f0
	adds	L =  -1, L
	}
	;;
	{ .mmf
	adds	C7 = 4 * SIZE, C3
	setf.d	f71  = r0
	mov	f79  = f0
	}
	{ .mfi
	setf.d	f87  = r0
	mov	f95  = f0
	mov	ar.lc = L
	}
	;;
	{ .mmf
	adds	C8 = 4 * SIZE, C4
	setf.d	f103 = r0
	mov	f111 = f0
	}
	{ .mfi
	setf.d	f119 = r0
	mov	f127 = f0
	cmp.eq	p3, p0 = r0, r0
	}
	;;
	.align 16

.L012:
/*  1 */
	{ .mfi
	lfetch.fault.nt1	[PREA],  16 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	(p12) cmp.ne p3, p0 =  0, L
	FMA	f72   = f32, f49, f72	// A1 * B2
	nop	__LINE__
	}
	;;
/*  2 */
	{ .mfi
	lfetch.nt1	[PREB],  16 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	cmp.ne	p4, p5 =  0, L
	FMA	f88   = f32, f51, f88	// A1 * B4
	nop	__LINE__
	}
	;;
/*  3 */
	{ .mfb
	(p3) LDFPD	f56, f57 = [BOFFSET],   2 * SIZE
	FMA	f96   = f32, f52, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
	FMA	f104  = f32, f53, f104	// A1 * B6
	nop	__LINE__
	}
	;;
/*  4 */
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f112  = f32, f54, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
	FMA	f120  = f32, f55, f120	// A1 * B8
	nop	__LINE__
	}
	;;
/*  5 */
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET],  2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f73   = f33, f49, f73	// A2 * B2
	nop	__LINE__
	}
	;;
/*  6 */
	{ .mfb
	(p3) LDFPD	f60, f61 = [BOFFSET], 2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	FMA	f89   = f33, f51, f89	// A2 * B4
	nop	__LINE__
	}
	;;
/*  7 */
	{ .mfb
	(p3) LDFPD	f62, f63 = [BOFFSET], 2 * SIZE
	FMA	f97   = f33, f52, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
	FMA	f105  = f33, f53, f105	// A2 * B6
	nop	__LINE__
	}
	;;
/*  8 */
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	FMA	f113  = f33, f54, f113	// A2 * B7
	nop	__LINE__
	}
	{ .mfb
	FMA	f121  = f33, f55, f121	// A2 * B8
	nop	__LINE__
	}
	;;
/*  9 */
	{ .mfb
	(p3) LDFPD	f44, f45 = [AOFFSET], 2 * SIZE
	FMA	f66   = f34, f48, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f74   = f34, f49, f74	// A3 * B2
	nop	__LINE__
	}
	;;
/* 10 */
	{ .mfb
	(p3) LDFPD	f46, f47 = [AOFFSET], 2 * SIZE
	FMA	f82   = f34, f50, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
	FMA	f90   = f34, f51, f90	// A3 * B4
	nop	__LINE__
	}
	;;
/* 11 */
	{ .mfb
	FMA	f98   = f34, f52, f98	// A3 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f106  = f34, f53, f106	// A3 * B6
	nop	__LINE__
	}
	;; 
/* 12 */
	{ .mfb
	FMA	f114  = f34, f54, f114	// A3 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f122  = f34, f55, f122	// A3 * B8
	nop	__LINE__
	}
	;;
/* 13 */
	{ .mfb
	nop	__LINE__
	FMA	f67   = f35, f48, f67	// A4 * B1
	}
	{ .mfb
	nop	__LINE__
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}
	;;
/* 14 */
	{ .mfb
	FMA	f83   = f35, f50, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f91   = f35, f51, f91	// A4 * B4
	nop	__LINE__
	}
	;;
/* 15 */
	{ .mfb
	FMA	f99   = f35, f52, f99	// A4 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f107  = f35, f53, f107	// A4 * B6
	nop	__LINE__
	}
	;;
/* 16 */
	{ .mfb
	FMA	f115  = f35, f54, f115	// A4 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f123  = f35, f55, f123	// A4 * B8
	nop	__LINE__
	}
	;;
/* 17 */
	{ .mfb
	nop	__LINE__
	FMA	f68   = f36, f48, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f76   = f36, f49, f76	// A5 * B2
	nop	__LINE__
	}
	;;
/* 18 */
	{ .mfb
	nop	__LINE__
	FMA	f84   = f36, f50, f84	// A5 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f92   = f36, f51, f92	// A5 * B4
	nop	__LINE__
	}
	;;
/* 19 */
	{ .mfb
	nop	__LINE__
	FMA	f100  = f36, f52, f100	// A5 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f108  = f36, f53, f108	// A5 * B6
	nop	__LINE__
	}
	;;
/* 20 */
	{ .mfb
	nop	__LINE__
	FMA	f116  = f36, f54, f116	// A5 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f124  = f36, f55, f124	// A5 * B8
	nop	__LINE__
	}
	;;
/* 21 */
	{ .mfb
	nop	__LINE__
	FMA	f69   = f37, f48, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f77   = f37, f49, f77	// A6 * B2
	nop	__LINE__
	}
	;;
/* 22 */
	{ .mfb
	nop	__LINE__
	FMA	f85   = f37, f50, f85	// A6 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f93   = f37, f51, f93	// A6 * B4
	nop	__LINE__
	}
	;;
/* 23 */
	{ .mfb
	nop	__LINE__
	FMA	f101  = f37, f52, f101	// A6 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f109  = f37, f53, f109	// A6 * B6
	nop	__LINE__
	}
	;;
/* 24 */
	{ .mfb
	nop	__LINE__
	FMA	f117  = f37, f54, f117	// A6 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f125  = f37, f55, f125	// A6 * B8
	nop	__LINE__
	}
	;;
/* 25 */
	{ .mfb
	nop	__LINE__
	FMA	f70   = f38, f48, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f78   = f38, f49, f78	// A7 * B2
	nop	__LINE__
	}
	;;
/* 26 */
	{ .mfb
	nop	__LINE__
	FMA	f86   = f38, f50, f86	// A7 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f94   = f38, f51, f94	// A7 * B4
	nop	__LINE__
	}
	;;
/* 27 */
	{ .mfb
	nop	__LINE__
	FMA	f102  = f38, f52, f102	// A7 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f110  = f38, f53, f110	// A7 * B6
	nop	__LINE__
	}
	;;
/* 28 */
	{ .mfb
	nop	__LINE__
	FMA	f118  = f38, f54, f118	// A7 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f126  = f38, f55, f126	// A7 * B8
	nop	__LINE__
	}
	;;
/* 29 */
	{ .mfb
	nop	__LINE__
	FMA	f71   = f39, f48, f71	// A8 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f79   = f39, f49, f79	// A8 * B2
	nop	__LINE__
	}
	;;
/* 30 */
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	FMA	f87   = f39, f50, f87	// A8 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f95   = f39, f51, f95	// A8 * B4
	nop	__LINE__
	}
	;;
/* 31 */
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	FMA	f103  = f39, f52, f103	// A8 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f111  = f39, f53, f111	// A8 * B6
	nop	__LINE__
	}
	;;
/* 32 */
	{ .mfb
	nop	__LINE__
	FMA	f119  = f39, f54, f119	// A8 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f127  = f39, f55, f127	// A8 * B8
	nop	__LINE__
	}
	;;
/* 33 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
/* 34 */
	{ .mfb
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	nop	__LINE__
	}
	;;
/* 35 */
	{ .mfb
	(p4) LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	(p3) FMA	f96   = f40, f60, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f104  = f40, f61, f104	// A1 * B6
	nop	__LINE__
	}
	;;
/* 36 */
	{ .mfb
	(p4) LDFPD	f54, f55 = [BOFFSET], 2 * SIZE
	(p3) FMA	f112  = f40, f62, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f120  = f40, f63, f120	// A1 * B8
	nop	__LINE__
	}
	;;
/* 37 */
	{ .mfb
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
/* 38 */
	{ .mfb
	(p4) LDFPD	f36, f37 = [AOFFSET], 2 * SIZE
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	nop	__LINE__
	}
	;;
/* 39 */
	{ .mfb
	(p4) LDFPD	f38, f39 = [AOFFSET], 2 * SIZE
	(p3) FMA	f97   = f41, f60, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f105  = f41, f61, f105	// A2 * B6
	nop	__LINE__
	}
	;;
/* 40 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f113  = f41, f62, f113	// A2 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f12 = [C5], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f121  = f41, f63, f121	// A2 * B8
	nop	__LINE__
	}
	;;
 /* 41 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f13 = [C5], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
/* 42 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10 = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f82   = f42, f58, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f14 = [C5], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f90   = f42, f59, f90	// A3 * B4
	nop	__LINE__
	}
	;;
/* 43 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11 = [C1], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f98   = f42, f60, f98	// A3 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f15 = [C5], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f106  = f42, f61, f106	// A3 * B6
	nop	__LINE__
	}
	;;
/* 44 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f16 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f114  = f42, f62, f114	// A3 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f20 = [C6], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f122  = f42, f63, f122	// A3 * B8
	nop	__LINE__
	}
	;;
/* 45 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f21 = [C6], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	nop	__LINE__
	}
	;;
/* 46 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f18 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f83   = f43, f58, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f22 = [C6], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f91   = f43, f59, f91	// A4 * B4
	nop	__LINE__
	}
	;;
/* 47 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f19 = [C2], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f99   = f43, f60, f99	// A4 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f23 = [C6], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f107  = f43, f61, f107	// A4 * B6
	nop	__LINE__
	}
	;;
/* 48 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f24 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f115  = f43, f62, f115	// A4 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f28 = [C7], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f123  = f43, f63, f123	// A4 * B8
	nop	__LINE__
	}
	;;
/* 49 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f25 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f68   = f44, f56, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f29 = [C7], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f76   = f44, f57, f76	// A5 * B2
	nop	__LINE__
	}
	;;
/* 50 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f26 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f84   = f44, f58, f84	// A5 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f30 = [C7], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f92   = f44, f59, f92	// A5 * B4
	nop	__LINE__
	}
	;;
/* 51 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f27 = [C3], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f100  = f44, f60, f100	// A5 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f31 = [C7], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f108  = f44, f61, f108	// A5 * B6
	nop	__LINE__
	}
	;;
/* 52 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f32 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f116  = f44, f62, f116	// A5 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f36 = [C8], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f124  = f44, f63, f124	// A5 * B8
	nop	__LINE__
	}
	;;
/* 53 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f33 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f69   = f45, f56, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f37 = [C8], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f77   = f45, f57, f77	// A6 * B2
	nop	__LINE__
	}
	;;
/* 54 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f34 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f85   = f45, f58, f85	// A6 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f38 = [C8], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f93   = f45, f59, f93	// A6 * B4
	nop	__LINE__
	}
	;;
/* 55 */
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f35 = [C4], -3 * SIZE
#else
	nop	__LINE__
#endif
 	(p3) FMA	f101  = f45, f60, f101	// A6 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f39 = [C8], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f109  = f45, f61, f109	// A6 * B6
	nop	__LINE__
	}
	;;
/* 56 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f117  = f45, f62, f117	// A6 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f125  = f45, f63, f125	// A6 * B8
	nop	__LINE__
	}
	;;
/* 57 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f70   = f46, f56, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f78   = f46, f57, f78	// A7 * B2
	nop	__LINE__
	}
	;;
/* 58 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f86   = f46, f58, f86	// A7 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f94   = f46, f59, f94	// A7 * B4
	nop	__LINE__
	}
	;;
/* 59 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f102  = f46, f60, f102	// A7 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f110  = f46, f61, f110	// A7 * B6
	nop	__LINE__
	}
	;;
/* 60 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f118  = f46, f62, f118	// A7 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f126  = f46, f63, f126	// A7 * B8
	nop	__LINE__
	}
	;;
/* 61 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f71   = f47, f56, f71	// A8 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f79   = f47, f57, f79	// A8 * B2
	nop	__LINE__
	}
	;;
/* 62 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f87   = f47, f58, f87	// A8 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f95   = f47, f59, f95	// A8 * B4
	nop	__LINE__
	}
	;;
/* 63 */
	{ .mfb
	nop	__LINE__
	(p3) FMA	f103  = f47, f60, f103	// A8 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f111  = f47, f61, f111	// A8 * B6
	nop	__LINE__
	}
	;;
/* 64 */
	{ .mfi
	nop	__LINE__
	(p3) FMA	f119  = f47, f62, f119	// A8 * B7
	adds	L = -1, L
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f127  = f47, f63, f127	// A8 * B8
	br.cloop.sptk.few .L012
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f68  = f68,  f77
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f69  = f69,  f76
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f70  = f70,  f79
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f71  = f71,  f78
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f84  = f84,  f93
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f85  = f85,  f92
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f82  = f82,  f91
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f86  = f86,  f95
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f83  = f83,  f90
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f87  = f87,  f94
	nop	__LINE__
	}


	{ .mfb
	nop	__LINE__
	FCALC_A	  f96  = f96,  f105
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f100 = f100, f109
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f97  = f97,  f104
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f101 = f101, f108
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f98  = f98,  f107
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f102 = f102, f111
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f99  = f99,  f106
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f103 = f103, f110
	nop	__LINE__
	}


	{ .mfb
	nop	__LINE__
	FCALC_A	  f112 = f112, f121
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f116 = f116, f125
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f113 = f113, f120
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f117 = f117, f124
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f114 = f114, f123
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f118 = f118, f127
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f115 = f115, f122
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f119 = f119, f126
	nop	__LINE__
	}
	;;

#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f12 = ALPHA_R, f68, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f14 = ALPHA_R, f70, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f15
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f20 = ALPHA_R, f84, f20
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f21 = ALPHA_R, f85, f21
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f22 = ALPHA_R, f86, f22
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f19
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f23 = ALPHA_R, f87, f23
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f24 = ALPHA_R, f96,  f24
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f28 = ALPHA_R, f100, f28
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f25
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f29 = ALPHA_R, f101, f29
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f26 = ALPHA_R, f98,  f26
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f30 = ALPHA_R, f102, f30
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f27 = ALPHA_R, f99,  f27
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f31 = ALPHA_R, f103, f31
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f32 = ALPHA_R, f112, f32
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f36 = ALPHA_R, f116, f36
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f33
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f37 = ALPHA_R, f117, f37
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f34 = ALPHA_R, f114, f34
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f38 = ALPHA_R, f118, f38
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f35 = ALPHA_R, f115, f35
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f39 = ALPHA_R, f119, f39
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f12, SIZE
	FCALC_D	f20 = ALPHA_I, f85, f20
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f13, SIZE
	FMA	f21 = ALPHA_I, f84, f21
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f10, SIZE
	FCALC_D	f18 = ALPHA_I, f83, f18
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f14, SIZE
	FCALC_D	f22 = ALPHA_I, f87, f22
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f11, 5 * SIZE
	FMA	f19 = ALPHA_I, f82, f19
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f15, 5 * SIZE
	FMA	f23 = ALPHA_I, f86, f23
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f16, SIZE
	FCALC_D	f24 = ALPHA_I, f97,  f24
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f20, SIZE
	FCALC_D	f28 = ALPHA_I, f101, f28
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f17, SIZE
	FMA	f25 = ALPHA_I, f96,  f25
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f21, SIZE
	FMA	f29 = ALPHA_I, f100, f29
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f18, SIZE
	FCALC_D	f26 = ALPHA_I, f99,  f26
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f22, SIZE
 	FCALC_D	f30 = ALPHA_I, f103, f30
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f19, 5 * SIZE
	FMA	f27 = ALPHA_I, f98,  f27
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f23, 5 * SIZE
	FMA	f31 = ALPHA_I, f102, f31
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f24, SIZE
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}
	{ .mfb
	STFD	[C7] = f28, SIZE
	FCALC_D	f36 = ALPHA_I, f117, f36
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f25, SIZE
	FMA	f33 = ALPHA_I, f112, f33
	nop	__LINE__
	}
	{ .mfb
	STFD	[C7] = f29, SIZE
	FMA	f37 = ALPHA_I, f116, f37
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f26, SIZE
	FCALC_D	f34 = ALPHA_I, f115, f34
	nop	__LINE__
	}
	{ .mfb
	STFD	[C7] = f30, SIZE
	FCALC_D	f38 = ALPHA_I, f119, f38
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f27, 5 * SIZE
	FMA	f35 = ALPHA_I, f114, f35
	cmp.ne	p6, p0 = 1, I
	}
	{ .mfb
	STFD	[C7] = f31, 5 * SIZE
	FMA	f39 = ALPHA_I, f118, f39
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f64  = f0
	adds	I = -1, I
	}
	{ .mfb
	STFD	[C8] = f36, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f80  = f0
	and	TEMP = 3, M
	}
	{ .mfb
	STFD	[C8] = f37, SIZE
	mov	f88  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C4] = f34, SIZE
	mov	f96  = f0
	cmp.ne	p8, p9 = r0, TEMP
	}
	{ .mfb
	STFD	[C8] = f38, SIZE
	mov	f104 = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C4] = f35, 5 * SIZE
	mov	f112 = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C8] = f39, 5 * SIZE
	mov	f120 = f0
	(p6)	br.cond.dptk .L011
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMPY	f6  = ALPHA_R, f64
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f12 = ALPHA_R, f68
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f10 = ALPHA_R, f66
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f14 = ALPHA_R, f70
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMPY	f16 = ALPHA_R, f80
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f20 = ALPHA_R, f84
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f21 = ALPHA_R, f85, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f18 = ALPHA_R, f82
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f22 = ALPHA_R, f86
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f23 = ALPHA_R, f87, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMPY	f24 = ALPHA_R, f96
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f28 = ALPHA_R, f100
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f29 = ALPHA_R, f101, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f26 = ALPHA_R, f98
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f30 = ALPHA_R, f102
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f27 = ALPHA_R, f99,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f31 = ALPHA_R, f103, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMPY	f32 = ALPHA_R, f112
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f36 = ALPHA_R, f116
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f37 = ALPHA_R, f117, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f34 = ALPHA_R, f114
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f38 = ALPHA_R, f118
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f35 = ALPHA_R, f115, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f39 = ALPHA_R, f119, f0
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f12, SIZE
	FCALC_D	f20 = ALPHA_I, f85, f20
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f13, SIZE
	FMA	f21 = ALPHA_I, f84, f21
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f10, SIZE
	FCALC_D	f18 = ALPHA_I, f83, f18
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f14, SIZE
	FCALC_D	f22 = ALPHA_I, f87, f22
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f11, 5 * SIZE
	FMA	f19 = ALPHA_I, f82, f19
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f15, 5 * SIZE
	FMA	f23 = ALPHA_I, f86, f23
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f16, SIZE
	FCALC_D	f24 = ALPHA_I, f97,  f24
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f20, SIZE
	FCALC_D	f28 = ALPHA_I, f101, f28
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f17, SIZE
	FMA	f25 = ALPHA_I, f96,  f25
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f21, SIZE
	FMA	f29 = ALPHA_I, f100, f29
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f18, SIZE
	FCALC_D	f26 = ALPHA_I, f99,  f26
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f22, SIZE
 	FCALC_D	f30 = ALPHA_I, f103, f30
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f19, 5 * SIZE
	FMA	f27 = ALPHA_I, f98,  f27
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f23, 5 * SIZE
	FMA	f31 = ALPHA_I, f102, f31
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f24, SIZE
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}
	{ .mfb
	STFD	[C7] = f28, SIZE
	FCALC_D	f36 = ALPHA_I, f117, f36
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f25, SIZE
	FMA	f33 = ALPHA_I, f112, f33
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C7] = f29, SIZE
	FMA	f37 = ALPHA_I, f116, f37
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f26, SIZE
	FCALC_D	f34 = ALPHA_I, f115, f34
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C7] = f30, SIZE
	FCALC_D	f38 = ALPHA_I, f119, f38
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C3] = f27, 5 * SIZE
	FMA	f35 = ALPHA_I, f114, f35
	cmp.ne	p6, p0 = 1, I
	}
	{ .mfi
	STFD	[C7] = f31, 5 * SIZE
	FMA	f39 = ALPHA_I, f118, f39
	adds	I = -1, I
	}
	;;
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C8] = f36, SIZE
	mov	f72  = f0
	and	TEMP = 3, M
	}
	;;
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f80  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 2, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C8] = f37, SIZE
	mov	f88  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 2, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C4] = f34, SIZE
	mov	f96  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 4, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C8] = f38, SIZE
	mov	f104 = f0
	cmp.ne	p8, p9 = r0, TEMP
	}
	;;
	{ .mfi
	STFD	[C4] = f35, 5 * SIZE
	mov	f112 = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C8] = f39, 5 * SIZE
	mov	f120 = f0
	(p6)	br.cond.dptk .L011
	}
	;;
#endif

.L020:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 2, KK
#else
	adds	L = 4, KK
#endif
#endif
	tbit.z	p6, p7 = M, 1
	(p6)	br.cond.dptk .L030
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f65  = f0
	nop	__LINE__
	}
	{ .mfi
	adds	BOFFSET = 2 * SIZE, B
	mov	f73  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mfi
	shladd	BOFFSET = KK8, 2, B
	mov	f65  = f0
	shladd	AOFFSET = KK8, 1, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f73  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	;;
	{ .mfi
	setf.d	f81  = r0
	mov	f89  = f0
	cmp.eq	p3, p0 = r0, r0
	}
	{ .mmi
	nop	__LINE__
	nop	__LINE__
	tbit.z	p12, p0 = L, 0
	}
	;;
	{ .mmf
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	setf.d	f97  = r0
	mov	f105 = f0
	}
	{ .mfi
	setf.d	f113 = r0
	mov	f121 = f0
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	}
	;;
	{ .mmf	
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	setf.d	f66  = r0
	mov	f74  = f0
	}
	{ .mfi
	setf.d	f82  = r0
	mov	f90  = f0
	shr	L = L, 1
	}
	;;
	{ .mmf
	LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	setf.d	f98   = r0
	mov	f106  = f0
	}
	{ .mfi
	setf.d	f114 = r0
	mov	f122 = f0
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f54, f55 = [BOFFSET], 2 * SIZE
	mov	f75  = f0
	adds	L =  -1, L
	}
	{ .mmf
	setf.d	f67  = r0
	setf.d	f83  = r0
	mov	f91  = f0
	}
	;;
	{ .mfi
	LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	mov	f107 = f0
	mov	ar.lc = L
	}
	{ .mmf
	setf.d	f99  = r0
	setf.d	f115 = r0
	mov	f123 = f0
	}
	;;
	.align 16

.L022:
	{ .mfb
	lfetch.nt1	[PREA],  8 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfi
	lfetch.nt1	[PREB],  16 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfb
	nop	__LINE__
	FMA	f88   = f32, f51, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f96   = f32, f52, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f104  = f32, f53, f104	// A1 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f56, f57 = [BOFFSET],   2 * SIZE
	FMA	f112  = f32, f54, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f120  = f32, f55, f120	// A1 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET],  2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f60, f61 = [BOFFSET], 2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f89   = f33, f51, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f62, f63 = [BOFFSET], 2 * SIZE
	FMA	f97   = f33, f52, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f105  = f33, f53, f105	// A2 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	FMA	f113  = f33, f54, f113	// A2 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f121  = f33, f55, f121	// A2 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f66   = f34, f48, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f74   = f34, f49, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f82   = f34, f50, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f90   = f34, f51, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f98   = f34, f52, f98	// A3 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f106  = f34, f53, f106	// A3 * B6
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f114  = f34, f54, f114	// A3 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f122  = f34, f55, f122	// A3 * B8
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f67   = f35, f48, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f83   = f35, f50, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f91   = f35, f51, f91	// A4 * B4
	nop	__LINE__
	}

	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	FMA	f99   = f35, f52, f99	// A4 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f107  = f35, f53, f107	// A4 * B6
	nop	__LINE__
	}

	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	FMA	f115  = f35, f54, f115	// A4 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f123  = f35, f55, f123	// A4 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f96   = f40, f60, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f16 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f104  = f40, f61, f104	// A1 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f112  = f40, f62, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f120  = f40, f63, f120	// A1 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f54, f55 = [BOFFSET], 2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	{ .mfb
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10 = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f97   = f41, f60, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f18 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f105  = f41, f61, f105	// A2 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11 = [C1], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f113  = f41, f62, f113	// A2 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f19 = [C2], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f121  = f41, f63, f121	// A2 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f24 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f32 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f25 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f82   = f42, f58, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f33 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f90   = f42, f59, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f26 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f98   = f42, f60, f98	// A3 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f34 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f106  = f42, f61, f106	// A3 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f27 = [C3], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f114  = f42, f62, f114	// A3 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f35 = [C4], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f122  = f42, f63, f122	// A3 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f83   = f43, f58, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f91   = f43, f59, f91	// A4 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f99   = f43, f60, f99	// A4 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f107  = f43, f61, f107	// A4 * B6
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	(p3) FMA	f115  = f43, f62, f115	// A4 * B7
	adds	L = -1, L
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f123  = f43, f63, f123	// A4 * B8
	br.cloop.sptk.few .L022
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f82  = f82,  f91
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f83  = f83,  f90
	nop	__LINE__
	}


	{ .mfb
	nop	__LINE__
	FCALC_A	  f96  = f96,  f105
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f112 = f112, f121
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f97  = f97,  f104
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f113 = f113, f120
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f98  = f98,  f107
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f114 = f114, f123
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f99  = f99,  f106
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f115 = f115, f122
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f19
	nop	__LINE__
	}


	{ .mfb
	nop	__LINE__
	FMA	f24 = ALPHA_R, f96,  f24
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f32 = ALPHA_R, f112, f32
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f25
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f33
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f26 = ALPHA_R, f98,  f26
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f34 = ALPHA_R, f114, f34
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f27 = ALPHA_R, f99,  f27
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f35 = ALPHA_R, f115, f35
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81,  f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	FMA	f17 = ALPHA_I, f80,  f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	FCALC_D	f18 = ALPHA_I, f83,  f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f19 = ALPHA_I, f82,  f19
	nop	__LINE__
	}
	;;

	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f24 = ALPHA_I, f97,  f24
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f25 = ALPHA_I, f96,  f25
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f17, SIZE
	FMA	f33 = ALPHA_I, f112, f33
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f10, SIZE
	FCALC_D	f26 = ALPHA_I, f99,  f26
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f18, SIZE
	FCALC_D	f34 = ALPHA_I, f115, f34
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f11, SIZE
	FMA	f27 = ALPHA_I, f98,  f27
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f19, SIZE
	FMA	f35 = ALPHA_I, f114, f35
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f24, SIZE
	mov	f64  = f0
	nop	__LINE__
	}
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f72  = f0
	}
	;;
	{ .mfb
	STFD	[C3] = f25, SIZE
	mov	f80  = f0
	nop	__LINE__
	}
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f88  = f0
	}
	;;
	{ .mfb
	STFD	[C3] = f26, SIZE
	mov	f96  = f0
	nop	__LINE__
	}
	{ .mfi
	STFD	[C4] = f34, SIZE
	mov	f104 = f0
	}
	;;
	{ .mfi
	STFD	[C3] = f27, SIZE
	mov	f112 = f0
	}
	{ .mfb
	STFD	[C4] = f35, SIZE
	mov	f120 = f0
	nop	__LINE__
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMPY	f6  = ALPHA_R, f64
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f16 = ALPHA_R, f80
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f10 = ALPHA_R, f66
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f18 = ALPHA_R, f82
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f0
	nop	__LINE__
	}


	{ .mfb
	nop	__LINE__
	FMPY	f24 = ALPHA_R, f96
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f32 = ALPHA_R, f112
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f26 = ALPHA_R, f98
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMPY	f34 = ALPHA_R, f114
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f27 = ALPHA_R, f99,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f35 = ALPHA_R, f115, f0
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81,  f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	FMA	f17 = ALPHA_I, f80,  f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	FCALC_D	f18 = ALPHA_I, f83,  f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f19 = ALPHA_I, f82,  f19
	nop	__LINE__
	}
	;;

	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f24 = ALPHA_I, f97,  f24
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f25 = ALPHA_I, f96,  f25
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f17, SIZE
	FMA	f33 = ALPHA_I, f112, f33
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	FCALC_D	f26 = ALPHA_I, f99,  f26
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C2] = f18, SIZE
	FCALC_D	f34 = ALPHA_I, f115, f34
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f11, SIZE
	FMA	f27 = ALPHA_I, f98,  f27
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C2] = f19, SIZE
	FMA	f35 = ALPHA_I, f114, f35
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C3] = f24, SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f25, SIZE
	mov	f80  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 1, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f88  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 2, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C3] = f26, SIZE
	mov	f96  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 2, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C4] = f34, SIZE
	mov	f104 = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f27, SIZE
	mov	f112 = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C4] = f35, SIZE
	mov	f120 = f0
	nop	__LINE__
	}
	;;
#endif
	.align 16

.L030:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 1, KK
#else
	adds	L = 4, KK
#endif
#endif
	tbit.z	p6, p7 = M, 0
	(p6)	br.cond.dptk .L049
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f65  = f0
	nop	__LINE__
	}
	{ .mfi
	adds	BOFFSET = 2 * SIZE, B
	mov	f73  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mfi
	shladd	BOFFSET = KK8, 2, B
	mov	f65  = f0
	add	AOFFSET = KK8, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f73  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	{ .mmi
	nop	__LINE__
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	tbit.z	p12, p0 = L, 0
	}
	;;
	{ .mfi
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	mov	f81  = f0
	shr	L = L, 1
	}
	{ .mfi
	(p7) LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	mov	f89  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	mov	f97  = f0
	adds	L =  -1, L
	}
	{ .mfb
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	mov	f105 = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	LDFPD	f54, f55 = [BOFFSET], 2 * SIZE	
	mov	f113 = f0
	mov	ar.lc = L
	}
	{ .mfi
	cmp.eq	p3, p0 = r0, r0
	mov	f121 = f0
	nop	__LINE__
	}
	;;
	.align 16

.L032:
	{ .mfb
	lfetch.nt1	[PREA],  4 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfi
	lfetch.nt1	[PREB],  16 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfb
	nop	__LINE__
	FMA	f88   = f32, f51, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f56, f57 = [BOFFSET],   2 * SIZE
	FMA	f96   = f32, f52, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f104  = f32, f53, f104	// A1 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f112  = f32, f54, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f120  = f32, f55, f120	// A1 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET],  2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f60, f61 = [BOFFSET], 2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f89   = f33, f51, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f62, f63 = [BOFFSET], 2 * SIZE
	FMA	f97   = f33, f52, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f105  = f33, f53, f105	// A2 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	FMA	f113  = f33, f54, f113	// A2 * B7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f121  = f33, f55, f121	// A2 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f52, f53 = [BOFFSET], 2 * SIZE
	(p3) FMA	f96   = f40, f60, f96	// A1 * B5
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f104  = f40, f61, f104	// A1 * B6
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f112  = f40, f62, f112	// A1 * B7
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f16 = [C2], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f120  = f40, f63, f120	// A1 * B8
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f54, f55 = [BOFFSET], 2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f24 = [C3], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f97   = f41, f60, f97	// A2 * B5
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f32 = [C4], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f105  = f41, f61, f105	// A2 * B6
	nop	__LINE__
	}
	;;
	{ .mfi
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f25 = [C3], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f113  = f41, f62, f113	// A2 * B7
	adds	L = -1, L
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f33 = [C4], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f121  = f41, f63, f121	// A2 * B8
	br.cloop.sptk.few .L032
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FCALC_A	  f96  = f96,  f105
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f112 = f112, f121
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f97  = f97,  f104
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f113 = f113, f120
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f24 = ALPHA_R, f96,  f24
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f32 = ALPHA_R, f112, f32
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f25
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f33
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81,  f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	FMA	f17 = ALPHA_I, f80,  f17
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f24 = ALPHA_I, f97,  f24
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f25 = ALPHA_I, f96,  f25
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f33 = ALPHA_I, f112, f33
	nop	__LINE__
	}
	;;

	{ .mfb
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	mov	f80  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f17, SIZE
	mov	f88  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f24, SIZE
	mov	f96  = f0
	nop	__LINE__
	}
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f104 = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C3] = f25, SIZE
	mov	f112 = f0
	nop	__LINE__
	}
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f120 = f0
	nop	__LINE__
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f24 = ALPHA_R, f96,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f32 = ALPHA_R, f112, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f25 = ALPHA_R, f97,  f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f33 = ALPHA_R, f113, f0
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81,  f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	FMA	f17 = ALPHA_I, f80,  f17
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f24 = ALPHA_I, f97,  f24
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f32 = ALPHA_I, f113, f32
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FMA	f25 = ALPHA_I, f96,  f25
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -1, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FMA	f33 = ALPHA_I, f112, f33
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f7,  SIZE
	mov	f80  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	add	AOFFSET = KK8, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C2] = f17, SIZE
	mov	f88  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 2, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C3] = f24, SIZE
	mov	f96  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 1, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C4] = f32, SIZE
	mov	f104 = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C3] = f25, SIZE
	mov	f112 = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C4] = f33, SIZE
	mov	f120 = f0
	nop	__LINE__
	}
	;;
#endif
	.align 16

.L049:
	{ .mmi
	mov	B = BOFFSET
	mov	AOFFSET = A
#if defined(TRMMKERNEL) && !defined(LEFT)
	adds	KK = 4, KK
#else
	nop	__LINE__
#endif
	}
	{ .mmb
	nop	__LINE__
	cmp.lt	p6, p0 = 0, J
	(p6)	br.cond.dptk .L010
	}
	;;
	.align 16

.L050:
	{ .mfi
#if defined(TRMMKERNEL) &&  defined(LEFT)
	mov	KK  = OFFSET
#else
	nop	__LINE__
#endif
	mov	f64  = f0
	tbit.z	p6, p0 = N, 1
	}
	{ .mfb
	mov	C1 = C
	mov	f72  = f0
	(p6)	br.cond.dpnt .L090
	}
	;;
	{ .mfi
	setf.d	f88  = r0
	mov	f65  = f0
	shr	I  = M, 2
	}
	{ .mfi
	add	C2 = LDC, C
	mov	f80  = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	setf.d	f81  = r0
	mov	f73  = f0
	cmp.eq	p6, p7 = 0, I
	} 
	{ .mfb
	shladd	C = LDC, 1, C
	mov	f89  = f0
	(p6)	br.cond.dpnt .L060
	}
	;;
	.align 16

.L052:
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f66  = f0
	nop	__LINE__
	}
	{ .mfb
	adds	BOFFSET = 2 * SIZE, B
	mov	f74  = f0
	nop	__LINE__
	}
	;;
#else
	{ .mfi
	shladd	BOFFSET = KK8, 1, B
	mov	f66  = f0
	shladd	AOFFSET = KK8, 2, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f74  = f0
	nop	__LINE__
	}
	;;
#endif
	{ .mmf
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	mov	f82  = f0
	}
	{ .mfi
	adds	PREC = CPREFETCHSIZE * SIZE, C1
	mov	f90  = f0
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 4, KK
#else
	adds	L = 2, KK
#endif
#endif
	}
	;;
	{ .mfi
	LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	mov	f67  = f0
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	}
	{ .mfi
	mov	f75  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
	{ .mfi
	LDFPD	f36, f37 = [AOFFSET], 2 * SIZE
	mov	f83  = f0
	tbit.z	p12, p0 = L, 0
	}
	{ .mmf
	CPREFETCH [PREC], LDC
	setf.d	f84  = r0
	mov	f68  = f0
	}
	;;
	{ .mfi
	LDFPD	f38, f39 = [AOFFSET], 2 * SIZE
	mov	f76  = f0
	cmp.eq	p3, p0 = r0, r0
	}
	{ .mmf
	setf.d	f91  = r0
	setf.d	f77  = r0
	mov	f92  = f0
	}
	;;
	{ .mfi
	CPREFETCH [PREC], LDC
	mov	f69  = f0
	shr	L = L, 1
	}
	{ .mfi
	setf.d	f85  = r0
	mov	f93  = f0
	adds	PREA = (PREFETCHSIZE + 8) * SIZE, AOFFSET
	}
	;;
	{ .mfi
	setf.d	f78  = r0
	mov	f70  = f0
	adds	L =  -1, L
	}
	{ .mmf
	setf.d	f86  = r0
	adds C5 = 4 * SIZE, C1
	mov	f94  = f0
	}
	;;
	{ .mfi
	setf.d	f79  = r0
	mov	f71  = f0
	mov	ar.lc = L
	}
	{ .mmf
	setf.d	f87  = r0
	adds C6 = 4 * SIZE, C2
	mov	f95  = f0
	}
	;;
	.align 16

.L053:
	{ .mfb
	lfetch.nt1	[PREA],  16 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfi
	lfetch.nt1	[PREB],   8 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfi
	nop	__LINE__
	FMA	f88   = f32, f51, f88	// A1 * B4
	}
	;;
	{ .mfi
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	}
	{ .mfi
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	}
	;;
	{ .mfi
	(p3) LDFPD	f56, f57 = [BOFFSET],  2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	}
	{ .mfb
	nop	__LINE__
	FMA	f89   = f33, f51, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET],  2 * SIZE
	FMA	f66   = f34, f48, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f74   = f34, f49, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	FMA	f82   = f34, f50, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f90   = f34, f51, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f44, f45 = [AOFFSET], 2 * SIZE
	FMA	f67   = f35, f48, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f46, f47 = [AOFFSET], 2 * SIZE
	FMA	f83   = f35, f50, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f91   = f35, f51, f91	// A4 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f68   = f36, f48, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f76   = f36, f49, f76	// A5 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f84   = f36, f50, f84	// A5 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f92   = f36, f51, f92	// A5 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f69   = f37, f48, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f77   = f37, f49, f77	// A6 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f85   = f37, f50, f85	// A6 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f93   = f37, f51, f93	// A6 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f70   = f38, f48, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f78   = f38, f49, f78	// A7 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FMA	f86   = f38, f50, f86	// A7 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f94   = f38, f51, f94	// A7 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	FMA	f71   = f39, f48, f71	// A8 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f79   = f39, f49, f79	// A8 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],  2 * SIZE
	FMA	f87   = f39, f50, f87	// A8 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f95   = f39, f51, f95	// A8 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f36, f37 = [AOFFSET], 2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f38, f39 = [AOFFSET], 2 * SIZE
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f12 = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f82   = f42, f58, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f13 = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f90   = f42, f59, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10 = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f14 = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11 = [C1 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f83   = f43, f58, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f15 = [C5 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f91   = f43, f59, f91	// A4 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f16 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f68   = f44, f56, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f20 = [C6 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f76   = f44, f57, f76	// A5 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f84   = f44, f58, f84	// A5 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f21 = [C6 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f92   = f44, f59, f92	// A5 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f18 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f69   = f45, f56, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f22 = [C6 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f77   = f45, f57, f77	// A6 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f19 = [C2 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f85   = f45, f58, f85	// A6 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f23 = [C6 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f93   = f45, f59, f93	// A6 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f70   = f46, f56, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f78   = f46, f57, f78	// A7 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f86   = f46, f58, f86	// A7 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f94   = f46, f59, f94	// A7 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f71   = f47, f56, f71	// A8 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f79   = f47, f57, f79	// A8 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	(p3) FMA	f87   = f47, f58, f87	// A8 * B3
	adds	L = -1, L
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f95   = f47, f59, f95	// A8 * B4
	br.cloop.sptk.few .L053
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f68  = f68,  f77
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f69  = f69,  f76
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f70  = f70,  f79
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f71  = f71,  f78
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f84  = f84,  f93
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f85  = f85,  f92
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f82  = f82,  f91
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f86  = f86,  f95
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f83  = f83,  f90
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f87  = f87,  f94
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f12 = ALPHA_R, f68, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f14 = ALPHA_R, f70, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f15
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f20 = ALPHA_R, f84, f20
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f21 = ALPHA_R, f85, f21
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f22 = ALPHA_R, f86, f22
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f19
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f23 = ALPHA_R, f87, f23
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f12, SIZE
	FCALC_D	f20 = ALPHA_I, f85, f20
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f13, SIZE
	FMA	f21 = ALPHA_I, f84, f21
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f10, SIZE
	FCALC_D	f18 = ALPHA_I, f83, f18
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f14, SIZE
	FCALC_D	f22 = ALPHA_I, f87, f22
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f11, 5 * SIZE
	FMA	f19 = ALPHA_I, f82, f19
	cmp.ne	p6, p0 = 1, I
	}
	{ .mfb
	STFD	[C5] = f15, 5 * SIZE
	FMA	f23 = ALPHA_I, f86, f23
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f16, SIZE
	mov	f64  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f20, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C2] = f17, SIZE
	mov	f80  = f0
	adds	I = -1, I
	}
	{ .mfb
	STFD	[C6] = f21, SIZE
	mov	f88  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f18, SIZE
	mov	f65  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f22, SIZE
	mov	f73  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C2] = f19, 5 * SIZE
	mov	f81  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C6] = f23, 5 * SIZE
	mov	f89 = f0
	(p6)	br.cond.dptk .L052
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f12 = ALPHA_R, f68, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f14 = ALPHA_R, f70, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f20 = ALPHA_R, f84, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f21 = ALPHA_R, f85, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f22 = ALPHA_R, f86, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f23 = ALPHA_R, f87, f0
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f12, SIZE
	FCALC_D	f20 = ALPHA_I, f85, f20
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f13, SIZE
	FMA	f21 = ALPHA_I, f84, f21
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	FCALC_D	f18 = ALPHA_I, f83, f18
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C5] = f14, SIZE
	FCALC_D	f22 = ALPHA_I, f87, f22
	cmp.ne	p6, p0 = 1, I
	}
	;;
	{ .mfi
	STFD	[C1] = f11, 5 * SIZE
	FMA	f19 = ALPHA_I, f82, f19
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C5] = f15, 5 * SIZE
	FMA	f23 = ALPHA_I, f86, f23
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C2] = f16, SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C6] = f20, SIZE
	mov	f72  = f0
	adds	I = -1, I
	}
	;;
	{ .mfi
	STFD	[C2] = f17, SIZE
	mov	f80  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 2, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C6] = f21, SIZE
	mov	f88  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 1, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C2] = f18, SIZE
	mov	f65  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 4, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C6] = f22, SIZE
	mov	f73  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C2] = f19, 5 * SIZE
	mov	f81  = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C6] = f23, 5 * SIZE
	mov	f89 = f0
	(p6)	br.cond.dptk .L052
	}
	;;
#endif
	.align 16

.L060:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 2, KK
#else
	adds	L = 2, KK
#endif
#endif
	tbit.z	p6, p7 = M, 1
	(p6)	br.cond.dptk .L070
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f66  = f0
	nop	__LINE__
	}
	{ .mfi
	adds	BOFFSET = 2 * SIZE, B
	mov	f74  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mfi
	shladd	BOFFSET = KK8, 1, B
	mov	f66  = f0
	shladd	AOFFSET = KK8, 1, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f74  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	{ .mfi
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	mov	f82  = f0
	tbit.z	p12, p0 = L, 0
	}
	{ .mfi
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	mov	f90  = f0
	shr	L = L, 1
	}
	;;
	{ .mfi
	LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	mov	f67  = f0
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	}
	{ .mfi
	nop	__LINE__
	mov	f75  = f0
	adds	L =  -1, L
	}
	;;
	{ .mfi
	nop	__LINE__
	mov	f83  = f0
	mov	ar.lc = L
	}
	{ .mfi
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	mov	f91  = f0
	cmp.eq	p3, p0 = r0, r0
	}
	;;
	.align 16

.L062:
	{ .mfi
	lfetch.nt1	[PREA],  8 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfb
	lfetch.nt1	[PREB],   8 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f88   = f32, f51, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f56, f57 = [BOFFSET],   2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f89   = f33, f51, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET],  2 * SIZE
	FMA	f66   = f34, f48, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f74   = f34, f49, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	FMA	f82   = f34, f50, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f90   = f34, f51, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	FMA	f67   = f35, f48, f67	// A4 * B1
	}
	{ .mfb
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}

	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	FMA	f83   = f35, f50, f83	// A4 * B3
	nop	__LINE__
	}
	{ .mfb
	FMA	f91   = f35, f51, f91	// A4 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f16 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10 = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f18 = [C2 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11 = [C1 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f82   = f42, f58, f82	// A3 * B3
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f19 = [C2 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f90   = f42, f59, f90	// A3 * B4
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	(p3) FMA	f83   = f43, f58, f83	// A4 * B3
	adds	L = -1, L
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f91   = f43, f59, f91	// A4 * B4
	br.cloop.sptk.few .L062
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f82  = f82,  f91
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f83  = f83,  f90
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f19
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f18 = ALPHA_I, f83, f18
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f19 = ALPHA_I, f82, f19
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	mov	f80  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C2] = f17, SIZE
	mov	f88  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	mov	f65  = f0
	adds	L =  1, K
	}
	{ .mfb
	STFD	[C2] = f18, SIZE
	mov	f73 = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f11, SIZE
	mov	f81 = f0
	shr	L = L, 1
	}
	{ .mfb
	STFD	[C2] = f19, SIZE
	mov	f89 = f0
	nop	__LINE__
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f0
	nop	__LINE__
	}

	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f18 = ALPHA_R, f82, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f19 = ALPHA_R, f83, f0
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f18 = ALPHA_I, f83, f18
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FMA	f19 = ALPHA_I, f82, f19
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C2] = f16, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f7,  SIZE
	mov	f80  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 1, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C2] = f17, SIZE
	mov	f88  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 1, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	mov	f65  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 2, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C2] = f18, SIZE
	mov	f73 = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f11, SIZE
	mov	f81 = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C2] = f19, SIZE
	mov	f89 = f0
	nop	__LINE__
	}
	;;
#endif
	.align 16

.L070:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 1, KK
#else
	adds	L = 2, KK
#endif
#endif
	tbit.z	p6, p7 = M, 0
	(p6)	br.cond.dptk .L089
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mmi
	LDFPD	f48, f49 = [B]
	adds	BOFFSET = 2 * SIZE, B
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mmi
	shladd	BOFFSET = KK8, 1, B
	add	AOFFSET = KK8, AOFFSET
	nop	__LINE__
	}
	;;
	{ .mmi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	nop	__LINE__
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	;;
	{ .mii
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	tbit.z	p12, p0 = L, 0
	shr	L = L, 1
	}
	;;
	{ .mmi
	LDFPD	f50, f51 = [BOFFSET], 2 * SIZE
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	adds	L =  -1, L
	}
	;;
	{ .mmi
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	cmp.eq	p3, p0 = r0, r0
	mov	ar.lc = L
	}
	;;
	.align 16

.L072:
	{ .mfb
	lfetch.nt1	[PREA],  4 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfi
	lfetch.nt1	[PREB],   8 * SIZE
	FMA	f80   = f32, f50, f80	// A1 * B3
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfb
	nop	__LINE__
	FMA	f88   = f32, f51, f88	// A1 * B4
	nop	__LINE__
	}
	;;
	{ .mfi
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	}
	{ .mfi
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	}
	;;
	{ .mfi
	(p3) LDFPD	f56, f57 = [BOFFSET], 2 * SIZE
	FMA	f81   = f33, f50, f81	// A2 * B3
	}
	{ .mmf
	FMA	f89   = f33, f51, f89	// A2 * B4
	}
	;;
	{ .mfb
	(p3) LDFPD	f58, f59 = [BOFFSET], 2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mmf
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	(p3) FMA	f80   = f40, f58, f80	// A1 * B3
	nop	__LINE__
	}
	{ .mmf
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1 ], SIZE
	(p5) LDFD	f16 = [C2 ], SIZE
#else
	nop	__LINE__
	nop	__LINE__
#endif
	(p3) FMA	f88   = f40, f59, f88	// A1 * B4
	}
	;;
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1 ], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
	(p4) LDFPD	f50, f51 = [BOFFSET],  2 * SIZE
	(p3) FMA	f81   = f41, f58, f81	// A2 * B3
	adds	L = -1, L
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f17 = [C2 ], - SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f89   = f41, f59, f89	// A2 * B4
	br.cloop.sptk.few .L072
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f80  = f80,  f89
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f81  = f81,  f88
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f17
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	FMA	f17 = ALPHA_I, f80, f17
	nop	__LINE__
	}
	;;
	{ .mmb
	STFD	[C1] = f6,  SIZE
	STFD	[C2] = f16, SIZE
	nop	__LINE__
	}
	;;	
	{ .mmi
	STFD	[C1] = f7,  SIZE
	STFD	[C2] = f17, SIZE
	mov	B = BOFFSET
	}
	;;
#else
	{ .mfi
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FMA	f16 = ALPHA_R, f80, f0
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -1, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FCALC_C	f17 = ALPHA_R, f81, f0
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	FCALC_D	f16 = ALPHA_I, f81, f16
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	add	AOFFSET = KK8, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FMA	f17 = ALPHA_I, f80, f17
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	BOFFSET = KK8, 1, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mmi
	STFD	[C1] = f6,  SIZE
	STFD	[C2] = f16, SIZE
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 1, KK
#else
	nop	__LINE__
#endif
	}
	;;	
	{ .mmi
	STFD	[C1] = f7,  SIZE
	STFD	[C2] = f17, SIZE
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	;;
#endif
	.align 16

.L089:
	{ .mmi
	mov	B = BOFFSET
	mov	AOFFSET = A
#if defined(TRMMKERNEL) && !defined(LEFT)
	adds	KK = 2, KK
#else
	nop	__LINE__
#endif
	}
	;;
	.align 16

.L090:
	{ .mfi
	mov	C1 = C
	mov	f64  = f0
	tbit.z	p6, p0 = N, 0
	}
	{ .mfi
#if defined(TRMMKERNEL) &&  defined(LEFT)
	mov	KK  = OFFSET
#else
	nop	__LINE__
#endif
	mov	f72  = f0
	shr	I  = M, 2
	}
	;;
	{ .mfi
	setf.d	f66  = r0
	mov	f65  = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	mov	AOFFSET = A
	mov	f73  = f0
	(p6)	br.cond.dpnt .L999
	}
	;;
	{ .mfi
	setf.d	f74  = r0
	mov	f67  = f0
	nop	__LINE__
	}
	{ .mfb
	cmp.eq	p6, p7 = 0, I
	mov	f75  = f0
	(p6)	br.cond.dpnt .L100
	}
	;;
	.align 16

.L092:
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mfb
	LDFPD	f48, f49 = [B]
	mov	f68  = f0
	nop	__LINE__
	}
	{ .mfi
	adds	BOFFSET = 2 * SIZE, B
	mov	f76  = f0
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 4, KK
#else
	adds	L = 1, KK
#endif
#endif
	}
	;;
#else
	{ .mfi
	add	BOFFSET = KK8, B
	mov	f68  = f0
	shladd	AOFFSET = KK8, 2, AOFFSET
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
	mov	f76  = f0
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 4, KK
#else
	adds	L = 1, KK
#endif
#endif
	}
	;;
#endif
	{ .mfi
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	mov	f79  = f0
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
	{ .mfi
	LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	mov	f69  = f0
	tbit.z	p12, p0 = L, 0
	}
	{ .mfi
	adds	PREC = CPREFETCHSIZE * SIZE, C1
	mov	f77  = f0
	shr	L = L, 1
	}
	;;
	{ .mfi
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	mov	f78  = f0
	adds	L =  -1, L
	}
	{ .mmf
	LDFPD	f36, f37 = [AOFFSET], 2 * SIZE
	CPREFETCH [PREC]
	mov	f70  = f0
	}
	;;
	{ .mfi
	LDFPD	f38, f39 = [AOFFSET], 2 * SIZE
	mov	f71  = f0
	mov	ar.lc = L
	}
	{ .mmi
	adds	C5 = 4 * SIZE, C1
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	cmp.eq	p3, p0 = r0, r0
	}
	;;
	.align 16

.L093:
/*  1 */
	{ .mfi
	lfetch.nt1	[PREA],  16 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfi
	nop	__LINE__
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfi
	lfetch.nt1	[PREB],   4 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	}
	{ .mfi
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	}
	;;
	{ .mfi
	(p3) LDFPD	f56, f57 = [BOFFSET],   2 * SIZE
	FMA	f66   = f34, f48, f66	// A3 * B1
	}
	{ .mfi
	nop	__LINE__
	FMA	f74   = f34, f49, f74	// A3 * B2
	}
	;;
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f67   = f35, f48, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	FMA	f68   = f36, f48, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f76   = f36, f49, f76	// A5 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f44, f45 = [AOFFSET], 2 * SIZE
	FMA	f69   = f37, f48, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f77   = f37, f49, f77	// A6 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f46, f47 = [AOFFSET], 2 * SIZE
	FMA	f70   = f38, f48, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f78   = f38, f49, f78	// A7 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	FMA	f71   = f39, f48, f71	// A8 * B1
	nop	__LINE__
	}
	{ .mfb
	FMA	f79   = f39, f49, f79	// A8 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],  2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f36, f37 = [AOFFSET], 2 * SIZE
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f38, f39 = [AOFFSET], 2 * SIZE
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6   = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f68   = f44, f56, f68	// A5 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f12  = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f76   = f44, f57, f76	// A5 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7   = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f69   = f45, f56, f69	// A6 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f13  = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f77   = f45, f57, f77	// A6 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f70   = f46, f56, f70	// A7 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f14  = [C5 ], SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f78   = f46, f57, f78	// A7 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11  = [C1 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f71   = f47, f56, f71	// A8 * B1
	adds	L = -1, L
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f15  = [C5 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f79   = f47, f57, f79	// A8 * B2
	br.cloop.sptk.few .L093
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f68  = f68,  f77
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f69  = f69,  f76
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f70  = f70,  f79
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f71  = f71,  f78
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f12 = ALPHA_R, f68, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f14 = ALPHA_R, f70, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f15
	nop	__LINE__
	}
	;;

	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
	cmp.ne	p6, p0 = 1, I
	}
	{ .mfb
	STFD	[C5] = f12, SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f7,  SIZE
	mov	f65  = f0
	adds	I = -1, I
	}
	{ .mfb
	STFD	[C5] = f13, SIZE
	mov	f73  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f10, SIZE
	mov	f66  = f0
	nop	__LINE__
	}
	{ .mfb
	STFD	[C5] = f14, SIZE
	mov	f74  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f11, 5 * SIZE
	mov	f67  = f0
	}
	{ .mfb
	STFD	[C5] = f15, 5 * SIZE
	mov	f75 = f0
	(p6)	br.cond.dptk .L092
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f12 = ALPHA_R, f68, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f13 = ALPHA_R, f69, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f14 = ALPHA_R, f70, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f15 = ALPHA_R, f71, f0
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f12 = ALPHA_I, f69, f12
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f13 = ALPHA_I, f68, f13
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f14 = ALPHA_I, f71, f14
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -4, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FMA	f15 = ALPHA_I, f70, f15
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -1, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C5] = f12, SIZE
	mov	f72  = f0
	cmp.ne	p6, p0 = 1, I
	}
	;;
	{ .mfi
	STFD	[C1] = f7,  SIZE
	mov	f65  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 2, AOFFSET
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C5] = f13, SIZE
	mov	f73  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	add	BOFFSET = KK8, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	mov	f66  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 4, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	STFD	[C5] = f14, SIZE
	mov	f74  = f0
	adds	I = -1, I
	}
	;;
	{ .mfi
	STFD	[C1] = f11, 5 * SIZE
	mov	f67  = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	STFD	[C5] = f15, 5 * SIZE
	mov	f75 = f0
	(p6)	br.cond.dptk .L092
	}
	;;
#endif
	.align 16

.L100:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 2, KK
#else
	adds	L = 1, KK
#endif
#endif
	tbit.z	p6, p7 = M, 1
	(p6)	br.cond.dptk .L110
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mmi
	LDFPD	f48, f49 = [B]
	adds	BOFFSET = 2 * SIZE, B
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mii
	add	BOFFSET = KK8, B
	shladd	AOFFSET = KK8, 1, AOFFSET
	nop	__LINE__
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	{ .mii
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	tbit.z	p12, p0 = L, 0
	shr	L = L, 1
	}
	;;
	{ .mmi
	LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	nop	__LINE__
	adds	L =  -1, L
	}
	;;
	{ .mmi
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	cmp.eq	p3, p0 = r0, r0
	mov	ar.lc = L
	}
	;;
	.align 16

.L102:
	{ .mfi
	lfetch.nt1	[PREA],  8 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfi
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mfb
	lfetch.nt1	[PREB],  4 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	FMA	f73   = f33, f49, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f56, f57 = [BOFFSET], 2 * SIZE
	FMA	f66   = f34, f48, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	FMA	f74   = f34, f49, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	FMA	f67   = f35, f48, f67	// A4 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f10 = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	FMA	f75   = f35, f49, f75	// A4 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p3) LDFPD	f42, f43 = [AOFFSET], 2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f11 = [C1 ], -3 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	nop	__LINE__
	}
	;;
	{ .mfb
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	(p3) FMA	f66   = f42, f56, f66	// A3 * B1
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f74   = f42, f57, f74	// A3 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
	(p4) LDFPD	f34, f35 = [AOFFSET], 2 * SIZE
	(p3) FMA	f67   = f43, f56, f67	// A4 * B1
	adds	L = -1, L
	}
	{ .mfb
	nop	__LINE__
	(p3) FMA	f75   = f43, f57, f75	// A4 * B2
	br.cloop.sptk.few .L102
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_A	  f66  = f66,  f75
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f67  = f67,  f74
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f11
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
	nop	__LINE__
	}
	;;
	{ .mfb
	STFD	[C1] = f7,  SIZE
	mov	f72  = f0
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	mov	f65  = f0
	adds	L =  1, K
	}
	;;
	{ .mfi
	STFD	[C1] = f11, SIZE
	mov	f73 = f0
	shr	L = L, 1
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FMA	f10 = ALPHA_R, f66, f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	sub	L = K, KK
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f11 = ALPHA_R, f67, f0
	nop	__LINE__
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
#if defined(TRMMKERNEL) && (defined(LEFT) &&  defined(TRANSA))
	adds	L = -2, L
#else
	nop	__LINE__
#endif
	}
	{ .mfi
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
#if defined(TRMMKERNEL) && (!defined(LEFT) && !defined(TRANSA))
	adds	L = -1, L
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	nop	__LINE__
	FCALC_D	f10 = ALPHA_I, f67, f10
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	KK8 = L, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	{ .mfb
	nop	__LINE__
	FMA	f11 = ALPHA_I, f66, f11
	nop	__LINE__
	}
	;;
	{ .mfi
	STFD	[C1] = f6,  SIZE
	mov	f64  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	shladd	AOFFSET = KK8, 1, AOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f7,  SIZE
	mov	f72  = f0
#if defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	add	BOFFSET = KK8, BOFFSET
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f10, SIZE
	mov	f65  = f0
#if defined(TRMMKERNEL) && defined(LEFT)
	adds	KK = 2, KK
#else
	nop	__LINE__
#endif
	}
	;;
	{ .mfi
	STFD	[C1] = f11, SIZE
	mov	f73 = f0
#ifdef TRMMKERNEL
	shladd	KK8 = KK, ZBASE_SHIFT, r0
#else
	nop	__LINE__
#endif
	}
	;;
#endif
	.align 16

.L110:
	{ .mib
#ifndef TRMMKERNEL
	nop	__LINE__
#else
#if (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	sub	L = K, KK
#elif defined(LEFT)
	adds	L = 1, KK
#else
	adds	L = 1, KK
#endif
#endif
	tbit.z	p6, p7 = M, 0
	(p6)	br.cond.dptk .L119
	}
	;;
#if !defined(TRMMKERNEL) || \
    defined(TRMMKERNEL) && \
    ((defined(LEFT) &&  defined(TRANSA)) || (!defined(LEFT) && !defined(TRANSA)))
	{ .mmi
	LDFPD	f48, f49 = [B]
	adds	BOFFSET = 2 * SIZE, B
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#else
	{ .mii
	add	BOFFSET = KK8, B
	add	AOFFSET = KK8, AOFFSET
	nop	__LINE__
	}
	;;
	{ .mfi
	LDFPD	f48, f49 = [BOFFSET], 2 * SIZE
#ifndef TRMMKERNEL
	adds	L =  1, K
#else
	adds	L =  1, L
#endif
	}
	;;
#endif
	;;
	{ .mii
	nop	__LINE__
	tbit.z	p12, p0 = L, 0
	shr	L = L, 1
	}
	;;
	{ .mmi
	LDFPD	f32, f33 = [AOFFSET], 2 * SIZE
	cmp.eq	p3, p0 = r0, r0
	adds	L =  -1, L
	}
	;;
	{ .mmi
	adds	PREA = (PREFETCHSIZE + 0) * SIZE, AOFFSET
	adds	PREB = (PREFETCHSIZE + 0) * SIZE, BOFFSET
	mov	ar.lc = L
	}
	;;
	.align 16

.L112:
	{ .mfi
	lfetch.nt1	[PREA],  4 * SIZE
	FMA	f64   = f32, f48, f64	// A1 * B1
	cmp.ne	p4, p5 =  0, L
	}
	{ .mfi
	lfetch.nt1	[PREB],   4 * SIZE
	FMA	f72   = f32, f49, f72	// A1 * B2
	(p12) cmp.ne p3, p0 =  0, L
	}
	;;
	{ .mmf
	(p3) LDFPD	f40, f41 = [AOFFSET], 2 * SIZE
	(p3) LDFPD	f56, f57 = [BOFFSET], 2 * SIZE
	FMA	f65   = f33, f48, f65	// A2 * B1
	}
	{ .mmf
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f6  = [C1 ], SIZE
#else
	nop	__LINE__
#endif
	nop	__LINE__
	FMA	f73   = f33, f49, f73	// A2 * B2
	}
	;;
	{ .mfb
	(p4) LDFPD	f32, f33 = [AOFFSET],   2 * SIZE
	(p3) FMA	f64   = f40, f56, f64	// A1 * B1
	nop	__LINE__
	}
	{ .mfb
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	(p5) LDFD	f7  = [C1 ], -1 * SIZE
#else
	nop	__LINE__
#endif
	(p3) FMA	f72   = f40, f57, f72	// A1 * B2
	nop	__LINE__
	}
	;;
	{ .mfi
	(p4) LDFPD	f48, f49 = [BOFFSET],   2 * SIZE
	(p3) FMA	f65   = f41, f56, f65	// A2 * B1
	adds	L = -1, L
	}
	{ .mfb
	(p3) FMA	f73   = f41, f57, f73	// A2 * B2
	br.cloop.sptk.few .L112
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_A	  f64  = f64,  f73
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_B	  f65  = f65,  f72
	nop	__LINE__
	}
	;;
#if! defined(TRMMKERNEL) && !defined(BETAZERO)
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f7
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	;;
	{ .mmb
	STFD	[C1] = f6,  SIZE
	nop	__LINE__
	}
	;;	
	{ .mmb
	STFD	[C1] = f7,  SIZE
	nop	__LINE__
	}
	;;
#else
	{ .mfb
	nop	__LINE__
	FMA	f6  = ALPHA_R, f64, f0
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FCALC_C	f7  = ALPHA_R, f65, f0
	nop	__LINE__
	}
	;;
	{ .mfb
	nop	__LINE__
	FCALC_D	f6  = ALPHA_I, f65, f6
	nop	__LINE__
	}
	{ .mfb
	nop	__LINE__
	FMA	f7  = ALPHA_I, f64, f7
	nop	__LINE__
	}
	;;
	{ .mmb
	STFD	[C1] = f6,  SIZE
	nop	__LINE__
	}
	;;	
	{ .mmb
	STFD	[C1] = f7,  SIZE
	nop	__LINE__
	}
	;;
#endif
	.align 16

.L119:
	{ .mmi
	mov	B = BOFFSET
	mov	AOFFSET = A
#if defined(TRMMKERNEL) && !defined(LEFT)
	adds	KK = 1, KK
#else
	nop	__LINE__
#endif
	}
	;;
	.align 16

.L999:
	mov	r8 = r0
	adds	r9 = 1 * 16, SP
	;;
	ldf.fill  f16 = [SP], 32
	ldf.fill  f17 = [r9], 32
	;;	
	ldf.fill  f18 = [SP], 32
	ldf.fill  f19 = [r9], 32
	;;	
	ldf.fill  f20 = [SP], 32
	ldf.fill  f21 = [r9], 32
	;;	
	ldf.fill  f22 = [SP], 32
	ldf.fill  f23 = [r9], 32
	mov	 ar.lc = ARLC
	;;
       ldf.fill  f24 = [SP], 32
       ldf.fill  f25 = [r9], 32
	mov pr    = PR, -276
       ;;
       ldf.fill  f26 = [SP], 32
       ldf.fill  f27 = [r9], 32
#ifdef TRMMKERNEL
	mov	ar.pfs = ARPFS
#endif
       ;;
       ldf.fill  f28 = [SP], 32
       ldf.fill  f29 = [r9], 32
       ;;
       ldf.fill  f30 = [SP], 32
       ldf.fill  f31 = [r9]
	br.ret.sptk.many b0
	EPILOGUE

